Package: lfl
Type: Package
Title: Linguistic Fuzzy Logic
Version: 2.4.0
Date: 2026-07-03
Authors@R: 
    person(given = "Michal",
           family = "Burda",
           role = c("aut", "cre"),
           email = "michal.burda@osu.cz",
           comment = c(ORCID = "0000-0002-4182-4407"))
Maintainer: Michal Burda <michal.burda@osu.cz>
Description: Various algorithms related to linguistic fuzzy logic: mining for
    linguistic fuzzy association rules, composition of fuzzy relations,
    performing Mamdani, implicative, and perception-based logical deduction
    (PbLD), and forecasting time-series using fuzzy rule-based ensemble (FRBE).
    The package also contains basic fuzzy-related algebraic functions capable of
    handling missing values in different styles (Bochvar, Sobocinski, Kleene
    etc.), computation of Sugeno integrals and the fuzzy transform.
License: GPL-3
Suggests: testthat, doMC, knitr, rmarkdown, R.rsp
Depends: R (>= 3.6)
Imports: Rcpp (>= 0.12.12), foreach, forecast (>= 5.5), plyr, tseries,
        e1071, tibble
LinkingTo: Rcpp
NeedsCompilation: yes
Encoding: UTF-8
Language: en-US
VignetteBuilder: R.rsp
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-03 07:15:20 UTC; michal
Author: Michal Burda [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4182-4407>)
Repository: CRAN
Date/Publication: 2026-07-03 08:50:10 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-07-03 12:40:23 UTC; unix
Archs: lfl.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",fsets)
S3method(as.ctx3,ctx3)
S3method(as.ctx3,ctx3bilat)
S3method(as.ctx3,ctx5)
S3method(as.ctx3,ctx5bilat)
S3method(as.ctx3,default)
S3method(as.ctx3bilat,ctx3)
S3method(as.ctx3bilat,ctx3bilat)
S3method(as.ctx3bilat,ctx5)
S3method(as.ctx3bilat,ctx5bilat)
S3method(as.ctx3bilat,default)
S3method(as.ctx5,ctx3)
S3method(as.ctx5,ctx3bilat)
S3method(as.ctx5,ctx5)
S3method(as.ctx5,ctx5bilat)
S3method(as.ctx5,default)
S3method(as.ctx5bilat,ctx3)
S3method(as.ctx5bilat,ctx3bilat)
S3method(as.ctx5bilat,ctx5)
S3method(as.ctx5bilat,ctx5bilat)
S3method(as.ctx5bilat,default)
S3method(as.data.frame,farules)
S3method(as.data.frame,fsets)
S3method(as.matrix,fsets)
S3method(as_tibble,fsets)
S3method(c,farules)
S3method(cbind,fsets)
S3method(fcut,data.frame)
S3method(fcut,default)
S3method(fcut,factor)
S3method(fcut,logical)
S3method(fcut,matrix)
S3method(fcut,numeric)
S3method(lcut,data.frame)
S3method(lcut,default)
S3method(lcut,factor)
S3method(lcut,logical)
S3method(lcut,matrix)
S3method(lcut,numeric)
S3method(plot,fsets)
S3method(print,algebra)
S3method(print,ctx3)
S3method(print,ctx3bilat)
S3method(print,ctx5)
S3method(print,ctx5bilat)
S3method(print,farules)
S3method(print,frbe)
S3method(print,fsets)
S3method(rbind,fsets)
S3method(str,fsets)
export("specs<-")
export("vars<-")
export(aggregateConsequents)
export(algebra)
export(allowed.lingexpr)
export(antecedents)
export(as.ctx3)
export(as.ctx3bilat)
export(as.ctx5)
export(as.ctx5bilat)
export(compose)
export(consequents)
export(ctx3)
export(ctx3bilat)
export(ctx5)
export(ctx5bilat)
export(defaultHedgeParams)
export(defuzz)
export(dragonfly)
export(equidist)
export(equifreq)
export(evalfrbe)
export(farules)
export(fcut)
export(fire)
export(frbe)
export(fsets)
export(ft)
export(ftinv)
export(goedel.biresiduum)
export(goedel.residuum)
export(goedel.tconorm)
export(goedel.tnorm)
export(goguen.biresiduum)
export(goguen.residuum)
export(goguen.tconorm)
export(goguen.tnorm)
export(hedge)
export(horizon)
export(infer)
export(invol.neg)
export(is.algebra)
export(is.ctx3)
export(is.ctx3bilat)
export(is.ctx5)
export(is.ctx5bilat)
export(is.farules)
export(is.frbe)
export(is.fsets)
export(is.ft)
export(is.specific)
export(kleene)
export(lcut)
export(lingexpr)
export(lowerBound)
export(lowerEst)
export(lukas.biresiduum)
export(lukas.residuum)
export(lukas.tconorm)
export(lukas.tnorm)
export(mase)
export(minmax)
export(mult)
export(nelson)
export(pbld)
export(perceive)
export(pgoedel.tconorm)
export(pgoedel.tnorm)
export(pgoguen.tconorm)
export(pgoguen.tnorm)
export(plukas.tconorm)
export(plukas.tnorm)
export(quantifier)
export(raisedcos)
export(raisedcosine)
export(rbcoverage)
export(reduce)
export(rmse)
export(searchrules)
export(slices)
export(smape)
export(sobocinski)
export(specs)
export(strict.neg)
export(sugeno)
export(triangle)
export(triangular)
export(upperBound)
export(vars)
importFrom(Rcpp,evalCpp)
importFrom(e1071,kurtosis)
importFrom(e1071,skewness)
importFrom(foreach,"%do%")
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(forecast,auto.arima)
importFrom(forecast,ets)
importFrom(forecast,forecast)
importFrom(forecast,rwf)
importFrom(forecast,thetaf)
importFrom(plyr,colwise)
importFrom(plyr,laply)
importFrom(stats,as.formula)
importFrom(stats,coef)
importFrom(stats,frequency)
importFrom(stats,lm)
importFrom(stats,model.matrix)
importFrom(stats,na.omit)
importFrom(stats,sd)
importFrom(stats,ts)
importFrom(stats,ts.plot)
importFrom(stats,weighted.mean)
importFrom(tibble,as_tibble)
importFrom(tseries,adf.test)
importFrom(utils,str)
useDynLib(lfl, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

[![CRAN
status](https://www.r-pkg.org/badges/version/lfl)](https://CRAN.R-project.org/package=lfl)
[![codecov](https://codecov.io/gh/beerda/lfl/branch/master/graph/badge.svg)](https://app.codecov.io/gh/beerda/lfl)

# lfl

The lfl package provides various algorithms related to linguistic fuzzy
logic: mining for linguistic fuzzy association rules, composition of
fuzzy relations, performing perception-based logical deduction (PbLD),
and forecasting time-series using fuzzy rule-based ensemble (FRBE). The
package also contains basic fuzzy-related algebraic functions capable of
handling missing values in different styles (Bochvar, Sobocinski, Kleene
etc.), computation of Sugeno integrals and fuzzy transform.

## Documentation

The complete documentation of the package is [available in this
vignette](https://github.com/beerda/lfl/blob/master/vignettes/main.pdf).

## Installation

To install the stable version from CRAN, simply issue the following
command within your R session:

``` r
install.packages("lfl")
```

If you want to install the development version instead, type:

``` r
install.packages("devtools")
devtools::install_github("beerda/lfl")
```

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Transform data into a `fsets` S3 class of linguistic fuzzy attributes
#'
#' This function creates a set of linguistic fuzzy attributes from crisp data.
#' Numeric vectors, matrix or data frame columns are transformed into a set of
#' fuzzy attributes, i.e., columns with membership degrees. Factors and other
#' data types are transformed to fuzzy attributes by calling the [fcut()]
#' function.
#'
#' The aim of this function is to transform numeric data into a set of fuzzy
#' attributes.  The resulting fuzzy attributes have direct linguistic
#' interpretation. This is a unique variant of fuzzification that is suitable
#' for the inference mechanism based on Perception-based Linguistic Description
#' (PbLD) -- see [pbld()].
#'
#' A numeric vector is transformed into a set of fuzzy attributes according to
#' the following scheme:
#'
#' \eqn{<hedge> <atomic expression>}
#'
#' where \eqn{<atomic expression>} is an atomic linguistic expression, a value
#' from the following possibilities (note that the allowance of atomic
#' expressions is influenced with `context` being used - see [ctx] for details):
#'
#'  * `neg.bi`: big negative (far from zero)
#'  * `neg.um`: upper medium negative (between medium negative and big negative)
#'  * `neg.me`: medium negative
#'  * `neg.lm`: lower medium negative (between medium negative and small
#'    negative)
#'  * `neg.sm`: small negative (close to zero)
#'  * `ze`: zero
#'  * `sm`: small
#'  * `lm`: lower medium
#'  * `me`: medium
#'  * `um`: upper medium
#'  * `bi`: big
#'
#' A \eqn{<hedge>} is a modifier that further concretizes the atomic expression
#' (note that not each combination of hedge and atomic expression is allowed -
#' see [allowed.lingexpr] for more details):
#'
#'  * `ex`: extremely,
#'  * `si`: significantly,
#'  * `ve`: very,
#'  * `ty`: typically,
#'  * `-`: empty hedge (no hedging),
#'  * `ml`: more or less,
#'  * `ro`: roughly,
#'  * `qr`: quite roughly,
#'  * `vr`: very roughly.
#'
#' Accordingly to the theory developed by Novak (2008), not every hedge is
#' suitable with each atomic expression (see the description of the `hedges`
#' argument).  The hedges to be used can be selected with the `hedges` argument.
#' Function takes care of not to use hedge together with an unapplicable atomic
#' expression by itself.
#'
#' Obviously, distinct data have different meaning of what is "small", "medium",
#' or "big" etc.  Therefore, a `context` has to be set that specifies sensible
#' values for these linguistic expressions.
#'
#' If a matrix (resp. data frame) is provided to this function instead of a
#' single vector, all columns are processed the same way.
#'
#' The function also sets up properly the [vars()] and [specs()] properties of
#' the result.
#'
#' @param x Data to be transformed: if it is a numeric vector, matrix, or data
#'   frame, then the creation of linguistic fuzzy attributes takes place. For
#'   other data types the [fcut()] function is called implicitly.
#' @param context A definition of context of a numeric attribute. It must be an
#'   instance of an S3 class [ctx3()], [ctx5()], [ctx3bilat()] or [ctx5bilat()].
#'
#'   If `x` is a matrix or data frame then `context` should be a named list of
#'   contexts for each `x`'s column.
#' @param atomic A vector of atomic linguistic expressions to be used for
#'   creation of fuzzy attributes.
#' @param hedges A vector of linguistic hedges to be used for creation of fuzzy
#'   attributes.
#' @param name A name to be added as a suffix to the created fuzzy attribute
#'   names. This parameter can be used only if `x` is a numeric or logical
#'   vector or a factor. If `x` is a matrix or data frame, `name` should be NULL
#'   because the fuzzy attribute names are taken from column names of parameter
#'   `x`. The `name` is also used as a value for the `vars` attribute of the
#'   resulting [fsets()] instance.
#' @param hedgeParams Parameters that determine the shape of the hedges
#' @param ...  Other parameters to some methods.
#' @return An object of S3 class `fsets` is returned, which is a numeric matrix
#'   with columns representing the fuzzy attributes. Each source column of the
#'   `x` argument corresponds to multiple columns in the resulting matrix.
#'   Columns will have names derived from used \eqn{hedges}, atomic expression,
#'   and \eqn{name} specified as the optional parameter.
#'
#'   The resulting object would also have set the [vars()] and [specs()]
#'   properties with the former being created from original column names (if `x`
#'   is a matrix or data frame) or the `name` argument (if `x` is a numeric
#'   vector). The [specs()] incidency matrix would be created to reflect the
#'   following order of the hedges: \eqn{"ex" < "si" < "ve" < "-" < "ml" < "ro"
#'   < "qr" < "vr"} and \eqn{"ty" < "" < "ml" < "ro" < "qr" < "vr"}.  Fuzzy
#'   attributes created from the same source numeric vector (or column) would be
#'   ordered that way, with other fuzzy attributes (from the other source) being
#'   incomparable.
#' @author Michal Burda
#' @seealso [fcut()], [fsets()], [vars()], [specs()]
#' @references V. Novak, A comprehensive theory of trichotomous evaluative
#'   linguistic expressions, Fuzzy Sets and Systems 159 (22) (2008) 2939--2969.
#' @keywords models robust multivariate
#' @examples
#'
#' # transform a single vector
#' x <- runif(10)
#' lcut(x, name='age')
#'
#' # transform single vector with a custom context
#' lcut(x, context=ctx5(0, 0.2, 0.5, 0.7, 1), name='age')
#'
#' # transform all columns of a data frame
#' # and do not use any hedges
#' data <- CO2[, c('conc', 'uptake')]
#' lcut(data)
#'
#'
#' # definition of custom contexts for different columns
#' # of a data frame while selecting only "ve" and "ro" hedges.
#' lcut(data,
#'      context=list(conc=minmax,
#'                   uptake=ctx3(0, 25, 50)),
#'      hedges=c('ve', 'ro'))
#'
#'
#' # lcut on non-numeric data is the same as fcut()
#' ff <- factor(substring("statistics", 1:10, 1:10), levels = letters)
#' lcut(ff)
#'
#' @export
#' @name lcut
lcut <- function(x, ...) {
    UseMethod('lcut')
}


#' @rdname lcut
#' @export
lcut.default <- function(x, ...) {
    .stop(paste0("'lcut' not implemented for variable of class '", class(x), "'"))
}


#' @rdname lcut
#' @export
lcut.factor <- function(x,
                        name=deparse(substitute(x)),
                        ...) {
    fcut(x, name=name, ...)
}


#' @rdname lcut
#' @export
lcut.logical <- function(x,
                         name=deparse(substitute(x)),
                         ...) {
    fcut(x, name=name, ...)
}


.allHedges=c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr')
.sharpness <- matrix(0,
                     nrow=length(.allHedges),
                     ncol=length(.allHedges),
                     dimnames=list(.allHedges, .allHedges))
.sharpness[row(.sharpness) < col(.sharpness)] <- 1
.sharpness[, 'ty'] <- 0


#' @rdname lcut
#' @export
lcut.numeric <- function(x,
                         context=minmax,
                         atomic=c('sm', 'me', 'bi', 'lm', 'um', 'ze',
                                  'neg.sm', 'neg.me', 'neg.bi', 'neg.lm', 'neg.um'),
                         hedges=c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr'),
                         name=deparse(substitute(x)),
                         hedgeParams=defaultHedgeParams,
                         ...) {
    atomic <- match.arg(atomic, several.ok=TRUE)
    hedges <- match.arg(hedges, several.ok=TRUE)

    .mustBeNumericVector(x)
    .mustNotBeZeroLength(atomic)
    .mustNotBeZeroLength(hedges)
    .mustNotBeNull(name)

    if (is.function(context)) {
        context <- context(x)
    }

    type <- NULL
    for (clazz in names(.horizonAllowedTable)) {
        if (inherits(context, clazz)) {
            type <- clazz
        }
    }
    .mustBe(!is.null(type),
            paste0("'context' must be an object of one of the following classes: '",
                   paste(names(.horizonAllowedTable), collapse="', '"), "' or a function returning such object"))

    allowedAtomic <- names(.horizonAllowedTable[[type]])
    if (!identical(atomic, eval(formals(lcut.numeric)$atomic))) {
        # non-default atomic argument
        wrongAtomic <- setdiff(atomic, allowedAtomic)
        if (length(wrongAtomic) > 0) {
            warning(paste0("The following atomic expressions are not allowed for context '",
                         type, "': ", paste(wrongAtomic, collapse=', ')))
        }
    }
    atomic <- intersect(atomic, allowedAtomic)

    res <- NULL
    for (a in atomic) {
        allowedHedges <- names(which(allowed.lingexpr[, a]))
        allowedHedges <- intersect(allowedHedges, hedges)
        if (length(allowedHedges) > 0) {
            m <- lapply(allowedHedges, function(h) {
                lingexpr(context, atomic=a, hedge=h, hedgeParams=hedgeParams)(x)
            })
            m <- matrix(unlist(m, use.names=FALSE), ncol=length(m), byrow=FALSE)

            n <- paste(allowedHedges, a, name, sep='.')
            n <- sub('-.', '', n, fixed=TRUE)
            colnames(m) <- n

            v <- rep(name, ncol(m))
            s <- .sharpness[allowedHedges, allowedHedges, drop=FALSE]
            f <- fsets(m, v, s)
            res <- cbind.fsets(res, f, warn=FALSE)
        }
    }
    return(res)
}


.byNameOrDefault <- function(x, name, default) {
    res <- x[[name]]
    if (is.null(res)) {
        res <- default
    }
    return(res)
}


#' @rdname lcut
#' @export
lcut.data.frame <- function(x,
                            context=minmax,
                            atomic=c('sm', 'me', 'bi', 'lm', 'um', 'ze',
                                     'neg.sm', 'neg.me', 'neg.bi', 'neg.lm', 'neg.um'),
                            hedges=c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr'),
                            ...) {
    .mustBeDataFrame(x)
    if (is.null(colnames(x))) {
        stop("Columns of 'x' must have names")
    }
    if (!is.list(context)) {
        context <- rep(list(context), ncol(x))
        names(context) <- colnames(x)
    }
    if (!is.list(atomic)) {
        atomic <- rep(list(atomic), ncol(x))
        names(atomic) <- colnames(x)
    }
    if (!is.list(hedges)) {
        hedges <- rep(list(hedges), ncol(x))
        names(hedges) <- colnames(x)
    }
    res <- NULL
    for (i in colnames(x)) {
        f <- lcut(x[[i]],
                  context=.byNameOrDefault(context, i, minmax),
                  atomic=.byNameOrDefault(atomic, i, c('sm', 'me', 'bi', 'lm', 'um', 'ze', 'neg.sm', 'neg.me', 'neg.bi', 'neg.lm', 'neg.um')),
                  hedges=.byNameOrDefault(hedges, i, c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr')),
                  name=i,
                  ...)
        res <- cbind.fsets(res, f)
    }
    return(res)
}


#' @rdname lcut
#' @export
lcut.matrix <- function(x, ...) {
    lcut(as.data.frame(x), ...)
}

test_that("infer mamdani on fsets", {
    # init fsets
    .vars <- c(rep('b', 3),
               rep('c', 3))
    names(.vars) <- paste(rep(c('ve.sm', 'sm', 'bi'), times=2),
                          rep(c('b', 'c'), each=3),
                          sep='.')


    .specs <- matrix(c(0,1,0, 0,0,0,
                       0,0,0, 0,0,0,
                       0,0,0, 0,0,0,

                       0,0,0, 0,1,0,
                       0,0,0, 0,0,0,
                       0,0,0, 0,0,0),
                     byrow=TRUE,
                     ncol=6)
    colnames(.specs) <- names(.vars)
    rownames(.specs) <- names(.vars)

    x <- matrix(runif(18), ncol=6, nrow=3)
    colnames(x) <- names(.vars)

    x <- fsets(x, vars=.vars, specs=.specs)

    # init rules
    rules <- list(c('sm.b', 've.sm.c'),
                  c('sm.b', 'sm.c'),
                  c('bi.b', 'bi.b'),
                  c('bi.b', 'sm.c', 'sm.b'))

    # init values
    values <- 0:10 / 10

    # init partition
    partition <- lcut(data.frame(b=values))

    res <- infer(x, rules, partition, alg = "goedel", type = "mamdani")
    expect_true(is.matrix(res))
    expect_equal(nrow(res), nrow(x))
    expect_equal(ncol(res), nrow(partition))
})


test_that("infer mamdani on custom matrix", {
    rules <- list(
        c("High",   "temp_high",   "pressure_high"),
        c("High",   "temp_high",   "humidity_low"),
        c("Medium", "temp_medium", "pressure_medium"),
        c("Low",    "temp_low"),
        c("Low",    "humidity_high", "pressure_low")
    )

    x <- matrix(
        c(
            0.9, 0.8, 0.7, 0.2, 0.3, 0.1, 0.2, 0.1,
            0.3, 0.2, 0.8, 0.7, 0.6, 0.2, 0.1, 0.4,
            0.1, 0.2, 0.2, 0.4, 0.5, 0.8, 0.7, 0.6,
            0.6, 0.5, 0.1, 0.5, 0.4, 0.3, 0.8, 0.7
        ),
        nrow = 4,
        byrow = TRUE
    )
    colnames(x) <- c("temp_high", "pressure_high", "humidity_low",
                     "temp_medium", "pressure_medium", "temp_low",
                     "humidity_high", "pressure_low")

    partition <- matrix(
        c(
            1.0, 0.2, 0.0,
            0.8, 0.5, 0.1,
            0.4, 1.0, 0.4,
            0.1, 0.5, 0.8,
            0.0, 0.1, 1.0
        ),
        nrow = 5,
        byrow = TRUE
    )
    colnames(partition) <- c("Low", "Medium", "High")
    rownames(partition) <- paste0("y", 1:5)

    expected <- matrix(
        c(
            0.2, 0.2, 0.4, 0.8, 0.8,
            0.2, 0.5, 0.6, 0.5, 0.3,
            0.8, 0.8, 0.4, 0.4, 0.1,
            0.7, 0.7, 0.4, 0.5, 0.5
        ),
        nrow = 4,
        byrow = TRUE
    )

    res <- infer(x, rules, partition, alg = "goedel", type = "mamdani")
    expect_equal(res, expected)
})


test_that("infer mamdani on empty rule set", {
    rules <- list()

    x <- matrix(
        c(
            0.9, 0.8, 0.7, 0.2, 0.3, 0.1, 0.2, 0.1,
            0.3, 0.2, 0.8, 0.7, 0.6, 0.2, 0.1, 0.4,
            0.1, 0.2, 0.2, 0.4, 0.5, 0.8, 0.7, 0.6,
            0.6, 0.5, 0.1, 0.5, 0.4, 0.3, 0.8, 0.7
        ),
        nrow = 4,
        byrow = TRUE
    )
    colnames(x) <- c("temp_high", "pressure_high", "humidity_low",
                     "temp_medium", "pressure_medium", "temp_low",
                     "humidity_high", "pressure_low")

    partition <- matrix(
        c(
            1.0, 0.2, 0.0,
            0.8, 0.5, 0.1,
            0.4, 1.0, 0.4,
            0.1, 0.5, 0.8,
            0.0, 0.1, 1.0
        ),
        nrow = 5,
        byrow = TRUE
    )
    colnames(partition) <- c("Low", "Medium", "High")
    rownames(partition) <- paste0("y", 1:5)

    expected <- matrix(
        c(
            0, 0, 0, 0, 0,
            0, 0, 0, 0, 0,
            0, 0, 0, 0, 0,
            0, 0, 0, 0, 0
        ),
        nrow = 4,
        byrow = TRUE
    )

    res <- infer(x, rules, partition, alg = "goedel", type = "mamdani")
    expect_equal(res, expected)
})


test_that("infer implicative on custom matrix", {
    rules <- list(
        c("High",   "temp_high",   "pressure_high"),
        c("High",   "temp_high",   "humidity_low"),
        c("Medium", "temp_medium", "pressure_medium"),
        c("Low",    "temp_low"),
        c("Low",    "humidity_high", "pressure_low")
    )

    x <- matrix(
        c(
            0.9, 0.8, 0.7, 0.2, 0.3, 0.1, 0.2, 0.1,
            0.3, 0.2, 0.8, 0.7, 0.6, 0.2, 0.1, 0.4,
            0.1, 0.2, 0.2, 0.4, 0.5, 0.8, 0.7, 0.6,
            0.6, 0.5, 0.1, 0.5, 0.4, 0.3, 0.8, 0.7
        ),
        nrow = 4,
        byrow = TRUE
    )
    colnames(x) <- c("temp_high", "pressure_high", "humidity_low",
                     "temp_medium", "pressure_medium", "temp_low",
                     "humidity_high", "pressure_low")

    partition <- matrix(
        c(
            1.0, 0.2, 0.0,
            0.8, 0.5, 0.1,
            0.4, 1.0, 0.4,
            0.1, 0.5, 0.8,
            0.0, 0.1, 1.0
        ),
        nrow = 5,
        byrow = TRUE
    )
    colnames(partition) <- c("Low", "Medium", "High")
    rownames(partition) <- paste0("y", 1:5)

    expected <- matrix(
        c(
            0,   0.1, 0.4, 1.0, 0,
            0,   0.1, 1.0, 0.1, 0,
            0,   1.0, 0.4, 0.1, 0,
            0,   0.1, 0.4, 0.1, 0
        ),
        nrow = 4,
        byrow = TRUE
    )

    res <- infer(x, rules, partition, alg = "goedel", type = "implicative")
    expect_equal(res, expected)
})

test_that('searchrules', {
    n <- 100
    d <- data.frame(a=1:n, b=n:1, c=runif(n))
    d <- lcut(d)

    rules <- searchrules(d)
    expect_true(is.farules(rules))
})


test_that('searchrules toy example', {
    expectRule <- function(d, i, rule, support, confidence) {
        expect_equal(rownames(d)[i], rule)
        expect_equal(d[i, 'support'], support, tolerance=1e-7)
        expect_equal(d[i, 'confidence'], confidence, tolerance=1e-7)
    }

    d <- data.frame(a=c(1, 0.9, 0.8, 0.2),
                    b=c(1, 0.7, 0, 0.3),
                    c=c(1, 0.6, 0, 0.4))
    f <- fsets(x=as.matrix(d), vars=letters[1:3])
    rules <- searchrules(f, lhs=1:3, rhs=1:3, minSupport=0, minConfidence=0, maxLength=100, tnorm='goedel', trie=FALSE)

    expect_true(is.farules(rules))
    expect_equal(length(rules$rules), 12)
    expect_equal(nrow(rules$statistics), 12)

    rulesOrder <- order(laply(rules$rules, function(r) paste(r, collapse='-')))
    drules <- as.data.frame(rules)
    drules <- drules[rulesOrder, ]

    expectRule(drules, 1, " => a", 0.725, 0.725)
    expectRule(drules, 2, "b => a", 0.475, 0.95)
    expectRule(drules, 3, "b & c => a", 0.45, 0.947368421)
    expectRule(drules, 4, "c => a", 0.45, 0.9)
    expectRule(drules, 5, " => b", 0.5, 0.5)
    expectRule(drules, 6, "a => b", 0.475, 0.655172414)
    expectRule(drules, 7, "a & c => b", 0.45, 1.0)
    expectRule(drules, 8, "c => b", 0.475, 0.95)
    expectRule(drules, 9, " => c", 0.5, 0.5)
    expectRule(drules, 10, "a => c", 0.45, 0.620689655)
    expectRule(drules, 11, "a & b => c", 0.45, 0.947368421)
    expectRule(drules, 12, "b => c", 0.475, 0.95)
})

test_that('perceive global', {
    f <- lcut(data.frame(a=0:1, b=0:1, c=0:1, d=0:1))

    runPerceive <- function(...) {
        ret <- perceive(list(...), fsets=f)
        return(ret)
    }

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('sm.d', 've.sm.a', 'bi.c'),
                             c('sm.d', 'sm.b', 'sm.c')),
                 c(F, T, T))

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('sm.d', 've.sm.a', 'sm.c'),
                             c('sm.d', 'sm.b', 'sm.c')),
                 c(T, T, T))

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('bi.d', 've.sm.a', 'bi.c'),
                             c('sm.d', 'sm.b', 'sm.c')),
                 c(F, T, T))
})


test_that('perceive local', {
    f <- lcut(data.frame(a=0:1, b=0:1, c=0:1, d=0:1))

    runPerceive <- function(..., fired) {
        return(perceive(list(...), fsets=f, type='local', fired=fired))
    }

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('sm.d', 've.sm.a', 'bi.c'),
                             fired=c(0.3, 0.4)),
                 c(F, T))

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('sm.d', 've.sm.a', 'bi.c'),
                             fired=c(0.4, 0.3)),
                 c(T, F))

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('sm.d', 've.sm.a', 'bi.c'),
                             fired=c(0.4, 0.4)),
                 c(F, T))

    expect_equal(runPerceive(c('sm.d', 'sm.a', 'bi.c'),
                             c('sm.d', 've.sm.a', 'bi.c'),
                             fired=c(1, 1)),
                 c(F, T))
})

test_that('pbld', {
    # init fsets
    .vars <- c(rep('b', 3),
               rep('c', 3))
    names(.vars) <- paste(rep(c('ve.sm', 'sm', 'bi'), times=2),
                          rep(c('b', 'c'), each=3),
                          sep='.')


    .specs <- matrix(c(0,1,0, 0,0,0,
                       0,0,0, 0,0,0,
                       0,0,0, 0,0,0,

                       0,0,0, 0,1,0,
                       0,0,0, 0,0,0,
                       0,0,0, 0,0,0),
                     byrow=TRUE,
                     ncol=6)
    colnames(.specs) <- names(.vars)
    rownames(.specs) <- names(.vars)

    x <- matrix(runif(18), ncol=6, nrow=3)
    colnames(x) <- names(.vars)

    x <- fsets(x, vars=.vars, specs=.specs)

    # init rules
    rules <- list(c('sm.b', 've.sm.c'),
                  c('sm.b', 'sm.c'),
                  c('bi.b', 'bi.b'),
                  c('bi.b', 'sm.c', 'sm.b'))

    # init values
    values <- 0:10 / 10

    # init partition
    partition <- lcut(data.frame(b=values))


    res <- pbld(x, rules, partition, values)
    expect_equal(nrow(x), length(res))
    #expect_true(FALSE)
})


test_that('pbld with empty rulebase', {
    # init fsets
    .vars <- c(rep('b', 3),
               rep('c', 3))
    names(.vars) <- paste(rep(c('ve.sm', 'sm', 'bi'), times=2),
                          rep(c('b', 'c'), each=3),
                          sep='.')


    .specs <- matrix(c(0,1,0, 0,0,0,
                       0,0,0, 0,0,0,
                       0,0,0, 0,0,0,

                       0,0,0, 0,1,0,
                       0,0,0, 0,0,0,
                       0,0,0, 0,0,0),
                     byrow=TRUE,
                     ncol=6)
    colnames(.specs) <- names(.vars)
    rownames(.specs) <- names(.vars)

    x <- matrix(runif(18), ncol=6, nrow=3)
    colnames(x) <- names(.vars)

    x <- fsets(x, vars=.vars, specs=.specs)

    # init rules
    rules <- list()

    # init values
    values <- 0:10 / 10

    # init partition
    partition <- lcut(data.frame(b=values))


    res <- pbld(x, rules, partition, values)
    expect_equal(nrow(x), length(res))
})


test_that('pbld complete test', {
    x <- matrix(seq(0, 0.4, 0.025), nrow=17, ncol=1)
    colnames(x) <- 'x'
    inputContext <- ctx3(0, 0.4, 1)
    input <- lcut(x, context=inputContext)

    rules <- list(c('bi.y', 'sm.x'),
                c('me.y', 'ro.sm.x'),
                c('ve.sm.y', 'ex.sm.x'))

    outputContext <- ctx3(0, 0.4, 1)
    v <- seq(outputContext[1], outputContext[3], length.out=1000)
    p <- lcut(v, name='y', context=outputContext)

    expect_equal(round(pbld(input, rules, p, v), 2),
                 c(0.03, 0.91, 0.91, 0.41, 0.41, 0.41, 0.41, 0.42, 0.44, 0.45, rep(0, 7)))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Algebra for Fuzzy Sets
#'
#' Compute triangular norms (t-norms), triangular conorms (t-conorms), residua,
#' bi-residua, and negations.
#'
#' `goedel.tnorm`, `lukas.tnorm`, and `goguen.tnorm` compute the
#' Goedel, Lukasiewicz, and Goguen triangular norm (t-norm) from all values in
#' the arguments. If the arguments are vectors they are combined together
#' firstly so that a numeric vector of length 1 is returned.
#'
#' `pgoedel.tnorm`, `plukas.tnorm`, and `pgoguen.tnorm` compute
#' the same t-norms, but in an element-wise manner. I.e. the values
#' with indices 1 of all arguments are used to compute the t-norm, then the
#' second values (while recycling the vectors if they do not have the same
#' size) so that the result is a vector of values.
#'
#' `goedel.tconorm`, `lukas.tconorm`, `goguen.tconorm`, are
#' similar to the previously mentioned functions, except that they compute
#' triangular conorms (t-conorms).  `pgoedel.tconorm`,
#' `plukas.tconorm`, and `pgoguen.tconorm` are their element-wise alternatives.
#'
#' `goedel.residuum`, `lukas.residuum`, and `goguen.residuum`
#' compute residua (i.e., implications) and `goedel.biresiduum`,
#' `lukas.biresiduum`, and `goguen.biresiduum` compute bi-residua. Residua and
#' bi-residua are computed in an element-wise manner, for each corresponding
#' pair of values in `x` and `y` arguments.
#'
#' `invol.neg` and `strict.neg` compute the involutive and strict
#' negation, respectively.
#'
#' Let \eqn{a}, \eqn{b} be values from the interval \eqn{[0, 1]}. The realized functions
#' can be defined as follows:
#' * Goedel t-norm: \eqn{min{a, b}};
#' * Goguen t-norm: \eqn{ab};
#' * Lukasiewicz t-norm: \eqn{max{0, a+b-1}};
#' * Goedel t-conorm: \eqn{max{a, b}};
#' * Goguen t-conorm: \eqn{a+b-ab};
#' * Lukasiewicz t-conorm: \eqn{min{1, a+b}};
#' * Goedel residuum (standard Goedel implication): \eqn{1} if \eqn{a \le b} and
#'   \eqn{b} otherwise;
#' * Goguen residuum (implication): \eqn{1} if \eqn{a \le b} and \eqn{b/a}
#'   otherwise;
#' * Lukasiewicz residuum (standard Lukasiewicz implication): \eqn{1} if
#'   \eqn{a \le b} and \eqn{1-a+b} otherwise;
#' * Involutive negation: \eqn{1-x};
#' * Strict negation: \eqn{1} if \eqn{x=0} and \eqn{0} otherwise.
#'
#' Bi-residuum \eqn{B} is derived from residuum \eqn{R}
#' as follows: \deqn{B(a, b) = inf(R(a, b), R(b, a)),}
#' where \eqn{inf} is the operation of infimum, which for all three algebras
#' corresponds to the \eqn{min} operation.
#'
#' The arguments have to be numbers from the interval \eqn{[0, 1]}. Values
#' outside that range cause an error. NaN values are treated as NAs.
#'
#' If some argument is NA or NaN, the result is NA. For other handling of missing values,
#' see [algebraNA].
#'
#' Selection of a t-norm may serve as a basis for definition of other operations.
#' From the t-norm, the operation of a residual implication may be defined, which
#' in turn allows the definition of a residual negation. If the residual negation
#' is not involutive, the involutive negation is often added as a new operation
#' and together with the t-norm can be used to define the t-conorm. Therefore,
#' the `algebra` function returns a named list of operations derived from the selected
#' Goedel, Goguen, or Lukasiewicz t-norm. Concretely:
#' * `algebra("goedel")`: returns the strict negation as the residual negation,
#'    the involutive negation, and also the Goedel t-norm, t-conorm, residuum, and bi-residuum;
#' * `algebra("goguen")`: returns the strict negation as the residual negation,
#'    the involutive negation, and also the Goguen t-norm, t-conorm, residuum, and bi-residuum;
#' * `algebra("lukasiewicz")`: returns involutive negation as both residual and involutive
#'    negation, and also the Lukasiewicz t-norm, t-conorm, residuum, and bi-residuum.
#'
#' Moreover, `algebra` returns the supremum and infimum functions computed as maximum and minimum,
#' respectively.
#'
#' `is.algebra` tests whether the given `a` argument is a valid
#' algebra, i.e., a list returned by the `algebra` function.
#'
#' @param name The name of the algebra to be created. Must be one of: "goedel",
#' "lukasiewicz", "goguen" (or an unambiguous abbreviation).
#' @param stdneg (Deprecated.) `TRUE` to force the use of a "standard" negation (i.e.,
#' involutive negation). Otherwise, the appropriate negation is used in the
#' algebra (e.g., strict negation in Goedel and Goguen algebras and involutive
#' negation in Lukasiewicz algebra).
#' @param ...  For t-norms and t-conorms, these arguments are numeric vectors
#' of values to compute t-norms or t-conorms from.  Values outside the
#' \eqn{[0,1]} interval cause an error. NA values are also permitted.
#'
#' For the `algebra()` function, these arguments are passed to the factory
#' functions that create the algebra. (Currently unused.)
#' @param a An object to be checked if it is a valid algebra (i.e., a list
#' returned by the `algebra` function).
#' @param x Numeric vector of values to compute a residuum or bi-residuum from.
#' Values outside the \eqn{[0,1]} interval cause an error. NA values are also
#' permitted.
#' @param y Numeric vector of values to compute a residuum or bi-residuum from.
#' Values outside the \eqn{[0,1]} interval cause an error. NA values are also
#' permitted.
#' @return Functions for t-norms and t-conorms (such as `goedel.tnorm`)
#' return a numeric vector of size 1 that is the result of the appropriate
#' t-norm or t-conorm applied on all values of all arguments.
#'
#' Element-wise versions of t-norms and t-conorms (such as `pgoedel.tnorm`)
#' return a vector of results after applying the appropriate t-norm or t-conorm
#' on argument in an element-wise (i.e., by indices) way. The
#' resulting vector is of length of the longest argument (shorter arguments are
#' recycled).
#'
#' Residua and bi-residua functions return a numeric vector of length of the
#' longest argument (shorter argument is recycled).
#'
#' `strict.neg` and `invol.neg` compute negations and return a
#' numeric vector of the same size as the argument `x`.
#'
#' `algebra` returns a list of functions of the requested algebra:
#' `"n"` (residual negation), `"ni"` (involutive negation), `"t"` (t-norm),
#' `"pt"` (element-wise t-norm),
#' `"c"` (t-conorm), `"pc"` (element-wise t-conorm), `"r"` (residuum),
#' `"b"` (bi-residuum), `"s"` (supremum),
#' `"ps"` (element-wise supremum), `"i"` (infimum), and
#' `"pi"` (element-wise infimum).
#'
#' For Lukasiewicz algebra, the elements `"n"` and `"ni"` are the same, i.e.
#' the `invol.neg` function. For Goedel and Goguen algebra, `"n"` (the residual
#' negation) equals `strict.neg` and `"ni"` (the involutive negation) equals
#' `invol.neg`.
#'
#' `"s"`, `"ps"`, `"i"`, `"pi"` are the same for each type of algebra:
#' `goedel.conorm`, `pgoedel.conorm`, `goedel.tnorm`, and `pgoedel.tnorm`.
#'
#' @author Michal Burda
#' @keywords models robust
#' @examples
#'     # direct and element-wise version of functions
#'     goedel.tnorm(c(0.3, 0.2, 0.5), c(0.8, 0.1, 0.5))  # 0.1
#'     pgoedel.tnorm(c(0.3, 0.2, 0.5), c(0.8, 0.1, 0.5)) # c(0.3, 0.1, 0.5)
#'
#'     # algebras
#'     x <- runif(10)
#'     y <- runif(10)
#'     a <- algebra('goedel')
#'     a$n(x)     # residual negation
#'     a$ni(x)    # involutive negation
#'     a$t(x, y)  # t-norm
#'     a$pt(x, y) # element-wise t-norm
#'     a$c(x, y)  # t-conorm
#'     a$pc(x, y) # element-wise t-conorm
#'     a$r(x, y)  # residuum
#'     a$b(x, y)  # bi-residuum
#'     a$s(x, y)  # supremum
#'     a$ps(x, y) # element-wise supremum
#'     a$i(x, y)  # infimum
#'     a$pi(x, y) # element-wise infimum
#'
#'     is.algebra(a) # TRUE
#' @export
algebra <- function(name, stdneg=FALSE, ...) {
    if (stdneg) {
      .Deprecated('algebra',
                  msg=paste('The "stdneg" argument is deprecated. If you need',
                            'the involutive negation, use the "ni" function',
                            'returned in list by the algebra() function.'))
    }
    name <- match.arg(name, names(.algebras))
    res <- .algebras[[name]](...)
    if (stdneg) {
        res[['n']] <- invol.neg
    }
    class(res) <- c('algebra', 'list')
    res$algebratype <- name
    return(res)
}


#' Print an instance of the [algebra()] S3 class in a human readable form.
#'
#' @param x An instance of the [algebra()] S3 class
#' @param ...  Unused.
#' @return None.
#' @author Michal Burda
#' @seealso [algebra()]
#' @keywords models robust
#' @export
print.algebra <- function(x, ...) {
  cat('Algebra:', paste0(x$algebratype, collapse='/'), '\n')
  x$algebratype <- NULL
  str(x, give.attr=FALSE, no.list=TRUE)
}


#' @rdname algebra
#' @export
is.algebra <- function(a) {
  return(is.list(a) &&
           inherits(a, 'algebra') &&
           is.function(a$n) &&
           is.function(a$ni) &&
           is.function(a$t) &&
           is.function(a$pt) &&
           is.function(a$c) &&
           is.function(a$pc) &&
           is.function(a$r) &&
           is.function(a$b) &&
           is.function(a$s) &&
           is.function(a$ps) &&
           is.function(a$i) &&
           is.function(a$pi) &&
           is.function(a$order))
}


###########################################################
# t-norms

#' @rdname algebra
#' @export
goedel.tnorm <- function(...) {
    vals <- as.numeric(c(...))
    .Call('_lfl_goedel_tnorm', vals, PACKAGE='lfl')
}

#' @rdname algebra
#' @export
lukas.tnorm <- function(...) {
    vals <- as.numeric(c(...))
    .Call('_lfl_lukas_tnorm', vals, PACKAGE='lfl')
}

#' @rdname algebra
#' @export
goguen.tnorm <- function(...) {
    vals <- as.numeric(c(...))
    .Call('_lfl_goguen_tnorm', vals, PACKAGE='lfl')
}

#' @rdname algebra
#' @export
pgoedel.tnorm <- function(...) {
  elts <- list(...)
  if (length(elts) <= 0L) {
    return(NULL);
  }
  vals <- lapply(elts, as.numeric)
  res <- .Call('_lfl_pgoedel_tnorm', vals, PACKAGE='lfl')
  mostattributes(res) <- attributes(elts[[1L]])
  res
}

#' @rdname algebra
#' @export
plukas.tnorm <- function(...) {
  elts <- list(...)
  if (length(elts) <= 0L) {
    return(NULL);
  }
  vals <- lapply(elts, as.numeric)
  res <- .Call('_lfl_plukas_tnorm', vals, PACKAGE='lfl')
  mostattributes(res) <- attributes(elts[[1L]])
  res
}


#' @rdname algebra
#' @export
pgoguen.tnorm <- function(...) {
  elts <- list(...)
  if (length(elts) <= 0L) {
    return(NULL);
  }
  vals <- lapply(elts, as.numeric)
  res <- .Call('_lfl_pgoguen_tnorm', vals, PACKAGE='lfl')
  mostattributes(res) <- attributes(elts[[1L]])
  res
}


###########################################################
# t-conorms

#' @rdname algebra
#' @export
goedel.tconorm <- function(...) {
    vals <- as.numeric(c(...))
    .Call('_lfl_goedel_tconorm', vals, PACKAGE='lfl')
}

#' @rdname algebra
#' @export
lukas.tconorm <- function(...) {
    vals <- as.numeric(c(...))
    .Call('_lfl_lukas_tconorm', vals, PACKAGE='lfl')
}

#' @rdname algebra
#' @export
goguen.tconorm <- function(...) {
    vals <- as.numeric(c(...))
    .Call('_lfl_goguen_tconorm', vals, PACKAGE='lfl')
}

#' @rdname algebra
#' @export
pgoedel.tconorm <- function(...) {
  elts <- list(...)
  if (length(elts) <= 0L) {
    return(NULL);
  }
  vals <- lapply(elts, as.numeric)
  res <- .Call('_lfl_pgoedel_tconorm', vals, PACKAGE='lfl')
  mostattributes(res) <- attributes(elts[[1L]])
  res
}

#' @rdname algebra
#' @export
plukas.tconorm <- function(...) {
  elts <- list(...)
  if (length(elts) <= 0L) {
    return(NULL);
  }
  vals <- lapply(elts, as.numeric)
  res <- .Call('_lfl_plukas_tconorm', vals, PACKAGE='lfl')
  mostattributes(res) <- attributes(elts[[1L]])
  res
}


#' @rdname algebra
#' @export
pgoguen.tconorm <- function(...) {
  elts <- list(...)
  if (length(elts) <= 0L) {
    return(NULL);
  }
  vals <- lapply(elts, as.numeric)
  res <- .Call('_lfl_pgoguen_tconorm', vals, PACKAGE='lfl')
  mostattributes(res) <- attributes(elts[[1L]])
  res
}


###########################################################
# residua

#' @rdname algebra
#' @export
goedel.residuum <- function(x, y) {
    .Call('_lfl_goedel_residuum', as.numeric(x), as.numeric(y), PACKAGE='lfl')
}

#' @rdname algebra
#' @export
lukas.residuum <- function(x, y) {
    .Call('_lfl_lukas_residuum', as.numeric(x), as.numeric(y), PACKAGE='lfl')
}

#' @rdname algebra
#' @export
goguen.residuum <- function(x, y) {
    .Call('_lfl_goguen_residuum', as.numeric(x), as.numeric(y), PACKAGE='lfl')
}


###########################################################
# bi-residua

#' @rdname algebra
#' @export
goedel.biresiduum <- function(x, y) {
    pgoedel.tnorm(goedel.residuum(x, y), goedel.residuum(y, x))
}

#' @rdname algebra
#' @export
lukas.biresiduum <- function(x, y) {
    pgoedel.tnorm(lukas.residuum(x, y), lukas.residuum(y, x))
}

#' @rdname algebra
#' @export
goguen.biresiduum <- function(x, y) {
    pgoedel.tnorm(goguen.residuum(x, y), goguen.residuum(y, x))
}


###########################################################
# negations

#' @rdname algebra
#' @export
invol.neg <- function(x) {
    vals <- as.numeric(c(x))
    res <- .Call('_lfl_invol_neg', vals, PACKAGE='lfl')
    mostattributes(res) <- attributes(x)
    return(res)
}

#' @rdname algebra
#' @export
strict.neg <- function(x) {
    vals <- as.numeric(c(x))
    res <- .Call('_lfl_strict_neg', vals, PACKAGE='lfl')
    mostattributes(res) <- attributes(x)
    return(res)
}


.tnorms <- list(goedel=goedel.tnorm,
                lukasiewicz=lukas.tnorm,
                goguen=goguen.tnorm)

.ptnorms <- list(goedel=pgoedel.tnorm,
                 lukasiewicz=plukas.tnorm,
                 goguen=pgoguen.tnorm)

.tconorms <- list(goedel=goedel.tconorm,
                  lukasiewicz=lukas.tconorm,
                  goguen=goguen.tconorm)

.ptconorms <- list(goedel=pgoedel.tconorm,
                  lukasiewicz=plukas.tconorm,
                  goguen=pgoguen.tconorm)

.residua <- list(goedel=goedel.residuum,
                 lukasiewicz=lukas.residuum,
                 goguen=goguen.residuum)

.biresidua <- list(goedel=goedel.biresiduum,
                   lukasiewicz=lukas.biresiduum,
                   goguen=goguen.biresiduum)

.negations <- list(involutive=invol.neg,
                   strict=strict.neg,
                   lukasiewicz=invol.neg,
                   goedel=strict.neg,
                   goguen=strict.neg)

.defaultOrder <- function(x, decreasing=FALSE) {
  order(x, decreasing=decreasing)
}

.algebras <- list('goedel'=function(...) {
                        list(n=strict.neg,
                             ni=invol.neg,
                             t=goedel.tnorm,
                             pt=pgoedel.tnorm,
                             c=goedel.tconorm,
                             pc=pgoedel.tconorm,
                             r=goedel.residuum,
                             b=goedel.biresiduum,
                             i=goedel.tnorm,
                             pi=pgoedel.tnorm,
                             s=goedel.tconorm,
                             ps=pgoedel.tconorm,
                             order=.defaultOrder)
                  },
                  'lukasiewicz'=function(...) {
                        list(n=invol.neg,
                             ni=invol.neg,
                             t=lukas.tnorm,
                             pt=plukas.tnorm,
                             c=lukas.tconorm,
                             pc=plukas.tconorm,
                             r=lukas.residuum,
                             b=lukas.biresiduum,
                             i=goedel.tnorm,
                             pi=pgoedel.tnorm,
                             s=goedel.tconorm,
                             ps=pgoedel.tconorm,
                             order=.defaultOrder)
                  },
                  'goguen'=function(...) {
                        list(n=strict.neg,
                             ni=invol.neg,
                             t=goguen.tnorm,
                             pt=pgoguen.tnorm,
                             c=goguen.tconorm,
                             pc=pgoguen.tconorm,
                             r=goguen.residuum,
                             b=goguen.biresiduum,
                             i=goedel.tnorm,
                             pi=pgoedel.tnorm,
                             s=goedel.tconorm,
                             ps=pgoedel.tconorm,
                             order=.defaultOrder)
                   })

set.seed(34523)

test_that('minimum (Goedel) t-norm', {
    expect_that(goedel.tnorm(0.2, 0.5, 0.1, 0.3), equals(0.1))
    expect_that(goedel.tnorm(0.4, 0.5, 0.3), equals(0.3))
    expect_that(goedel.tnorm(0.2, 0.5, 0.9), equals(0.2))
    expect_that(goedel.tnorm(0.2, 0.5, 0.0), equals(0))
    expect_that(goedel.tnorm(1, 1, 1, 1), equals(1))
    expect_that(goedel.tnorm(1, 0.9, 1, 1), equals(0.9))

    expect_that(goedel.tnorm(c(0.2, 0.5, 0.1, 0.3)), equals(0.1))
    expect_that(goedel.tnorm(c(0.4, 0.5, 0.3)), equals(0.3))
    expect_that(goedel.tnorm(c(0.2, 0.5, 0.9)), equals(0.2))
    expect_that(goedel.tnorm(c(0.2, 0.5, 0.0)), equals(0))
    expect_that(goedel.tnorm(c(1, 1, 1, 1)), equals(1))
    expect_that(goedel.tnorm(c(1, 0.9, 1, 1)), equals(0.9))
})


test_that('lukasiewicz t-norm', {
    expect_that(lukas.tnorm(0.2, 0.5, 0.1, 0.3), equals(0))
    expect_that(lukas.tnorm(0.8, 0.5, 0.9), equals(0.2))
    expect_that(lukas.tnorm(1, 1, 1, 1), equals(1))
    expect_that(lukas.tnorm(1, 0.9, 1, 1), equals(0.9))
    expect_that(lukas.tnorm(1, 0.9, 0.8, 1), equals(0.7))
    expect_that(lukas.tnorm(0.2, 0.5, 0.0), equals(0))

    expect_that(lukas.tnorm(c(0.2, 0.5, 0.1, 0.3)), equals(0))
    expect_that(lukas.tnorm(c(0.8, 0.5, 0.9)), equals(0.2))
    expect_that(lukas.tnorm(c(1, 1, 1, 1)), equals(1))
    expect_that(lukas.tnorm(c(1, 0.9, 1, 1)), equals(0.9))
    expect_that(lukas.tnorm(c(0.2, 0.5, 0.0)), equals(0))
})


test_that('product (goguen) t-norm', {
    expect_that(goguen.tnorm(0.2, 0.5, 0.1, 0.3), equals(0.2 * 0.5 * 0.1 * 0.3))
    expect_that(goguen.tnorm(0.8, 0.5, 0.9), equals(0.8 * 0.5 * 0.9))
    expect_that(goguen.tnorm(1, 1, 1, 1), equals(1))
    expect_that(goguen.tnorm(1, 0.9, 1, 1), equals(0.9))
    expect_that(goguen.tnorm(0.2, 0.5, 0.0), equals(0))

    expect_that(goguen.tnorm(c(0.2, 0.5, 0.1, 0.3)), equals(0.2 * 0.5 * 0.1 * 0.3))
    expect_that(goguen.tnorm(c(0.8, 0.5, 0.9)), equals(0.8 * 0.5 * 0.9))
    expect_that(goguen.tnorm(c(1, 1, 1, 1)), equals(1))
    expect_that(goguen.tnorm(c(1, 0.9, 1, 1)), equals(0.9))
    expect_that(goguen.tnorm(c(0.2, 0.5, 0.0)), equals(0))
})


for (ttt in names(.tnorms)) {
    test_that(paste(ttt, 't-norm borders'), {
        tnorm <- .tnorms[[ttt]]

        expect_that(tnorm(), equals(NA_real_))
        expect_that(tnorm(0.2, NA, 1), equals(NA_real_))
        expect_that(tnorm(0.2, NA, 0), equals(NA_real_))
        expect_that(tnorm(0.2, NaN, 1), equals(NA_real_))
        expect_that(tnorm(0.2, NaN, 0), equals(NA_real_))
        expect_that(tnorm(0.2, Inf, 0), throws_error('argument out of range 0..1'))
        expect_that(tnorm(0.2, -Inf, 0), throws_error('argument out of range 0..1'))
        expect_that(tnorm(0.2, 3, 0), throws_error('argument out of range 0..1'))
        expect_that(tnorm(0.2, -3, 0), throws_error('argument out of range 0..1'))

        expect_that(tnorm(c()), equals(NA_real_))
        expect_that(tnorm(c(0.2, NA, 1)), equals(NA_real_))
        expect_that(tnorm(c(0.2, NA, 0)), equals(NA_real_))
        expect_that(tnorm(c(0.2, NaN, 1)), equals(NA_real_))
        expect_that(tnorm(c(0.2, NaN, 0)), equals(NA_real_))
        expect_that(tnorm(c(0.2, Inf, 0)), throws_error('argument out of range 0..1'))
        expect_that(tnorm(c(0.2, -Inf, 0)), throws_error('argument out of range 0..1'))
        expect_that(tnorm(c(0.2, 3, 0)), throws_error('argument out of range 0..1'))
        expect_that(tnorm(c(0.2, -3, 0)), throws_error('argument out of range 0..1'))
    })
}


test_that('parallel minimum t-norm', {
    expect_that(pgoedel.tnorm(c(0.2, 0.5, 0.4, 0.9, 1),
                    c(0.5, 0.9, 0.8, 1.0, 1),
                    c(0.6, 0.4, 0.0, 1.0, 1),
                    c(0.3, 0.7, 0.5, 1.0, 1)),
                equals(c(0.2, 0.4, 0.0, 0.9, 1)))

    expect_that(pgoedel.tnorm(0.2, 0.5), equals(0.2))
    expect_that(pgoedel.tnorm(0.2, 0.5, 0.0), equals(0))
    expect_that(pgoedel.tnorm(c(0.2, 0.5, 0.0)), equals(c(0.2, 0.5, 0.0)))
})


test_that('parallel lukasiewicz t-norm', {
    expect_that(plukas.tnorm(c(0.2, 0.8, 0.4, 0.9, 1),
                    c(0.5, 0.9, 0.8, 1.0, 1),
                    c(0.6, 0.5, 0.0, 1.0, 1),
                    c(0.3, 0.9, 0.5, 1.0, 1)),
                equals(c(0, 0.1, 0.0, 0.9, 1)))

    expect_that(plukas.tnorm(0.7, 0.8, 0.6), equals(0.1))
    expect_that(plukas.tnorm(0.2, 0.5, 0.0), equals(0))
    expect_that(plukas.tnorm(c(0.2, 0.5, 0.0)), equals(c(0.2, 0.5, 0.0)))
})


test_that('parallel product t-norm', {
    expect_that(pgoguen.tnorm(c(0.2, 0.5, 0.4, 0.9, 1),
                    c(0.5, 0.9, 0.8, 1.0, 1),
                    c(0.6, 0.4, 0.0, 1.0, 1),
                    c(0.3, 0.7, 0.5, 1.0, 1)),
                equals(c(0.2 * 0.5 * 0.6 * 0.3,
                        0.5 * 0.9 * 0.4 * 0.7,
                        0.0, 0.9, 1)))

    expect_that(pgoguen.tnorm(0.2, 0.5), equals(0.2 * 0.5))
    expect_that(pgoguen.tnorm(0.2, 0.5, 0.0), equals(0))
    expect_that(pgoguen.tnorm(c(0.2, 0.5, 0.0)), equals(c(0.2, 0.5, 0.0)))
})


for (ttt in names(.ptnorms)) {
    test_that(paste('parallel', ttt, 't-norm borders'), {
        tnorm <- .ptnorms[[ttt]]

        expect_true(is.null(tnorm()))
        expect_that(tnorm(c(0.2, NA, 1), c(0.8, 0.6, NA))[2:3], equals(as.numeric(c(NA, NA))))
        expect_that(tnorm(c(0.2, NA, 0), c(0.8, 0, NA))[2:3], equals(as.numeric(c(NA, NA))))
        expect_that(tnorm(c(0.2, NaN, 1), c(0.8, 0.6, NA))[2:3], equals(as.numeric(c(NA, NA))))
        expect_that(tnorm(c(0.2, NaN, 0), c(0.8, 0, NaN))[2:3], equals(as.numeric(c(NA, NA))))
        expect_that(tnorm(c(0.2, 0.9, 0), c(0.8, 0, Inf)), throws_error('argument out of range 0..1'))
        expect_that(tnorm(c(0.2, 0.9, 0), c(0.8, 0, -Inf)), throws_error('argument out of range 0..1'))
        expect_that(tnorm(c(0.2, 0.9, 0), c(0.8, 0, 3)), throws_error('argument out of range 0..1'))
        expect_that(tnorm(c(0.2, 0.9, 0), c(0.8, 0, -3)), throws_error('argument out of range 0..1'))

        mr <- matrix(runif(12), nrow=3, ncol=4)
        colnames(mr) <- LETTERS[1:4]
        rownames(mr) <- letters[1:3]

        m0 <- matrix(0, nrow=3, ncol=4)
        colnames(m0) <- colnames(mr)
        rownames(m0) <- rownames(mr)
        expect_that(tnorm(mr, 0), equals(m0))
        expect_that(tnorm(mr, m0), equals(m0))

        m1 <- matrix(1, nrow=3, ncol=4)
        expect_that(tnorm(mr, 1), equals(mr))
        expect_that(tnorm(mr, m1), equals(mr))

        mx <- matrix(tnorm(c(mr), c(mr)), nrow=3, ncol=4)
        colnames(mx) <- colnames(mr)
        rownames(mx) <- rownames(mr)
        expect_that(tnorm(mr, mr), equals(mx))
    })
}


test_that('goedel t-conorm', {
    expect_that(goedel.tconorm(0.2, 0.5, 0.1, 0.3), equals(0.5))
    expect_that(goedel.tconorm(0.4, 0.5, 0.8), equals(0.8))
    expect_that(goedel.tconorm(0.9, 0.5, 0.2), equals(0.9))
    expect_that(goedel.tconorm(0.2, 1, 0.0), equals(1))
    expect_that(goedel.tconorm(0, 0, 0, 0), equals(0))

    expect_that(goedel.tconorm(c(0.2, 0.5, 0.1, 0.3)), equals(0.5))
    expect_that(goedel.tconorm(c(0.4, 0.5, 0.8)), equals(0.8))
    expect_that(goedel.tconorm(c(0.9, 0.5, 0.2)), equals(0.9))
    expect_that(goedel.tconorm(c(0.2, 1, 0.0)), equals(1))
    expect_that(goedel.tconorm(c(0, 0, 0, 0)), equals(0))
})


test_that('lukasiewicz t-conorm', {
    expect_that(lukas.tconorm(0.2, 0.5, 0.1, 0.0), equals(0.8))
    expect_that(lukas.tconorm(0.4, 0.5, 0.8), equals(1))
    expect_that(lukas.tconorm(1, 1, 1), equals(1))
    expect_that(lukas.tconorm(0, 0, 0, 0), equals(0))

    expect_that(lukas.tconorm(c(0.2, 0.5, 0.1, 0.0)), equals(0.8))
    expect_that(lukas.tconorm(c(0.4, 0.5, 0.8)), equals(1))
    expect_that(lukas.tconorm(c(1, 1, 1)), equals(1))
    expect_that(lukas.tconorm(c(0, 0, 0, 0)), equals(0))
})


test_that('goguen t-conorm', {
    expect_that(goguen.tconorm(0.2, 0.5, 0.1, 0.3), equals(0.748))
    expect_that(goguen.tconorm(0.2, 1, 0.0), equals(1))
    expect_that(goguen.tconorm(0, 0, 0, 0), equals(0))

    expect_that(goguen.tconorm(c(0.2, 0.5, 0.1, 0.3)), equals(0.748))
    expect_that(goguen.tconorm(c(0.2, 1, 0.0)), equals(1))
    expect_that(goguen.tconorm(c(0, 0, 0, 0)), equals(0))
})


for (ttt in names(.tconorms)) {
    test_that(paste(ttt, 't-conorm borders'), {
        tconorm <- .tconorms[[ttt]]

        expect_that(tconorm(), equals(NA_real_))
        expect_that(tconorm(0.2, NA, 0), equals(NA_real_))
        expect_that(tconorm(0.2, NA, 1), equals(NA_real_))
        expect_that(tconorm(0.2, Inf, 0), throws_error('argument out of range 0..1'))
        expect_that(tconorm(0.2, -Inf, 0), throws_error('argument out of range 0..1'))
        expect_that(tconorm(0.2, 3, 0), throws_error('argument out of range 0..1'))
        expect_that(tconorm(0.2, -3, 0), throws_error('argument out of range 0..1'))

        expect_that(tconorm(c()), equals(NA_real_))
        expect_that(tconorm(c(0.2, NA, 0)), equals(NA_real_))
        expect_that(tconorm(c(0.2, NA, 1)), equals(NA_real_))
        expect_that(tconorm(c(0.2, Inf, 0)), throws_error('argument out of range 0..1'))
        expect_that(tconorm(c(0.2, -Inf, 0)), throws_error('argument out of range 0..1'))
        expect_that(tconorm(c(0.2, 3, 0)), throws_error('argument out of range 0..1'))
        expect_that(tconorm(c(0.2, -3, 0)), throws_error('argument out of range 0..1'))
    })
}


test_that('goedel residuum', {
    expect_that(goedel.residuum(c(0, 0.2, 0.8, 1), 1), equals(c(1, 1, 1, 1)))
    expect_that(goedel.residuum(c(0, 0.2, 0.8, 1), 0), equals(c(1, 0, 0, 0)))
    expect_that(goedel.residuum(c(0, 0.2, 0.8, 1), 0.5), equals(c(1, 1, 0.5, 0.5)))
    expect_that(goedel.residuum(c(0, 0.2, 0.8, 1), c(0.3, 0.9)), equals(c(1, 1, 0.3, 0.9)))
})


test_that('lukasiewicz residuum', {
    expect_that(lukas.residuum(c(0, 0.2, 0.8, 1), 1), equals(c(1, 1, 1, 1)))
    expect_that(lukas.residuum(c(0, 0.2, 0.8, 1), 0), equals(c(1, 0.8, 0.2, 0)))
    expect_that(lukas.residuum(c(0, 0.2, 0.8, 1), 0.5), equals(c(1, 1, 0.7, 0.5)))
    expect_that(lukas.residuum(c(0, 0.2, 0.8, 1), c(0.3, 0.9)), equals(c(1, 1, 0.5, 0.9)))
})


test_that('goguen residuum', {
    expect_that(goguen.residuum(c(0, 0.2, 0.8, 1), 1), equals(c(1, 1, 1, 1)))
    expect_that(goguen.residuum(c(0, 0.2, 0.8, 1), 0), equals(c(1, 0, 0, 0)))
    expect_that(goguen.residuum(c(0, 0.2, 0.8, 1), 0.5), equals(c(1, 1, 0.625, 0.5)))
    expect_that(goguen.residuum(c(0, 0.2, 0.8, 1), c(0.3, 0.9)), equals(c(1, 1, 0.375, 0.9)))
})


test_that('goedel bi-residuum', {
    expect_that(goedel.biresiduum(c(0, 0.2, 0.8, 1), 1), equals(c(0, 0.2, 0.8, 1)))
    expect_that(goedel.biresiduum(c(0, 0.2, 0.8, 1), 0), equals(c(1, 0, 0, 0)))
    expect_that(goedel.biresiduum(c(0, 0.2, 0.8, 1), 0.5), equals(c(0, 0.2, 0.5, 0.5)))
    expect_that(goedel.biresiduum(c(0, 0.2, 0.8, 1), c(0.3, 0.9)), equals(c(0, 0.2, 0.3, 0.9)))
})


test_that('lukasiewicz bi-residuum', {
    expect_that(lukas.biresiduum(c(0, 0.2, 0.8, 1), 1), equals(c(0, 0.2, 0.8, 1)))
    expect_that(lukas.biresiduum(c(0, 0.2, 0.8, 1), 0), equals(c(1, 0.8, 0.2, 0)))
    expect_that(lukas.biresiduum(c(0, 0.2, 0.8, 1), 0.5), equals(c(0.5, 0.7, 0.7, 0.5)))
    expect_that(lukas.biresiduum(c(0, 0.2, 0.8, 1), c(0.3, 0.9)), equals(c(0.7, 0.3, 0.5, 0.9)))
})


test_that('goguen bi-residuum', {
    expect_that(goguen.biresiduum(c(0, 0.2, 0.8, 1), 1), equals(c(0, 0.2, 0.8, 1)))
    expect_that(goguen.biresiduum(c(0, 0.2, 0.8, 1), 0), equals(c(1, 0, 0, 0)))
    expect_that(goguen.biresiduum(c(0, 0.2, 0.8, 1), 0.5), equals(c(0, 2/5, 5/8, 0.5)))
    expect_that(goguen.biresiduum(c(0, 0.2, 0.8, 1), c(0.3, 0.9)), equals(c(0, 2/9, 3/8, 0.9)))
})


for (ttt in names(.residua)) {
    test_that(paste(ttt, 'residua borders'), {
        resid <- .residua[[ttt]]

        expect_that(resid(0, NA), equals(NA_real_))
        expect_that(resid(0.4, NA), equals(NA_real_))
        expect_that(resid(1, NA), equals(NA_real_))
        expect_that(resid(0.2, Inf), throws_error('argument out of range 0..1'))
        expect_that(resid(0.2, -Inf), throws_error('argument out of range 0..1'))
        expect_that(resid(0.2, 3), throws_error('argument out of range 0..1'))
        expect_that(resid(0.2, -3), throws_error('argument out of range 0..1'))

        expect_that(resid(NA, 0), equals(NA_real_))
        expect_that(resid(NA, 0.2), equals(NA_real_))
        expect_that(resid(NA, 1), equals(NA_real_))
        expect_that(resid(Inf, 0.2), throws_error('argument out of range 0..1'))
        expect_that(resid(-Inf, 0.2), throws_error('argument out of range 0..1'))
        expect_that(resid(3, 0.2), throws_error('argument out of range 0..1'))
        expect_that(resid(-3, 0.2), throws_error('argument out of range 0..1'))
    })
}


test_that('involutive negation', {
    expect_that(invol.neg(c(0, 0.2, NA, 0.8, 1, NaN)), equals(c(1, 0.8, NA, 0.2, 0, NA)))

    m <- matrix(c(0, 0.2, NA, 0.8, 1, 0.3), nrow=2)
    colnames(m) <- letters[1:3]
    rownames(m) <- letters[1:2]
    r <- matrix(c(1, 0.8, NA, 0.2, 0, 0.7), nrow=2)
    colnames(r) <- letters[1:3]
    rownames(r) <- letters[1:2]
    expect_that(invol.neg(m), equals(r))

    expect_that(invol.neg(c(-3, 0.2)), throws_error('argument out of range 0..1'))
    expect_that(invol.neg(c(3, 0.2)), throws_error('argument out of range 0..1'))
    expect_that(invol.neg(c(-Inf, 0.2)), throws_error('argument out of range 0..1'))
    expect_that(invol.neg(c(Inf, 0.2)), throws_error('argument out of range 0..1'))
})


test_that('strict negation', {
    expect_that(strict.neg(c(0, 0.2, NA, 0.8, 1, NaN)), equals(c(1, 0, NA, 0, 0, NA)))

    m <- matrix(c(0, 0.2, NA, 0.8, 1, 0.3), nrow=2)
    colnames(m) <- letters[1:3]
    rownames(m) <- letters[1:2]
    r <- matrix(c(1, 0, NA, 0, 0, 0), nrow=2)
    colnames(r) <- letters[1:3]
    rownames(r) <- letters[1:2]
    expect_that(strict.neg(m), equals(r))

    expect_that(strict.neg(c(-3, 0.2)), throws_error('argument out of range 0..1'))
    expect_that(strict.neg(c(3, 0.2)), throws_error('argument out of range 0..1'))
    expect_that(strict.neg(c(-Inf, 0.2)), throws_error('argument out of range 0..1'))
    expect_that(strict.neg(c(Inf, 0.2)), throws_error('argument out of range 0..1'))
})


test_that('algebra', {
  o <- c(0.5, 1, 0, 0.8, NA, 0.3)
  oi <- c(3, 6, 1, 4, 2, 5)
  od <- c(2, 4, 1, 6, 3, 5)

  a <- algebra('goe')
  expect_true(inherits(a, 'algebra'))
  expect_true(inherits(a, 'list'))
  expect_true(is.algebra(a))
  expect_that(a$t, equals(goedel.tnorm))
  expect_that(a$pt, equals(pgoedel.tnorm))
  expect_that(a$c, equals(goedel.tconorm))
  expect_that(a$pc, equals(pgoedel.tconorm))
  expect_that(a$r, equals(goedel.residuum))
  expect_that(a$b, equals(goedel.biresiduum))
  expect_that(a$n, equals(strict.neg))
  expect_that(a$ni, equals(invol.neg))
  expect_that(a$s, equals(goedel.tconorm))
  expect_that(a$ps, equals(pgoedel.tconorm))
  expect_that(a$i, equals(goedel.tnorm))
  expect_that(a$pi, equals(pgoedel.tnorm))
  expect_that(a$algebratype, equals('goedel'))
  expect_equal(a$order(o), oi)
  expect_equal(a$order(o, decreasing=TRUE), od)

  a <- algebra('lukas')
  expect_true(inherits(a, 'algebra'))
  expect_true(inherits(a, 'list'))
  expect_true(is.algebra(a))
  expect_that(a$t, equals(lukas.tnorm))
  expect_that(a$pt, equals(plukas.tnorm))
  expect_that(a$c, equals(lukas.tconorm))
  expect_that(a$pc, equals(plukas.tconorm))
  expect_that(a$r, equals(lukas.residuum))
  expect_that(a$b, equals(lukas.biresiduum))
  expect_that(a$n, equals(invol.neg))
  expect_that(a$ni, equals(invol.neg))
  expect_that(a$s, equals(goedel.tconorm))
  expect_that(a$ps, equals(pgoedel.tconorm))
  expect_that(a$i, equals(goedel.tnorm))
  expect_that(a$pi, equals(pgoedel.tnorm))
  expect_that(a$algebratype, equals('lukasiewicz'))
  expect_equal(a$order(o), oi)
  expect_equal(a$order(o, decreasing=TRUE), od)

  a <- algebra('gog')
  expect_true(inherits(a, 'algebra'))
  expect_true(inherits(a, 'list'))
  expect_true(is.algebra(a))
  expect_that(a$t, equals(goguen.tnorm))
  expect_that(a$pt, equals(pgoguen.tnorm))
  expect_that(a$c, equals(goguen.tconorm))
  expect_that(a$pc, equals(pgoguen.tconorm))
  expect_that(a$r, equals(goguen.residuum))
  expect_that(a$b, equals(goguen.biresiduum))
  expect_that(a$n, equals(strict.neg))
  expect_that(a$ni, equals(invol.neg))
  expect_that(a$s, equals(goedel.tconorm))
  expect_that(a$ps, equals(pgoedel.tconorm))
  expect_that(a$i, equals(goedel.tnorm))
  expect_that(a$pi, equals(pgoedel.tnorm))
  expect_that(a$algebratype, equals('goguen'))
  expect_equal(a$order(o), oi)
  expect_equal(a$order(o, decreasing=TRUE), od)
})

set.seed(34523)

test_that('sobocinski', {
    a <- sobocinski(algebra('luk'))
    expect_that(a$algebratype, equals(c('lukasiewicz', 'sobocinski')))


    a <- sobocinski(algebra('gog'))

    expect_that(a$algebratype, equals(c('goguen', 'sobocinski')))
    expect_that(a$t(), equals(NA_real_))
    expect_that(a$t(NA), equals(NA_real_))
    expect_that(a$t(NaN), equals(NA_real_))
    expect_that(a$t(0.5, NA, NaN, 0.8, NA), equals(0.4))
    expect_that(a$t(c(0.5, NA, NA, 0.8, NA)), equals(0.4))

    expect_true(is.null(a$pt()))
    expect_that(a$pt(NA, 0.8), equals(0.8))
    expect_that(a$pt(NaN, 0.8), equals(0.8))
    expect_that(a$pt(c(0.5, NA, 0.4), c(NA, 0.4, 0.5)), equals(c(0.5, 0.4, 0.2)))

    expect_that(a$c(), equals(NA_real_))
    expect_that(a$c(NA), equals(NA_real_))
    expect_that(a$c(NaN), equals(NA_real_))
    expect_that(a$c(0.5, NA, NaN, 0.8, NA), equals(0.9))
    expect_that(a$c(c(0.5, NA, NA, 0.8, NA)), equals(0.9))

    expect_true(is.null(a$pc()))
    expect_that(a$pc(NA, 0.8), equals(0.8))
    expect_that(a$pc(NaN, 0.8), equals(0.8))
    expect_that(a$pc(c(0.5, NA, 0.4), c(NA, 0.4, 0.5)), equals(c(0.5, 0.4, 0.7)))

    expect_that(a$r(c(0.5, 0.8,    NA, 0.3, NA),
                    c(0.8, 0.5,   0.4,  NA, NA)),
                equals(c(1.0, 0.625, 0.4,   0, NA_real_)))

    expect_that(a$b(c(0.5,   0.8,    NA,  0.4,  NA),
                    c(0.8,   0.5,   0.4,   NA,  NA)),
                equals(c(0.625, 0.625,   0,    0, NA_real_)))

    expect_that(a$i(), equals(NA_real_))
    expect_that(a$i(NA), equals(NA_real_))
    expect_that(a$i(NaN), equals(NA_real_))
    expect_that(a$i(0.5, NA, NaN, 0.8, NA), equals(0.5))
    expect_that(a$i(c(0.5, NA, NA, 0.8, NA)), equals(0.5))

    expect_true(is.null(a$pi()))
    expect_that(a$pi(NA, 0.8), equals(0.8))
    expect_that(a$pi(NaN, 0.8), equals(0.8))
    expect_that(a$pi(c(0.5, NA, 0.4), c(NA, 0.4, 0.5)), equals(c(0.5, 0.4, 0.4)))

    expect_that(a$s(), equals(NA_real_))
    expect_that(a$s(NA), equals(NA_real_))
    expect_that(a$s(NaN), equals(NA_real_))
    expect_that(a$s(0.5, NA, NaN, 0.8, NA), equals(0.8))
    expect_that(a$s(c(0.5, NA, NA, 0.8, NA)), equals(0.8))

    expect_true(is.null(a$ps()))
    expect_that(a$ps(NA, 0.8), equals(0.8))
    expect_that(a$ps(NaN, 0.8), equals(0.8))
    expect_that(a$ps(c(0.5, NA, 0.4), c(NA, 0.4, 0.5)), equals(c(0.5, 0.4, 0.5)))

    expect_that(a$n(NA), equals(0))
    expect_that(a$n(1), equals(0))
    expect_that(a$n(0), equals(1))
    expect_that(a$n(0.8), equals(a$r(0.8, 0)))

    expect_that(a$ni(NA), equals(NA_real_))
    expect_that(a$ni(1), equals(0))
    expect_that(a$ni(0), equals(1))
    expect_that(a$ni(0.8), equals(0.2))
})


test_that('sobocinski 2', {
    a <- sobocinski(algebra('gog'))

    expect_equal(a$t(0.4, 0.5), 0.2)
    expect_equal(a$t(0, NA), 0)
    expect_equal(a$t(0.5, NA), 0.5)
    expect_equal(a$t(1, NA), 1)
    expect_equal(a$t(NA, 0), 0)
    expect_equal(a$t(NA, 0.5), 0.5)
    expect_equal(a$t(NA, 1), 1)
    expect_equal(a$t(NA, NA), NA_real_)

    expect_equal(a$pt(0.4, 0.5), 0.2)
    expect_equal(a$pt(0, NA), 0)
    expect_equal(a$pt(0.5, NA), 0.5)
    expect_equal(a$pt(1, NA), 1)
    expect_equal(a$pt(NA, 0), 0)
    expect_equal(a$pt(NA, 0.5), 0.5)
    expect_equal(a$pt(NA, 1), 1)
    expect_equal(a$pt(NA, NA), NA_real_)

    expect_equal(a$c(0.4, 0.5), 0.7)
    expect_equal(a$c(0, NA), 0)
    expect_equal(a$c(0.5, NA), 0.5)
    expect_equal(a$c(1, NA), 1)
    expect_equal(a$c(NA, 0), 0)
    expect_equal(a$c(NA, 0.5), 0.5)
    expect_equal(a$c(NA, 1), 1)
    expect_equal(a$c(NA, NA), NA_real_)

    expect_equal(a$pc(0.4, 0.5), 0.7)
    expect_equal(a$pc(0, NA), 0)
    expect_equal(a$pc(0.5, NA), 0.5)
    expect_equal(a$pc(1, NA), 1)
    expect_equal(a$pc(NA, 0), 0)
    expect_equal(a$pc(NA, 0.5), 0.5)
    expect_equal(a$pc(NA, 1), 1)
    expect_equal(a$pc(NA, NA), NA_real_)

    expect_equal(a$i(0.4, 0.5), 0.4)
    expect_equal(a$i(0, NA), 0)
    expect_equal(a$i(0.5, NA), 0.5)
    expect_equal(a$i(1, NA), 1)
    expect_equal(a$i(NA, 0), 0)
    expect_equal(a$i(NA, 0.5), 0.5)
    expect_equal(a$i(NA, 1), 1)
    expect_equal(a$i(NA, NA), NA_real_)

    expect_equal(a$pi(0.4, 0.5), 0.4)
    expect_equal(a$pi(0, NA), 0)
    expect_equal(a$pi(0.5, NA), 0.5)
    expect_equal(a$pi(1, NA), 1)
    expect_equal(a$pi(NA, 0), 0)
    expect_equal(a$pi(NA, 0.5), 0.5)
    expect_equal(a$pi(NA, 1), 1)
    expect_equal(a$pi(NA, NA), NA_real_)

    expect_equal(a$s(0.4, 0.5), 0.5)
    expect_equal(a$s(0, NA), 0)
    expect_equal(a$s(0.5, NA), 0.5)
    expect_equal(a$s(1, NA), 1)
    expect_equal(a$s(NA, 0), 0)
    expect_equal(a$s(NA, 0.5), 0.5)
    expect_equal(a$s(NA, 1), 1)
    expect_equal(a$s(NA, NA), NA_real_)

    expect_equal(a$ps(0.4, 0.5), 0.5)
    expect_equal(a$ps(0, NA), 0)
    expect_equal(a$ps(0.5, NA), 0.5)
    expect_equal(a$ps(1, NA), 1)
    expect_equal(a$ps(NA, 0), 0)
    expect_equal(a$ps(NA, 0.5), 0.5)
    expect_equal(a$ps(NA, 1), 1)
    expect_equal(a$ps(NA, NA), NA_real_)

    expect_equal(a$n(c(0, 0.5, 1, NA)),
                     c(1,   0, 0,  0))
    expect_equal(a$ni(c(0, 0.5, 1, NA)),
                      c(1, 0.5, 0, NA_real_))

    expect_equal(a$b(0.4, 0.5), 4/5)
    expect_equal(a$b(0, NA), 0)
    expect_equal(a$b(0.5, NA), 0)
    expect_equal(a$b(1, NA), 0)
    expect_equal(a$b(0.5, 0.4), 4/5)
    expect_equal(a$b(NA, 0), 0)
    expect_equal(a$b(NA, 0.5), 0)
    expect_equal(a$b(NA, 1), 0)
    expect_equal(a$b(NA, NA), NA_real_)
    expect_equal(a$b(0.5, 0.5), 1)

    expect_equal(a$r(0.4, 0.5), 1)
    expect_equal(a$r(0, NA), 1)
    expect_equal(a$r(0.5, NA), 0)
    expect_equal(a$r(1, NA), 0)
    expect_equal(a$r(NA, 0), 0)
    expect_equal(a$r(NA, 0.5), 0.5)
    expect_equal(a$r(NA, 1), 1)
    expect_equal(a$r(NA, NA), NA_real_)

    a <- sobocinski(algebra('lukasiewicz'))
    expect_equal(a$r(0.5, NA), 0.5)
    expect_equal(a$b(0.5, NA), 0.5)
})


test_that('sobocinski order', {
    a <- sobocinski(algebra('lukasiewicz'))
    expect_error(a$order(0))
})

set.seed(34523)

test_that('upperBound', {
    a <- upperBound(algebra('luk'))
    expect_that(a$algebratype, equals(c('lukasiewicz', 'upperBound')))


    a <-upperBound(algebra('gog'))

    expect_that(a$algebratype, equals(c('goguen', 'upperBound')))
    expect_that(a$t(), equals(NA_real_))
    expect_that(a$t(NA), equals(NA_real_))
    expect_that(a$t(NaN), equals(NA_real_))
    expect_that(a$t(0.5, NA), equals(0.5))
    expect_that(a$t(0.5, NaN), equals(0.5))
    expect_that(a$t(0, NA), equals(0))
    expect_that(a$t(1, NA), equals(NA_real_))
    expect_that(a$t(NA, 0), equals(0))
    expect_that(a$t(NA, 0.5), equals(0.5))
    expect_that(a$t(NA, 1), equals(NA_real_))
    expect_that(a$t(0.5, 0.4), equals(0.2))
    expect_that(a$t(c(0.5, NA, NA, 0.8, NA)), equals(0.4))
    expect_that(a$t(c(0, 0.5, NA, NA, 0.8, NA)), equals(0))

    expect_true(is.null(a$pt()))
    expect_that(a$pt(NA, 0.8), equals(0.8))
    expect_that(a$pt(NaN, 0.8), equals(0.8))
    expect_that(a$pt(c(0.5, NA, 0.4, NA, NA, 1, 1),
                     c(NA, 0.4, 0.5, NA, 0, NA, 0.3)),
                equals(c(0.5, 0.4, 0.2, NA_real_, 0, NA_real_, 0.3)))

    expect_that(a$c(), equals(NA_real_))
    expect_that(a$c(NA), equals(NA_real_))
    expect_that(a$c(NaN), equals(NA_real_))
    expect_that(a$c(NA, 0), equals(NA_real_))
    expect_that(a$c(NA, 0.5), equals(NA_real_))
    expect_that(a$c(NA, 1), equals(1))
    expect_that(a$c(0, NA), equals(NA_real_))
    expect_that(a$c(0.5, NA), equals(NA_real_))
    expect_that(a$c(1, NA), equals(1))
    expect_that(a$c(0.5, NA, NaN, 0.8, NA), equals(NA_real_))
    expect_that(a$c(c(0.5, NA, NA, 0.8, NA)), equals(NA_real_))
    expect_that(a$c(c(1, 0.5, NA, NA, 0.8, NA)), equals(1))

    expect_true(is.null(a$pc()))
    expect_that(a$pc(NA, 0.8), equals(NA_real_))
    expect_that(a$pc(NaN, 0.8), equals(NA_real_))
    expect_that(a$pc(NA, 0), equals(NA_real_))
    expect_that(a$pc(NA, 1), equals(1))
    expect_that(a$pc(c(0.5, NA, 0.4, 1),
                     c(NA, 0.4, 0.5, NA)),
                equals(c(NA_real_, NA_real_, 0.7, 1)))

    expect_that(a$r(c(0.5,   0.8,  NA, 0.3, NA,        1, NA, NA,  0,   0, 1, 0.6),
                    c(0.8,   0.5, 0.4,  NA, NA,       NA,  1,  0, NA, 0.3, 0.3, 0)),
             equals(c(1.0, 0.625, 0.4,   1,  1, NA_real_,  1,  0,  1,   1, 0.3, 0)))

    expect_that(a$b(c(0.5,   0.8,   NA,  0.4, NA,  0,       1, NA, NA),
                    c(0.8,   0.5,   0.4,  NA, NA, NA,      NA,  0,  1)),
             equals(c(0.625, 0.625, 0.4, 0.4,  1,  0, NA_real_, 0, NA_real_)))

    expect_that(a$i(), equals(NA_real_))
    expect_that(a$i(NA), equals(NA_real_))
    expect_that(a$i(NaN), equals(NA_real_))
    expect_that(a$i(0.5, NA), equals(0.5))
    expect_that(a$i(0.5, NaN), equals(0.5))
    expect_that(a$i(0, NA), equals(0))
    expect_that(a$i(1, NA), equals(NA_real_))
    expect_that(a$i(NA, 0), equals(0))
    expect_that(a$i(NA, 0.5), equals(0.5))
    expect_that(a$i(NA, 1), equals(NA_real_))
    expect_that(a$i(0.5, 0.4), equals(0.4))
    expect_that(a$i(c(0.5, NA, NA, 0.8, NA)), equals(0.5))
    expect_that(a$i(c(0, 0.5, NA, NA, 0.8, NA)), equals(0))

    expect_true(is.null(a$pi()))
    expect_that(a$pi(NA, 0.8), equals(0.8))
    expect_that(a$pi(NaN, 0.8), equals(0.8))
    expect_that(a$pi(c(0.5, NA, 0.4, NA, NA),
                     c(NA, 0.4, 0.5, NA, 0)),
                equals(c(0.5, 0.4, 0.4, NA_real_, 0)))

    expect_that(a$s(), equals(NA_real_))
    expect_that(a$s(NA), equals(NA_real_))
    expect_that(a$s(NaN), equals(NA_real_))
    expect_that(a$s(0.5, NA, NaN, 0.8, NA), equals(NA_real_))
    expect_that(a$s(1, 0.5, NA, NaN, 0.8, NA), equals(1))
    expect_that(a$s(c(0.5, NA, NA, 0.8, NA)), equals(NA_real_))
    expect_that(a$s(c(0.5, 1, NA, NA, 0.8, NA)), equals(1))

    expect_true(is.null(a$ps()))
    expect_that(a$ps(NA, 0.8), equals(NA_real_))
    expect_that(a$ps(NaN, 0.8), equals(NA_real_))
    expect_that(a$ps(NA, 0), equals(NA_real_))
    expect_that(a$ps(NA, 1), equals(1))
    expect_that(a$ps(c(0.5, NA, 0.4, NA, NA),
                     c(NA, 0.4, 0.5, 0, 1)),
                equals(c(NA_real_, NA_real_, 0.5, NA_real_, 1)))

    expect_that(a$n(NA), equals(0))
    expect_that(a$n(1), equals(0))
    expect_that(a$n(0), equals(1))
    expect_that(a$n(0.8), equals(a$r(0.8, 0)))

    expect_that(a$ni(NA), equals(NA_real_))
    expect_that(a$ni(1), equals(0))
    expect_that(a$ni(0), equals(1))
    expect_that(a$ni(0.8), equals(0.2))
})


test_that('upperBound order', {
    a <- upperBound(algebra('lukasiewicz'))

    expect_equal(a$order(c(NA), decreasing=FALSE), c(1))
    expect_equal(a$order(c(0), decreasing=FALSE), c(1))
    expect_equal(a$order(c(0.3), decreasing=FALSE), c(1))
    expect_equal(a$order(c(0.3, 0), decreasing=FALSE), c(2, 1))
    expect_equal(a$order(c(0.3, 0, 0.8), decreasing=FALSE), c(2, 1, 3))
    expect_equal(a$order(c(0.3, NA, 0, 1), decreasing=FALSE), c(3, 1, 2, 4))
    expect_equal(a$order(c(1, 0.3, NA, 0, 0.8), decreasing=FALSE), c(4, 2, 5, 3, 1))
    expect_equal(a$order(c(NA, 0), decreasing=FALSE), c(2, 1))
    expect_equal(a$order(c(NA, 0.5), decreasing=FALSE), c(2, 1))
    expect_equal(a$order(c(NA, 1), decreasing=FALSE), c(1, 2))

    expect_equal(a$order(c(NA), decreasing=TRUE), c(1))
    expect_equal(a$order(c(0), decreasing=TRUE), c(1))
    expect_equal(a$order(c(0.3), decreasing=TRUE), c(1))
    expect_equal(a$order(c(0.3, 0), decreasing=TRUE), c(1, 2))
    expect_equal(a$order(c(0.3, 0, 0.8), decreasing=TRUE), c(3, 1, 2))
    expect_equal(a$order(c(0.3, NA, 0, 1), decreasing=TRUE), c(4, 2, 1, 3))
    expect_equal(a$order(c(1, 0.3, NA, 0, 0.8), decreasing=TRUE), c(1, 3, 5, 2, 4))
    expect_equal(a$order(c(NA, 0), decreasing=TRUE), c(1, 2))
    expect_equal(a$order(c(NA, 0.5), decreasing=TRUE), c(1, 2))
    expect_equal(a$order(c(NA, 1), decreasing=TRUE), c(2, 1))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' S3 class representing a set of fuzzy sets on the fixed universe
#'
#' The aim of the `fsets` S3 class is to store several fuzzy sets in the
#' form of numeric matrix where columns represent fuzzy sets, rows are
#' elements from the universe, and therefore a value of `i`-th row and `j`-th
#' column is a membership degree of `i`-th element of the universe to `j`-th fuzzy
#' set. The \code{fsets} object also stores the information about
#' the origin of the fuzzy sets as well as a relation of specificity among
#' them.
#'
#' The `fsets()` function is a constructor of an object of type `fsets`.
#' Each object stores two attributes: `vars` and `specs`. The functions [vars()]
#' and [specs()]). can be used to access these attributes.
#'
#' It is assumed that the fuzzy sets
#' are derived from some raw variables, e.g., numeric vectors or factors. `vars`
#' attribute is a character vector of names of raw variables with size equal
#' to the number of fuzzy sets in `fsets` object. It is then assumed that
#' two fuzzy sets with the same name in [vars()] attribute are derived from
#' the same variable.
#'
#' `specs` attribute gives a square numeric matrix of size equal to the number
#' of fuzzy sets in `fsets`. `specs[i][j] == 1` if and only if the `i`-th fuzzy
#' set is more specific than `j`-th fuzzy set. Specificity of fuzzy sets means
#' the nestedness of fuzzy set: for instance, `very small` is more specific than
#' `small`; similarly, `extremely big` is more specific than `very big`; on the
#' other hand, `very big` and `extremely small` are incomparable. A necessary
#' condition for specificity is subsethood.
#'
#' @param x A matrix of membership degrees. Columns of the matrix represent
#'   fuzzy sets, colnames are names of the fuzzy sets (and must not be NULL). Rows
#'   of the matrix represent elements of the universe.
#' @param vars A character vector that must correspond to the
#'   columns of `x`. It is a vector of names of original variables that the
#'   fuzzy sets were created from. In other words, the `vars` vector should
#'   contain the same value for each `x`'s column that corresponds to the same
#'   variable. Names of the `vars` vector are ignored.
#'   For instance, an [fcut()] function can transform a single numeric
#'   vector into several different fuzzy sets. To indicate that all of them in
#'   fact describe the same original variable, the same name is stored on
#'   appropriate positions of the `vars` vector.
#' @param specs A square numeric matrix containing values from `{0, 1}`.
#'   It is a specificity matrix, for which both rows and columns correspond to
#'   `x`'s columns and where `specs[i][j] == 1` if and only if `i`-th fuzzy
#'   set (i.e. `x[, i]`) is more specific (is a subset or equal to) than `j`-th fuzzy
#'   set (i.e. `x[, j]`).
#' @param f An instance of S3 class `fsets`.
#' @param value Attribute values to be set to the object.
#' @return [fsets()] returns an object of S3 class `fsets`. [vars()] returns
#'   a vector of original variable names of the `fsets` object. `specs`
#'   returns the specificity matrix.
#' @author Michal Burda
#' @seealso [fcut()], [lcut()], [is.specific()]
#' @keywords models robust
#' @examples
#'     # create a matrix of random membership degrees
#'     m <- matrix(runif(30), ncol=5)
#'     colnames(m) <- c('a1', 'a2', 'a12', 'b1', 'b2')
#'
#'     # create vars - first three (a1, a2, a3) and next two (b1, b2)
#'     # fuzzy sets originate from the same variable
#'     v <- c('a', 'a', 'a', 'b', 'b')
#'     names(v) <- colnames(m)
#'
#'     # create specificity matrix - a1 and a2 are more specific than a12,
#'     # the rest is incomparable
#'     s <- matrix(c(0, 0, 1, 0, 0,
#'                   0, 0, 1, 0, 0,
#'                   0, 0, 0, 0, 0,
#'                   0, 0, 0, 0, 0,
#'                   0, 0, 0, 0, 0), byrow=TRUE, ncol=5)
#'     colnames(s) <- colnames(m)
#'     rownames(s) <- colnames(m)
#'
#'     # create a valid instance of the fsets class
#'     o <- fsets(m, v, s)
#'
#' @export
fsets <- function(x,
                  vars=rep(deparse(substitute(x)), ncol(x)),
                  specs=matrix(0, nrow=ncol(x), ncol=ncol(x))) {
    .mustBeNumericMatrix(x)

    .mustBeCharacterVector(vars)
    .mustNotHaveNA(vars)
    .mustBe(ncol(x) == length(vars), "'vars' must be a vector of size equal to 'ncol(x)'")

    .mustBeNumericMatrix(specs)
    .mustNotHaveNA(specs)
    .mustBe(all(specs == 0 | specs == 1), "'specs' must be a binary matrix (only 0 and 1 are allowed)")
    .mustBe(identical(ncol(x), ncol(specs)) && identical(ncol(specs), nrow(specs)),
            "'specs' must be a square numeric matrix of size equal to 'ncol(x)'")

    return(structure(x,
                     class=c('fsets', 'matrix'),
                     vars=as.vector(vars),   # remove all attributes
                     specs=matrix(specs, nrow=nrow(specs))))   # remove all attributes
}


#' @rdname fsets
#' @export
vars <- function(f) {
    .mustBe(is.fsets(f), "'f' must be an instance of S3 'fsets' class")
    attr(f, 'vars')
}


#' @rdname fsets
#' @export
`vars<-` <- function(f, value) {
    .mustBe(is.fsets(f), "'f' must be an instance of S3 'fsets' class")
    .mustBeCharacterVector(value)
    .mustNotHaveNA(value)
    .mustBe(ncol(f) == length(value), "'value' must be a vector of size equal to fsets' ncol")
    attr(f, 'vars') <- as.vector(value)
    return(f)
}


#' @rdname fsets
#' @export
specs <- function(f) {
    .mustBe(is.fsets(f), "'f' must be an instance of S3 'fsets' class")
    attr(f, 'specs')
}


#' @rdname fsets
#' @export
`specs<-` <- function(f, value) {
    .mustBe(is.fsets(f), "'f' must be an instance of S3 'fsets' class")
    .mustBeNumericMatrix(value)
    .mustNotHaveNA(value)
    .mustBe(all(value == 0 | value == 1), "'value' must be a binary matrix (only 0 and 1 are allowed)")
    .mustBe(identical(ncol(f), ncol(value)) && identical(ncol(value), nrow(value)),
            "'value' must be a square numeric matrix of size equal to fsets' ncol")
    attr(f, 'specs') <- matrix(value, nrow=nrow(value))
    return(f)
}


#' Test whether `x` is a valid object of the S3 `fsets` class
#'
#' This function tests whether `x` inherits from S3 `fsets` class.
#'
#' @param x An object being tested.
#' @return TRUE if `x` is a valid `fsets` object and FALSE otherwise.
#' @author Michal Burda
#' @seealso [fsets()]
#' @keywords models robust
#' @export
is.fsets <- function(x) {
    (return(inherits(x, 'fsets') &&
            is.matrix(x)))
}


#' Convert an object of `fsets` class into a matrix or data frame
#' This function converts an instance of S3 class [fsets] into a
#' matrix or a data frame. The [vars()] and [specs()] attributes
#' of the original object are deleted.
#'
#' @param x An instance of class [fsets] to be converted
#' @param ... arguments further passed to `as.data.frame` after converting
#'   to matrix in `as.data.frame.fsets`. Unused in `as.matrix.fsets`.
#' @return  A numeric matrix or data frame of membership degrees.
#' @author Michal Burda
#' @seealso [fsets()], [fcut()], [lcut()]
#' @keywords models robust
#' @export
as.data.frame.fsets <- function(x, ...) {
    .mustBe(is.fsets(x), "'x' must be an instance of S3 class 'fsets'")
    return(as.data.frame(as.matrix(x), ...))
}


#' @rdname as.data.frame.fsets
#' @export
as.matrix.fsets <- function(x, ...) {
    .mustBe(is.fsets(x), "'x' must be an instance of S3 class 'fsets'")
    return(matrix(x, nrow=nrow(x), dimnames=dimnames(x)))
}


#' @export
as_tibble.fsets <- function(x, ...) {
    .mustBe(is.fsets(x), "'x' must be an instance of S3 class 'fsets'")
    return(as_tibble(as.matrix(x), ...))
}


#' Combine several 'fsets' objects into a single one
#'
#' Take a sequence of objects of class 'fsets' and combine them by columns.
#' This version of cbind takes care of the [vars()] and [specs()]
#' attributes of the arguments and merges them to the result. If some argument
#' does not inherit from the 'fsets' class, an error is thrown.
#'
#' The [vars()] attribute is merged by concatenating the [vars()] attributes
#' of each argument. Also the [specs()] attributes of the arguments are merged together.
#'
#' @param ...  A sequence of objects of class 'fsets' to be merged by columns.
#' @param deparse.level This argument has currently no function and is added
#' here only for compatibility with generic [cbind()] function.
#' @param warn Whether to issue warning when combining two fsets having the same vars
#' about the fact that specs may not be accurate
#' @return An object of class 'fsets' that is created by merging the arguments
#' by columns.  Also the arguments' attributes [vars()] and [specs()] are merged together.
#' @author Michal Burda
#' @seealso [vars()], [specs()], [fcut()], [lcut()]
#' @keywords models robust
#'
#' @examples
#'     d1 <- fcut(CO2[, 1:2])
#'     d2 <- fcut(CO2[, 3:4], breaks=list(conc=1:4*1000/4))
#'     r <- cbind(d1, d2)
#'
#'     print(colnames(d1))
#'     print(colnames(d2))
#'     print(colnames(r))
#'
#'     print(vars(d1))
#'     print(vars(d2))
#'     print(vars(r))
#'
#'     print(specs(d1))
#'     print(specs(d2))
#'     print(specs(r))
#'
#' @export
cbind.fsets <- function(..., deparse.level = 1, warn = TRUE) {
    dots <- list(...)
    m <- NULL
    v <- NULL
    s <- NULL
    for (i in seq_along(dots)) {
        arg <- dots[[i]]
        argName <- names(dots)[i]

        if (!is.null(arg)) {
            .mustBe(is.fsets(arg),
                    "Cannot cbind arguments that are not valid 'fsets' objects")
            if (is.null(m)) {
                v <- vars(arg)
                s <- specs(arg)
                class(arg) <- 'matrix'
                m <- arg
            } else {
                if (warn && length(intersect(v, vars(arg))) > 0) {
                    warning("Binding fsets with the same 'vars' - resulting 'specs' may be incorrect")
                    warn <- FALSE
                }
                v <- c(v, vars(arg))
                o1 <- matrix(0, nrow=nrow(s), ncol=ncol(specs(arg)))
                o2 <- matrix(0, nrow=nrow(specs(arg)), ncol=ncol(s))
                s <- rbind(cbind(s, o1), cbind(o2, specs(arg)))
                class(arg) <- 'matrix'
                m <- cbind(m, arg)
            }
        }
    }
    return(fsets(m, v, s))
}


#' @export
rbind.fsets <- function(..., deparse.level = 1) {
    dots <- list(...)
    m <- NULL
    v <- NULL
    s <- NULL
    for (i in seq_along(dots)) {
        arg <- dots[[i]]

        if (!is.null(arg)) {
            .mustBe(is.fsets(arg),
                    "Cannot rbind arguments that are not valid 'fsets' objects")
            if (is.null(m)) {
                v <- vars(arg)
                s <- specs(arg)
                class(arg) <- 'matrix'
                m <- arg
            } else {
                .mustBe(identical(colnames(m), colnames(arg)),
                        "Cannot rbind fsets having not equal column names")
                .mustBe(identical(v, vars(arg)),
                        "Cannot rbind fsets with unequal 'vars'");
                .mustBe(identical(s, specs(arg)),
                        "Cannot rbind fsets with unequal 'specs'");
                class(arg) <- 'matrix'
                m <- rbind(m, arg)
            }
        }
    }

    return(fsets(m, v, s))
}


#' @export
`[.fsets` <- function(x, i, j, drop=FALSE) {
    if (drop) warning('drop ignored')
    v <- vars(x)
    s <- specs(x)
    m <- as.matrix(x)

    names(v) <- colnames(x)
    colnames(s) <- colnames(x)
    rownames(s) <- colnames(x)

    return(fsets(m[i, j, drop=FALSE],
                 vars=v[j],
                 specs=s[j, j, drop=FALSE]))
}


#' Print an instance of the [fsets()] class
#'
#' Format an object of the [fsets()] class into human readable form
#' and print it to the output.
#'
#' @param x An instance of the [fsets()] class
#' @param ...  Unused.
#' @return Nothing
#' @author Michal Burda
#' @seealso [fsets()], [fcut()], [lcut()]
#' @keywords models robust
#' @examples
#'     d <- fcut(CO2[, 1:2])
#'     print(d)
#'
#' @export
print.fsets <- function(x, ...) {
    .mustBe(is.fsets(x), "'x' is not a valid 'fsets' object")
    v <- vars(x)
    s <- specs(x)
    print(as.matrix(x))
}


#' Compactly display the internal structure of an instance of the [fsets()] class
#'
#' This function is a method for the generic function [str()] for objects of the
#' [fsets()] class. It is a compact way to display the internal structure of an
#' instance of the [fsets()] class.
#'
#' @param object An instance of the [fsets()] class
#' @param ...  Unused.
#' @return Nothing
#' @author Michal Burda
#' @seealso [fsets()], [fcut()], [lcut()]
#' @keywords models robust
#' @export
str.fsets <- function(object, ...) {
    .mustBe(is.fsets(object), "'object' is not a valid 'fsets' object")
    cat("fsets object:")
    str(as.matrix(object), ...)
    cat(" - vars(*) :")
    str(vars(object), nest.level=1, ...)
    cat(" - specs(*) :")
    str(specs(object), nest.level=1, ...)
}


#' Plot membership degrees stored in the instance of the S3 class
#' [fsets()] as a line diagram.
#'
#' This function plots the membership degrees stored in the instance of the
#' [fsets()] class. Internally, the membership degrees are
#' transformed into a time-series object and viewed in a plot using the
#' [ts.plot()] function. This function is useful mainly to see the
#' shape of fuzzy sets on regularly sampled inputs.
#'
#' @param x An instance of class [fsets()]
#' @param ...  Other arguments that are passed to the underlying [ts.plot()]
#' function.
#' @return Result of the [ts.plot()] method.
#' @author Michal Burda
#' @seealso [fsets()], [fcut()], [lcut()], [ts.plot()]
#' @keywords models robust multivariate
#' @examples
#' d <- lcut(0:1000/1000, name='x')
#' plot(d)
#'
#' # Additional arguments are passed to the ts.plot method
#' # Here thick lines represent atomic linguistic expressions,
#' # i.e. ``small'', ``medium'', and ``big''.
#' plot(d,
#'      ylab='membership degree',
#'      xlab='values',
#'      gpars=list(lwd=c(rep(1, 3), 5, rep(1, 5), 5, rep(1, 7), 5, rep(1,4))))
#' @export
plot.fsets <- function(x, ...) {
    n <- nrow(x)
    args <- list(...)
    x <- as.list(as.data.frame(x))
    x <- plyr::llply(x, ts, start=0, frequency=n)
    x <- c(x, args)
    if (is.null(args[['xlab']])) {
        x[['xlab']] <- ''
    }
    if (is.null(args[['ylab']])) {
        x[['ylab']] <- ''
    }
    do.call('ts.plot', x)
}

test_that('fcut for factor', {
    x <- factor(c('a', 'b', 'a', 'c', 'c', 'b', 'c'))
    res <- fcut(x)

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 3)
    expect_equal(nrow(res), 7)
    expect_equal(colnames(res), c('x=a', 'x=b', 'x=c'))
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), rep('x', 3))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(rep(0, 3*3),
                                     nrow=3,
                                     ncol=3))
    expect_true(is.fsets(res))

    expect_equivalent(as.matrix(res)[1, 1], 1)
    expect_equivalent(as.matrix(res)[1, 2], 0)
    expect_equivalent(as.matrix(res)[1, 3], 0)

    expect_equivalent(as.matrix(res)[2, 1], 0)
    expect_equivalent(as.matrix(res)[2, 2], 1)
    expect_equivalent(as.matrix(res)[2, 3], 0)

    expect_equivalent(as.matrix(res)[3, 1], 1)
    expect_equivalent(as.matrix(res)[3, 2], 0)
    expect_equivalent(as.matrix(res)[3, 3], 0)

    expect_equivalent(as.matrix(res)[4, 1], 0)
    expect_equivalent(as.matrix(res)[4, 2], 0)
    expect_equivalent(as.matrix(res)[4, 3], 1)

    expect_equivalent(as.matrix(res)[5, 1], 0)
    expect_equivalent(as.matrix(res)[5, 2], 0)
    expect_equivalent(as.matrix(res)[5, 3], 1)

    expect_equivalent(as.matrix(res)[6, 1], 0)
    expect_equivalent(as.matrix(res)[6, 2], 1)
    expect_equivalent(as.matrix(res)[6, 3], 0)

    expect_equivalent(as.matrix(res)[7, 1], 0)
    expect_equivalent(as.matrix(res)[7, 2], 0)
    expect_equivalent(as.matrix(res)[7, 3], 1)
})


test_that('fcut for logical', {
    x <- c(TRUE, FALSE, FALSE, TRUE)
    res <- fcut(x)

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 2)
    expect_equal(nrow(res), 4)
    expect_equal(colnames(res), c('x', 'not.x'))
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), rep('x', 2))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(0, nrow=2, ncol=2))
    expect_true(is.fsets(res))

    expect_equivalent(as.matrix(res)[1, 1], 1)
    expect_equivalent(as.matrix(res)[1, 2], 0)

    expect_equivalent(as.matrix(res)[2, 1], 0)
    expect_equivalent(as.matrix(res)[2, 2], 1)

    expect_equivalent(as.matrix(res)[3, 1], 0)
    expect_equivalent(as.matrix(res)[3, 2], 1)

    expect_equivalent(as.matrix(res)[4, 1], 1)
    expect_equivalent(as.matrix(res)[4, 2], 0)
})


test_that('fcut of numeric by single triangle', {
    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 50, 100),
                type='triangle')

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 1)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), 'x=1')
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), 'x')
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(0,
                                    nrow=1,
                                    ncol=1))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[26, 1], 0.5)
    expect_equivalent(as.matrix(res)[51, 1], 1)
    expect_equivalent(as.matrix(res)[76, 1], 0.5)
    expect_equivalent(as.matrix(res)[101, 1], 0)
    expect_true(is.fsets(res))
})


test_that('fcut of numeric by multiple triangles', {
    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 25, 50, 75, 100),
                type='triangle')

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 3)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), c('x=1', 'x=2', 'x=3'))
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), rep('x', 3))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(rep(0, 3*3),
                                    nrow=3,
                                    ncol=3))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[26, 1], 1)
    expect_equivalent(as.matrix(res)[51, 1], 0)
    expect_equivalent(as.matrix(res)[76, 1], 0)
    expect_equivalent(as.matrix(res)[101, 1], 0)

    expect_equivalent(as.matrix(res)[1, 2], 0)
    expect_equivalent(as.matrix(res)[26, 2], 0)
    expect_equivalent(as.matrix(res)[51, 2], 1)
    expect_equivalent(as.matrix(res)[76, 2], 0)
    expect_equivalent(as.matrix(res)[101, 2], 0)

    expect_equivalent(as.matrix(res)[1, 3], 0)
    expect_equivalent(as.matrix(res)[26, 3], 0)
    expect_equivalent(as.matrix(res)[51, 3], 0)
    expect_equivalent(as.matrix(res)[76, 3], 1)
    expect_equivalent(as.matrix(res)[101, 3], 0)
    expect_true(is.fsets(res))
})


test_that('fcut of numeric with merge 1:3', {
    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 25, 50, 75, 100),
                type='triangle',
                merge=1:3)

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 6)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), c('x=1', 'x=2', 'x=3',
                                  'x=1|x=2', 'x=2|x=3',
                                  'x=1|x=2|x=3'))
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), rep('x', 6))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(c(0,0,0,1,0,1,
                                      0,0,0,1,1,1,
                                      0,0,0,0,1,1,
                                      0,0,0,0,0,1,
                                      0,0,0,0,0,1,
                                      0,0,0,0,0,0),
                                    byrow=TRUE,
                                    nrow=6))
    expect_true(is.fsets(res))
})


test_that('fcut of numeric with merge 2', {
    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 25, 50, 75, 100),
                type='triangle',
                merge=2)

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 2)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), c('x=1|x=2', 'x=2|x=3'))
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), rep('x', 2))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(rep(0, 2*2),
                                    nrow=2,
                                    ncol=2))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[26, 1], 1)
    expect_equivalent(as.matrix(res)[30, 1], 1)
    expect_equivalent(as.matrix(res)[51, 1], 1)
    expect_equivalent(as.matrix(res)[76, 1], 0)
    expect_equivalent(as.matrix(res)[101, 1], 0)

    expect_equivalent(as.matrix(res)[1, 2], 0)
    expect_equivalent(as.matrix(res)[26, 2], 0)
    expect_equivalent(as.matrix(res)[51, 2], 1)
    expect_equivalent(as.matrix(res)[60, 2], 1)
    expect_equivalent(as.matrix(res)[76, 2], 1)
    expect_equivalent(as.matrix(res)[101, 2], 0)
    expect_true(is.fsets(res))
})


test_that('fcut of numeric with merge 1,3', {
    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 25, 50, 75, 100),
                type='triangle',
                merge=c(1,3))

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 4)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), c('x=1', 'x=2', 'x=3',
                                  'x=1|x=2|x=3'))
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), rep('x', 4))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(c(0,0,0,1,
                                      0,0,0,1,
                                      0,0,0,1,
                                      0,0,0,0),
                                    byrow=TRUE,
                                    nrow=4))
    expect_true(is.fsets(res))
})


test_that('fcut of matrix', {
    x <- matrix(1:100, byrow=TRUE, ncol=4)
    colnames(x) <- letters[1:4]

    res <- fcut(x,
                breaks=c(1, 30, 60, 100),
                type='triangle')

    expect_true(is.fsets(res))
    expect_equal(ncol(res), 8)
    expect_equal(nrow(res), 25)
    expect_equal(colnames(res), c('a=1', 'a=2', 'b=1', 'b=2', 'c=1', 'c=2', 'd=1', 'd=2'))
    expect_equal(vars(res), c(rep('a', 2), rep('b', 2), rep('c', 2), rep('d', 2)))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(0,
                                    nrow=8,
                                    ncol=8))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[8, 3], 1)
    expect_equivalent(as.matrix(res)[15, 7], 0)
})



test_that('fcut of data frame', {
    x <- matrix(1:100, byrow=TRUE, ncol=4)
    colnames(x) <- letters[1:4]

    res <- fcut(as.data.frame(x),
                breaks=c(1, 30, 60, 100),
                type='triangle')

    expect_true(is.fsets(res))
    expect_equal(ncol(res), 8)
    expect_equal(nrow(res), 25)
    expect_equal(colnames(res), c('a=1', 'a=2', 'b=1', 'b=2', 'c=1', 'c=2', 'd=1', 'd=2'))
    expect_equal(vars(res), c(rep('a', 2), rep('b', 2), rep('c', 2), rep('d', 2)))
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(0,
                                    nrow=8,
                                    ncol=8))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[8, 3], 1)
    expect_equivalent(as.matrix(res)[15, 7], 0)
})


test_that('fcut for custom function factory', {
    func <- function(a, b, c) {
        f <- triangular(a, b, c)
        return(function(x) f(x)^2)
    }

    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 50, 100),
                type=func)

    expect_true(is.fsets(res))
    expect_equal(ncol(res), 1)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), 'x=1')
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), 'x')
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(0,
                                    nrow=1,
                                    ncol=1))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[26, 1], 0.25)
    expect_equivalent(as.matrix(res)[51, 1], 1)
    expect_equivalent(as.matrix(res)[76, 1], 0.25)
    expect_equivalent(as.matrix(res)[101, 1], 0)
    expect_true(is.fsets(res))
})


test_that('fcut for custom function', {
    func <- function(x, a, b, c) {
        f <- triangular(a, b, c)
        return(f(x)^2)
    }

    x <- 0:100
    res <- fcut(x,
                breaks=c(0, 50, 100),
                type=func)

    expect_true(is.fsets(res))
    expect_equal(ncol(res), 1)
    expect_equal(nrow(res), 101)
    expect_equal(colnames(res), 'x=1')
    expect_true(inherits(res, 'fsets'))
    expect_equivalent(vars(res), 'x')
    expect_equal(names(vars(res)), NULL)
    expect_equal(specs(res), matrix(0,
                                    nrow=1,
                                    ncol=1))

    expect_equivalent(as.matrix(res)[1, 1], 0)
    expect_equivalent(as.matrix(res)[26, 1], 0.25)
    expect_equivalent(as.matrix(res)[51, 1], 1)
    expect_equivalent(as.matrix(res)[76, 1], 0.25)
    expect_equivalent(as.matrix(res)[101, 1], 0)
    expect_true(is.fsets(res))
})

test_that('is.specific', {
    is.specifi2 <- function(x, y, f) {
        is.specific(rev(x), y, f)
    }

    is.specifi3 <- function(x, y, f) {
        is.specific(x, rev(y), f)
    }

    is.specifi4 <- function(x, y, f) {
        is.specific(rev(x), rev(y), f)
    }

    v <- c(rep('a', 3), rep('b', 3), rep('c', 3), rep('d', 3))
    s <- matrix(c(0,1,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,

                  0,0,0, 0,1,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,

                  0,0,0, 0,0,0, 0,1,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,

                  0,0,0, 0,0,0, 0,0,0, 0,1,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0),
                byrow=TRUE,
                ncol=12)
    m <- matrix(0, nrow=1, ncol=12)
    colnames(m) <- paste(rep(c('VeSm', 'Sm', 'Bi'), times=4),
                         rep(c('a', 'b', 'c', 'd'), each=3),
                         sep='.')
    f <- fsets(m, v, s)

    # equal rules
    expect_true(is.specific(c('VeSm.a', 'Bi.c'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi2(c('VeSm.a', 'Bi.c'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi3(c('VeSm.a', 'Bi.c'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi4(c('VeSm.a', 'Bi.c'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    # the same but single item that is more specific
    expect_true(is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c', 'Sm.d'),
                            f))

    expect_true(is.specifi2(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c', 'Sm.d'),
                            f))

    expect_true(is.specifi3(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c', 'Sm.d'),
                            f))

    expect_true(is.specifi4(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c', 'Sm.d'),
                            f))

    # "any" is less specific
    expect_true(is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi2(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi3(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi4(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('VeSm.a', 'Bi.c'),
                            f))

    # "any" + other
    expect_true(is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi2(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi3(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c'),
                            f))

    expect_true(is.specifi4(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            c('Sm.a', 'Bi.c'),
                            f))

    # everything is more specific than empty rule
    expect_true(is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                            character(),
                            f))

    # null rules are <=
    expect_true(is.specific(character(),
                            character(),
                            f))

    # the same but single item that is more specific
    expect_false(is.specific(c('Sm.a', 'Bi.c', 'Sm.d'),
                             c('VeSm.a', 'Bi.c', 'Sm.d'),
                             f))

    # "any" is less specific
    expect_false(is.specific(c('VeSm.a', 'Bi.c'),
                             c('VeSm.a', 'Bi.c', 'Sm.d'),
                             f))

    # "any" + other
    expect_false(is.specific(c('Sm.a', 'Bi.c'),
                             c('VeSm.a', 'Bi.c', 'Sm.d'),
                             f))

    # everything is more specific than empty rule
    expect_false(is.specific(character(),
                             c('VeSm.a', 'Bi.c', 'Sm.d'),
                             f))

    # different .vars are incomparable
    expect_false(is.specific(c('Sm.a'),
                             c('Bi.c'),
                             f))
    expect_false(is.specific(c('Bi.c'),
                             c('Sm.a'),
                             f))



    expect_false(is.specific(c('VeSm.a', 'Sm.c'),
                             c('Sm.a', 'Bi.c'),
                             f))
    expect_false(is.specific(c('Sm.b', 'Sm.d'),
                             c('Sm.a', 'Bi.c'),
                             f))
})

test_that('cbind of two fsets', {
    x <- 0:10 * 10
    y <- 0:10 * 10
    a <- fcut(x, breaks=c(0, 25, 50, 75, 100), merge=1:2)    #x1, x2, x3, x1|2, x2|3
    b <- fcut(y, breaks=c(0, 30, 70, 80, 100), merge=c(1,3)) #y1, y2, y3, y1|2|3

    expect_silent(res <- cbind(a, b))


    expectedNames <- c('x=1', 'x=2', 'x=3',
                       'x=1|x=2', 'x=2|x=3',
                       'y=1', 'y=2', 'y=3',
                       'y=1|y=2|y=3')

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 9)
    expect_equal(nrow(res), 11)
    expect_equal(colnames(res), expectedNames)
    expect_true(inherits(res, 'fsets'))

    expectedVars <- c(rep('x', 5), rep('y', 4))

    expect_equal(vars(res), expectedVars)
    expect_equal(specs(res), matrix(c(0,0,0,1,0, 0,0,0,0,
                                      0,0,0,1,1, 0,0,0,0,
                                      0,0,0,0,1, 0,0,0,0,
                                      0,0,0,0,0, 0,0,0,0,
                                      0,0,0,0,0, 0,0,0,0,

                                      0,0,0,0,0, 0,0,0,1,
                                      0,0,0,0,0, 0,0,0,1,
                                      0,0,0,0,0, 0,0,0,1,
                                      0,0,0,0,0, 0,0,0,0),
                                    byrow=TRUE,
                                    nrow=9))
    expect_equal(res[, 1:5], a[, 1:ncol(a)])  # 'a[, ...]' removes all attributes from 'a'
    expect_equal(res[, 6:9], b[, 1:ncol(b)])  # 'b[, ...]' removes all attributes from 'b'
    expect_true(is.fsets(res))
})



test_that('cbind of two fsets and NULL', {
    x <- 0:10 * 10
    y <- 0:10 * 10
    a <- fcut(x, breaks=c(0, 25, 50, 75, 100), merge=1:2)    #x1, x2, x3, x1|2, x2|3
    b <- fcut(y, breaks=c(0, 30, 70, 80, 100), merge=c(1,3)) #y1, y2, y3, y1|2|3

    expect_silent(res <- cbind(a, b, NULL))

    expectedNames <- c('x=1', 'x=2', 'x=3',
                       'x=1|x=2', 'x=2|x=3',
                       'y=1', 'y=2', 'y=3',
                       'y=1|y=2|y=3')

    expect_true(is.matrix(res))
    expect_equal(ncol(res), 9)
    expect_equal(nrow(res), 11)
    expect_equal(colnames(res), expectedNames)
    expect_true(inherits(res, 'fsets'))

    expectedVars <- c(rep('x', 5), rep('y', 4))

    expect_equal(vars(res), expectedVars)
    expect_equal(specs(res), matrix(c(0,0,0,1,0, 0,0,0,0,
                                      0,0,0,1,1, 0,0,0,0,
                                      0,0,0,0,1, 0,0,0,0,
                                      0,0,0,0,0, 0,0,0,0,
                                      0,0,0,0,0, 0,0,0,0,

                                      0,0,0,0,0, 0,0,0,1,
                                      0,0,0,0,0, 0,0,0,1,
                                      0,0,0,0,0, 0,0,0,1,
                                      0,0,0,0,0, 0,0,0,0),
                                    byrow=TRUE,
                                    nrow=9))
    expect_equal(res[, 1:5], a[, 1:ncol(a)])  # 'a[, ...]' removes all attributes from 'a'
    expect_equal(res[, 6:9], b[, 1:ncol(b)])  # 'b[, ...]' removes all attributes from 'b'
    expect_true(is.fsets(res))
})


test_that('cbind of the same vars has to produce warning', {
    set.seed(3335)

    m1 <- matrix(runif(12), ncol=3)
    m2 <- matrix(runif(12), ncol=3)
    s <- matrix(0, nrow=3, ncol=3)
    diag(s) <- 1
    f1 <- fsets(m1, vars=letters[1:3], specs=s)
    f2 <- fsets(m2, vars=letters[3:5], specs=s)

    expect_warning(cbind(f1, f2))
})

test_that('reduce', {
    x <- matrix(c(1,1,1,1,0,0,1,1,0,0,
                  0,0,1,1,1,0,0,0,1,1,
                  1,1,1,1,1,1,0,0,0,0),
                byrow=FALSE, ncol=3)
    colnames(x) <- c('a', 'b', 'c')
    
    v <- colnames(x)
    names(v) <- colnames(x)

    s <- matrix(0, nrow=3, ncol=3)
    colnames(s) <- colnames(x)
    rownames(s) <- colnames(x)

    d <- fsets(x, v, s)



    rules <- farules(rules=list(c('c', 'a'),
                                c('c', 'b')),
                     statistics=matrix(as.numeric(1:2), nrow=2, ncol=1))

    expect_equal(reduce(d, rules, 0.66),
                 farules(rules=list(c('c', 'a')),
                         statistics=matrix(as.numeric(1), nrow=1, ncol=1)))

    expect_equal(reduce(d, rules, 0.67),
                 farules(rules=list(c('c', 'a'),
                                    c('c', 'b')),
                         statistics=matrix(as.numeric(1:2), nrow=2, ncol=1)))
})

test_that('fsets without colnames', {
    set.seed(35423)
    m <- matrix(runif(12), ncol=3)
    v <- letters[1:3]
    s <- matrix(0, nrow=3, ncol=3)
    diag(s) <- 1

    f <- fsets(m, vars=v, specs=s)
    expect_true(is.fsets(f))
    expect_that(class(f), equals(c('fsets', 'matrix')))
    expect_that(vars(f), equals(letters[1:3]))
    expect_that(specs(f), equals(s))

    mat <- as.matrix(f)
    expect_true(is.matrix(mat))
    expect_that(mat, equals(m))
    expect_error(vars(mat))
    expect_error(specs(mat))

    df <- as.data.frame(f)
    expect_true(is.data.frame(df))
    expect_that(ncol(df), equals(3))
    expect_that(nrow(df), equals(4))
    expect_that(df[, 1], equals(m[, 1]))
    expect_that(df[, 2], equals(m[, 2]))
    expect_that(df[, 3], equals(m[, 3]))
    expect_error(vars(df))
    expect_error(specs(df))

    #---------
    v2 <- v
    names(v2) <- LETTERS[1:3]
    s2 <- s
    colnames(s2) <- LETTERS[1:3]
    rownames(s2) <- LETTERS[1:3]

    f2 <- f
    vars(f2) <- v
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- f
    specs(f2) <- s
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- f
    vars(f2) <- v2
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- f
    specs(f2) <- s2
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- fsets(m, vars=v2, specs=s)
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- fsets(m, vars=v, specs=s2)
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))
})

test_that('fsets with colnames', {
    set.seed(35423)
    m <- matrix(runif(12), ncol=3)
    colnames(m) <- LETTERS[1:3]
    v <- letters[1:3]
    s <- matrix(0, nrow=3, ncol=3)
    diag(s) <- 1

    f <- fsets(m, vars=v, specs=s)
    expect_true(is.fsets(f))
    expect_that(colnames(f), equals(LETTERS[1:3]))
    expect_that(class(f), equals(c('fsets', 'matrix')))
    expect_that(vars(f), equals(letters[1:3]))
    expect_that(specs(f), equals(s))

    mat <- as.matrix(f)
    expect_true(is.matrix(mat))
    expect_that(mat, equals(m))
    expect_that(colnames(mat), equals(LETTERS[1:3]))
    expect_error(vars(mat))
    expect_error(specs(mat))

    df <- as.data.frame(f)
    expect_true(is.data.frame(df))
    expect_that(ncol(df), equals(3))
    expect_that(nrow(df), equals(4))
    expect_that(colnames(df), equals(LETTERS[1:3]))
    expect_that(df[, 1], equals(m[, 1]))
    expect_that(df[, 2], equals(m[, 2]))
    expect_that(df[, 3], equals(m[, 3]))
    expect_error(vars(df))
    expect_error(specs(df))

    #---------
    v2 <- v
    names(v2) <- LETTERS[1:3]
    s2 <- s
    colnames(s2) <- LETTERS[1:3]
    rownames(s2) <- LETTERS[1:3]

    f2 <- f
    vars(f2) <- v
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_that(colnames(f2), equals(LETTERS[1:3]))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- f
    specs(f2) <- s
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_that(colnames(f2), equals(LETTERS[1:3]))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- f
    vars(f2) <- v2
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_that(colnames(f2), equals(LETTERS[1:3]))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- f
    specs(f2) <- s2
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_that(colnames(f2), equals(LETTERS[1:3]))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- fsets(m, vars=v2, specs=s)
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_that(colnames(f2), equals(LETTERS[1:3]))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))

    f2 <- fsets(m, vars=v, specs=s2)
    expect_true(is.fsets(f2))
    expect_that(class(f2), equals(c('fsets', 'matrix')))
    expect_that(vars(f2), equals(letters[1:3]))
    expect_that(specs(f2), equals(s))
    expect_that(colnames(f2), equals(LETTERS[1:3]))
    expect_null(names(vars(f2)))
    expect_null(colnames(specs(f2)))
    expect_null(rownames(specs(f2)))
})


test_that('non fsets', {
    expect_false(is.fsets(matrix(3, ncol=3, nrow=3)))
})


test_that('fsets [ ]', {
    orig <- matrix(runif(100), ncol=10)
    colnames(orig) <- letters[1:10]
    x <- fsets(orig,
               vars=as.character(1:10),
               specs=matrix(0, nrow=10, ncol=10))

    res <- x[1:5, ]
    expect_equal(as.matrix(res), orig[1:5, ])
    expect_equal(vars(res), vars(x))
    expect_equal(specs(res), specs(x))

    res <- x[-3, ]
    expect_equal(as.matrix(res), orig[-3, ])
    expect_equal(vars(res), vars(x))
    expect_equal(specs(res), specs(x))

    res <- x[3, ]
    expect_equal(as.matrix(res), orig[3, , drop=FALSE])
    expect_equal(vars(res), vars(x))
    expect_equal(specs(res), specs(x))

    res <- x[, 2:4]
    expect_equal(as.matrix(res), orig[, 2:4])
    expect_equal(vars(res), vars(x)[2:4])
    expect_equal(specs(res), specs(x)[2:4, 2:4])

    res <- x[1:5, 2:4]
    expect_equal(as.matrix(res), orig[1:5, 2:4])
    expect_equal(vars(res), vars(x)[2:4])
    expect_equal(specs(res), specs(x)[2:4, 2:4])

    res <- x[5, 4]
    expect_equal(as.matrix(res), orig[5, 4, drop=FALSE])
    expect_equal(vars(res), vars(x)[4])
    expect_equal(specs(res), specs(x)[4, 4, drop=FALSE])

    res <- x[, letters[2:4]]
    expect_equal(as.matrix(res), orig[, 2:4])
    expect_equal(vars(res), vars(x)[2:4])
    expect_equal(specs(res), specs(x)[2:4, 2:4])

    res <- x[1:5, letters[2:4]]
    expect_equal(as.matrix(res), orig[1:5, 2:4])
    expect_equal(vars(res), vars(x)[2:4])
    expect_equal(specs(res), specs(x)[2:4, 2:4])

    res <- x[5, letters[4]]
    expect_equal(as.matrix(res), orig[5, 4, drop=FALSE])
    expect_equal(vars(res), vars(x)[4])
    expect_equal(specs(res), specs(x)[4, 4, drop=FALSE])
})


test_that('rbind of two fsets', {
    set.seed(3335)

    m1 <- matrix(runif(12), ncol=3)
    m2 <- matrix(runif(24), ncol=3)
    s <- matrix(0, nrow=3, ncol=3)
    diag(s) <- 1
    f1 <- fsets(m1, vars=letters[1:3], specs=s)
    f2 <- fsets(m2, vars=letters[1:3], specs=s)

    r <- rbind(f1, f2)
    expect_equal(as.matrix(r), rbind(m1, m2))
    expect_equal(vars(r), letters[1:3])
    expect_equal(specs(r), s)
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#.u2 <- function(predicted, real, rw) {
    #sqrt(sum((real - predicted)^2) / length(real)) / sqrt(sum((real - rw)^2) / length(real))
#}




#' Evaluate the performance of the FRBE forecast
#'
#' Take a FRBE forecast and compare it with real values using arbitrary error
#' function.
#'
#' Take a FRBE forecast and compare it with real values by evaluating a given
#' error measure.  FRBE forecast should be made for a horizon of the same value
#' as length of the vector of real values.
#'
#' @param fit A FRBE model of class `frbe` as returned by the [frbe()] function.
#' @param real A numeric vector of real (known) values. The vector must
#' correspond to the values being forecasted, i.e., the length must be the same
#' as the horizon forecasted by [frbe()].
#' @param error Error measure to be computed. It can be either Symmetric Mean
#' Absolute Percentage Error (`smape`), Mean Absolute Scaled Error (`mase`), or
#' Root Mean Squared Error (`rmse`). See also [smape()], [mase()], and [rmse()]
#' for more details.
#' @return Function returns a data.frame with single row and columns
#' corresponding to the error of the individual forecasting methods that the
#' FRBE is computed from. Additionally to this, a column "avg" is added with
#' error of simple average of the individual forecasting methods and a column
#' "frbe" with error of the FRBE forecasts.
#' @author Michal Burda
#' @seealso [frbe()], [smape()], [mase()], [rmse()]
#' @references Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
#' Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
#' SYST. 2015.
#' @keywords models robust
#' @examples
#'
#'   # prepare data (from the forecast package)
#'   library(forecast)
#'   horizon <- 10
#'   train <- wineind[-1 * (length(wineind)-horizon+1):length(wineind)]
#'   test <- wineind[(length(wineind)-horizon+1):length(wineind)]
#'   f <- frbe(ts(train, frequency=frequency(wineind)), h=horizon)
#'   evalfrbe(f, test)
#'
#' @export
evalfrbe <- function(fit,
                     real,
                     error=c('smape', 'mase', 'rmse')) {
    .mustBe(is.frbe(fit), "'fit' must be an instance of class 'frbe'")
    .mustBeNumericVector(real)

    error <- match.arg(error)
    if (error == 'smape') {
        errorFunc <- smape
    } else if (error == 'mase') {
        errorFunc <- mase
    } else if (error == 'rmse') {
        errorFunc <- rmse
    } else {
        stop("Unknown error function name")
    }

    d <- fit$forecasts
    d$avg <- rowSums(d) / ncol(d)
    d$frbe <- fit$mean

    if (length(real) > nrow(d)) {
        length(real) <- nrow(d)
    } else {
        d <- d[seq_along(real), ]
    }

    r <- colwise(function(col) { errorFunc(col, real) })(d)
    #u2 <- colwise(function(col) { .u2(col, real, fit$forecasts$randomWalk) })(d)
    #names(u2) <- paste('u2', names(u2))
    #return(cbind(r, u2))
    return(r)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Compute Symmetric Mean Absolute Percentage Error (SMAPE)
#'
#' SMAPE is computed as \eqn{mean(abs(forecast - validation) / ((abs(forecast) + abs(validation)) / 2))}.
#'
#' @param forecast A numeric vector of forecasted values
#' @param validation A numeric vector of actual (real) values
#' @return A Symmetric Mean Absolute Percentage Error (SMAPE)
#' @author Michal Burda
#' @seealso [rmse()], [mase()], [frbe()]
#' @export
smape <- function(forecast, validation) {
  .mustBeNumericVector(forecast)
  .mustBeNumericVector(validation)
  .mustNotHaveNA(forecast)
  .mustNotHaveNA(validation)
  .mustBe(length(forecast) == length(validation), "Length of 'forecast' and 'validation' must be equal")
  .mustBe(length(forecast) > 1, "Length of both 'forecast' and 'validation' must be greater than zero")

  return(mean(abs(forecast - validation) / ((abs(forecast) + abs(validation)) / 2)))
}

test_that('smape', {
    f <- 1:10 + 0.2
    r <- 1:10

    expect_equal(smape(f, r), mean(0.2 / ((f+r) / 2)))
    expect_error(smape(1:10, 1:9))
    expect_error(smape(1, 2))
    expect_error(smape(c(1:5, NA), 1:6))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Return vector of values from given interval
#'
#' Returns an ordered vector of values from given interval, of given size,
#' generated by equal steps.
#'
#' Returns a vector of values from \code{from} to \code{to} (inclusive), with
#' equal difference between two consecutive values, with total length \code{n}.
#' Function is useful e.g., together with the \code{\link{pbld}} or
#' \code{\link{defuzz}} functions (for the \code{values} argument; see also
#' \code{\link{lcut}} or \code{\link{fcut}}) or \code{\link{defuzz}}).
#'
#' @param from The lower bound of the interval.
#' @param to The upper bound of the interval.
#' @param n The length of the vector to be produced.
#' @return A vector of numbers in the given interval and size.
#' @author Michal Burda
#' @seealso \code{\link{pbld}}, \code{\link{defuzz}}, \code{\link{fcut}},
#' \code{\link{lcut}}
#' @keywords models robust
#' @examples
#' \dontrun{
#'     slices(1, 5, 10) # 1, 1.5, 2, 2.5, 3, 3.5 4, 4.5, 5
#' }
#' # is the same as
#' seq(1, 5, length.out=10)
#'
#'
#' @export slices
slices <- function(from, to, n) {
    .Deprecated('base::seq')
    if (!is.numeric(from) && length(from) != 1) {
        stop("'from' must be a numeric value")
    }
    if (!is.numeric(to) && length(to) != 1) {
        stop("'to' must be a numeric value")
    }
    if (!is.numeric(n) && length(n) != 1) {
        stop("'n' must be a numeric value")
    }
    if (n < 2) {
        stop("'n' must be at least 2")
    }
    res <- 0:(n-1) / (n-1) * (to - from) + from
    return(res)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Creator of functions representing linguistic expressions
#'
#' A linguistic expression represents vague human terms such as "very small", "extremely big" etc. Such notions are
#' always reasoned within a given context. `lingexpr` returns a function that models a selected linguistic expression.
#' Accordingly to the given `context`, `atomic` expression (such as "small", "big") and a linguistic `hedge` (such as
#' `very`, `extremely`), the returned function transforms numeric values into degrees (from `[0, 1]` interval),
#' at which the values correspond to the expression.
#'
#' Based on the context type, the following atomic expressions are allowed:
#' * [ctx3()] (trichotomy): small, medium, big;
#' * [ctx5()] (pentachotomy): small, lower medium, medium, upper medium, big;
#' * [ctx3bilat()] (bilateral trichotomy): negative big, negative medium, negative small,
#'   zero, small, medium, big;
#' * [ctx5bilat()] (bilateral pentachotomy): negative big, negative medium, negative
#'   small, zero, small, medium, big.
#'
#' The values of the `atomic` parameter have the following meaning (in ascending order):
#' * `neg.bi`: big negative (far from zero)
#' * `neg.um`: upper medium negative (between medium negative and big negative)
#' * `neg.me`: medium negative
#' * `neg.lm`: lower medium negative (between medium negative and small negative)
#' * `neg.sm`: small negative (close to zero)
#' * `ze`: zero
#' * `sm`: small
#' * `lm`: lower medium
#' * `me`: medium
#' * `um`: upper medium
#' * `bi`: big
#'
#' `hedge` parameter has the following meaning:
#' * `ex`: extremely,
#' * `si`: significantly,
#' * `ve`: very,
#' * `ty`: typically,
#' * `-`: empty hedge,
#' * `ml`: more or less,
#' * `ro`: roughly,
#' * `qr`: quite roughly,
#' * `vr`: very roughly.
#'
#' Accordingly to the theory of linguistic expressions by Novak, not every hedge is applicable to each atomic
#' expression. The combinations of allowed pairs can be found in [allowed.lingexpr]. Trying to create forbidden
#' combination results in error.
#'
#' @param context A context of linguistic expressions (see [ctx3()], [ctx5()], [ctx3bilat()] or [ctx5bilat()])
#' @param atomic An atomic expression whose horizon we would like to obtain
#' @param hedge The type of the required linguistic hedge ('-' for no hedging)
#' @param negated Negate the expression? (For instance, "not very small".) Negation
#'   is done using the [invol.neg()] function.
#' @param hedgeParams Parameters that determine the shape of the hedges
#' @return Returns a function with a single argument, which has to be a numeric vector.
#' @author Michal Burda
#' @seealso [horizon()], [hedge()], [fcut()], [lcut()], [ctx()]
#' @keywords models robust
#' @examples
#'     small <- lingexpr(ctx3(0, 0.5, 1), atomic='sm', hedge='-')
#'     small(0)   # 1
#'     small(0.8) # 0
#'     plot(small)
#'
#'     verySmall <- lingexpr(ctx3(0, 0.5, 1), atomic='sm', hedge='ve')
#'     plot(verySmall)
#' @export
lingexpr <- function(context,
                     atomic=c('sm', 'me', 'bi', 'lm', 'um', 'ze',
                              'neg.sm', 'neg.me', 'neg.bi', 'neg.lm', 'neg.um'),
                     hedge=c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr'),
                     negated=FALSE,
                     hedgeParams=defaultHedgeParams) {
    .mustBeLogicalScalar(negated)
    atomic <- match.arg(atomic)
    h <- match.arg(hedge)
    if (!allowed.lingexpr[h, atomic]) {
        stop(paste0("Hedge '", hedge, "' is not applicable on atomic expression '", atomic, "'"))
    }
    rm(hedge) # in order to access global hedge() function

    hor <- horizon(context, atomic)
    hed <- hedge(h, hedgeParams=hedgeParams)
    neg <- if (negated) invol.neg else identity

    return(function(x) {
        neg(hed(hor(x)))
    })
}

#' @rdname lingexpr
#' @export
allowed.lingexpr <- array(
    # ex si ve ty -  ml ro qr vr
    c(T, T, T, F, T, T, T, T, T,  # neg.bi
      F, F, F, F, T, F, F, F, F,  # neg.um
      F, F, F, T, T, T, T, T, T,  # neg.me
      F, F, F, F, T, F, F, F, F,  # neg.lm
      T, T, T, F, T, T, T, T, T,  # neg.sm
      F, F, F, F, T, T, T, F, F,  # ze
      T, T, T, F, T, T, T, T, T,  # sm
      F, F, F, F, T, F, F, F, F,  # lm
      F, F, F, T, T, T, T, T, T,  # me
      F, F, F, F, T, F, F, F, F,  # um
      T, T, T, F, T, T, T, T, T), # bi
    dim=c(9, 11),
    dimnames=list(c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr'),
                  c('neg.bi', 'neg.um', 'neg.me', 'neg.lm', 'neg.sm', 'ze', 'sm', 'lm', 'me', 'um', 'bi')))

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.ft.weights <- function(xmemb, antecedents) {
    res <- apply(xmemb, 1, function(row) {
        fire <- sapply(seq_len(ncol(antecedents)),
                       function(i) row[antecedents[, i], drop = FALSE])
        apply(fire, 1, prod)
    })

    dimnames(res) <- NULL

    res
}


#' Fuzzy transform
#'
#' Compute a fuzzy tranform of the given input matrix `x`.
#'
#' @param x the numeric matrix of input values
#' @param xmemb the partitioning of input values, i.e., a `fsets` object with membership degrees
#'     (see [fcut()])
#' @param y the numeric vector of target values
#' @param order the order of the fuzzy transform (0, 1, 2, ...)
#' @return the instance of the S3 class `ft`
#' @author Michal Burda
#' @seealso [ftinv()], [is.ft()]
#' @keywords models robust
#' @references Perfilieva I. Fuzzy transforms: Theory and applications. FUZZY SET SYST,
#'     volume 157, issue 8, p. 993-1023. 2006.
#' @examples
#'
#' # create the fuzzy transform object
#' y <- (1:30)^2
#' x <- as.matrix(data.frame(a = 1:30, b = 30:1))
#' xmemb <- fcut(x,
#'               breaks = list(a = equidist(x[, 'a'], 3),
#'                             b = equidist(x[, 'b'], 3)))
#' fit <- ft(x, xmemb, y, order = 1)
#'
#' # obtain function values
#' x2 <- as.matrix(data.frame(a = 10:20, b = 20:10))
#' xmemb2 <- fcut(x2,
#'                breaks = list(a = equidist(x[, 'a'], 3),
#'                              b = equidist(x[, 'b'], 3)))
#' y2 <- ftinv(fit, x2, xmemb2)
#' print(y2)
#'
#' # compare original values with those obtained by the fuzzy transform
#' y[10:20] - y2
#'
#' @export
ft <- function(x,
               xmemb,
               y,
               order = 1) {
    .mustBeNumericMatrix(x)
    .mustBe(ncol(x) >= 1, "'x' must have at least 1 column")
    .mustBe(nrow(x) >= 1, "'x' must not be empty")

    .mustBe(is.fsets(xmemb), "'xmemb' must be an instance of class 'fsets'")
    .mustBe(nrow(x) == nrow(xmemb), "the number of rows in 'x' must equal to the number of rows in 'xmemb")
    .mustBe(all(colnames(x) %in% vars(xmemb)), "colnames(x) must be equal to vars(xmemb)")
    .mustBe(all(vars(xmemb) %in% colnames(x)), "colnames(x) must be equal to vars(xmemb)")

    .mustBeNumericVector(y)
    .mustBeNumericScalar(order)
    .mustBe(order >= 0, "'order' must be greater or equal to 0")
    .mustBe(nrow(x) == length(y), "the number of rows in 'x' must be equal to the length of vector 'y'")

    inputs <- colnames(x)
    partitions <- lapply(inputs, function(var) colnames(xmemb)[vars(xmemb) == var])

    antecedents <- do.call(expand.grid, c(partitions, list(KEEP.OUT.ATTRS = FALSE)))
    antecedents <- as.matrix(antecedents)
    dimnames(antecedents) <- NULL

    weights <- .ft.weights(xmemb, antecedents)

    d <- as.data.frame(x)
    colnames(d) <- paste0('c', seq_along(d))
    lmvars <- colnames(d)
    d$y <- y

    form <- 'y ~ 1'
    for (o in seq_len(order)) {
        form <- paste0(form,  ' + ', paste0('I(', lmvars, '^', o, ')', collapse = ' + '))
    }

    .fit_lm <- function(ante) {
        w <- weights[ante, ]
        form <- as.formula(form)
        fit <- lm(formula = form, data = d, weights = w)
        as.vector(coef(fit))
    }

    consequents <- sapply(seq_len(nrow(weights)), .fit_lm)
    consequents[is.na(consequents)] <- 0    # this occurs if an input is linear combination of other inputs
    if (!is.matrix(consequents)) {
        # order 0 ft
        consequents <- matrix(consequents, nrow=1)
    }


    structure(list(inputs = inputs,
                   partitions = partitions,
                   order = order,
                   antecedents = antecedents,
                   consequents = consequents),
              class = 'ft')
}


#' Test whether `x` is a valid object of the S3 `ft` class
#'
#' Test whether `x` has a valid format for objects of the S3 `ft` class that represents
#' the Fuzzy Transform.
#'
#' This function tests whether `x` is an instance of the `ft` class and whether it is a list
#' with the following elements: `inputs` character vector, `partitions` list, `order` number,
#' `antecedents` matrix and `consequents` matrix.
#'
#' @param x An object to be tested
#' @return `TRUE` if `x` is a valid `ft` object and `FALSE` otherwise.
#' @author Michal Burda
#' @seealso [ft()], [ftinv()]
#' @keywords models robust
#' @export
is.ft <- function(x) {
    inherits(x, 'ft') &&
        is.list(x)
}

test_that('ft', {
    y <- (1:30)^2
    x <- data.frame(a = 1:30, b = 30:1)
    x <- as.matrix(x)

    xmemb <- fcut(x,
                  breaks = list(a = equidist(x[, 'a'], 3),
                                b = equidist(x[, 'b'], 3)))

    res <- ft(x, xmemb, y, order = 2)
    expect_true(is.ft(res))
    expect_true(is.list(res))
    expect_equal(res$inputs, c('a', 'b'))
    expect_equal(res$partitions,
                 list(c('a=1', 'a=2', 'a=3'), c('b=1', 'b=2', 'b=3')))
    expect_equal(res$order, 2)
    expect_equal(nrow(res$antecedents), 9)
    expect_equal(ncol(res$antecedents), 2)
    expect_equal(nrow(res$consequents), 5)
    expect_equal(ncol(res$consequents), 9)

    yy <- ftinv(res, x, xmemb)
    expect_equal(yy, y)
})


test_that('ft extrapolation 1', {
    y <- (1:30)
    x <- data.frame(a = 1:30)
    x <- as.matrix(x)
    xbreaks <- equidist(x[, 'a'], 3, left='infinity', right='infinity')
    xmemb <- fcut(x, breaks = list(a = xbreaks))

    res <- ft(x, xmemb, y, order = 1)

    x <- data.frame(a = 20:50)
    x <- as.matrix(x)
    xmemb <- fcut(x, breaks = list(a = xbreaks))

    yy <- ftinv(res, x, xmemb)
    expect_equal(yy, 20:50)
})


test_that('ft extrapolation 2', {
    y <- c(1:10, 101:110, 201:210)
    x <- data.frame(a = 1:30)
    x <- as.matrix(x)
    xbreaks <- equidist(x[, 'a'], 9, left='infinity', right='infinity')
    xmemb <- fcut(x, breaks = list(a = xbreaks))

    res <- ft(x, xmemb, y, order = 1)

    x <- data.frame(a = 21:50)
    x <- as.matrix(x)
    xmemb <- fcut(x, breaks = list(a = xbreaks))

    yy <- ftinv(res, x, xmemb)
    expect_equal(yy[11:30], 211:230)
})


test_that('ft extrapolation to NaN', {
    y <- (1:30)
    x <- data.frame(a = 1:30)
    x <- as.matrix(x)
    xbreaks <- equidist(x[, 'a'], 3, left='same', right='same')
    xmemb <- fcut(x, breaks = list(a = xbreaks))

    res <- ft(x, xmemb, y, order = 1)

    x <- data.frame(a = 20:50)
    x <- as.matrix(x)
    xmemb <- fcut(x, breaks = list(a = xbreaks))

    yy <- ftinv(res, x, xmemb)
    expect_equal(yy, c(20:30, rep(NaN, 20)))


})


test_that('ft order 0', {
    y <- rep(10, 30)
    x <- data.frame(a = 1:30, b = 30:1)
    x <- as.matrix(x)

    xmemb <- fcut(x,
                  breaks = list(a = equidist(x[, 'a'], 3),
                                b = equidist(x[, 'b'], 3)))

    res <- ft(x, xmemb, y, order = 0)
    expect_true(is.ft(res))
    expect_true(is.list(res))
    expect_equal(res$inputs, c('a', 'b'))
    expect_equal(res$partitions,
                 list(c('a=1', 'a=2', 'a=3'), c('b=1', 'b=2', 'b=3')))
    expect_equal(res$order, 0)
    expect_equal(nrow(res$antecedents), 9)
    expect_equal(ncol(res$antecedents), 2)
    expect_equal(nrow(res$consequents), 1)
    expect_equal(ncol(res$consequents), 9)

    yy <- ftinv(res, x, xmemb)
    expect_equal(yy, y)
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.rollapply <- function(x, window, f, ...) {
    if (length(x) < window) {
        return(NULL)
    }

    res <- sapply(seq_len(length(x) - window + 1), function(i) {
        ii <- seq(from = i, to = i + window - 1)
        f(x[ii], ...)
    })

    if (is.matrix(res)) {
        res <- t(res)
    }

    res
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Evaluate rules and obtain truth-degrees
#'
#' Given truth degrees of predicates, compute the truth value of given list of rules.
#'
#' The aim of this function is to compute the truth value of each rule in a
#' `rules` list by assigning truth values to rule's predicates given by data `x`.
#'
#' `x` is a numeric vector or numeric matrix of truth values of predicates. If
#' `x` is vector then `names(x)` must correspond to the predicate names
#' in `rules`. If `x` is a matrix then each column should represent a predicate
#' and thus `colnames(x)` must correspond to predicate names in `rules`.
#' Values of `x` are interpreted as truth values, i.e., they must be from the
#' interval \eqn{[0, 1]}. If matrix is given, the resulting truth values are
#' computed row-wisely.
#'
#' `rules` may be a list of character vectors or an instance of the S3 class
#' [farules()]. The character vectors in the `rules` list represent formulae
#' in conjunctive form. If `onlyAnte=FALSE`, `fire()` treats the rule as
#' a conjunction of all predicates, i.e., a conjunction of all predicates is
#' computed. If `onlyAnte=TRUE`, the first element of each rule is removed
#' prior evaluation, i.e., a conjunction of all predicates except the first
#' are computed: this is useful if `rules` is a [farules()] object, since
#' [farules()] objects save a rule's consequent as the first element (see also
#' [antecedents()] and [consequents()] functions).
#'
#' The type of  conjunction to be computed can be specified with the `tnorm` parameter.
#'
#' @param x Truth degrees of predicates. `x` could be either a numeric
#' matrix or a numeric vector. If vector is given then each named element represents
#' a truth value of a predicate. If matrix is given then each row of the matrix
#' is evaluated sequentially as a vector. The values must be in the interval
#' \eqn{[0, 1]}.
#' @param rules Either an object of S3 class [farules()] or a list of character
#' vectors where each vector is a rule in a conjunctive form. Elements of these
#' character vectors (i.e., predicate names) must correspond to
#' the `x`'s names (of elements resp. columns if `x` is a vector resp. matrix).
#' @param tnorm A character string representing a triangular norm to be used
#' (either `"goedel"`, `"goguen"`, or `"lukasiewicz"`) or an arbitrary function
#' that performs element-wise computation on arbitrary number of vector parameters
#' similarly as e.g. [pgoedel.tnorm()], [pgoguen.tnorm()] or [plukas.tnorm()].
#' @param onlyAnte `TRUE` is useful if rules store both the antecedent and consequent
#' and if only the antecedent-part of a rule should be included into the evaluated
#' conjunction. Antecedent-part of a rule are all predicates in the vector starting from
#' the 2nd position. `TRUE` value in this parameter causes the first element of each
#' rule to be ignored.
#'
#' If `FALSE`, all predicates in a rule will be included in the conjunction.
#' @param parallel Deprecated parameter. Computation is done sequentially.
#' @return If `x` is a matrix then the result of this function is a list
#' of numeric vectors with truth values of each rule, i.e., each element of the
#' resulting list corresponds to a rule and each value of the vector in the resulting
#' list corresponds to a row of the original data matrix `x`.
#'
#' `x` as a vector is treated as a single-row matrix.
#' @author Michal Burda
#' @seealso [aggregateConsequents()], [defuzz()], [perceive()], [pbld()], [infer()], [fcut()], [lcut()], [farules()]
#' @keywords models robust multivariate
#' @examples
#'
#' # fire whole rules on a vector
#' x <- 1:10 / 10
#' names(x) <- letters[1:10]
#' rules <- list(c('a', 'c', 'e'),
#'               c('b'),
#'               c('d', 'a'),
#'               c('c', 'a', 'b'))
#' fire(x, rules, tnorm='goguen', onlyAnte=FALSE)
#'
#' # fire antecedents of the rules on a matrix
#' x <- matrix(1:20 / 20, nrow=2)
#' colnames(x) <- letters[1:10]
#' rules <- list(c('a', 'c', 'e'),
#'               c('b'),
#'               c('d', 'a'),
#'               c('c', 'a', 'b'))
#' fire(x, rules, tnorm='goedel', onlyAnte=TRUE)
#'
#' # the former command should be equal to
#' fire(x, antecedents(rules), tnorm='goedel', onlyAnte=FALSE)
#'
#' @export
fire <- function(x,
                 rules,
                 tnorm=c("goedel", "goguen", "lukasiewicz"),
                 onlyAnte=TRUE,
                 parallel=FALSE) {
    if (parallel) {
        warning('"parallel=TRUE" is deprecated. Computing sequentially.')
    }
    if (is.vector(x)) {
        x <- matrix(x, nrow=1, dimnames=list(NULL, names(x)))
    }
    .mustBe(is.matrix(x), "'x' must be a vector or matrix")

    if (is.farules(rules)) {
        rules <- rules$rules
    } else if (is.vector(rules) && is.character(rules)) {
        rules <- list(rules)
    }

    .mustBe(is.list(rules), "'rules' must be a list of rules, a valid farules object, or a character vector")

    if (!is.function(tnorm)) {
        tnorm <- match.arg(tnorm)
        alg <- algebra(tnorm)
        tnorm <- alg$pt
    }

    if (onlyAnte) {
        if (any(lengths(rules) <= 0)) {
            .stop("Cannot extract antecedent from an empty rule (onlyAnte is set to TRUE)")
        }
        rules <- antecedents(rules)
    }

    unknown <- unique(unlist(rules))
    unknown <- setdiff(unknown, colnames(x))
    if (length(unknown) > 0) {
        .stop(paste0("Unknown predicates in rules: ", paste(unknown, collapse=', ')))
    }

    xx <- as.matrix(x)
    xx <- split(xx, c(col(xx)))  # convert matrix to a list of columns
    names(xx) <- colnames(x)
    res <- lapply(rules, function(rule) {
        if (length(rule) <= 0) {
            return(rep(1, nrow(x)))
        } else {
            return(do.call(tnorm, xx[rule]))
        }
    })

    return(res)
}

test_that('fire rules on data vector', {
    x <- 1:10 / 10
    names(x) <- letters[1:10]
    rules <- list(c('c', 'e'),
                  character(),
                  c('a'),
                  c('a', 'b'))

    res <- fire(x, rules, pgoguen.tnorm, onlyAnte=FALSE)
    expect_equal(res, list(0.15, 1, 0.1, 0.02))

    res <- fire(x, rules, 'goguen', onlyAnte=FALSE)
    expect_equal(res, list(0.15, 1, 0.1, 0.02))
})


test_that('fire single rule on data matrix', {
    x <- matrix(1:20 / 20, nrow=2)
    colnames(x) <- letters[1:10]
    rules <- c('c', 'e')

    res <- fire(x, rules, pgoguen.tnorm, onlyAnte=FALSE)
    expect_equal(res, list(c(0.1125, 0.15)))

    res <- fire(x, rules, 'goguen', onlyAnte=FALSE)
    expect_equal(res, list(c(0.1125, 0.15)))
})


test_that('fire rules on data matrix', {
    x <- matrix(1:20 / 20, nrow=2)
    colnames(x) <- letters[1:10]
    rules <- list(c('c', 'e'),
                  character(),
                  c('a'),
                  c('a', 'b'))

    res <- fire(x, rules, 'goguen', onlyAnte=FALSE)
    expect_equal(res, list(c(0.1125, 0.15),
                           c(1, 1),
                           c(0.05, 0.1),
                           c(0.0075, 0.02)))
})


test_that('fire on farules', {
    x <- matrix(1:20 / 20, nrow=2)
    colnames(x) <- letters[1:10]
    rules <- list(c('a', 'c', 'e'),
                  c('b'),
                  c('d', 'a'),
                  c('c', 'a', 'b'))
    frules <- list(rules=rules, statistics=matrix(0, ncol=1, nrow=1))
    class(frules) <- c('farules', 'list')
    expect_true(is.farules(frules))

    res <- fire(x, frules, 'goguen', onlyAnte=TRUE)
    expect_equal(res, list(c(0.1125, 0.15),
                           c(1, 1),
                           c(0.05, 0.1),
                           c(0.0075, 0.02)))
})


test_that('fire rules on NA data matrix', {
    x <- matrix(1:20 / 20, nrow=2)
    colnames(x) <- letters[1:10]
    x[2, 1] <- NA

    rules <- list(c('c', 'e'),
                  character(),
                  c('a'),
                  c('a', 'b'))

    res <- fire(x, rules, 'goguen', onlyAnte=FALSE)
    expect_equal(res, list(c(0.1125, 0.15),
                           c(1, 1),
                           c(0.05, NA),
                           c(0.0075, NA)))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Extract antecedent-part (left-hand side) of rules in a list
#'
#' Given a list of rules or an instance of the S3 [farules()] class,
#' the function returns a list of their antecedents (i.e.
#' left-hand side of rules).
#'
#' This function assumes `rules` to be a valid [farules()] object or
#' a list of character vectors where
#' the first element of each vector is a consequent part and the
#' rest is an antecedent part of rules. Function returns a list of
#' antecedents.
#'
#' @param rules Either a list of character vectors or an object of class [farules()].
#' @return A list of character vectors.
#' @author Michal Burda
#' @seealso [consequents()], [farules()], [searchrules()]
#' @keywords models robust
#' @export
#' @examples
#'     rules <- list(c('a', 'b', 'c'), c('d'), c('a', 'e'))
#'     antecedents(rules)
antecedents <- function(rules) {
    .mustBe(is.farules(rules) ||
                (is.list(rules) && all(sapply(rules, function(r) { is.character(r) && length(r) > 0 }))),
            "'rules' must be either a list of non-empty character vectors or a valid farules object")
    r <- rules
    if (is.farules(r)) {
        r <- r$rules
    }
    return(lapply(r, function(rule) rule[-1]))
}

test_that('antecedents', {
    expect_equal(antecedents(list(c('a', 'b', 'c'),
                                  c('d'),
                                  c('a', 'e'))),
                 list(c('b', 'c'),
                      character(0),
                      c('e')))
    expect_error(antecedents(list(c(1, 2, 3))))
    expect_error(antecedents(list(character())))
})

test_that('antecedents (single rule)', {
    expect_equal(antecedents(list(c('a', 'b', 'c'))),
                 list(c('b', 'c')))
})

test_that('antecedents (empty list)', {
    expect_equal(antecedents(list()),
                 list())
})


test_that('antecedents (farules)', {
    f <- farules(list(letters[1:3], letters[2:5]),
                 matrix(0, nrow=2, ncol=3))
    expect_equal(antecedents(f),
                 list(letters[2:3], letters[3:5]))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.horizonAllowedTable <- list('ctx3'=list('sm'=1:3, 'me'=2:4, 'bi'=3:5),
                             'ctx5'=list('sm'=c(1,2,4), 'lm'=2:4, 'me'=c(2,4,6), 'um'=4:6, 'bi'=c(4,6,7)),
                             'ctx3bilat'=list('neg.bi'=1:3, 'neg.me'=2:4, 'neg.sm'=c(3,4,4),
                                           'ze'=3:5, 'sm'=c(4,4,5), 'me'=4:6, 'bi'=5:7),
                             'ctx5bilat'=list('neg.bi'=c(1,2,4), 'neg.um'=2:4, 'neg.me'=c(2,4,6),
                                           'neg.lm'=4:6, 'neg.sm'=c(4,6,6), 'ze'=c(4,6,8),
                                           'sm'=c(6,6,8), 'lm'=6:8, 'me'=c(6,8,10), 'um'=8:10,
                                           'bi'=c(8,10,11)))

#' Create a function that computes linguistic horizons
#'
#' Based on given `context` and `atomic` expression, this function returns a function that computes a linguistic
#' horizon, i.e., a triangular function representing basic limits of what humans treat as "small", "medium", "big" etc.
#' within given `context`. Linguistic horizon stands as a base for creation of linguistic expressions. A linguistic
#' expression is created by applying a [hedge()] on horizon. (Atomic linguistic expression is created from horizon by
#' applying an empty (`-`) hedge).
#'
#' The values of the `atomic` parameter have the following meaning (in ascending order):
#' * `neg.bi`: big negative (far from zero)
#' * `neg.um`: upper medium negative (between medium negative and big negative)
#' * `neg.me`: medium negative
#' * `neg.lm`: lower medium negative (between medium negative and small negative)
#' * `neg.sm`: small negative (close to zero)
#' * `ze`: zero
#' * `sm`: small
#' * `lm`: lower medium
#' * `me`: medium
#' * `um`: upper medium
#' * `bi`: big
#'
#' Based on the context type, the following atomic expressions are allowed:
#' * [ctx3()] (trichotomy): small, medium, big;
#' * [ctx5()] (pentachotomy): small, lower medium, medium, upper medium, big;
#' * [ctx3bilat()] (bilateral trichotomy): negative big, negative medium, negative small,
#'   zero, small, medium, big;
#' * [ctx5bilat()] (bilateral pentachotomy): negative big, negative medium, negative
#'   small, zero, small, medium, big.
#'
#' This function is quite low-level. Perhaps a more convenient way to create linguistic expressions
#' is to use the [lingexpr()] function.
#'
#' @param context A context of linguistic expressions (see [ctx3()], [ctx5()], [ctx3bilat()] or [ctx5bilat()])
#' @param atomic An atomic expression whose horizon we would like to obtain
#' @return A function of single argument that must be a numeric vector
#' @author Michal Burda
#' @seealso [ctx3()], [ctx5()], [ctx3bilat()], [ctx5bilat()], [hedge()], [fcut()], [lcut()]
#' @keywords models robust
#' @examples
#'     plot(horizon(ctx3(), 'sm'), from=-1, to=2)
#'     plot(horizon(ctx3(), 'me'), from=-1, to=2)
#'     plot(horizon(ctx3(), 'bi'), from=-1, to=2)
#'
#'     a <- horizon(ctx3(), 'sm')
#'     plot(a)
#'     h <- hedge('ve')
#'     plot(h)
#'     verySmall <- function(x) h(a(x))
#'     plot(verySmall)
#' @export
horizon <- function(context,
                    atomic=c('sm', 'me', 'bi', 'lm', 'um', 'ze',
                             'neg.sm', 'neg.me', 'neg.bi', 'neg.lm', 'neg.um')) {
    type <- NULL
    for (clazz in names(.horizonAllowedTable)) {
        if (inherits(context, clazz)) {
            type <- clazz
        }
    }
    .mustBe(!is.null(type),
            paste0("'context' must be an object of one of the following classes: '",
                   paste(names(.horizonAllowedTable), collapse="', '"), "'"))

    allowed <- .horizonAllowedTable[[type]]
    atomic <- match.arg(atomic)
    .mustBeOneOf(atomic, names(allowed))

    space <- c(-Inf, context, Inf)
    arg <- space[allowed[[atomic]]]
    return(triangular(arg[1], arg[2], arg[3]))
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Inverse of the fuzzy transform
#'
#'
#' Compute an inverse of fuzzy transform `fit` for values `x` with corresponding
#' membership degrees `xmemb`.
#'
#' @param fit The fuzzy transform object as the instance of the `ft` S3 class
#' @param x The numeric matrix of input values, for which the inverse fuzzy transform
#'     has to be computed
#' @param xmemb the partitioning of input values, i.e., a `fsets` object with membership degrees
#'     (see [fcut()]). Such partitioning must correspond to the `xmemb` partitioning used to
#'     create `fit` using the [ft()] function.
#' @return The inverse of the fuzzy transform `fit`, i.e., the approximated values
#'     of the original function that was the subject of the fuzzy transform
#' @author Michal Burda
#' @seealso [ft()], [is.ft()]
#' @keywords models robust
#' @references Perfilieva I. Fuzzy transforms: Theory and applications. FUZZY SET SYST,
#'     volume 157, issue 8, p. 993-1023. 2006.
#' @examples
#'
#' # create the fuzzy transform object
#' y <- (1:30)^2
#' x <- as.matrix(data.frame(a = 1:30, b = 30:1))
#' xmemb <- fcut(x,
#'               breaks = list(a = equidist(x[, 'a'], 3),
#'                             b = equidist(x[, 'b'], 3)))
#' fit <- ft(x, xmemb, y, order = 1)
#'
#' # obtain function values
#' x2 <- as.matrix(data.frame(a = 10:20, b = 20:10))
#' xmemb2 <- fcut(x2,
#'                breaks = list(a = equidist(x[, 'a'], 3),
#'                              b = equidist(x[, 'b'], 3)))
#' y2 <- ftinv(fit, x2, xmemb2)
#' print(y2)
#'
#' # compare original values with those obtained by the fuzzy transform
#' y - y2
#'
#' @export
ftinv <- function(fit, x, xmemb) {
    .mustBe(is.ft(fit), "'fit' must be an instance of the S3 'ft' class")

    .mustBe(is.matrix(x))
    .mustBe(ncol(x) >= 1, "'x' must have at least 1 column")
    .mustBe(nrow(x) >= 1, "'x' must not be empty")

    .mustBe(is.fsets(xmemb), "'xmemb' must be an instance of class 'fsets'")
    .mustBe(nrow(x) == nrow(xmemb), "the number of rows in 'x' must equal to the number of rows in 'xmemb")
    .mustBe(all(colnames(x) %in% vars(xmemb)), "colnames(x) must be equal to vars(xmemb)")
    .mustBe(all(vars(xmemb) %in% colnames(x)), "colnames(x) must be equal to vars(xmemb)")


    weights <- .ft.weights(xmemb, fit$antecedents)
    inputs <- rep(1, nrow(x))
    for (o in seq_len(fit$order)) {
        inputs <- cbind(inputs, x^o)
    }

    res <- inputs %*% fit$consequents

    sapply(seq_len(nrow(res)), function(i) weighted.mean(res[i, ], weights[, i]))
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.computeLength <- function(d) {
  res <- length(d)
  return(res)
}


.computeTrendStrength <- function(d) {
  # computing as a p-value of the zero slope in linear regression model
  # TODO: nebylo by lepsi vracet nejak logaritmus pval?)
  # TODO: prokladame primkou (linearni regrese) - co kdybychom zkusili prokladat taky necim jinym
  #       a definovat tak vice ruznych trendStrength-u?
  t <- 1:length(d)
  pval <- summary(lm(d ~ t))$coefficients[2, 4]
  return(1 - pval)
}


.computeSeasonStrength <- function(d) {
  # TODO: nebylo by lepsi vracet nejak logaritmus pval?)
  if (frequency(d) == 1) {
    # perioda = 12 mesicu, rocni casovka automaticky neni sezonni
    return(0)
  } else {
    time <- 1:length(d)
    num <- frequency(d) - 1
    vec <- rep(c(rep(0, num), 1), length(d))
    flags <- matrix(vec[1:(length(d) * num)], byrow=TRUE, ncol=num, nrow=length(d))
    form <- paste("d ~ time +", paste("flags[,", 1:num, "]", sep="", collapse=" + "))
    pval <- min(summary(lm(as.formula(form)))$coefficients[c(-1, -2), 4])
    return(1 - pval)
  }
}


.computeSkewness <- function(d) {
  # TODO: mozna by stalo za to  to transformovat nejak nelinearne
  s <- abs(skewness(d, type=1))
  return(s)
}


.computeKurtosis <- function(d) {
  # TODO: asi urcite by stalo za to  to transformovat nejak nelinearne
  k <- 3 + kurtosis(d, type=1)
  return(k)
}


.computeVarcoef <- function(d) {
  # TODO: uprava i pro zaporna data (tj. data s nulovym prumerem)
  return(sd(d) / mean(d))
}


.computeStationarity <- function(d) {
  # TODO: nebylo by lepsi vracet nejak logaritmus pval?)

  # on very short time series, this sometimes causes error
  # (see https://stackoverflow.com/questions/17282788/r-error-with-adf-test-in-time-series-lapply)
  # to recover that, we return 0.5
  pval <- try(suppressWarnings(adf.test(d)$p.value))
  if (inherits(pval, 'try-error')) {
      return(0.5)
  }
  return(1 - pval)
}


.computeFrequency <- function(d) {
  return(1 / frequency(d))
}







#' Fuzzy Rule-Based Ensemble (FRBE) of time-series forecasts
#'
#' This function computes the fuzzy rule-based ensemble of time-series
#' forecasts. Several forecasting methods are used to predict future values of
#' a given time-series and a weighted sum is computed from them with weights
#' being determined from a fuzzy rule base.
#'
#' This function computes the fuzzy rule-based ensemble of time-series
#' forecasts. The evaluation comprises the following steps:
#' 1. Several features are extracted from the given time-series `d`:
#'    * length of the time-series;
#'    * strength of trend;
#'    * strength of seasonality;
#'    * skewness;
#'    * kurtosis;
#'    * variation coefficient;
#'    * stationarity;
#'    * frequency.
#'    These features are used later to infer weights of the forecasting methods.
#' 1. Several forecasting methods are applied on the given time-series `d` to
#'    obtain forecasts. Actually, the following methods are used:
#'    * ARIMA - by calling [forecast::auto.arima()];
#'    * Exponential Smoothing - by calling [forecast::ets()];
#'    * Random Walk with Drift - by calling [forecast::rwf()];
#'    * Theta - by calling [forecast::thetaf()].
#' 1. Computed features are input to the fuzzy rule-based inference mechanism
#'    which yields weights of the forecasting methods. The fuzzy rule base is
#'    hardwired in this package and it was obtained by performing data mining with
#'    the use of the [farules()] function.
#' 1. A weighted sum of forecasts is computed and returned as a result.
#'
#' @param d A source time-series in the ts time-series format. Note that the
#' frequency of the time-series must be set properly.
#' @param h A forecasting horizon, i.e., the number of values to forecast.
#' @return Result is a list of class `frbe` with the following elements:
#' * `features` - a data frame with computed features of the given time-series;
#' * `forecasts` - a data frame with forecasts to be ensembled;
#' * `weights` - weights of the forecasting methods as inferred from the features
#'   and the hard-wired fuzzy rule base;
#' * `mean` - the resulting ensembled forecast (computed as a weighted sum
#' of forecasts).
#'
#' @author Michal Burda
#' @seealso [evalfrbe()]
#' @references Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
#' Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
#' SYST. 2015.
#' @keywords models robust
#' @examples
#'   # prepare data (from the forecast package)
#'   library(forecast)
#'   horizon <- 10
#'   train <- wineind[-1 * (length(wineind)-horizon+1):length(wineind)]
#'   test <- wineind[(length(wineind)-horizon+1):length(wineind)]
#'
#'   # perform FRBE
#'   f <- frbe(ts(train, frequency=frequency(wineind)), h=horizon)
#'
#'   # evaluate FRBE forecasts
#'   evalfrbe(f, test)
#'
#'   # display forecast results
#'   f$mean
#'
#' @export
frbe <- function(d, h=10) {
    .mustBeTs(d);
    .mustBeNumericScalar(h)
    .mustBe(h >= 1, "'h' must be positive")

    result <- list()
    result$data <- d

    result$forecasts <- data.frame(
                arima=as.numeric(forecast(auto.arima(d, stepwise=FALSE), h=h)$mean),
                expSmooth=as.numeric(forecast(ets(d), h=h)$mean),
                randomWalk=as.numeric(rwf(d, drift=FALSE, h=h)$mean),
                theta=as.numeric(thetaf(d, h=h)$mean))


    result$features <- data.frame(length=.computeLength(d),
                                  trendStrength=.computeTrendStrength(d),
                                  seasonStrength=.computeSeasonStrength(d),
                                  skewness=.computeSkewness(d),
                                  kurtosis=.computeKurtosis(d),
                                  varcoef=.computeVarcoef(d),
                                  stationarity=.computeStationarity(d),
                                  frequency=.computeFrequency(d))

    #f <- lcut3(result$features, context=.frbemodel$featuresContext)
    ctx <- lapply(.frbemodel$featuresContext, function(x) {
        do.call('ctx3', as.list(x))
    })
    atomic <- c("sm", "me", "bi")
    hedges <- c("ex", "si", "ve", "-", "ml", "ro", "qr", "vr")
    f <- lcut(result$features, context=ctx, atomic=atomic, hedges=hedges)

    result$weights <- sapply(names(.frbemodel$model),
                             function(n) {
                                 ctx <- do.call('ctx3', as.list(.frbemodel$weightContext[[n]]))
                                 vals <- seq(from=ctx[1], to=ctx[3], length.out=1000)
                                 parts <- lcut(vals, name='weight', context=ctx, atomic=atomic, hedges=hedges)
                                 pbld(f, .frbemodel$model[[n]], parts, vals, type='global')
                             })
    result$weights <- result$weights[colnames(result$forecasts)]

    if (sum(result$weights) == 0) {
        result$weights <- rep(1, ncol(result$forecasts))
        names(result$weights) <- colnames(result$forecasts)
    }

    result$mean <- apply(result$forecasts, 1,
                         function(row) {
                             sum(row * result$weights) / sum(result$weights)
                         })

    class(result) <- c('frbe', class(result))
    return(result)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Return equidistant breaks
#'
#' If both `left` and `right` equal to `"none"`, the function returns a vector of `n` values from `x`
#' that divide the range of values in `x` into `n - 1` equidistant intervals.
#'
#' If the `left` (resp. `right`) argument equals to `"infinity"`, `-Inf` (resp. `Inf`) is prepended
#' (resp. appended) to the result. If it equals to `"same"`, the first (resp. last) value is doubled.
#' Such functionality is beneficial if using the result of this function with e.g., the [fcut()] function:
#' `Inf` values at the beginning (resp. at the end) of the vector of breaks means that the fuzzy set
#' partition starts with a fuzzy set with kernel going to negative (resp. positive) infinity; the doubled
#' value at the beginning (resp. end) results in half-cut (trimmed) fuzzy set.
#'
#' @param x A numeric vector of input values
#' @param n The number of breaks of `x` to find (`n` must be at least 2)
#' @param left The left border of the returned vector of breaks: `"infinity"`, `"same"` or `"none"`
#'     (see the description below)
#' @param right The right border of the returned vector of breaks: `"infinity"`, `"same"` or `"none"`
#'     (see the description below)
#' @return A vector of equidistant breaks, which can be used e.g., in [fcut()]
#' @seealso [equifreq()], [fcut()]
#' @author Michal Burda
#' @keywords models robust
#' @export
equidist <- function(x,
                     n,
                     left = c('infinity', 'same', 'none'),
                     right = c('infinity', 'same', 'none')) {
    left <- match.arg(left)
    right <- match.arg(right)

    .mustBeNumericVector(x)
    .mustBeNumericScalar(n)
    .mustBe(n >= 2, "'n' must be greater or equal to 2")
    .mustBeCharacterScalar(left)
    .mustBeCharacterScalar(right)


    res <- seq(from = min(x, na.rm = TRUE),
               to = max(x, na.rm = TRUE),
               length.out = n)

    if (left == 'infinity') {
        res <- c(-Inf, res)
    } else if (left == 'same') {
        res <- c(res[1], res)
    }

    if (right == 'infinity') {
        res <- c(res, Inf)
    } else if (right == 'same') {
        res <- c(res, res[length(res)])
    }

    res
}

test_that('equidist', {
    expect_equal(equidist(10:20, 2, left = 'none', right = 'none'),
                 c(10, 20))

    expect_equal(equidist(10:20, 3, left = 'none', right = 'none'),
                 c(10, 15, 20))
    expect_equal(equidist(10:20, 3, left = 'infinity', right = 'none'),
                 c(-Inf, 10, 15, 20))
    expect_equal(equidist(10:20, 3, left = 'same', right = 'none'),
                 c(10, 10, 15, 20))
    expect_equal(equidist(10:20, 3, left = 'none', right = 'infinity'),
                 c(10, 15, 20, Inf))
    expect_equal(equidist(10:20, 3, left = 'none', right = 'same'),
                 c(10, 15, 20, 20))
    expect_equal(equidist(10:20, 3, left = 'infinity', right = 'infinity'),
                 c(-Inf, 10, 15, 20, Inf))
    expect_equal(equidist(10:20, 3, left = 'same', right = 'same'),
                 c(10, 10, 15, 20, 20))

    expect_equal(equidist(10:20, 11, left = 'none', right = 'none'),
                 10:20)
})


test_that('equidist 2', {
    expect_equal(equidist(c(10, 18, 20), 2, left = 'none', right = 'none'),
                 c(10, 20))

    expect_equal(equidist(c(10, 18, 20), 3, left = 'none', right = 'none'),
                 c(10, 15, 20))
    expect_equal(equidist(c(10, 18, 20), 3, left = 'infinity', right = 'none'),
                 c(-Inf, 10, 15, 20))
    expect_equal(equidist(c(10, 18, 20), 3, left = 'same', right = 'none'),
                 c(10, 10, 15, 20))
    expect_equal(equidist(c(10, 18, 20), 3, left = 'none', right = 'infinity'),
                 c(10, 15, 20, Inf))
    expect_equal(equidist(c(10, 18, 20), 3, left = 'none', right = 'same'),
                 c(10, 15, 20, 20))
    expect_equal(equidist(c(10, 18, 20), 3, left = 'infinity', right = 'infinity'),
                 c(-Inf, 10, 15, 20, Inf))
    expect_equal(equidist(c(10, 18, 20), 3, left = 'same', right = 'same'),
                 c(10, 10, 15, 20, 20))

    expect_equal(equidist(c(10, 18, 20), 11, left = 'none', right = 'none'),
                 10:20)
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Aggregation of fired consequents into a resulting fuzzy set
#'
#' Take a character vector of consequent names, a numeric vector representing
#' the degree of consequents' firing and a matrix that models fuzzy sets
#' corresponding to the consequent names, and perform an aggregation of the
#' consequents into a resulting fuzzy set.
#'
#' This function is typically used within an inference mechanism, after a set of
#' firing rules is determined and membership degrees of their antecedents are
#' computed, to combine the consequents of the firing rules into a resulting
#' fuzzy set. The result of this function is then typically defuzzified
#' (see [defuzz()]) to  obtain a crisp result of the inference.
#'
#' Function assumes a set of rules with antecedents firing at degrees given in
#' `degrees` and with consequents in `conseq`. The meaning of the consequents is
#' modeled with fuzzy sets whose membership degree values are captured in the
#' `partition` matrix.
#'
#' With default values of `firing` and `aggreg` parameters, the function
#' computes a fuzzy set that results from a conjunction (Goedel minimum t-norm)
#' of all provided implicative (Lukasiewicz residuum) rules.
#'
#' In detail, the function first computes the fuzzy set of each fired consequent
#' by calling `part\[i\] <- firing(degrees\[i\], partition\[, conseq\[i\]\])` for each
#' `i`-th consequent and the results are aggregated using the `aggreg`
#' parameter: `aggreg(part\[1\], part\[2\], ...)`. In order to aggregate consequents
#' in a Mamdani-Assilian's fashion, set `firing` to [pgoedel.tnorm()] and `aggreg`
#' to [pgoedel.tconorm()].
#'
#' @param conseq A character vector of consequents. Each value in the vector
#'   must correspond to a name of some column of the `partition` matrix.
#'   The length of this vector must be the same as of the `degrees`
#'   argument.
#' @param degrees A numeric vector of membership degrees at which the
#'   corresponding consequents (see the `conseq` argument) are fired.
#' @param partition A matrix of membership degrees that describes the meaning of
#'   the consequents in vector `conseq`: each column of the matrix
#'   corresponds to a fuzzy set that models a single consequent (of a name given
#'   by column names of the matrix), each row corresponds to a single crisp
#'   value (which is not important for this function), hence each cell
#'   corresponds to a membership degree in which the crisp value is a member of
#'   a fuzzy set modeling the consequent.  Each consequent in `conseq`
#'   must correspond to some column of this matrix. Such matrix may be created
#'   e.g., by using [fcut()] or [lcut()] functions.
#' @param firing A two-argument function used to compute the resulting truth value of the consequent.
#'   Function is evaluated for each consequent in `conseq`, with corresponding `degrees` value
#'   as the first argument and corresponding truth-value of the consequent (from `partition`)
#'   as the second argument. In default, the Lukasiewicz residuum ([lukas.residuum()]) is
#'   evaluated that way.
#' @param aggreg An aggregation function to be used to combine fuzzy sets resulting from firing
#'   the consequents with the `firing` function. The function should accept multiple
#'   numeric vectors of membership degrees as its arguments.
#'   In default, the Goedel t-norm ([pgoedel.tnorm()]) is evaluated.
#' @return A vector of membership degrees of fuzzy set elements that correspond
#'   to rows in the `partition` matrix. If empty vector of consequents is
#'   provided, vector of 1's is returned. The length of the resulting vector
#'   equals to the number of rows of the `partition` matrix.
#'
#' @author Michal Burda
#' @seealso [fire()], [perceive()], [pbld()], [infer()], [defuzz()], [fcut()], [lcut()]
#' @keywords models robust
#' @examples
#'     # create a partition matrix
#'     partition <- matrix(c(0:10/10, 10:0/10, rep(0, 5),
#'                           rep(0, 5), 0:10/10, 10:0/10,
#'                           0:12/12, 1, 12:0/12),
#'                         byrow=FALSE,
#'                         ncol=3)
#'     colnames(partition) <- c('a', 'b', 'c')
#'
#'     # the result of aggregation is equal to:
#'     # pmin(1, partition[, 1] + (1 - 0.5), partition[, 2] + (1 - 0.8))
#'     aggregateConsequents(c('a', 'b'), c(0.5, 0.8), partition)
#'
#' @export aggregateConsequents
aggregateConsequents <- function(conseq,
                                 degrees,
                                 partition,
                                 firing=lukas.residuum,
                                 aggreg=pgoedel.tnorm) {
    .mustBeCharacterVector(conseq)
    .mustBeNumericVector(degrees)
    .mustBe(length(conseq) == length(degrees),
            "The length of 'conseq' and 'degrees' must be the same")
    .mustBeNumericMatrix(partition)
    .mustBe(nrow(partition) > 0 && ncol(partition) > 0,
            "'partition' must not be empty matrix")
    .mustBe(!is.null(colnames(partition)) && is.character(colnames(partition)),
            "'partition' must have colnames")

    uniqConseq <- unique(conseq)
    .mustBe(length(intersect(uniqConseq, colnames(partition))) == length(uniqConseq),
            "Not all consequents are present in 'partition'")

    if (length(conseq) == 0) {
      return(rep(1, nrow(partition)))
    }

    res <- lapply(seq_along(conseq), function(i) {
      firing(degrees[i], partition[, conseq[i]])
    })
    return(do.call(aggreg, res))
}

test_that('aggregateConsequents', {
    partition <- matrix(c(0:10/10, 10:0/10, rep(0, 5),
                          rep(0, 5), 0:10/10, 10:0/10,
                          0:12/12, 1, 12:0/12),
                        byrow=FALSE,
                        ncol=3)
    colnames(partition) <- c('a', 'b', 'c')

    expect_equal(aggregateConsequents(c('a', 'b'),
                           c(1, 1),
                           partition),
                 pmin(partition[, 1], partition[, 2]))

    expect_equal(aggregateConsequents(c('a', 'b'),
                           c(0, 0),
                           partition),
                 rep(1, 27))

    expect_equal(aggregateConsequents(c('a', 'b'),
                           c(0.5, 0.8),
                           partition),
                 pmin(1, partition[, 1] + 0.5, partition[, 2] + 0.2))
    
    expect_equal(aggregateConsequents(c('a', 'b'),
                           c(0.5, 0.8),
                           partition,
                           firing=pgoedel.tnorm,
                           aggreg=pgoedel.tconorm),
                 pmax(pmin(partition[, 1], 0.5), pmin(partition[, 2], 0.8)))
    
    expect_equal(aggregateConsequents(as.character(c()), as.numeric(c()), partition),
                 rep(1, nrow(partition)))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Composition of Fuzzy Relations
#'
#' Composition of Fuzzy Relations
#'
#' Function composes a fuzzy relation `x` (i.e., a numeric matrix of size
#' \eqn{(u,v)}) with a fuzzy relation `y` (i.e., a numeric matrix of size
#' \eqn{(v,w)}) and possibly with the deprecated use of an exclusion fuzzy relation
#' `e` (i.e., a numeric matrix of size \eqn{(v,w)}).
#'
#' The style of composition is determined by the algebra `alg`, the
#' composition type `type`, and possibly also by a deprecated `quantifier`.
#'
#' This function performs four main composition types, the basic composition (
#' also known as direct product), the Bandler-Kohout subproduct (also subdirect
#' product), the Bandler-Kohout superproduct (also supdirect product), and finally,
#' the Bandler-Kohout square product. More complicated composition operations
#' may be performed by using the [mult()] function and/or by combining multiple
#' composition results with the [algebra()] operations.
#'
#' @param x A first fuzzy relation to be composed. It must be a numeric matrix
#' with values within the \eqn{[0,1]} interval. The number of columns must
#' match with the number of rows of the `y` matrix.
#' @param y A second fuzzy relation to be composed. It must be a numeric matrix
#' with values within the \eqn{[0,1]} interval. The number of columns must
#' match with the number of rows of the `x` matrix.
#' @param e Deprecated. An excluding fuzzy relation. If not NULL,
#' it must be a numeric matrix with dimensions equal to the `y` matrix.
#' @param alg An algebra to be used for composition. It must be one of
#' `'goedel'` (default), `'goguen'`, or `'lukasiewicz'`, or an instance of class `algebra`
#' (see [algebra()]).
#' @param type A type of a composition to be performed. It must be one of
#' `'basic'` (default), `'sub'`, `'super'`, or `'square'`.
#' @param quantifier Deprecated. If not NULL, it must be a function taking a single
#' argument, a vector of relative cardinalities, that would be translated into
#' membership degrees. A result of the [lingexpr()] function is a
#' good candidate for that. Note that the vector of relative cardinalities contains also
#' two attributes, `x` and `y`, which carry the original `R`'s data row (in `x`) and `S`'s
#' feature column (in `y`). These attributes are accessible using the standard [base::attr()]
#' function. Find examples below that define some quantifiers.
#' @param sorting Deprecated. Sorting function used within quantifier application. The given function
#' must sort the membership degrees and allow the `decreasing` argument as in [base::sort()].
#' This function have to be explicitly specified typically if performing compositions that
#' handle `NA` values.
#' @return A matrix with \eqn{v} rows and \eqn{w} columns, where \eqn{v} is the
#' number of rows of `x` and \eqn{w} is the number of columns of `y`.
#' @author Michal Burda
#' @seealso [algebra(), [mult()], [lingexpr()]
#' @keywords models robust multivariate
#' @examples
#'     R <- matrix(c(0.1, 0.6, 1, 0, 0, 0,
#'                   0, 0.3, 0.7, 0.9, 1, 1,
#'                   0, 0, 0.6, 0.8, 1, 0,
#'                   0, 1, 0.5, 0, 0, 0,
#'                   0, 0, 1, 1, 0, 0), byrow=TRUE, nrow=5)
#'
#'     S <- matrix(c(0.9, 1, 0.9, 1,
#'                   1, 1, 1, 1,
#'                   0.1, 0.2, 0, 0.2,
#'                   0, 0, 0, 0,
#'                   0.7, 0.6, 0.5, 0.4,
#'                   1, 0.9, 0.7, 0.6), byrow=TRUE, nrow=6)
#'
#'     RS <- matrix(c(0.6, 0.6, 0.6, 0.6,
#'                    1, 0.9, 0.7, 0.6,
#'                    0.7, 0.6, 0.5, 0.4,
#'                    1, 1, 1, 1,
#'                    0.1, 0.2, 0, 0.2), byrow=TRUE, nrow=5)
#'
#'     compose(R, S, alg='goedel', type='basic') # should be equal to RS
#' @export compose
compose <- function(x,
                    y,
                    e=NULL,
                    alg=c('goedel', 'goguen', 'lukasiewicz'),
                    type=c('basic', 'sub', 'super', 'square'),
                    quantifier=NULL,
                    sorting=sort) {

    .mustBeNumericMatrix(x)
    .mustBeNumericMatrix(y)
    .mustBe(nrow(x) > 0, "'x' must be a matrix with at least 1 row")
    .mustBe(ncol(y) > 0, "'y' must be a matrix with at least 1 column")
    .mustBe(ncol(x) == nrow(y), "The number of columns of 'x' must be equal to the number of rows of 'y'")

    if (!is.null(e)) {
        .Deprecated('mult',
                    msg=paste('The "e" argument is deprecated. Complex compositions',
                                'have to be computed by using the "compose()" or "mult()"',
                                'functions and combined with algebra() operations.'))
        .mustBeNumericMatrix(e)
        .mustBe(nrow(y) == nrow(e) && ncol(y) == ncol(e), "'e' must have the same dimensions as 'y'")

    }

    if (is.character(alg)) {
        alg <- match.arg(alg)
        alg <- algebra(alg)
    }
    .mustBe(is.algebra(alg), "'alg' must be either one of 'goedel', 'goguen', lukasiewicz', or an instance of class 'algebra'")

    if (is.character(type)) {
        type <- match.arg(type)
        if (type == 'basic') {
            type <- alg$pt
            merge <- alg$s
        } else if (type == 'sub') {
            type <- alg$r
            merge <- alg$i
        } else if (type == 'super') {
            type <- function(x, y) { alg$r(y, x) }
            merge <- alg$i
        } else if (type == 'square') {
            type <- alg$b
            merge <- alg$i
        } else {
            stop('Unrecognized composition type')
        }
    } else if (is.function(type)) {
        .mustBe(is.function(quantifier), "if 'type' is a function then 'quantifier' must be a function too")
    }

    .mustBe(is.function(type), "'type' must be either one of 'basic', 'sub', 'super', 'square', or a function with 1 argument")

    f <- NULL
    # TODO: the 'quantifier' argument is not a real quantifier! It is more a linguistic expression giving degrees to relative
    # cardinalitites. The quantifiers are tightly related to algebras, e.g. the 'sort' function is a property of algebra (especially
    # if considering algebras that work with NAs)
    if (is.function(quantifier)) {
        .Deprecated('mult',
                    msg=paste('The "quantifier" argument is deprecated. Complex compositions',
                                'have to be computed by using the "compose()" or "mult()"',
                                'functions and combined with algebra() operations.'))
        merge <- function(val, x, y) {
            res <- sorting(val, decreasing=TRUE)
            relcard <- seq_along(res) / length(res)
            attr(relcard, 'x') <- x
            attr(relcard, 'y') <- y
            alg$s(alg$pt(res, quantifier(relcard)))
        }
        f <- function(x, y) {
          merge(type(x, y), x, y)
        }

    } else if (!is.null(quantifier)) {
        stop("'quantifier' must be a function or NULL")

    } else {
      f <- function(x, y) {
          merge(type(x, y))
      }
    }

    res <- mult(x, y, f)
    if (!is.null(e)) {
      # TODO: test excluding features!
      # TODO: excluding should not be inside of this function - it should be a separate function
      fe <- function(x, y) {
        alg$s(alg$pt(x, y))
      }
      e <- mult(x, e, fe)
      res <- alg$pt(res, alg$n(e))
    }
    return(res)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.printCtx <- function(name) {
    function(x, ...) {
        clazz <- class(x)[1]
        cat('Linguistic context: ', name, ' (', clazz, ')', '\n', sep='')
        class(x) <- NULL
        print(x)
    }
}

#' Print the linguistic context
#'
#' Format an object of the [ctx3()], [ctx5()], [ctx3bilat()] and the [ctx5bilat()]
#' class into human readable form and print it to the output.
#'
#' @param x A linguistic context to be printed
#' @param ... Unused.
#' @return Nothing.
#' @author Michal Burda
#' @seealso [ctx3()], [ctx5()], [ctx3bilat()], [ctx5bilat()], [minmax()]
#' @keywords models robust
#' @examples
#'
#'   context <- ctx3()
#'   print(context)
#'
#' @rdname print.ctx
#' @export
print.ctx3 <- .printCtx('unilateral trichotomy')

#' @rdname print.ctx
#' @export
print.ctx5 <- .printCtx('unilateral pentachotomy')

#' @rdname print.ctx
#' @export
print.ctx3bilat <- .printCtx('bilateral trichotomy')

#' @rdname print.ctx
#' @export
print.ctx5bilat <- .printCtx('bilateral pentachotomy')

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Create an instance of S3 class `farules` which represents a set of fuzzy
#' association rules and their statistical characteristics.
#'
#' This function is a constructor that returns an instance of the `farules` S3 class.
#' To search for fuzzy association rules, refer to the [searchrules()] function.
#'
#' @param rules A list of character vectors, where each vector represents a
#' rule and each value of the vector represents a predicate. The first value of
#' the vector is assumed to be a rule's consequent, the rest is a rule's antecedent.
#' @param statistics A numeric matrix of various statistical characteristics of
#' the rules. Each column of that matrix corresponds to some statistic (such as
#' support, confidence, etc.) and each row corresponds to a rule in the list of
#' `rules`.
#' @return Returns an object of class `farules`.
#' @author Michal Burda
#' @seealso [searchrules()]
#' @keywords models robust
#' @export
farules <- function(rules, statistics) {
    .mustBe(is.list(rules) &&
                all(sapply(rules, function(r) { is.character(r) && length(r) > 0 })),
            "'rules' must be a list of non-empty character vectors")
    .mustBeNumericMatrix(statistics)
    .mustBe(nrow(statistics) == length(rules), "The length of 'rules' must be the same as 'nrow(statistics)'")

    res <- structure(list(rules=rules,
                          statistics=statistics),
                     class=c('farules', 'list'))
    return(res)
}


#' Test whether `x` inherits from the S3 `farules` class.
#'
#' @param x An object being tested.
#' @return `TRUE` if `x` is a valid [farules()] object and `FALSE` otherwise.
#' @author Michal Burda
#' @seealso \code{\link{farules}}
#' @keywords models robust
#' @export
is.farules <- function(x) {
    return(inherits(x, 'farules') &&
               is.list(x) &&
               is.list(x$rules) &&
               is.matrix(x$statistics))
}


#' Convert the instance of the [farules()] S3 class into a data frame.
#' Empty [farules()] object is converted into an empty [data.frame()].
#'
#' @param x An instance of class [farules()] to be transformed.
#' @param ...  Unused.
#' @return A data frame of statistics of the rules that are stored in the given
#' [farules()] object. Row names of the resulting data frame are in
#' the form: `A1 & A2 & ... & An => C`, where `Ai` are antecedent
#' predicates and `C` is a consequent. An empty [farules()] object
#' is converted into an empty [data.frame()] object.
#' @author Michal Burda
#' @seealso [farules()], [searchrules()]
#' @keywords models robust multivariate
#' @export
as.data.frame.farules <- function(x, ...) {
    if (length(x$rules) <= 0) {
        return(data.frame())
    } else {
        r <- x$statistics
        rownames(r) <- sapply(x$rules, function(rule) {
            ante <- rule[-1]
            conseq <- rule[1]
            paste(paste(ante, collapse=' & '), conseq, sep=' => ')
        })
        return(as.data.frame(r))
    }
}


#' Print an instance of the [farules()] S3 class in a human readable form.
#'
#' @param x An instance of the [farules()] S3 class
#' @param ...  Unused.
#' @return None.
#' @author Michal Burda
#' @seealso [farules()], [searchrules()]
#' @keywords models robust
#' @export
print.farules <- function(x, ...) {
    if (length(x$rules) <= 0) {
        cat('Empty rule base\n')
    } else {
        print(as.data.frame(x))
    }
}


#' Take a sequence of instances of S3 class [farules()] and combine them into a single
#' object. An error is thrown if some argument does not inherit from the [farules()]
#' class.
#'
#' @param ...  A sequence of objects of class [farules()] to be concatenated.
#' @param recursive This argument has currently no function and is added here
#' only for compatibility with generic \code{\link{c}} function.
#' @return An object of class [farules()] that is created by merging the
#' arguments together, i.e.  by concatenating the rules and row-binding the
#' statistics of given objects.
#' @author Michal Burda
#' @seealso [farules()], [searchrules()]
#' @keywords models robust
#' @examples
#'     ori1 <- farules(rules=list(letters[1:3],
#'                                letters[2:5]),
#'                     statistics=matrix(runif(16), nrow=2))
#'     ori2 <- farules(rules=list(letters[4],
#'                                letters[3:8]),
#'                     statistics=matrix(runif(16), nrow=2))
#'     res <- c(ori1, ori2)
#'     print(res)
#' @export
c.farules <- function(..., recursive=FALSE) {
    dots <- list(...)

    .mustBe(all(sapply(dots, is.farules)),
            "Cannot concatenate arguments that are not valid 'farules' objects")
    if (recursive) {
        warning("Recursion not implemented in 'c.farules'")
    }

    rules <- do.call('c', lapply(dots, '[[', 'rules'))
    statistics <- do.call('rbind', lapply(dots, '[[', 'statistics'))
    return(farules(rules, statistics))
}

test_that('consequents', {
    expect_equal(consequents(list(c('a', 'b', 'c'),
                                  c('d'),
                                  c('a', 'e'))),
                 list('a', 'd', 'a'))
    expect_error(consequents(list(c(1, 2, 3))))
    expect_error(consequents(list(character())))
})


test_that('consequents (single rule)', {
    expect_equal(consequents(list(c('a', 'b', 'c'))),
                 list('a'))
})

test_that('consequents (empty list)', {
    expect_equal(consequents(list()),
                 list())
})


test_that('consequents (farules)', {
    f <- farules(list(letters[1:3], letters[2:5]),
                 matrix(0, nrow=2, ncol=3))
    expect_equal(consequents(f),
                 list(letters[1], letters[2]))
})

test_that("farules", {
    r <- list(letters[1:3], letters[1:4])
    m <-  matrix(1:6, nrow=2)
    f <- farules(r, m)

    expect_true(is.farules(f))
    expect_true(inherits(f, 'farules'))
    expect_true(is.list(f))
    expect_true(is.list(f$rules))
    expect_true(is.matrix(f$statistics))
    expect_equal(f$rules, r)
    expect_equal(f$statistics, m)
})


test_that("as.data.frame.farules", {
    stats <- matrix(1:9, nrow=3);
    colnames(stats) <- c('support', 'confidence', 'lift')
    res <- data.frame(stats)
    rownames(res) <- c('a & b => c', 'x & y & z => d', ' => e')

    f <- farules(list(c('c', 'a', 'b'),
                      c('d', 'x', 'y', 'z'),
                      c('e')),
                 stats)
    expect_true(is.farules(f))

    d <- as.data.frame(f)
    expect_equal(d, res)
    expect_false(is.farules(d))
})


test_that('c.farules', {
    set.seed(3345)
    ori1 <- farules(rules=list(letters[1:3],
                               letters[2:5]),
                    statistics=matrix(runif(16), nrow=2))
    ori2 <- farules(rules=list(letters[4],
                               letters[3:8]),
                    statistics=matrix(runif(16), nrow=2))

    res <- c(ori1, ori2)
    expect_true(is.farules(res))
    expect_true(is.list(res))
    expect_true(is.list(res$rules))
    expect_true(is.matrix(res$statistics))
    expect_equal(res$rules, c(ori1$rules, ori2$rules))
    expect_equal(res$statistics, rbind(ori1$statistics, ori2$statistics))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Factories for functions that convert numeric data into membership degrees of fuzzy sets
#'
#' These functions create functions with a single argument `x` that compute membership degrees of `x` to a fuzzy set
#' of either triangular or raised-cosine shape that is defined by `lo`, `center`, and `hi`.
#'
#' The arguments must satisfy `lo <= center <= hi`. Functions compute membership degrees of triangular or
#' raised-cosine fuzzy sets. `x` values equal to `center obtain membership degree equal to 1, `x` values lower than `lo`
#' or greater than `hi` obtain membership degree equal to 0. A transition of the triangular (resp. raised cosine) shape
#' (with peak at `center` is computed for `x` values between `lo` and `hi`.
#'
#' If `lo == -Inf` then any value that is lower or equal to center gets membership degree 1.  Similarly, if `hi == Inf`
#' then any value that is greater or equal to center gets membership degree 1. `NA` and `NaN` values remain unchanged.
#'
#' `triangular()` produces fuzzy sets of a triangular shape (with peak at `center`), `raisedcosine()` produces
#' fuzzy sets defined as a raised cosine hill.
#'
#' @aliases triangular raisedcosine
#' @param lo A lower bound (can be -Inf).
#' @param center A peak value.
#' @param hi An upper bound (can be Inf).
#' @return A function with single argument `x` that should be a numeric vector to be converted.
#' @author Michal Burda
#' @seealso [fcut()]
#' @keywords models robust multivariate
#' @examples
#'
#' tr <- triangular(1, 2, 3)
#' tr(1:30 / 3)
#'
#' rc <- raisedcosine(1, 2, 3)
#' rc(1:30 / 3)
#'
#' plot(triangular(-1, 0, 1), from=-2, to=3)
#' plot(triangular(-1, 0, 2), from=-2, to=3)
#' plot(triangular(-Inf, 0, 1), from=-2, to=3)
#' plot(triangular(-1, 0, Inf), from=-2, to=3)
#'
#' plot(raisedcosine(-1, 0, 1), from=-2, to=3)
#' plot(raisedcosine(-1, 0, 2), from=-2, to=3)
#' plot(raisedcosine(-Inf, 0, 1), from=-2, to=3)
#' plot(raisedcosine(-1, 0, Inf), from=-2, to=3)
#'
#' @export
triangular <- function(lo, center, hi) {
    .mustBeNumericScalar(lo)
    .mustBeNumericScalar(center)
    .mustBeNumericScalar(hi)

    if (lo > center || lo > hi) {
        .stop('"lo" must be the lower-bound of the interval <lo, hi>')
    }
    if (hi < center || hi < lo) {
        .stop('"hi" must be the upper-bound of the interval <lo, hi>')
    }
    if (center < lo || center > hi) {
        .stop('"center" must be within the interval <lo, hi>')
    }

    ctx <- as.numeric(c(lo, center, hi))
    return(function(x) {
        .Call('_lfl_triangle', as.numeric(as.vector(x)), ctx, PACKAGE='lfl')
    })
}

#' @rdname triangular
#' @export
raisedcosine <- function(lo, center, hi) {
    .mustBeNumericScalar(lo)
    .mustBeNumericScalar(center)
    .mustBeNumericScalar(hi)

    if (lo > center || lo > hi) {
        .stop('"lo" must be the lower-bound of the interval <lo, hi>')
    }
    if (hi < center || hi < lo) {
        .stop('"hi" must be the upper-bound of the interval <lo, hi>')
    }
    if (center < lo || center > hi) {
        .stop('"center" must be within the interval <lo, hi>')
    }

    ctx <- as.numeric(c(lo, center, hi))
    return(function(x) {
        .Call('_lfl_raisedcos', as.numeric(as.vector(x)), ctx, PACKAGE='lfl')
    })
}

test_that("triangular", {
    expect_equal(triangular(0, 1, 2)(-5), 0)
    expect_equal(triangular(0, 1, 2)(-1), 0)
    expect_equal(triangular(0, 1, 2)(0), 0)
    expect_equal(triangular(0, 1, 2)(0.5), 0.5)
    expect_equal(triangular(0, 1, 2)(1), 1)
    expect_equal(triangular(0, 1, 2)(1.5), 0.5)
    expect_equal(triangular(0, 1, 2)(2), 0)
    expect_equal(triangular(0, 1, 2)(10), 0)
    expect_equal(triangular(0, 1, 2)(-Inf), 0)
    expect_equal(triangular(0, 1, 2)(Inf), 0)

    expect_equal(triangular(1, 1, 2)(-1), 0)
    expect_equal(triangular(1, 1, 2)(0), 0)
    expect_equal(triangular(1, 1, 2)(1), 1)
    expect_equal(triangular(1, 1, 2)(1.5), 0.5)
    expect_equal(triangular(1, 1, 2)(2), 0)
    expect_equal(triangular(1, 1, 2)(10), 0)
    expect_equal(triangular(1, 1, 2)(-Inf), 0)
    expect_equal(triangular(1, 1, 2)(Inf), 0)

    expect_equal(triangular(0, 1, 1)(-1), 0)
    expect_equal(triangular(0, 1, 1)(0), 0)
    expect_equal(triangular(0, 1, 1)(0.5), 0.5)
    expect_equal(triangular(0, 1, 1)(1), 1)
    expect_equal(triangular(0, 1, 1)(2), 0)
    expect_equal(triangular(0, 1, 1)(10), 0)
    expect_equal(triangular(0, 1, 1)(-Inf), 0)
    expect_equal(triangular(0, 1, 1)(Inf), 0)

    expect_equal(triangular(1, 1, 1)(0), 0)
    expect_equal(triangular(1, 1, 1)(1), 1)
    expect_equal(triangular(1, 1, 1)(2), 0)
    expect_equal(triangular(1, 1, 1)(-Inf), 0)
    expect_equal(triangular(1, 1, 1)(Inf), 0)

    expect_equal(triangular(-Inf, 1, 2)(-5), 1)
    expect_equal(triangular(-Inf, 1, 2)(0), 1)
    expect_equal(triangular(-Inf, 1, 2)(1), 1)
    expect_equal(triangular(-Inf, 1, 2)(1.5), 0.5)
    expect_equal(triangular(-Inf, 1, 2)(2), 0)
    expect_equal(triangular(-Inf, 1, 2)(5), 0)
    expect_equal(triangular(-Inf, 1, 2)(-Inf), 1)
    expect_equal(triangular(-Inf, 1, 2)(Inf), 0)

    expect_equal(triangular(0, 1, Inf)(0), 0)
    expect_equal(triangular(0, 1, Inf)(0.5), 0.5)
    expect_equal(triangular(0, 1, Inf)(1), 1)
    expect_equal(triangular(0, 1, Inf)(2), 1)
    expect_equal(triangular(0, 1, Inf)(5), 1)
    expect_equal(triangular(0, 1, Inf)(-Inf), 0)
    expect_equal(triangular(0, 1, Inf)(Inf), 1)

    expect_equal(triangular(-Inf, 1, Inf)(-5), 1)
    expect_equal(triangular(-Inf, 1, Inf)(0), 1)
    expect_equal(triangular(-Inf, 1, Inf)(1), 1)
    expect_equal(triangular(-Inf, 1, Inf)(2), 1)
    expect_equal(triangular(-Inf, 1, Inf)(5), 1)
    expect_equal(triangular(-Inf, 1, Inf)(-Inf), 1)
    expect_equal(triangular(-Inf, 1, Inf)(Inf), 1)

    expect_equal(triangular(-Inf, -Inf, Inf)(1), 1)
    expect_equal(triangular(-Inf, Inf, Inf)(1), 1)
    expect_equal(triangular(-Inf, Inf, Inf)(-Inf), 1)
    expect_equal(triangular(-Inf, Inf, Inf)(Inf), 1)

    expect_equal(triangular(0, 1, 2)(NA), as.numeric(NA))
    expect_equal(triangular(0, 1, 2)(NaN), as.numeric(NaN))
})


test_that("raisedcosine", {
    expect_equal(raisedcosine(0, 1, 2)(-5), 0)
    expect_equal(raisedcosine(0, 1, 2)(-1), 0)
    expect_equal(raisedcosine(0, 1, 2)(0), 0)
    expect_equal(raisedcosine(0, 1, 2)(0.5), 0.5)
    expect_equal(raisedcosine(0, 1, 2)(1), 1)
    expect_equal(raisedcosine(0, 1, 2)(1.5), 0.5)
    expect_equal(raisedcosine(0, 1, 2)(2), 0)
    expect_equal(raisedcosine(0, 1, 2)(10), 0)
    expect_equal(raisedcosine(0, 1, 2)(-Inf), 0)
    expect_equal(raisedcosine(0, 1, 2)(Inf), 0)

    expect_equal(raisedcosine(1, 1, 2)(-1), 0)
    expect_equal(raisedcosine(1, 1, 2)(0), 0)
    expect_equal(raisedcosine(1, 1, 2)(1), 1)
    expect_equal(raisedcosine(1, 1, 2)(1.5), 0.5)
    expect_equal(raisedcosine(1, 1, 2)(2), 0)
    expect_equal(raisedcosine(1, 1, 2)(10), 0)
    expect_equal(raisedcosine(1, 1, 2)(-Inf), 0)
    expect_equal(raisedcosine(1, 1, 2)(Inf), 0)

    expect_equal(raisedcosine(0, 1, 1)(-1), 0)
    expect_equal(raisedcosine(0, 1, 1)(0), 0)
    expect_equal(raisedcosine(0, 1, 1)(0.5), 0.5)
    expect_equal(raisedcosine(0, 1, 1)(1), 1)
    expect_equal(raisedcosine(0, 1, 1)(2), 0)
    expect_equal(raisedcosine(0, 1, 1)(10), 0)
    expect_equal(raisedcosine(0, 1, 1)(-Inf), 0)
    expect_equal(raisedcosine(0, 1, 1)(Inf), 0)

    expect_equal(raisedcosine(1, 1, 1)(0), 0)
    expect_equal(raisedcosine(1, 1, 1)(1), 1)
    expect_equal(raisedcosine(1, 1, 1)(2), 0)
    expect_equal(raisedcosine(1, 1, 1)(-Inf), 0)
    expect_equal(raisedcosine(1, 1, 1)(Inf), 0)

    expect_equal(raisedcosine(-Inf, 1, 2)(-5), 1)
    expect_equal(raisedcosine(-Inf, 1, 2)(0), 1)
    expect_equal(raisedcosine(-Inf, 1, 2)(1), 1)
    expect_equal(raisedcosine(-Inf, 1, 2)(1.5), 0.5)
    expect_equal(raisedcosine(-Inf, 1, 2)(2), 0)
    expect_equal(raisedcosine(-Inf, 1, 2)(5), 0)
    expect_equal(raisedcosine(-Inf, 1, 2)(-Inf), 1)
    expect_equal(raisedcosine(-Inf, 1, 2)(Inf), 0)

    expect_equal(raisedcosine(0, 1, Inf)(0), 0)
    expect_equal(raisedcosine(0, 1, Inf)(0.5), 0.5)
    expect_equal(raisedcosine(0, 1, Inf)(1), 1)
    expect_equal(raisedcosine(0, 1, Inf)(2), 1)
    expect_equal(raisedcosine(0, 1, Inf)(5), 1)
    expect_equal(raisedcosine(0, 1, Inf)(-Inf), 0)
    expect_equal(raisedcosine(0, 1, Inf)(Inf), 1)

    expect_equal(raisedcosine(-Inf, 1, Inf)(-5), 1)
    expect_equal(raisedcosine(-Inf, 1, Inf)(0), 1)
    expect_equal(raisedcosine(-Inf, 1, Inf)(1), 1)
    expect_equal(raisedcosine(-Inf, 1, Inf)(2), 1)
    expect_equal(raisedcosine(-Inf, 1, Inf)(5), 1)
    expect_equal(raisedcosine(-Inf, 1, Inf)(Inf), 1)
    expect_equal(raisedcosine(-Inf, 1, Inf)(-Inf), 1)

    expect_equal(raisedcosine(-Inf, -Inf, Inf)(1), 1)
    expect_equal(raisedcosine(-Inf, Inf, Inf)(1), 1)
    expect_equal(raisedcosine(-Inf, Inf, Inf)(Inf), 1)
    expect_equal(raisedcosine(-Inf, Inf, Inf)(-Inf), 1)

    expect_equal(raisedcosine(0, 1, 2)(NA), as.numeric(NA))
    expect_equal(raisedcosine(0, 1, 2)(NaN), as.numeric(NaN))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.perceiveGlobal <- function(rules, n, v, specs) {
    len <- length(rules)
    if (len <= 1) {
        return(rep(TRUE, len))
    }
    res <- rep(TRUE, len)
    for (i in 1:(len-1)) {
        if (res[i]) {
            ri <- rules[[i]]
            ii <- match(ri[-1], n)
            for (j in (i+1):len) {
                if (res[j]) {
                    rj <- rules[[j]]
                    ji <- match(rj[-1], n)
                    if (.is.specific(ii, ji, v, specs)) {
                        res[j] <- FALSE
                    } else if (.is.specific(ji, ii, v, specs)) {
                        res[i] <- FALSE
                    }
                }
            }
        }
    }
    return(res)
}


.perceiveLocal <- function(rules, n, v, specs, fired) {
    len <- length(rules)
    if (len <= 1) {
        return(rep(TRUE, len))
    }
    res <- rep(TRUE, len)
    for (i in 1:(len-1)) {
        if (res[i]) {
            ri <- rules[[i]]
            ii <- match(ri[-1], n)
            for (j in (i+1):len) {
                if (res[j]) {
                    rj <- rules[[j]]
                    ji <- match(rj[-1], n)
                    rispec <- .is.specific(ii, ji, v, specs)
                    rjspec <- .is.specific(ji, ii, v, specs)
                    if (rispec || rjspec) {
                        if (fired[i] > fired[j]) {
                            res[j] <- FALSE
                        } else if (fired[i] < fired[j]) {
                            res[i] <- FALSE
                        } else {
                            if (rispec) {
                                res[j] <- FALSE
                            } else {
                                res[i] <- FALSE
                            }
                        }
                    }
                }
            }
        }
    }
    return(res)
}


#' From a set of rules, remove each rule for which another rule exists that is
#' more specific.
#'
#' Examine rules in a list and remove all of them for which other more specific
#' rules are present in the list. The specificity is determined by calling the
#' [is.specific()] function.  This operation is a part of the
#' [pbld()] inference mechanism.
#'
#' In other words, for each rule `x` in the `rules` list, it searches for another
#' rule `y` such that `is.specific(y, x)` returns TRUE. If yes then
#' `x` is removed from the list.
#'
#' @param rules A list of character vectors where each element is a fuzzy set
#' name (a predicate) and thus each such vector forms a rule.
#' @param fsets A valid instance of the [fsets()] class such that all predicates
#' in `rules` (i.e., all values of all character vectors in `rules$rules`)
#' can be found in `colnames(fsets)`
#' @param type The type of perception to use. It can be either `"local"`
#' or `"global"` (default).
#' @param fired If `type=="global"` then this argument can be NULL. If
#' `type` is `"local"` then `fired` must be a numeric vector of
#' values in the interval \eqn{[0,1]} indicating the truth values of all rules,
#' i.e., the length of the vector must be equal to the number of rules in the
#' `rules` argument.
#' @param vars A deprecated parameter that must be `NULL`. Formerly, it was
#' a named (typically character) vector that determined which
#' predicates originate from the same variable, i.e., which of them semantically
#' deal with the same property.  For that purpose, each value from any vector
#' stored in the `rules` list must be present in `names(vars)`.  See
#' also [vars()] function of the [fsets()] class.
#' @param specs A deprecated parameter that must be `NULL`. Formerly, it was
#' a square numeric matrix containing values from \eqn{\{0, 1\}}.
#' It is a specificity matrix for which each row and column corresponds to an
#' `rules`'es predicate `specs[i][j] = 1` if and only if the
#' \eqn{i}-th predicate is more specific (i.e., the corresponding fuzzy set is a
#' subset of) than the \eqn{j}-th predicate (i.e. `x[, j]`).  See also
#' [specs()] function of the [fsets()] class.
#' @return A modified list of rules for which no other more specific rule
#' exists. (Each rule is a vector.)
#' @author Michal Burda
#' @seealso [is.specific()], [fsets()], [fcut()], [lcut()]
#' @keywords models robust
#' @examples
#' # prepare fsets
#' f <- lcut(data.frame(a=0:1, b=0:1, c=0:1, d=0:1))
#'
#' # run perceive function: (sm.a, bi.c) has
#' # more specific rule (ve.sm.a, bi.c)
#' perceive(list(c('sm.a', 'bi.c'),
#'               c('ve.sm.a', 'bi.c'),
#'               c('sm.b', 'sm.d')),
#'          f)
#'
#' @export perceive
perceive <- function(rules,
                     fsets,
                     type=c('global', 'local'),
                     fired=NULL,
                     vars=NULL,
                     specs=NULL) {
    if (!(is.null(vars) && is.null(specs))) {
        .stop('"vars" and "specs" parameters are defunct. Specify "fsets" parameter instead.')
    }
    type <- match.arg(type)
    if (!is.list(rules) && !is.null(rules)) {
        stop("'rules' must be a list of rules")
    }
    .mustBe(is.fsets(fsets), '"fsets" must be a valid instance of the "fsets" S3 class')
    unlisted <- unique(unlist(antecedents(rules)))
    if (length(intersect(unlisted, colnames(fsets))) != length(unlisted)) {
        stop('All predicates in "rules" must be from colnames of "fsets"')
    }

    n <- colnames(fsets)
    v <- as.integer(factor(vars(fsets)))
    specs <- specs(fsets)

    if (type == 'local') {
        if (!is.vector(fired) || !is.numeric(fired)) {
            stop("If type of perception is 'local' then 'fired' must be a numeric vector")
        }
        return(.perceiveLocal(rules, n, v, specs, fired))
    } else {
        return(.perceiveGlobal(rules, n, v, specs))
    }
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Creating linguistic context directly from values
#'
#' This function creates a context (i.e., an instance of S3 class
#' [ctx3()], [ctx3bilat()], [ctx5()], or [ctx5bilat()]) based on values
#' of the numeric vector `x`. In default, the context is based on minimum
#' and maximum value of `x` in the following way:
#' * `ctx3`, `ctx5`: low = minimum, high = maximum value of `x`;
#' * `ctx3bilat`, `ctx5bilat`: negMax = minimum, max = maximum value of `x`,
#'    origin = mean of minimum and maximum.
#'
#' Other values are computed according to the defaults as defined in the constructors
#' [ctx3()], [ctx3bilat()], [ctx5()], and [ctx5bilat()]).
#'
#' @param x A numeric vector to compute the context from
#' @param type A type of the context to be returned. Must be one of:
#'   `ctx3`, `ctx5`, `ctx3bilat` or `ctx5bilat`
#' @param ... other parameters to be passed to the appropriate constructor
#'   ([ctx3()], [ctx3bilat()], [ctx5()], and [ctx5bilat()]) that is called internally.
#'   These values overwrite the defaults computed by `minmax` -- see the examples.
#' @examples
#'   minmax(0:100)                # returns ctx3: 0, 50, 100
#'   minmax(0:100, high=80)       # returns ctx3: 0, 40, 80
#'   minmax(0:100, relCenter=0.4) # returns ctx3: 0, 40, 100
#'   minmax(0:100, type='ctx5')   # returns ctx5: 0, 25, 50, 75, 100
#' @export
minmax <- function(x,
                   type=c('ctx3', 'ctx5', 'ctx3bilat', 'ctx5bilat'),
                   ...) {
    .mustBeNumericVector(x)
    type <- match.arg(type)
    dots <- list(...)
    mini <- min(x, na.rm=TRUE)
    maxi <- max(x, na.rm=TRUE)
    if (type %in% c('ctx3', 'ctx5')) {
        if (is.null(dots[['low']])) {
            dots[['low']] <- mini
        }
        if (is.null(dots[['high']])) {
            dots[['high']] <- maxi
        }
    } else {
        if (is.null(dots[['negMax']])) {
            dots[['negMax']] <- mini
        }
        if (is.null(dots[['max']])) {
            dots[['max']] <- maxi
        }
        if (is.null(dots[['origin']])) {
            dots[['origin']] <- mean(c(mini, maxi))
        }
    }
    do.call(type, dots)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.defuzzEmptyRulebase <- function(values) {
    return(defuzz(rep(1, length(values)),
                    values,
                    type='dee'))
}




#' Perform a Perception-based Logical Deduction (PbLD) with given rule-base on
#' given dataset
#'
#' Take a set of rules (a rule-base) and perform a Perception-based Logical
#' Deduction (PbLD) on each row of a given [fsets()] object.
#'
#' Perform a Perception-based Logical Deduction (PbLD) with given rule-base
#' `rules` on each row of input `x`. Columns of `x` are truth
#' values of predicates that appear in the antecedent part of `rules`,
#' `partition` together with `values` determine the shape of
#' predicates in consequents: each element in `values` corresponds to a
#' row of membership degrees in `partition`.
#'
#' @param x Input to the inference. It should be an object of class
#' [fsets()] (e.g., created by using the [fcut()] or [lcut()] functions).
#' It is basically a matrix with columns representing fuzzy sets.
#'
#' Each row represents a single case of inference. Columns should be named
#' after predicates in rules' antecedents.
#' @param rules A rule-base (a.k.a. linguistic description) either in the form
#' of the [farules()] object or as a list of character vectors where
#' each element is a fuzzy set name (a predicate) and thus each such vector
#' forms a rule.
#' @param partition A [fsets()] object with columns that are
#' consequents in `rules`. These membership degrees must correspond to
#' `values`.
#' @param values Crisp values that correspond to rows of membership degrees in
#' the `partition` matrix.  Function assumes that the values are sorted in
#' the ascending order.
#' @param type The type of inference to use. It can be either `"local"` or
#' `"global"` (default).
#' @param parallel Whether the processing should be run in parallel or not.
#' Parallelization is implemented using the [foreach::foreach()]
#' package. The parallel environment must be set properly in advance, e.g., with
#' the [doMC::registerDoMC()] function.
#' @return A vector of inferred defuzzified values. The number of resulting
#' values corresponds to the number of rows of the `x` argument.
#' @author Michal Burda
#' @seealso [infer()], [lcut()], [searchrules()], [fire()], [aggregateConsequents()], [defuzz()]
#' @references A. Dvořák, M. Štěpnička, On perception-based logical deduction
#' and its variants, in: Proc. 16th World Congress of the International Fuzzy
#' Systems Association and 9th Conference of the European Society for Fuzzy
#' Logic and Technology (IFSA-EUSFLAT 2015), Advances in Intelligent Systems
#' Research, Atlantic Press, Gijon, 2015.
#' @keywords models robust
#' @examples
#'
#' # --- TRAINING PART ---
#' # custom context of the RHS variable
#' uptakeContext <- ctx3(7, 28.3, 46)
#'
#' # convert data into fuzzy sets
#' d <- lcut(CO2, context=list(uptake=uptakeContext))
#'
#' # split data into the training and testing set
#' testingIndices <- 1:5
#' trainingIndices <- setdiff(seq_len(nrow(CO2)), testingIndices)
#' training <- d[trainingIndices, ]
#' testing <- d[testingIndices, ]
#'
#' # search for rules
#' r <- searchrules(training, lhs=1:38, rhs=39:58, minConfidence=0.5)
#'
#' # --- TESTING PART ---
#' # prepare values and partition
#' v <- seq(uptakeContext[1], uptakeContext[3], length.out=1000)
#' p <- lcut(v, name='uptake', context=uptakeContext)
#'
#' # do the inference
#' pbld(testing, r, p, v)
#'
#' @export pbld
pbld <- function(x,
                 rules,
                 partition,
                 values,
                 type=c('global', 'local'),
                 parallel=FALSE) {
    .mustBe(is.fsets(x), "'x' must be an instance of the 'fsets' class")

    if (is.farules(rules)) {
        rules <- rules$rules
    } else if (is.vector(rules) && is.character(rules)) {
        rules <- list(rules)
    }

    .mustBe(is.list(rules), "'rules' must be a list of rules or an instance of the 'farules' class")
    .mustBeNumericMatrix(partition)
    .mustNotBeZeroLength(partition)
    .mustBe(!is.null(colnames(partition)) && is.character(colnames(partition)), "'partition' must have colnames")
    .mustBeNumericVector(values)
    .mustBe(nrow(partition) == length(values), "The length of 'values' must be equal to the number of rows of 'partition'")

    diffAttr <- setdiff(unique(unlist(antecedents(rules))), colnames(x))
    if (length(diffAttr) > 0) {
        stop(paste0("'rules' contains predicates in antecedent-part that are not present in 'colnames(x)': ",
                    paste(diffAttr, collapse=', ', sep='')))
    }

    type <- match.arg(type)

    if (length(rules) <= 0) {
        return(rep(.defuzzEmptyRulebase(values), nrow(x)))
    }

    n <- colnames(x)
    v <- as.integer(factor(vars(x)))
    specs <- specs(x)

    if (type == 'global') {
        innerLoop <- function(row) {
            fired <- fire(row, rules, tnorm=goedel.tnorm, onlyAnte=TRUE)
            fired <- unlist(fired)
            maxFired <- max(fired)
            most <- which(fired == maxFired)
            selected <- .perceiveGlobal(rules[most], n, v, specs)
            if (!any(selected)) {
                return(.defuzzEmptyRulebase(values))
            }
            conseq <- unlist(consequents(rules[most][selected]))
            degrees <- aggregateConsequents(conseq, rep(maxFired, length(conseq)), partition)
            res <- defuzz(degrees, values, type='dee')
            return(res)
        }
    } else {
        innerLoop <- function(row) {
            fired <- fire(row, rules, tnorm=goedel.tnorm, onlyAnte=TRUE)
            fired <- unlist(fired)
            most <- which(fired > 0)
            selected <- .perceiveLocal(rules[most], n, v, specs, fired[most])
            if (!any(selected)) {
                return(.defuzzEmptyRulebase(values))
            }
            conseq <- unlist(consequents(rules[most][selected]))
            degrees <- aggregateConsequents(conseq, fired[most][selected], partition)
            res <- defuzz(degrees, values, type='dee')
            return(res)
        }
    }

    i <- NULL
    if (parallel) {
        result <- foreach(i=seq_len(nrow(x))) %dopar% { innerLoop(x[i, , drop=FALSE]) }
    } else {
        result <- foreach(i=seq_len(nrow(x))) %do% { innerLoop(x[i, , drop=FALSE]) }
    }

    return(unlist(result))
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Deprecated functions to compute membership degrees of numeric fuzzy sets
#'
#' These functions compute membership degrees of numeric fuzzy sets with
#' triangular or raised-cosine shape. These functions are *deprecated*.
#' Please use [triangular()] or [raisedcosine()] functions instead.
#'
#' @param x A numeric vector to be transformed.
#' @param lo A lower bound (can be -Inf).
#' @param center A peak value.
#' @param hi An upper bound (can be Inf).
#' @return A numeric vector of membership degrees of `x` to a fuzzy set with the shape
#' determined with `lo`, `center`, `hi`.
#' @seealso [triangular()], [raisedcosine()]
#' @author Michal Burda
#' @export
triangle <- function(x, lo, center, hi) {
    .Deprecated('triangular')
    if (lo > center || lo > hi) {
        stop('"lo" must be the lower-bound of the interval <lo, hi>')
    }
    if (hi < center || hi < lo) {
        stop('"hi" must be the upper-bound of the interval <lo, hi>')
    }
    if (center < lo || center > hi) {
        stop('"center" must be within the interval <lo, hi>')
    }

    return(sapply(x, .singleTriangle, lo, center, hi))
}


#' @rdname triangle
#' @export
raisedcos <- function(x, lo, center, hi) {
    .Deprecated('raisedcosine')
    if (lo > center || lo > hi) {
        stop('"lo" must be the lower-bound of the interval <lo, hi>')
    }
    if (hi < center || hi < lo) {
        stop('"hi" must be the upper-bound of the interval <lo, hi>')
    }
    if (center < lo || center > hi) {
        stop('"center" must be within the interval <lo, hi>')
    }

    return(sapply(x, .singleRaisedcos, lo, center, hi))
}


.singleTriangle <- function(x, lo, center, hi) {
    if (x < center) {
        if (lo == -Inf) {
            return(1)
        }
        return(pmax(0, (x - lo) / (center - lo)))
    } else if (x == center) {
        return(1)
    } else {
        if (hi == Inf) {
            return(1)
        }
        return(pmax(0, (hi - x) / (hi - center)))
    }
}


.singleRaisedcos <- function(x, lo, center, hi) {
    if (x < lo || x > hi) {
        return(0)
    } else if (x < center) {
        if (lo == -Inf) {
            return(1)
        }
        return((cos((x - center) * pi / (center - lo)) + 1) / 2)
    } else if (x == center) {
        return(1)
    } else {
        if (hi == Inf) {
            return(1)
        }
        return((cos((x - center) * pi / (hi - center)) + 1) / 2)
    }
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Callback-based Multiplication of Matrices
#'
#' Perform a custom multiplication of the matrices `x` and `y` by
#' using the callback function `f`.
#'
#' For a matrix `x` of size \eqn{(u,v)} and a matrix `y` of size
#' \eqn{(v,w)}, `mult` calls the function `f` \eqn{uw}-times to
#' create a resulting matrix of size \eqn{(u,w)}.  Each \eqn{(i,j)}-th element
#' of the resulting matrix is obtained from a call of the function `f`
#' with `x`'s \eqn{i}-th row and `y`'s \eqn{j}-th column passed as its arguments.
#'
#' @param x A first matrix. The number of columns must match with the number of
#' rows of the `y` matrix.
#' @param y A second matrix. The number of rows must match with the number of
#' columns of the `x` matrix.
#' @param f A function to be applied to the matrices in order to compute the
#' multiplication.  It must accept at least two arguments.
#' @param ...  Additional arguments that are passed to the function `f`.
#' @return A matrix with \eqn{v} rows and \eqn{w} columns, where \eqn{v} is the
#' number of rows of `x` and \eqn{w} is the number of columns of `y`.
#' @author Michal Burda
#' @seealso [compose()]
#' @keywords models robust multivariate
#' @examples
#'     x <- matrix(runif(24, -100, 100), ncol=6)
#'     y <- matrix(runif(18, -100, 100), nrow=6)
#'
#'     mult(x, y, function(xx, yy) sum(xx * yy)) # the same as "x %*% y"
#'
#' @export mult
mult <- function(x, y, f, ...) {
    .mustBeMatrix(x)
    .mustBeMatrix(y)
    .mustNotBeZeroLength(x)
    .mustNotBeZeroLength(y)
    .mustBe(ncol(x) == nrow(y), "Number of columns of 'x' must equal to the number of rows of 'y'")
    .mustBeFunction(f)
    .mustBe(length(formals(f)) >= 2, "'f' must be a function with at least 2 arguments")

    if (is.null(colnames(x))) {
        colnames(x) <- rownames(y)
    }
    if (is.null(rownames(y))) {
        rownames(y) <- colnames(x)
    }
    if (!is.null(colnames(x))) {
        .mustBe(all(colnames(x) == rownames(y)), "Column names of 'x' must be equal to row names of 'y'")
    }

    ff <- f
    if (length(list(...)) > 0) {
        ff <- function(xx, yy) {
            f(xx, yy, ...)
        }
    }
    res <- .Call('_lfl_mult', x, y, ff, PACKAGE='lfl')
    colnames(res) <- colnames(y)
    rownames(res) <- rownames(x)
    res
}

test_that('mult', {
    set.seed(4523)
    x <- matrix(runif(24, -100, 100), ncol=6)
    y <- matrix(runif(18, -100, 100), nrow=6)
    res <- mult(x, y, function(xx, yy) sum(xx * yy))
    expect_true(is.numeric(res))
    expect_true(is.matrix(res))
    expect_equal(res, x %*% y)

    rownames(x) <- rev(LETTERS[seq_len(nrow(x))])
    colnames(x) <- letters[seq_len(ncol(x))]
    rownames(y) <- letters[seq_len(nrow(y))]
    colnames(y) <- LETTERS[seq_len(ncol(y))]
    res <- mult(x, y, function(xx, yy) sum(xx * yy))
    expect_equal(colnames(res), colnames(y))
    expect_equal(rownames(res), rownames(x))

    colnames(x) <- rev(letters[seq_len(ncol(x))])
    expect_error(mult(x, y, function(xx, yy) sum(xx * yy)))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Searching for fuzzy association rules
#'
#' This function searches the given [fsets()] object `d` for all
#' fuzzy association rules that satisfy defined constraints. It returns a list
#' of fuzzy association rules together with some statistics characterizing them
#' (such as support, confidence etc.).
#'
#' The function searches data frame `d` for fuzzy association rules that
#' satisfy conditions specified by the parameters.
#'
#' @param d An object of class [fsets()] - it is basically a matrix
#' where columns represent the fuzzy sets and values are the membership
#' degrees. For creation of such object, use [fcut()] or [lcut()] function.
#' @param lhs Indices of fuzzy attributes that may appear on the left-hand-side
#' (LHS) of association rules, i.e., in the antecedent.
#' @param rhs Indices of fuzzy attributes that may appear on the
#' right-hand-side (RHS) of association rules, i.e., in the consequent.
#' @param tnorm A t-norm to be used for computation of conjunction of fuzzy
#' attributes. (Only the starting letters of "lukasiewicz",
#' "goedel" and "goguen" are allowed).
#' @param n The non-negative number of rules to be found. If zero, the function
#' returns all rules satisfying the given conditions. If positive, only the
#' `n` best rules are returned. The criterion of what is "best" is
#' specified with the `best` argument.
#' @param best Specifies the measure according to which the rules are ordered
#' from best to worst. This argument is used mainly in combination with the
#' `n` argument. Currently, only a single value ("confidence") can be used.
#' @param minSupport The minimum support degree of a rule. Rules with support
#' below that number are filtered out. It must be a numeric value from interval
#' \eqn{[0, 1]}. See below for details on how the support degree is computed.
#' @param minConfidence The minimum confidence degree of a rule. Rules with
#' confidence below that number are filtered out.  It must be a numeric value
#' from interval \eqn{[0, 1]}. See below for details on how the confidence degree is
#' computed.
#' @param maxConfidence Maximum confidence threshold. After finding a rule that
#' has confidence degree above the `maxConfidence` threshold, no other
#' rule is produced based on adding some additional attribute to its antecedent
#' part. I.e., if "Sm.age & Me.age => Sm.height" has confidence above the
#' `maxConfidence` threshold, no other rule containing "Sm.age & Me.age"
#' will be produced regardless of its interest measures.
#'
#' If you want to disable this feature, set `maxConfidence` to 1.
#' @param maxLength Maximum allowed length of the rule, i.e., maximum
#' number of predicates that are allowed on the left-hand + right-hand side of the rule. If
#' negative, the maximum length of rules is unlimited.
#' @param numThreads Number of threads used to perform the algorithm in
#' parallel. If greater than 1, the OpenMP library (not to be confused with
#' Open MPI) is used for parallelization. Please note that there are known
#' problems of using OpenMP together with other means of parallelization that
#' may be used within R. Therefore, if you plan to use the `searchrules`
#' function with some external parallelization mechanisms such as the
#' `doMC` library, make sure that `numThreads` equals 1. This
#' feature is available only on systems that have the OpenMP library installed.
#' @param trie Whether or not to use the internal mechanism of Tries. If FALSE,
#' then a rule may appear in the output that is a descendant of a rule that
#' has confidence above the `maxConfidence` threshold.
#'
#' Tries consume a large amount of memory, so if you encounter problems with
#' insufficient memory, set this argument to FALSE. On the other hand, the size
#' of the result (if `n` is set to 0) can be very high if trie is set to
#' FALSE.
#' @return A list of the following elements: `rules` and `statistics`.
#'
#' `rules` is a list of mined fuzzy association rules. Each element of
#' that list is a character vector with consequent attribute being on the first
#' position.
#'
#' `statistics` is a data frame of statistical characteristics about mined
#' rules. Each row corresponds to a rule in the `rules` list. Let us
#' consider a rule "a & b => c", let \eqn{\otimes} be a t-norm specified with
#' the `tnorm` parameter and \eqn{i} goes over all rows of a data table
#' `d`. Then columns of the `statistics` data frame are as follows:
#' * support: a rule's support degree: \eqn{1/nrow(d) * \sum_{\forall i} a(i) \otimes b(i) \otimes c(i)}
#' * lhsSupport: a support of rule's antecedent (LHS): \eqn{1/nrow(d) * \sum_{\forall i} a(i) \otimes b(i)}
#' * rhsSupport: a support of rule's consequent (RHS): \eqn{1/nrow(d) * \sum_{\forall i} c(i)}
#' * confidence: a rule's confidence degree: \eqn{support / lhsSupport}
#' @author Michal Burda
#' @seealso [fcut()], [lcut()], [farules()], [fsets()], [pbld()]
#' @keywords models robust multivariate
#' @examples
#'
#'   d <- lcut(CO2)
#'   searchrules(d, lhs=1:ncol(d), rhs=1:ncol(d))
#'
#' @export searchrules
searchrules <- function(d,
                        lhs=2:ncol(d),
                        rhs=1,
                        tnorm=c("goedel", "goguen", "lukasiewicz"),
                        n=100,
                        best=c("confidence"),
                        minSupport=0.02,
                        minConfidence=0.75,
                        maxConfidence=1,
                        maxLength=4,
                        numThreads=1,
                        trie=(maxConfidence < 1)) {

    .mustBe(is.fsets(d), "'d' must be an instance of class 'fsets'")
    .mustBe(ncol(d) >= 2, "'d' must have at least 2 columns")
    .mustBe(nrow(d) >= 1, "'d' must not be empty")
    .mustBeNumericVector(lhs)
    .mustBe(min(lhs) >= 1 && max(lhs) <= ncol(d), "'lhs' must contain valid indexes of the columns of 'd'")
    .mustBeNumericVector(rhs)
    .mustBe(min(rhs) >= 1 && max(rhs) <= ncol(d), "'rhs' must contain valid indexes of the columns of 'd'")
    .mustBeNumericScalar(n)
    .mustBe(n >= 0, "'n' must be a non-negative number (zero means unlimited number of rules)")
    .mustBeNumericScalar(minSupport)
    .mustBeInInterval(minSupport, 0, 1)
    .mustBeNumericScalar(minConfidence)
    .mustBeInInterval(minConfidence, 0, 1)
    .mustBeNumericScalar(maxConfidence)
    .mustBeInInterval(maxConfidence, 0, 1)
    .mustBe(maxConfidence >= minConfidence, "'maxConfidence' must be greater or equal to 'minConfidence'")

    if (maxLength < 0) {
        maxLength <- -1
    }

    .mustBeNumericScalar(numThreads)
    .mustBe(numThreads >= 1, "'numThreads' must be positive integer number")

    tnorm <- match.arg(tnorm)
    if (tnorm == 'goedel') {
        tnorm <- 'minimum'
    } else if (tnorm == 'goguen') {
        tnorm <- 'product'
    }

    best = match.arg(best)

    config <- list(vars=as.numeric(as.factor(vars(d))),
                   minSupport=minSupport,
                   minConfidence=minConfidence,
                   maxConfidence=maxConfidence,
                   maxLength=maxLength,
                   lhs=lhs - 1,
                   rhs=rhs - 1,
                   tnorm=tnorm,
                   n=n,
                   best=best,
                   numThreads=numThreads,
                   trie=trie)

    result <- .Call("_lfl_search", d, config, PACKAGE="lfl")

    map <- colnames(d)
    names(map) <- 1:ncol(d)

    rules <- lapply(result$rules, function(x) {
                        x <- x + 1
                        x[] <- map[unlist(x)]
                        return(x)
                    })
    if (is.null(result$statistics) || length(result$statistics) <= 0) {
        stats <- matrix(0, nrow=0, ncol=0)
    } else {
        stats <-matrix(unlist(result$statistics),
                       byrow=TRUE,
                       nrow=length(result$statistics))
        colnames(stats) <- c('support', 'lhsSupport', 'rhsSupport', 'confidence', 'lift', 'loLift', 'hiLift')

        stats <- stats[, c('support', 'lhsSupport', 'rhsSupport', 'confidence'), drop=FALSE]

        if (tnorm == 'product') {
            stats <- cbind(stats,
                           lift=stats[, 'confidence'] / stats[, 'rhsSupport'],
                           conviction=(1 - stats[, 'rhsSupport']) / (1 - stats[, 'confidence']))
        }
    }
    res <- farules(rules=rules,
                   statistics=stats)
    return(res)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Context for linguistic expressions
#'
#' A context describes a range of allowed values for a data column.
#'
#' A context describes a range of allowed values for a data column. For that,
#' only the borders of the interval, i.e., minimum and maximum, are usually
#' needed, but we use contexts to hold more additional information that is
#' crucial for the construction of linguistic expressions.
#'
#' Currently, four different contexts are supported that determine the types of
#' possible linguistic expressions, as constructed with [lingexpr()].
#' Unilateral or bilateral context is allowed in the variants of trichotomy or
#' pentachotomy. Trichotomy distinguishes three points in the interval: the
#' lowest value, highest value, and center. Pentachotomy adds lower center and
#' upper center to them. As opposite to unilateral, the bilateral context
#' handles explicitly the negative values. That is, bilateral context expects
#' some middle point, the origin (usually 0), around which the positive and
#' negative values are placed.
#'
#' Concretely, the type of the context determines the allowed atomic
#' expressions as follows:
#' * `ctx3`: trichotomy (low, center, high) enables atomic expressions:
#'   small, medium, big;
#' * `ctx5`: pentachotomy (low, lowerCenter, center, upperCenter, high) enables
#'   atomic expressions: small, lower medium, medium, upper medium, big;
#' * `ctx3bilat`: bilateral trichotomy (negMax, negCenter, origin, center, max)
#'   enables atomic expressions: negative big, negative medium, negative small,
#'   zero, small, medium, big;
#' * `ctx5bilat`: bilateral pentachotomy (negMax, negCenter, origin, center,
#'   max) enables atomic expressions: negative big, negative medium, negative
#'   small, zero, small, medium, big.
#'
#' The `as.ctx*` functions return instance of the appropriate class. The
#' functions perform the conversion so that missing points of the new context
#' are computed from the old context that is being transformed. In the
#' subsequent table, rows represent compatible values of different context
#' types:
#'
#' \tabular{llll}{
#' ctx3   \tab ctx5        \tab ctx3bilat \tab ctx5bilat      \cr
#'        \tab             \tab negMax    \tab negMax         \cr
#'        \tab             \tab           \tab negUpperCenter \cr
#'        \tab             \tab negCenter \tab negCenter      \cr
#'        \tab             \tab           \tab negLowerCenter \cr
#' low    \tab low         \tab origin    \tab origin         \cr
#'        \tab lowerCenter \tab           \tab lowerCenter    \cr
#' center \tab center      \tab center    \tab center \cr
#'        \tab upperCenter \tab           \tab upperCenter \cr
#' high   \tab high        \tab max       \tab max
#' }
#'
#' The `as.ctx*` conversion is performed by replacing values by rows, as
#' indicated in the table above.  When converting from a context with less
#' points to a context with more points (e.g., from unilateral to bilateral, or
#' from trichotomy to pentachotomy), missing points are computed as follows:
#' * `center` is computed as a mean of `origin` (or `low`) and `max` (or `high`).
#' * `lowerCenter` is computed as a mean of `origin` (or `low`) and `center`.
#' * `upperCenter` is computed as a mean of `mas` (or `high`) and `center`.
#' * negative points (such as `negMax`, `negCenter` etc.) are computed
#' symmetrically around `origin` to the corresponding positive points.
#'
#' The code `as.ctx*` functions allow the parameter to be also a numeric
#' vector of size equal to the number of points required for the given context
#' type, i.e. 3 (`ctx3`), 5 (`ctx3bilat`, `ctx5`), or 9 (`ctx5bilat`).
#'
#' @param negMax Lowest negative value of a bilateral context.
#' @param negUpperCenter A typical negative value between `negMax` and
#' `negCenter`.
#' @param negCenter A negative middle value.
#' @param negLowerCenter A typical negative value between `negCenter` and
#' `negOrigin`.
#' @param origin Origin, i.e., the initial point of the bilateral context. It is
#' typically a value of zero.
#' @param lowerCenter A typical positive value between origin and center.
#' @param center A positive middle value of a bilateral context, or simply a
#' middle value of an unilateral context.
#' @param upperCenter A typical positive value between center and maximum.
#' @param max Highest value of a bilateral context.
#' @param low Lowest value of an unilateral context.
#' @param high Highest value of an unilateral context.
#' @param relCenter A relative quantity used to compute the `negCenter`
#' and/or `center`, if they are not specified explicitly. The sensible
#' value is 0.5 for context symmetric around center, or 0.42 as proposed by
#' Novak.
#' @param x A value to be examined or converted. For `as.ctx*`, it can be
#' an instance of any `ctx*` class or a numeric vector of size equal to
#' the number of points required for the given context type.
#' @return `ctx*` and `as.ctx*` return an instance of the appropriate
#' class. `is.ctx*` returns `TRUE` or `FALSE`.
#' @author Michal Burda
#' @seealso [minmax()], [lingexpr()], [horizon()], [hedge()], [fcut()], [lcut()]
#' @keywords models robust
#' @examples
#'     ctx3(low=0, high=10)
#'     as.ctx3bilat(ctx3(low=0, high=10))
#' @name ctx
NULL


#' @rdname ctx
#' @export
ctx3 <- function(low=0,
                 center=low+(high-low)*relCenter,
                 high=1,
                 relCenter=0.5) {
    .mustBeNumericScalar(low)
    .mustBeNumericScalar(center)
    .mustBeNumericScalar(high)
    .mustBeNumericScalar(relCenter)
    .mustBe(0 < relCenter && relCenter < 1, "'relCenter' must be a number between 0 and 1")
    .mustBe(low < center, "'low' must be lower than 'center'")
    .mustBe(center < high, "'center' must be lower than 'high'")

    structure(c(low, center, high),
              names=c('low', 'center', 'high'),
              class=c('ctx3', 'numeric'))
}


#' @rdname ctx
#' @export
ctx3bilat <- function(negMax=-1,
                      negCenter=origin+(negMax-origin)*relCenter,
                      origin=0,
                      center=origin+(max-origin)*relCenter,
                      max=1,
                      relCenter=0.5) {
    .mustBeNumericScalar(negMax)
    .mustBeNumericScalar(negCenter)
    .mustBeNumericScalar(origin)
    .mustBeNumericScalar(center)
    .mustBeNumericScalar(max)
    .mustBeNumericScalar(relCenter)
    .mustBe(0 < relCenter && relCenter < 1, "'relCenter' must be a number between 0 and 1")
    .mustBe(negMax < negCenter, "'negMax' must be lower than 'negCenter'")
    .mustBe(negCenter < origin, "'negCenter' must be lower than 'origin'")
    .mustBe(origin < center, "'origin' must be lower than 'center'")
    .mustBe(center < max, "'center' must be lower than 'max'")

    structure(c(negMax, negCenter, origin, center, max),
              names=c('negMax', 'negCenter', 'origin', 'center', 'max'),
              class=c('ctx3bilat', 'numeric'))
}


#' @rdname ctx
#' @export
ctx5 <- function(low=0,
                 lowerCenter=mean(c(low, center)),
                 center=low+(high-low)*relCenter,
                 upperCenter=mean(c(center, high)),
                 high=1,
                 relCenter=0.5) {
    .mustBeNumericScalar(low)
    .mustBeNumericScalar(lowerCenter)
    .mustBeNumericScalar(center)
    .mustBeNumericScalar(upperCenter)
    .mustBeNumericScalar(high)
    .mustBeNumericScalar(relCenter)
    .mustBe(0 < relCenter && relCenter < 1, "'relCenter' must be a number between 0 and 1")
    .mustBe(low < lowerCenter, "'low' must be lower than 'lowerCenter'")
    .mustBe(lowerCenter < center, "'lowerCenter' must be lower than 'center'")
    .mustBe(center < upperCenter, "'center' must be lower than 'upperCenter'")
    .mustBe(upperCenter < high, "'upperCenter' must be lower than 'high'")

    structure(c(low, lowerCenter, center, upperCenter, high),
              names=c('low', 'lowerCenter', 'center', 'upperCenter', 'high'),
              class=c('ctx5', 'numeric'))
}


#' @rdname ctx
#' @export
ctx5bilat <- function(negMax=-1,
                      negUpperCenter=mean(c(negCenter, negMax)),
                      negCenter=origin+(negMax-origin)*relCenter,
                      negLowerCenter=mean(c(origin, negCenter)),
                      origin=0,
                      lowerCenter=mean(c(origin, center)),
                      center=origin+(max-origin)*relCenter,
                      upperCenter=mean(c(center, max)),
                      max=1,
                      relCenter=0.5) {
    .mustBeNumericScalar(negMax)
    .mustBeNumericScalar(negUpperCenter)
    .mustBeNumericScalar(negCenter)
    .mustBeNumericScalar(negLowerCenter)
    .mustBeNumericScalar(origin)
    .mustBeNumericScalar(lowerCenter)
    .mustBeNumericScalar(center)
    .mustBeNumericScalar(upperCenter)
    .mustBeNumericScalar(max)
    .mustBeNumericScalar(relCenter)
    .mustBe(0 < relCenter && relCenter < 1, "'relCenter' must be a number between 0 and 1")
    .mustBe(negMax < negUpperCenter, "'negMax' must be lower than 'negUpperCenter'")
    .mustBe(negUpperCenter < negCenter, "'negUpperCenter' must be lower than 'negCenter'")
    .mustBe(negCenter < negLowerCenter, "'negCenter' must be lower than 'negLowerCenter'")
    .mustBe(negLowerCenter < origin, "'negLowerCenter' must be lower than 'origin'")
    .mustBe(origin < lowerCenter, "'origin' must be lower than 'lowerCenter'")
    .mustBe(lowerCenter < center, "'lowerCenter' must be lower than 'center'")
    .mustBe(center < upperCenter, "'center' must be lower than 'upperCenter'")
    .mustBe(upperCenter < max, "'upperCenter' must be lower than 'max'")

    structure(c(negMax, negUpperCenter, negCenter, negLowerCenter, origin, lowerCenter, center, upperCenter, max),
              names=c('negMax', 'negUpperCenter', 'negCenter', 'negLowerCenter', 'origin',
                      'lowerCenter', 'center', 'upperCenter', 'max'),
              class=c('ctx5bilat', 'numeric'))
}


#' @rdname ctx
#' @export
as.ctx3 <- function(x) UseMethod("as.ctx3")

#' @rdname ctx
#' @export
as.ctx3.ctx3 <- identity

#' @rdname ctx
#' @export
as.ctx3.ctx3bilat <- function(x) ctx3(x['origin'], x['center'], x['max'])

#' @rdname ctx
#' @export
as.ctx3.ctx5 <- function(x) ctx3(x['low'], x['center'], x['high'])

#' @rdname ctx
#' @export
as.ctx3.ctx5bilat <- function(x) ctx3(x['origin'], x['center'], x['max'])

#' @rdname ctx
#' @export
as.ctx3.default <- function(x) {
    x <- as.vector(x)
    .mustBe(length(x) == 3, "'x' must be of length 3")
    ctx3(x[1], x[2], x[3])
}


#' @rdname ctx
#' @export
as.ctx3bilat <- function(x) UseMethod("as.ctx3bilat")

#' @rdname ctx
#' @export
as.ctx3bilat.ctx3bilat <- identity

#' @rdname ctx
#' @export
as.ctx3bilat.ctx3 <- function(x) ctx3bilat(2*x['low']-x['high'],
                                           2*x['low']-x['center'],
                                           x['low'],
                                           x['center'],
                                           x['high'])

#' @rdname ctx
#' @export
as.ctx3bilat.ctx5 <- function(x) as.ctx3bilat(as.ctx3(x))

#' @rdname ctx
#' @export
as.ctx3bilat.ctx5bilat <- function(x) ctx3bilat(x['negMax'], x['negCenter'], x['origin'], x['center'], x['max'])

#' @rdname ctx
#' @export
as.ctx3bilat.default <- function(x) {
    x <- as.vector(x)
    .mustBe(length(x) == 5, "'x' must be of length 5")
    ctx3bilat(x[1], x[2], x[3], x[4], x[5])
}


#' @rdname ctx
#' @export
as.ctx5 <- function(x) UseMethod("as.ctx5")

#' @rdname ctx
#' @export
as.ctx5.ctx5 <- identity

#' @rdname ctx
#' @export
as.ctx5.ctx3 <- function(x) ctx5(x['low'],
                                 mean(c(x['low'], x['center'])),
                                 x['center'],
                                 mean(c(x['center'], x['high'])), x['high'])

#' @rdname ctx
#' @export
as.ctx5.ctx3bilat <- function(x) as.ctx5(as.ctx3(x))

#' @rdname ctx
#' @export
as.ctx5.ctx5bilat <- function(x) ctx5(x['origin'], x['lowerCenter'], x['center'], x['upperCenter'], x['max'])

#' @rdname ctx
#' @export
as.ctx5.default <- function(x) {
    x <- as.vector(x)
    .mustBe(length(x) == 5, "'x' must be of length 5")
    ctx5(x[1], x[2], x[3], x[4], x[5])
}


#' @rdname ctx
#' @export
as.ctx5bilat <- function(x) UseMethod("as.ctx5bilat")

#' @rdname ctx
#' @export
as.ctx5bilat.ctx5bilat <- identity

#' @rdname ctx
#' @export
as.ctx5bilat.ctx3 <- function(x) as.ctx5bilat(as.ctx5(x))

#' @rdname ctx
#' @export
as.ctx5bilat.ctx3bilat <- function(x) ctx5bilat(x['negMax'],
                                                mean(c(x['negMax'], x['negCenter'])),
                                                x['negCenter'],
                                                mean(c(x['negCenter'], x['origin'])),
                                                x['origin'],
                                                mean(c(x['origin'], x['center'])),
                                                x['center'], mean(c(x['center'], x['max'])),
                                                x['max'])

#' @rdname ctx
#' @export
as.ctx5bilat.ctx5 <- function(x) {
    shift <- 2*x['low']
    ctx5bilat(shift-x['high'], shift-x['upperCenter'], shift-x['center'], shift-x['lowerCenter'],
              x['low'], x['lowerCenter'], x['center'], x['upperCenter'], x['high'])
}

#' @rdname ctx
#' @export
as.ctx5bilat.default <- function(x) {
    x <- as.vector(x)
    .mustBe(length(x) == 9, "'x' must be of length 9")
    ctx5bilat(x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9])
}

#' @rdname ctx
#' @export
is.ctx3 <- function(x) {
    return(inherits(x, 'ctx3') &&
           is.numeric(x) &&
           length(x) == 3)
}


#' @rdname ctx
#' @export
is.ctx3bilat <- function(x) {
    return(inherits(x, 'ctx3bilat') &&
           is.numeric(x) &&
           length(x) == 5)
}


#' @rdname ctx
#' @export
is.ctx5 <- function(x) {
    return(inherits(x, 'ctx5') &&
           is.numeric(x) &&
           length(x) == 5)
}


#' @rdname ctx
#' @export
is.ctx5bilat <- function(x) {
    return(inherits(x, 'ctx5bilat') &&
           is.numeric(x) &&
           length(x) == 9)
}

test_that("minmax ctx3", {
    r <- minmax(0:100)
    expect_true(is.ctx3(r))
    expect_equal(as.vector(r), c(0, 50, 100))
    expect_equal(names(r), c('low', 'center', 'high'))
})


test_that("minmax ctx3 with custom value", {
    r <- minmax(0:100, high=80)
    expect_true(is.ctx3(r))
    expect_equal(as.vector(r), c(0, 40, 80))
    expect_equal(names(r), c('low', 'center', 'high'))
})


test_that("minmax ctx3 with custom value 2", {
    r <- minmax(0:100, relCenter=0.4)
    expect_true(is.ctx3(r))
    expect_equal(as.vector(r), c(0, 40, 100))
    expect_equal(names(r), c('low', 'center', 'high'))
})


test_that("minmax ctx3bilat", {
    r <- minmax(0:100, type='ctx3bilat')
    expect_true(is.ctx3bilat(r))
    expect_equal(as.vector(r), c(0, 25, 50, 75, 100))
    expect_equal(names(r), c('negMax', 'negCenter', 'origin', 'center', 'max'))
})


test_that("minmax ctx5", {
    r <- minmax(0:100, type='ctx5')
    expect_true(is.ctx5(r))
    expect_equal(as.vector(r), c(0, 25, 50, 75, 100))
    expect_equal(names(r), c('low', 'lowerCenter', 'center', 'upperCenter', 'high'))
})


test_that("minmax ctx5bilat", {
    r <- minmax(0:100, type='ctx5bilat')
    expect_true(is.ctx5bilat(r))
    expect_equal(as.vector(r), c(0, 12.5, 25, 37.5, 50, 62.5, 75, 87.5, 100))
    expect_equal(names(r), c('negMax', 'negUpperCenter', 'negCenter', 'negLowerCenter', 'origin', 'lowerCenter', 'center', 'upperCenter', 'max'))
})

test_that("ctx3", {
    ex <- function(x1, x2, x3) {
        structure(c(low=x1, center=x2, high=x3),
                  class=c('ctx3', 'numeric'))
    }

    context <- ctx3()
    expect_true(is.ctx3(context))
    expect_false(is.ctx3bilat(context))
    expect_false(is.ctx5(context))
    expect_false(is.ctx5bilat(context))

    context <- ctx3(1, 2, 4)
    expect_true(is.ctx3(context))
    expect_false(is.ctx3bilat(context))
    expect_false(is.ctx5(context))
    expect_false(is.ctx5bilat(context))

    expect_equal(ctx3(),
                 ex(0, 0.5, 1))
    expect_equal(ctx3(high=5),
                 ex(0, 2.5, 5))
    expect_equal(ctx3(high=5, relCenter=0.8),
                 ex(0, 4, 5))
    expect_equal(ctx3(low=3, high=5),
                 ex(3, 4, 5))
    expect_equal(ctx3(low=3, high=5, relCenter=0.75),
                 ex(3, 4.5, 5))
    expect_equal(ctx3(low=3, center=3.1, high=5),
                 ex(3, 3.1, 5))

    expect_error(ctx3(0, 0, 1))
    expect_error(ctx3(0, 1, 1))
    expect_error(ctx3(2, 1, 0))
    expect_error(ctx3(low=-10, high=10, relCenter=0))
    expect_error(ctx3(low=-10, high=10, relCenter=1))
    expect_error(ctx3(low=-10, high=10, relCenter=-0.5))
    expect_error(ctx3(low=-10, high=10, relCenter=1.5))

    expect_equal(as.ctx3(ctx3(1, 2, 4)),
                 ex(1, 2, 4))
    expect_equal(as.ctx3(ctx5(1, 2, 4, 8, 16)),
                 ex(1, 4, 16))
    expect_equal(as.ctx3(ctx5bilat(1, 2, 4, 8, 16, 32, 64, 128, 256)),
                 ex(16, 64, 256))
    expect_equal(as.ctx3(ctx3bilat(1, 2, 4, 8, 16)),
                 ex(4, 8, 16))

    expect_equal(as.ctx3(c(1, 2, 3)),
                 ex(1, 2, 3))
    expect_error(as.ctx3(runif(100)))
})


test_that("ctx5", {
    ex <- function(x1, x2, x3, x4, x5) {
        structure(c(low=x1, lowerCenter=x2, center=x3, upperCenter=x4, high=x5),
                  class=c('ctx5', 'numeric'))
    }

    context <- ctx5()
    expect_false(is.ctx3(context))
    expect_false(is.ctx3bilat(context))
    expect_true(is.ctx5(context))
    expect_false(is.ctx5bilat(context))

    context <- ctx5(1, 2, 4, 5, 8)
    expect_false(is.ctx3(context))
    expect_false(is.ctx3bilat(context))
    expect_true(is.ctx5(context))
    expect_false(is.ctx5bilat(context))

    expect_equal(ctx5(),
                 ex(0, 0.25, 0.5, 0.75, 1))
    expect_equal(ctx5(high=5),
                 ex(0, 1.25, 2.5, 3.75, 5))
    expect_equal(ctx5(high=5, relCenter=0.8),
                 ex(0, 2, 4, 4.5, 5))
    expect_equal(ctx5(low=3, high=5),
                 ex(3, 3.5, 4, 4.5, 5))
    expect_equal(ctx5(low=3, high=5, relCenter=0.75),
                 ex(3, 3.75, 4.5, 4.75, 5))
    expect_equal(ctx5(low=3, center=3.1, high=5),
                 ex(3, 3.05, 3.1, 4.05, 5))
    expect_equal(ctx5(1, 2, 3, 4, 5),
                 ex(1, 2, 3, 4, 5))

    expect_error(ctx5(0, 0, 1, 2, 3))
    expect_error(ctx5(0, 1, 1, 2, 3))
    expect_error(ctx5(0, 1, 2, 2, 3))
    expect_error(ctx5(0, 1, 2, 3, 3))
    expect_error(ctx5(low=-10, high=10, relCenter=0))
    expect_error(ctx5(low=-10, high=10, relCenter=1))
    expect_error(ctx5(low=-10, high=10, relCenter=-0.5))
    expect_error(ctx5(low=-10, high=10, relCenter=1.5))

    expect_equal(as.ctx5(ctx5(1, 1.5, 2, 3, 4)),
                 ex(1, 1.5, 2, 3, 4))
    expect_equal(as.ctx5(ctx3(1, 2, 4)),
                 ex(1, 1.5, 2, 3, 4))
    expect_equal(as.ctx5(ctx5bilat(1, 2, 4, 8, 16, 32, 64, 128, 256)),
                 ex(16, 32, 64, 128, 256))
    expect_equal(as.ctx5(ctx3bilat(1, 2, 4, 8, 16)),
                 ex(4, 6, 8, 12, 16))

    expect_equal(as.ctx5(c(1, 2, 3, 4, 5)),
                 ex(1, 2, 3, 4, 5))
    expect_error(as.ctx3(runif(100)))
})


test_that("ctx3bilat", {
    ex <- function(x1, x2, x3, x4, x5) {
        structure(c(negMax=x1, negCenter=x2, origin=x3, center=x4, max=x5),
                  class=c('ctx3bilat', 'numeric'))
    }

    context <- ctx3bilat()
    expect_false(is.ctx3(context))
    expect_true(is.ctx3bilat(context))
    expect_false(is.ctx5(context))
    expect_false(is.ctx5bilat(context))

    context <- ctx3bilat(1, 2, 4, 5, 8)
    expect_false(is.ctx3(context))
    expect_true(is.ctx3bilat(context))
    expect_false(is.ctx5(context))
    expect_false(is.ctx5bilat(context))

    expect_equal(ctx3bilat(),
                 ex(-1, -0.5, 0, 0.5, 1))
    expect_equal(ctx3bilat(max=5),
                 ex(-1, -0.5, 0, 2.5, 5))
    expect_equal(ctx3bilat(negMax=-5),
                 ex(-5, -2.5, 0, 0.5, 1))
    expect_equal(ctx3bilat(max=5, relCenter=0.8),
                 ex(-1, -0.8, 0, 4, 5))
    expect_equal(ctx3bilat(negMax=-5, relCenter=0.8),
                 ex(-5, -4, 0, 0.8, 1))
    expect_equal(ctx3bilat(origin=3, max=5),
                 ex(-1, 1, 3, 4, 5))
    expect_equal(ctx3bilat(negMax=-5, origin=3, max=5),
                 ex(-5, -1, 3, 4, 5))
    expect_equal(ctx3bilat(negMax=3, negCenter=3.1, origin=4, center=4.1, max=5),
                 ex(3, 3.1, 4, 4.1, 5))

    expect_error(ctx3bilat(0, 0, 1, 2, 3))
    expect_error(ctx3bilat(0, 1, 1, 2, 3))
    expect_error(ctx3bilat(0, 1, 2, 2, 3))
    expect_error(ctx3bilat(0, 1, 2, 3, 3))
    expect_error(ctx3bilat(5, 1, 2, 3, 4))
    expect_error(ctx3bilat(relCenter=0))
    expect_error(ctx3bilat(relCenter=1))
    expect_error(ctx3bilat(relCenter=-0.5))
    expect_error(ctx3bilat(relCenter=1.5))

    expect_equal(as.ctx3bilat(ctx3bilat(-2, 0, 1, 2, 4)),
                 ex(-2, 0, 1, 2, 4))
    expect_equal(as.ctx3bilat(ctx3(1, 2, 4)),
                 ex(-2, 0, 1, 2, 4))
    expect_equal(as.ctx3bilat(ctx5bilat(1, 2, 4, 8, 16, 32, 64, 128, 256)),
                 ex(1, 4, 16, 64, 256))
    expect_equal(as.ctx3bilat(ctx5(16, 32, 64, 128, 256)),
                 ex(-224, -32, 16, 64, 256))

    expect_equal(as.ctx3bilat(c(1, 2, 3, 4, 5)),
                 ex(1, 2, 3, 4, 5))
    expect_error(as.ctx3bilat(runif(100)))
})


test_that("ctx5bilat", {
    ex <- function(x1, x2, x3, x4, x5, x6, x7, x8, x9) {
        structure(c(negMax=x1, negUpperCenter=x2, negCenter=x3, negLowerCenter=x4, origin=x5,
                    lowerCenter=x6, center=x7, upperCenter=x8, max=x9),
                  class=c('ctx5bilat', 'numeric'))
    }

    context <- ctx5bilat()
    expect_false(is.ctx3(context))
    expect_false(is.ctx3bilat(context))
    expect_false(is.ctx5(context))
    expect_true(is.ctx5bilat(context))

    context <- ctx5bilat(1, 2, 4, 5, 8, 9, 10, 11, 12)
    expect_false(is.ctx3(context))
    expect_false(is.ctx3bilat(context))
    expect_false(is.ctx5(context))
    expect_true(is.ctx5bilat(context))

    expect_equal(ctx5bilat(),
                 ex(-1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1))
    expect_equal(ctx5bilat(max=5),
                 ex(-1, -0.75, -0.5, -0.25, 0, 1.25, 2.5, 3.75, 5))
    expect_equal(ctx5bilat(negMax=-5),
                 ex(-5, -3.75, -2.5, -1.25, 0, 0.25, 0.5, 0.75, 1))
    expect_equal(ctx5bilat(max=5, relCenter=0.8),
                 ex(-1, -0.9, -0.8, -0.4, 0, 2, 4, 4.5, 5))
    expect_equal(ctx5bilat(negMax=-5, relCenter=0.8),
                 ex(-5, -4.5, -4, -2, 0, 0.4, 0.8, 0.9, 1))
    expect_equal(ctx5bilat(origin=3, max=5),
                 ex(-1, 0, 1, 2, 3, 3.5, 4, 4.5, 5))
    expect_equal(ctx5bilat(negMax=-5, origin=3, max=5),
                 ex(-5, -3, -1, 1, 3, 3.5, 4, 4.5, 5))
    expect_equal(ctx5bilat(negMax=1, negUpperCenter=1.5, negCenter=2,
                        negLowerCenter=2.5, origin=4, lowerCenter=4.2,
                        center=4.5, upperCenter=4.8, max=5),
                 ex(1, 1.5, 2, 2.5, 4, 4.2, 4.5, 4.8, 5))

    expect_error(ctx5bilat(0, 0, 1, 2, 3, 4, 5, 6, 7))
    expect_error(ctx5bilat(0, 1, 1, 2, 3, 4, 5, 6, 7))
    expect_error(ctx5bilat(0, 1, 2, 2, 3, 4, 5, 6, 7))
    expect_error(ctx5bilat(0, 1, 2, 3, 3, 4, 5, 6, 7))
    expect_error(ctx5bilat(0, 1, 2, 3, 4, 4, 5, 6, 7))
    expect_error(ctx5bilat(0, 1, 2, 3, 4, 5, 5, 6, 7))
    expect_error(ctx5bilat(0, 1, 2, 3, 4, 5, 6, 6, 7))
    expect_error(ctx5bilat(0, 1, 2, 3, 4, 5, 6, 7, 7))
    expect_error(ctx5bilat(relCenter=0))
    expect_error(ctx5bilat(relCenter=1))
    expect_error(ctx5bilat(relCenter=-0.5))
    expect_error(ctx5bilat(relCenter=1.5))

    expect_equal(as.ctx5bilat(ctx5bilat(1, 2, 4, 8, 16, 32, 64, 128, 256)),
                 ex(1, 2, 4, 8, 16, 32, 64, 128, 256))
    expect_equal(as.ctx5bilat(ctx3(1, 2, 4)),
                 ex(-2, -1, 0, 0.5, 1, 1.5, 2, 3, 4))
    expect_equal(as.ctx5bilat(ctx3bilat(1, 2, 4, 8, 16)),
                 ex(1, 1.5, 2, 3, 4, 6, 8, 12, 16))
    expect_equal(as.ctx5bilat(ctx5(16, 32, 64, 128, 256)),
                 ex(-224, -96, -32, 0, 16, 32, 64, 128, 256))

    expect_equal(as.ctx5bilat(c(1, 2, 3, 4, 5, 6, 7, 8, 9)),
                 ex(1, 2, 3, 4, 5, 6, 7, 8, 9))
    expect_error(as.ctx5bilat(runif(100)))
})

test_that('horizon ctx3', {
    ctx <- ctx3(0, 0.5, 1)
    expect_equal(horizon(ctx, 'sm')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(1, 1, 1, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'me')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'bi')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 1, 1, 1, NA, NaN))

    expect_equal(horizon(ctx, 'sm')(seq(0, 1, length.out=11)),
                 c(seq(1, 0, length.out=6), rep(0, 5)))
    expect_equal(horizon(ctx, 'me')(seq(0, 1, length.out=11)),
                 c(seq(0, 1, length.out=6), seq(0.8, 0, length.out=5)))
    expect_equal(horizon(ctx, 'bi')(seq(0, 1, length.out=11)),
                 c(rep(0, 5), seq(0, 1, length.out=6)))

    expect_error(horizon(ctx, 'lm'))
    expect_error(horizon(ctx, 'um'))
    expect_error(horizon(ctx, 'neg.sm'))
    expect_error(horizon(ctx, 'neg.lm'))
    expect_error(horizon(ctx, 'neg.um'))
    expect_error(horizon(ctx, 'neg.me'))
    expect_error(horizon(ctx, 'neg.bi'))
    expect_error(horizon(ctx, 'neg.lm'))
    expect_error(horizon(ctx, 'neg.um'))
    expect_error(horizon(ctx, 'ze'))
})


test_that('horizon ctx5', {
    ctx <- ctx5(0, 0.25, 0.5, 0.75, 1)
    expect_equal(horizon(ctx, 'sm')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(1, 1, 1, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'lm')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'me')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'um')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'bi')(c(-Inf, -10, 0, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 1, 1, 1, NA, NaN))

    expect_equal(horizon(ctx, 'sm')(seq(0, 1, length.out=21)),
                 c(seq(1, 0, length.out=11), rep(0, 10)))
    expect_equal(horizon(ctx, 'lm')(seq(0, 1, length.out=21)),
                 c(seq(0, 1, length.out=6), seq(0.8, 0, length.out=5), rep(0, 10)))
    expect_equal(horizon(ctx, 'me')(seq(0, 1, length.out=21)),
                 c(seq(0, 1, length.out=11), seq(0.9, 0, length.out=10)))
    expect_equal(horizon(ctx, 'um')(seq(0, 1, length.out=21)),
                 c(rep(0, 10), seq(0, 1, length.out=6), seq(0.8, 0, length.out=5)))
    expect_equal(horizon(ctx, 'bi')(seq(0, 1, length.out=21)),
                 c(rep(0, 10), seq(0, 1, length.out=11)))

    expect_error(horizon(ctx, 'neg.sm'))
    expect_error(horizon(ctx, 'neg.lm'))
    expect_error(horizon(ctx, 'neg.um'))
    expect_error(horizon(ctx, 'neg.me'))
    expect_error(horizon(ctx, 'neg.bi'))
    expect_error(horizon(ctx, 'ze'))
})


test_that('horizon ctx3bilat', {
    ctx <- ctx3bilat(-1, -0.5, 0, 0.5, 1)
    expect_equal(horizon(ctx, 'neg.bi')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(1, 1, 1, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'neg.me')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'neg.sm')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'ze')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'sm')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'me')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'bi')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 1, 1, 1, NA, NaN))

    expect_equal(horizon(ctx, 'neg.bi')(seq(-1, 1, length.out=21)),
                 c(seq(1, 0, length.out=6), rep(0, 15)))
    expect_equal(horizon(ctx, 'neg.me')(seq(-1, 1, length.out=21)),
                 c(seq(0, 1, length.out=6), seq(0.8, 0, length.out=5), rep(0, 10)))
    expect_equal(horizon(ctx, 'neg.sm')(seq(-1, 1, length.out=21)),
                 c(rep(0, 5), seq(0, 1, length.out=6), rep(0, 10)))
    expect_equal(horizon(ctx, 'ze')(seq(-1, 1, length.out=21)),
                 c(rep(0, 5), seq(0, 1, length.out=6), seq(0.8, 0, length.out=5), rep(0, 5)))
    expect_equal(horizon(ctx, 'sm')(seq(-1, 1, length.out=21)),
                 c(rep(0, 10), seq(1, 0, length.out=6), rep(0, 5)))
    expect_equal(horizon(ctx, 'me')(seq(-1, 1, length.out=21)),
                 c(rep(0, 10), seq(0, 1, length.out=6), seq(0.8, 0, length.out=5)))
    expect_equal(horizon(ctx, 'bi')(seq(-1, 1, length.out=21)),
                 c(rep(0, 15), seq(0, 1, length.out=6)))

    expect_error(horizon(ctx, 'lm'))
    expect_error(horizon(ctx, 'um'))
    expect_error(horizon(ctx, 'neg.lm'))
    expect_error(horizon(ctx, 'neg.um'))
})


test_that('horizon ctx5bilat', {
    ctx <- ctx5bilat(-1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1)
    expect_equal(horizon(ctx, 'neg.bi')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(1, 1, 1, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'neg.me')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'neg.sm')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'ze')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'sm')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'me')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 0, 0, 0, NA, NaN))
    expect_equal(horizon(ctx, 'bi')(c(-Inf, -10, -1, 1, 10, Inf, NA, NaN)), 
                 c(0, 0, 0, 1, 1, 1, NA, NaN))

    expect_equal(horizon(ctx, 'neg.bi')(seq(-1, 1, length.out=41)),
                 c(seq(1, 0, length.out=11), rep(0, 30)))
    expect_equal(horizon(ctx, 'neg.um')(seq(-1, 1, length.out=41)),
                 c(seq(0, 1, length.out=6), seq(0.8, 0, length.out=5), rep(0, 30)))
    expect_equal(horizon(ctx, 'neg.me')(seq(-1, 1, length.out=41)),
                 c(seq(0, 1, length.out=11), seq(0.9, 0, length.out=10), rep(0, 20)))
    expect_equal(horizon(ctx, 'neg.lm')(seq(-1, 1, length.out=41)),
                 c(rep(0, 10), seq(0, 1, length.out=6), seq(0.8, 0, length.out=5), rep(0, 20)))
    expect_equal(horizon(ctx, 'neg.sm')(seq(-1, 1, length.out=41)),
                 c(rep(0, 10), seq(0, 1, length.out=11), rep(0, 20)))
    expect_equal(horizon(ctx, 'ze')(seq(-1, 1, length.out=41)),
                 c(rep(0, 10), seq(0, 1, length.out=11), seq(0.9, 0, length.out=10), rep(0, 10)))
    expect_equal(horizon(ctx, 'sm')(seq(-1, 1, length.out=41)),
                 c(rep(0, 20), seq(1, 0, length.out=11), rep(0, 10)))
    expect_equal(horizon(ctx, 'lm')(seq(-1, 1, length.out=41)),
                 c(rep(0, 20), seq(0, 1, length.out=6), seq(0.8, 0, length.out=5), rep(0, 10)))
    expect_equal(horizon(ctx, 'me')(seq(-1, 1, length.out=41)),
                 c(rep(0, 20), seq(0, 1, length.out=11), seq(0.9, 0, length.out=10)))
    expect_equal(horizon(ctx, 'um')(seq(-1, 1, length.out=41)),
                 c(rep(0, 30), seq(0, 1, length.out=6), seq(0.8, 0, length.out=5)))
    expect_equal(horizon(ctx, 'bi')(seq(-1, 1, length.out=41)),
                 c(rep(0, 30), seq(0, 1, length.out=11)))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Compute rule base coverage of data
#'
#' This function computes rule base coverage, i.e., a an average of maximum
#' membership degree at which each row of data fires the rules in rule base.
#'
#' Let \eqn{f_{ij}} be a truth value of \eqn{i}-th rule on \eqn{j}-th row of
#' data \code{x}.  Then \eqn{m_j = max(f_{.j})} is a maximum truth value that
#' is reached for the \eqn{j}-th data row with the rule base. Then the rule
#' base coverage is a mean of that truth values, i.e.  \eqn{rbcoverage =
#' mean(m_.)}.
#'
#' @param x Data for the rules to be evaluated on. Could be either a numeric
#' matrix or numeric vector.  If matrix is given then the rules are evaluated
#' on rows. Each value of the vector or column of the matrix represents a
#' predicate - it's numeric value represents the truth values (values in the
#' interval \eqn{[0, 1]}).
#' @param rules Either an object of class "farules" or list of character
#' vectors where each vector is a rule with consequent being the first element
#' of the vector. Elements of the vectors (predicate names) must correspond to
#' the `x`'s names (of columns if `x` is a matrix).
#' @param tnorm A character string representing a triangular norm to be used
#' (either `"goedel"`, `"goguen"`, or `"lukasiewicz"`) or an
#' arbitrary function that takes a vector of truth values and returns a t-norm
#' computed of them.
#' @param onlyAnte TRUE if only antecedent-part of a rule should be evaluated.
#' Antecedent-part of a rule are all predicates in rule vector starting from
#' the 2nd position. (First element of a rule is the consequent - see above.)
#'
#' If FALSE, then the whole rule will be evaluated (antecedent part together
#' with consequent).
#' @return A numeric value of the rule base coverage of given data.
#' @author Michal Burda
#' @seealso [fire()], [reduce()]
#' @references M. Burda, M. Štěpnička, Reduction of Fuzzy Rule Bases Driven by
#' the Coverage of Training Data, in: Proc. 16th World Congress of the
#' International Fuzzy Systems Association and 9th Conference of the European
#' Society for Fuzzy Logic and Technology (IFSA-EUSFLAT 2015), Advances in
#' Intelligent Systems Research, Atlantic Press, Gijon, 2015.
#' @keywords models robust multivariate
#' @examples
#'
#'     x <- matrix(1:20 / 20, nrow=2)
#'     colnames(x) <- letters[1:10]
#'
#'     rules <- list(c('a', 'c', 'e'),
#'                   c('b'),
#'                   c('d', 'a'),
#'                   c('c', 'a', 'b'))
#'     rbcoverage(x, rules, "goguen", TRUE)  # returns 1
#'
#'
#'     rules <- list(c('d', 'a'),
#'                   c('c', 'a', 'b'))
#'     rbcoverage(x, rules, "goguen", TRUE)  # returns 0.075)
#'
#' @export
rbcoverage <- function(x,
                       rules,
                       tnorm=c("goedel", "goguen", "lukasiewicz"),
                       onlyAnte=TRUE) {
    if (is.vector(x)) {
        x <- matrix(x, nrow=1, dimnames=list(NULL, names(x)))
    }
    .mustBe(is.matrix(x) && is.numeric(x), "'x' must be a numeric vector or matrix")
    .mustBeInInterval(x, 0, 1)

    if (is.farules(rules)) {
        rules <- rules$rules
    } else if (is.vector(rules) && is.character(rules)) {
        rules <- list(rules)
    }

    .mustBe(is.list(rules), "'rules' must be a list of rules, a valid farules object, or a character vector")

    fired <- fire(x, rules, tnorm, onlyAnte)
    return(mean(do.call('pmax', fired)))
}

test_that('rbcoverage on data vector', {
    x <- 1:10 / 10
    names(x) <- letters[1:10]
    rules <- list(c('a', 'c', 'e'),
                  c('b'),
                  c('d', 'a'),
                  c('c', 'a', 'b'))

    expect_equal(rbcoverage(x, rules, "goguen", TRUE),
                 1)
})


test_that('rbcoverage on data matrix', {
    x <- matrix(1:20 / 20, nrow=2)
    colnames(x) <- letters[1:10]

    rules <- list(c('a', 'c', 'e'),
                  c('b'),
                  c('d', 'a'),
                  c('c', 'a', 'b'))
    expect_equal(rbcoverage(x, rules, "goguen", TRUE),
                 1)

    rules <- list(c('a', 'c', 'e'),
                  c('d', 'a'),
                  c('c', 'a', 'b'))
    expect_equal(rbcoverage(x, rules, "goguen", TRUE),
                 0.13125)

    rules <- list(c('d', 'a'),
                  c('c', 'a', 'b'))
    expect_equal(rbcoverage(x, rules, "goguen", TRUE),
                 0.075)

    rules <- list(c('d', 'a'))
    expect_equal(rbcoverage(x, rules, "goguen", TRUE),
                 0.075)
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Print an instance of the [frbe()] class
#'
#' Format an object of the [frbe()] class into human readable form
#' and print it to the output.
#'
#' Format an object of the [frbe()] class into human readable form
#' and print it to the output.
#'
#' @param x An instance of [frbe()] class
#' @param ... Unused.
#' @return None.
#' @author Michal Burda
#' @seealso [frbe()]
#' @references Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
#' Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
#' SYST. 2015.
#' @keywords models robust
#' @examples
#'
#'   # prepare data (from the forecast package)
#'   library(forecast)
#'   horizon <- 10
#'   train <- wineind[-1 * (length(wineind)-horizon+1):length(wineind)]
#'   test <- wineind[(length(wineind)-horizon+1):length(wineind)]
#'   f <- frbe(ts(train, frequency=frequency(wineind)), h=horizon)
#'   print(f)
#'   print(test)
#'
#' @export
print.frbe <- function(x, ...) {
    .mustBe(is.frbe(x), "'x' is not a valid 'frbe' object")
    xx <- x
    class(xx) <- NULL
    print(xx, ...)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Compute Mean Absolute Scaled Error (MASE)
#'
#' MASE is computed as \eqn{sum(abs(validation - forecast)) / sum(abs(validation[-1] - validation[-n])) / (n/(n-1))}.
#' @param forecast A numeric vector of forecasted values
#' @param validation A numeric vector of actual (real) values
#' @return A Mean Absolute Scaled Error (MASE)
#' @author Michal Burda
#' @seealso [rmse()], [smape()], [frbe()]
#' @export
mase <- function(forecast, validation) {
  .mustBeNumericVector(forecast)
  .mustBeNumericVector(validation)
  .mustNotHaveNA(forecast)
  .mustNotHaveNA(validation)
  .mustBe(length(forecast) == length(validation), "Length of 'forecast' and 'validation' must be equal")
  .mustBe(length(forecast) > 1, "Length of both 'forecast' and 'validation' must be greater than zero")

  n <- length(validation)

  #mase <- mean(abs(validation - forecast)) / mean(abs(history[-length(history)] - history[-1]))
  mase <- sum(abs(validation - forecast)) / sum(abs(validation[-1] - validation[-n])) / (n/(n-1))

  return(mase)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Test whether `x` is a valid object of the S3 `frbe` class
#'
#' Test whether `x` has a valid format for the objects of the S3 `frbe` class.
#'
#' This function tests whether `x` inherits from `frbe` i.e., whether
#' it is a list with the following elements: `forecasts` data frame,
#' `features` data frame, `weights` vector, and `mean` vector.
#' Instances of the S3 class `frbe` are usually created by the [frbe()] function.
#'
#' @param x An object being tested.
#' @return `TRUE` if `x` is a valid `frbe` object and `FALSE` otherwise.
#' @author Michal Burda
#' @seealso [frbe()]
#' @references Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
#' Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
#' SYST. 2015.
#' @keywords models robust
#' @export
is.frbe <- function(x) {
    return(inherits(x, 'frbe') &&
           is.list(x) &&
           is.data.frame(x$forecasts) &&
           is.data.frame(x$features) &&
           is.vector(x$weights) &&
           is.vector(x$mean))
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Convert fuzzy set into a crisp numeric value
#'
#' Take a discretized fuzzy set (i.e., a vector of membership degrees and a vector
#' of numeric values that correspond to that degrees) and perform a selected
#' type of defuzzification, i.e., conversion of the fuzzy set into a single
#' crisp value.
#'
#' Function converts input fuzzy set into a crisp value. The definition of
#' input fuzzy set is provided by the arguments `degrees` and
#' `values`. These arguments should be numeric vectors of the same length,
#' the former containing membership degrees in the interval \eqn{[0, 1]} and
#' the latter containing the corresponding crisp values: i.e., `values[i]`
#' has a membership degree `degrees[i]`.
#'
#' @param degrees A fuzzy set in the form of a numeric vector of membership
#' degrees of values provided as the `values` argument.
#' @param values A universe for the fuzzy set.
#' @param type Type of the requested defuzzification method. The possibilities are:
#' * `'mom'`: Mean of Maxima - maximum membership degrees are
#'   found and a mean of values that correspond to that degrees is returned;
#' * `'fom'`: First of Maxima - first value with maximum membership
#'   degree is returned;
#' * `'lom'`: Last of Maxima - last value with maximum membership degree
#'    is returned;
#' * `'dee'`: Defuzzification of Evaluative Expressions - method used
#'   by the [pbld()] inference mechanism that combines the former three
#'   approaches according to the shape of the `degrees` vector:
#'   If `degrees` is non-increasing then `'lom'` type is used,
#'   if it is non-decreasing then `'fom'` is applied, else `'mom'` is selected;
#' * `'cog'`: Center of Gravity - the result is a mean of `values` weighted by `degrees`;
#' * `'exp1'`: Experimental 1.
#' @return A defuzzified value.
#' @author Michal Burda
#' @seealso [fire()], [aggregateConsequents()], [perceive()], [pbld()], [infer()], [fcut()], [lcut()]
#' @keywords models robust
#' @examples
#'
#' # returns mean of maxima, i.e., mean of 6, 7, 8
#' defuzz(c(0, 0, 0, 0.1, 0.3, 0.9, 0.9, 0.9, 0.2, 0),
#'        1:10,
#'        type='mom')
#'
#' @export defuzz
defuzz <- function(degrees,
                   values,
                   type=c('mom', 'fom', 'lom', 'dee', 'cog', 'expun', 'expw1', 'expw2')) {
    .mustBeNumericVector(degrees)
    .mustBeNumericVector(values)
    .mustBe(length(degrees) == length(values), "The length of 'degrees' and 'values' must be the same")
    .mustBe(length(degrees) >= 3, "The length of 'degrees' must be at least 3")
    .mustBe(min(degrees) >= 0 && max(degrees) <= 1, "Values of 'degrees' must be truth values in the interval [0,1]")

    type <- match.arg(type)

    if (type == 'cog') {
        return(weighted.mean(values, degrees))

    } else if (type %in% c('expun', 'expw1', 'expw2')) {
        alpha <- unique(degrees)
        alpha <- setdiff(alpha, 0)
        alpha <- sort(alpha)
        center <- sapply(alpha, function(a) mean(values[degrees >= a]))
        p <- seq(1, length(alpha) - 1)
        n <- seq(2, length(alpha))
        alphadiff <- alpha[n] - alpha[p]
        mins <- pmin(center[p], center[n])
        maxs <- pmax(center[p], center[n])

        if (type == 'expw1') {
            weight <- alphadiff * alpha[n]
        } else  if (type == 'expw2') {
            weight <- alphadiff * (alpha[n])^2
        } else {
            weight <- alphadiff
        }

        return(weighted.mean((mins + maxs) / 2, weight))

    } else {
        o <- order(values)
        values <- values[o]
        degrees <- degrees[o]
        i <- which(degrees == max(degrees))

        if (type == 'dee') {
            l <- length(degrees)
            diff <- degrees[-1] - degrees[-l]
            if (all(diff >= 0)) {
                type <- 'fom'
            } else if (all(diff <= 0)) {
                type <- 'lom'
            } else {
                type <- 'mom'
            }
        }

        if (type == 'fom') {
            return(values[min(i)])

        } else if (type == 'lom') {
            return(values[max(i)])

        } else if (type == 'mom') {
            return(mean(values[i]))

        } else {
            stop("Unknown defuzzification type")
        }
    }
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2026 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Perform Mamdani or implicative fuzzy inference with given rule-base on given
#' dataset
#'
#' Take a set of rules (a rule-base) and perform Mamdani or implicative
#' inference on each row of a given matrix (or the [fsets()] object) of truth
#'  values.
#'
#' Perform Mamdani or implicative fuzzy inference with given rule-base `rules`
#' on each row of input `x`. Columns of `x` are truth values of predicates that
#' appear in the antecedent part of `rules`. The `partition` matrix determines
#' the shape of predicates in consequents: each row represents a point in the
#' output fuzzy set for which membership degrees are computed.
#'
#' For both inference types, the predicates in the antecedent part of rules are
#' combined using a t-norm (`alg$pt`) of the given algebra `alg` (see also the
#' `fire()` function).
#'
#' For the Mamdani inference (`type = "mamdani"`), the antecedent firing
#' degrees are combined with the consequent partition using a t-norm (`alg$pt`)
#' of the selected algebra `alg`, and the resulting fuzzy sets are aggregated
#' with a t-conorm (`alg$ps`).
#'
#' For the implicative inference (`type = "implicative"`), the antecedent
#' firing degrees are combined with the consequent partition using a residuum
#' (`alg$r`) of the selected algebra `alg`, and the resulting fuzzy sets are
#' aggregated with a t-norm (`alg$pi`).
#'
#' @param x Input to the inference. It must be a numeric matrix with columns
#' representing fuzzy sets (truth values of predicates). Each row represents a
#' single case of inference. Columns should be named after predicates in rules'
#' antecedents. The values must be in the interval \eqn{[0, 1]}. `x` can also be
#' an object of class [fsets()] (e.g., created by using the [fcut()] or [lcut()]
#' functions).
#' @param rules A rule-base (a.k.a. linguistic description) either in the form
#' of the [farules()] object or as a list of character vectors where
#' each element is a fuzzy set name (a predicate) and thus each such vector
#' forms a rule. The first element of each rule is the consequent predicate,
#' and the remaining elements are antecedent predicates.
#' @param partition A numeric matrix (or a [fsets()] object) with columns that
#' are consequents in `rules`. The column names of `partition` must correspond
#' to consequent predicate names used in `rules`. Each row represents a point in the
#' output domain.
#' @param alg The algebra to use for the inference. It can be either a character
#' string (`"goedel"`, `"goguen"`, or `"lukasiewicz"`) or an instance of the
#' [algebra()] class.
#' @param type The type of inference to use. It can be either `"mamdani"` or
#' `"implicative"`.
#' @param parallel Whether the processing should be run in parallel or not.
#' Parallelization is implemented using the [foreach::foreach()]
#' package. The parallel environment must be set properly in advance, e.g., with
#' the [doMC::registerDoMC()] function.
#' @return A numeric matrix of inferred membership degrees. The rows of the
#' result correspond to rows of the `x` argument, the columns of the result
#' correspond to rows of the `partition` argument.
#' @author Michal Burda
#' @seealso [pbld()], [fire()], [aggregateConsequents()], [defuzz()], [algebra()], [farules()], [lcut()]
#' @keywords models robust
#' @examples
#' # Create rules: first element is the consequent, the rest are antecedents
#' rules <- list(
#'     c("High",   "temp_high",   "pressure_high"),
#'     c("Medium", "temp_medium", "pressure_medium"),
#'     c("Low",    "temp_low")
#' )
#'
#' # Create input data (fuzzy set membership degrees)
#' # (Alternatively, fcut() or lcut() functions can be used)
#' x <- matrix(
#'     c(0.9, 0.8, 0.2, 0.3, 0.1,
#'       0.1, 0.2, 0.4, 0.5, 0.8),
#'     nrow = 2,
#'     byrow = TRUE
#' )
#' colnames(x) <- c("temp_high", "pressure_high",
#'                   "temp_medium", "pressure_medium", "temp_low")
#'
#' # Create output partition
#' # (Alternatively, fcut() or lcut() functions can be used)
#' partition <- matrix(
#'     c(1.0, 0.2, 0.0,
#'       0.8, 0.5, 0.1,
#'       0.4, 1.0, 0.4,
#'       0.1, 0.5, 0.8,
#'       0.0, 0.1, 1.0),
#'     nrow = 5,
#'     byrow = TRUE
#' )
#' colnames(partition) <- c("Low", "Medium", "High")
#'
#' # Mamdani inference with Goedel algebra
#' infer(x, rules, partition, alg = "goedel", type = "mamdani")
#'
#' # Implicative inference with Lukasiewicz algebra
#' infer(x, rules, partition, alg = "lukasiewicz", type = "implicative")
#'
#' @export
infer <- function(x,
                  rules,
                  partition,
                  alg=c('goedel', 'goguen', 'lukasiewicz'),
                  type=c('mamdani', 'implicative'),
                  parallel=FALSE) {
    .mustBe(is.matrix(x), "'x' must be a matrix")

    if (is.farules(rules)) {
        rules <- rules$rules
    } else if (is.vector(rules) && is.character(rules)) {
        rules <- list(rules)
    }

    .mustBe(is.list(rules), "'rules' must be a list of rules or an instance of the 'farules' class")
    .mustBeNumericMatrix(partition)
    .mustNotBeZeroLength(partition)
    .mustBe(!is.null(colnames(partition)) && is.character(colnames(partition)), "'partition' must have colnames")

    diffAttr <- setdiff(unique(unlist(antecedents(rules))), colnames(x))
    if (length(diffAttr) > 0) {
        stop(paste0("'rules' contains predicates in antecedent-part that are not present in 'colnames(x)': ",
                    paste(diffAttr, collapse=', ', sep='')))
    }

    if (is.character(alg)) {
        alg <- match.arg(alg)
        alg <- algebra(alg)
    }
    .mustBe(is.algebra(alg), "'alg' must be either one of 'goedel', 'goguen', lukasiewicz', or an instance of class 'algebra'")

    type <- match.arg(type)
    emptyResult <- NULL
    innerOp <- NULL
    outerOp <- NULL
    if (type == 'mamdani') {
        emptyResult <- rep(0, nrow(partition))
        innerOp <- alg$pt
        outerOp <- alg$ps
    } else if (type == 'implicative') {
        emptyResult <- rep(1, nrow(partition))
        innerOp <- alg$r
        outerOp <- alg$pi
    } else {
        stop("Unknown inference type")
    }

    if (length(rules) <= 0) {
        return(matrix(rep(emptyResult, nrow(x)), byrow=TRUE, nrow=nrow(x)))
    }

    innerLoop <- function(row) {
        fired <- fire(row, rules, tnorm=alg$pt, onlyAnte=TRUE)
        fired <- unlist(fired)
        if (all(fired == 0)) {
            return(emptyResult)
        }
        conseq <- unlist(consequents(rules))
        res <- aggregateConsequents(conseq,
                                    fired,
                                    partition,
                                    firing = innerOp,
                                    aggreg = outerOp)
        return(res)
    }

    i <- NULL
    if (parallel) {
        result <- foreach(i=seq_len(nrow(x))) %dopar% { innerLoop(x[i, , drop=FALSE]) }
    } else {
        result <- foreach(i=seq_len(nrow(x))) %do% { innerLoop(x[i, , drop=FALSE]) }
    }

    matrix(unlist(result), byrow = TRUE, nrow = nrow(x))
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Determine whether the first set `x` of predicates is more specific (or equal)
#' than `y` with respect to `vars` and `specs`.
#'
#' The function takes two character vectors of predicates and determines whether
#' `x` is more specific (or equal w.r.t. the specificity) than `y`. The
#' specificity relation is fully determined with the values of the [vars()] vector
#' and the [specs()] incidence matrix that is encapsulated in the given `fsets` object.
#'
#' Let \eqn{x_i} and \eqn{y_j} represent some predicates of vectors `x` and `y`,
#' respectively. Function assumes that each vector `x` and `y` does not
#' contain two or more predicates with the same value of [vars()].
#'
#' This function returns TRUE iff all of the following conditions hold:
#' * for any \eqn{y_j} there exists \eqn{x_i} such that \eqn{vars[y_j] = vars[x_i]};
#' * for any \eqn{x_i} there either does not exist \eqn{y_j} such that
#'   \eqn{vars[x_i] = vars[y_j]}, or \eqn{x_i = y_j}, or \eqn{specs[x_i, y_j] = 1}.
#'
#' @param x The first character vector of predicates.
#' @param y The second character vector of predicates.
#' @param fsets A valid instance of the [fsets()] class such that all values in `x` and `y`
#' can be found in `colnames(fsets)`
#' @param vars Deprecated parameter must be `NULL`.
#' @param specs Deprecated parameter must be `NULL`.
#' @return TRUE or FALSE (see description).
#' @author Michal Burda
#' @seealso [perceive()], [pbld()], [fsets()], [vars()], [specs()]
#' @keywords models robust
#' @examples
#'     # prepare fsets object
#'     v <- c(rep('a', 3), rep('b', 3), rep('c', 3), rep('d', 3))
#'     s <- matrix(c(0,1,0, 0,0,0, 0,0,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'
#'                   0,0,0, 0,1,0, 0,0,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'
#'                   0,0,0, 0,0,0, 0,1,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'
#'                   0,0,0, 0,0,0, 0,0,0, 0,1,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0,
#'                   0,0,0, 0,0,0, 0,0,0, 0,0,0),
#'                 byrow=TRUE,
#'                 ncol=12)
#'     m <- matrix(0, nrow=1, ncol=12)
#'     colnames(m) <- paste(rep(c('VeSm', 'Sm', 'Bi'), times=4),
#'                          rep(c('a', 'b', 'c', 'd'), each=3),
#'                          sep='.')
#'     f <- fsets(m, v, s)
#'
#'
#'     # returns TRUE
#'     is.specific(c('VeSm.a', 'Bi.c'),
#'                 c('VeSm.a', 'Bi.c'),
#'                 f)
#'
#'     # returns TRUE (x and y swapped return FALSE)
#'     is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
#'                 c('Sm.a', 'Bi.c', 'Sm.d'),
#'                 f)
#'
#'     # returns TRUE (x and y swapped return FALSE)
#'     is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
#'                 c('VeSm.a', 'Bi.c'),
#'                 f)
#'
#'     # returns TRUE (x and y swapped return FALSE)
#'     is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
#'                 character(),
#'                 f)
#'
#'     # returns FALSE
#'     is.specific(c('Sm.a'), c('Bi.c'), f)
#'
#'     # returns FALSE
#'     is.specific(c('VeSm.a', 'Sm.c'),
#'                 c('Sm.a', 'Bi.c'),
#'                 f)
#' @export
is.specific <- function(x, y, fsets, vars=NULL, specs=NULL) {
    if (!(is.null(vars) && is.null(specs))) {
        .stop('"vars" and "specs" parameters are defunct. Specify "fsets" parameter instead.')
    }
    .mustBeCharacterVector(x)
    .mustBeCharacterVector(y)
    .mustBe(is.fsets(fsets), '"fsets" must be a valid instance of the "fsets" S3 class')

    n <- colnames(fsets)
    xi <- match(x, n)
    yi <- match(y, n)

    .mustBe(!any(is.na(xi)), 'All values in "x" must be from colnames of "fsets"')
    .mustBe(!any(is.na(yi)), 'All values in "y" must be from colnames of "fsets"')

    if ((length(unique(xi)) != length(xi)) || (length(unique(yi)) != length(yi))) {
        .stop('Unable to work with rules containing the same predicate more times')
    }

    v <- as.integer(factor(vars(fsets)))
    .is.specific(xi, yi, v, specs(fsets))
}


.is.specific <- function(xi, yi, v, specs) {
    return(.Call('_lfl_specificity',
                 xi-1, yi-1,  # C++ array indexing is from 0
                 v, specs, PACKAGE='lfl'))
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


# Create an (intermediate) fuzzy quantifier
#
#' A quantifier is a function that computes a fuzzy truth value of a claim about
#' the quantity. This function creates the <1>-type quantifier. (See the examples
#' below on how to use it as a quantifier of the <1,1> type.)
#'
#' @param quantity the quantity to be evaluated. 'all' computes the degree of
#'   truth to which all elements of the universe have the given property,
#'   'almost.all', #'   'most', and 'many' evaluate whether the property is
#'   present in extremely big, very big, or not small number of elements from
#'   the universe, where these linguistic expressions are internally modeled
#'   using the [lingexpr()] function. 'at.least' quantity requires the 'n'
#'   argument to be specified, as it computes the truth value that at least
#'   \eqn{n} elements from the universe have the given property.
#' @param n the number of elements in the 'at.least n' quantifier
#' @param alg the underlying algebra in which to compute the quantifier.
#'   Note that the algebra must have properly defined the `order` function,
#'   as in the case of 'goedel', 'goguen', or 'lukasiewicz' algebra, (see the
#'   [algebra()] function) or as in the [dragonfly()] or [lowerEst()] algebra.
#' @return A two-argument function, which expects two numeric vectors of equal length
#'   (the vector elements are recycled to ensure equal lengths). The first argument, `x`,
#'   is a vector of membership degrees to be measured, the second argument, `w`, is
#'   the vector of weights to which the element belongs to the universe.
#'
#'   Let \eqn{U} be the set of input vector indices (1 to `length(x)`). Then the quantifier
#'   computes the truth values according to the following formula:
#'   \eqn{\vee_{z \subseteq U} \wedge_{u \in z} (x[u]) \wedge measure(m_z)},
#'   where
#'   \eqn{m_z = sum(w)} for `"some"` and `"at.least` and \eqn{m_z = sum(w[z]) / sum(w)} otherwise.
#'   See [sugeno()] for more details on how the quantifier is evaluated.
#'
#'   Setting `w` to 1 yields to operation of the <1> quantifier as developed by Dvořák et al.
#'   To compute the <1,1> quantifier as developed by Dvořák et al., e.g. "almost all A are B", `w` must
#'   be set again to 1 and `x` to the result of the implication \eqn{A \Rightarrow B}.
#'   To compute the <1,1> quantifier as proposed by Murinová et al., e.g. "almost all A are B",
#'   `x` must be set to the result of the implication \eqn{A \Rightarrow B} and `w` to the membership
#'   degrees of \eqn{A}. See the examples below.
#' @references Dvořák, A., Holčapek, M. L-fuzzy quantifiers of type <1> determined by fuzzy measures.
#'   Fuzzy Sets and Systems vol.160, issue 23, 3425-3452, 2009.
#' @references Dvořák, A., Holčapek, M. Type <1,1> fuzzy quantifiers determined by fuzzy measures.
#'   IEEE International Conference on Fuzzy Systems (FuzzIEEE), 2010.
#' @references Murinová, P., Novák, V. The theory of intermediate quantifiers in fuzzy natural logic
#'   revisited and the model of "Many". Fuzzy Sets and Systems, vol 388, 2020.
#' @author Michal Burda
#' @keywords models robust
#' @seealso [sugeno()], [lingexpr()]
#' @examples
#'   # Dvorak <1> "almost all" quantifier
#'   f <- quantifier('almost.all')
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   f(x=a, w=1)
#'
#'   # Dvorak <1,1> "almost all" quantifier (w set to 1)
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   b <- c(0.2, 1, 0, 0.5, 0.8)
#'   f <- quantifier('almost.all')
#'   f(x=lukas.residuum(a, b), w=1)
#'
#'   # Murinová <1,1> "almost all" quantifier (note w set to a)
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   b <- c(0.2, 1, 0, 0.5, 0.8)
#'   f <- quantifier('almost.all')
#'   f(x=lukas.residuum(a, b), w=a)
#'
#'   # Murinová <1,1> "some" quantifier
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   b <- c(0.2, 1, 0, 0.5, 0.8)
#'   f <- quantifier('some')
#'   f(x=plukas.tnorm(a, b), w=a)
#' @export
quantifier <- function(quantity=c('all', 'almost.all', 'most', 'many', 'few', 'several', 'some', 'at.least'),
                       n=NULL,
                       alg=c('lukasiewicz', 'goedel', 'goguen')) {
    quantity <- match.arg(quantity)
    if (is.character(alg)) {
        alg <- match.arg(alg)
        alg <- algebra(alg)
    }
    .mustBe(is.algebra(alg), "'alg' must be either one of 'goedel', 'goguen', lukasiewicz', or an instance of class 'algebra'")

    m <- NULL
    r <- NULL
    if (quantity == 'all') {
        m <- function(x) { (x >= 1) + 0 }
        r <- TRUE
    } else if (quantity == 'almost.all') {
        m <- lingexpr(context=ctx3(), atomic='bi', hedge='ex')
        r <- TRUE
    } else if (quantity == 'most') {
        m <- lingexpr(context=ctx3(), atomic='bi', hedge='ve')
        r <- TRUE
    } else if (quantity == 'many') {
        m <- lingexpr(context=ctx3(), atomic='sm', hedge='-', negated=TRUE)
        r <- TRUE
    } else if (quantity == 'few') {
        m <- lingexpr(context=ctx3(), atomic='sm', hedge='si', negated=TRUE)
        r <- TRUE
    } else if (quantity == 'several') {
        m <- lingexpr(context=ctx3(), atomic='sm', hedge='ve', negated=TRUE)
        r <- TRUE

    # decreasing measure function makes no sense here
    #} else if (quantity == 'at.most') {
        #m <- function(x) { (x <= n) + 0 }
        #r <- FALSE

    } else if (quantity == 'some') {
        m <- function(x) { (x > 0) + 0 }
        r <- FALSE
    } else if (quantity == 'at.least') {
        .mustBeNumericScalar(n)
        .mustBe(n >= 0)
        m <- function(x) { (x >= n) + 0 }
        r <- FALSE
    } else {
        .stop(paste0('Unknown quantity: ', quantity))
    }

    sugeno(measure=m, relative=r, strong=FALSE, alg=alg)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

.goedel.tnorm <- function(vals) {
    .Call(`_lfl_goedel_tnorm`, vals)
}

.pgoedel.tnorm <- function(list) {
    .Call(`_lfl_pgoedel_tnorm`, list)
}

.lukas.tnorm <- function(vals) {
    .Call(`_lfl_lukas_tnorm`, vals)
}

.plukas.tnorm <- function(list) {
    .Call(`_lfl_plukas_tnorm`, list)
}

.goguen.tnorm <- function(vals) {
    .Call(`_lfl_goguen_tnorm`, vals)
}

.pgoguen.tnorm <- function(list) {
    .Call(`_lfl_pgoguen_tnorm`, list)
}

.goedel.tconorm <- function(vals) {
    .Call(`_lfl_goedel_tconorm`, vals)
}

.pgoedel.tconorm <- function(list) {
    .Call(`_lfl_pgoedel_tconorm`, list)
}

.lukas.tconorm <- function(vals) {
    .Call(`_lfl_lukas_tconorm`, vals)
}

.plukas.tconorm <- function(list) {
    .Call(`_lfl_plukas_tconorm`, list)
}

.goguen.tconorm <- function(vals) {
    .Call(`_lfl_goguen_tconorm`, vals)
}

.pgoguen.tconorm <- function(list) {
    .Call(`_lfl_pgoguen_tconorm`, list)
}

.goedel.residuum <- function(x, y) {
    .Call(`_lfl_goedel_residuum`, x, y)
}

.lukas.residuum <- function(x, y) {
    .Call(`_lfl_lukas_residuum`, x, y)
}

.goguen.residuum <- function(x, y) {
    .Call(`_lfl_goguen_residuum`, x, y)
}

.invol.neg <- function(x) {
    .Call(`_lfl_invol_neg`, x)
}

.strict.neg <- function(x) {
    .Call(`_lfl_strict_neg`, x)
}

.triangle <- function(x, ctx) {
    .Call(`_lfl_triangle`, x, ctx)
}

.raisedcos <- function(x, ctx) {
    .Call(`_lfl_raisedcos`, x, ctx)
}

.hedge <- function(x, p) {
    .Call(`_lfl_hedge`, x, p)
}

.mult <- function(x, y, f) {
    .Call(`_lfl_mult`, x, y, f)
}

.reduce <- function(config) {
    .Call(`_lfl_reduce`, config)
}

.search <- function(rcppData, config) {
    .Call(`_lfl_search`, rcppData, config)
}

.specificity <- function(x, y, vars, specs) {
    .Call(`_lfl_specificity`, x, y, vars, specs)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' A factory function for creation of sugeno-integrals.
#'
#' @param measure A non-decreasing function that assigns a truth value from the
#'   \eqn{[0, 1]} interval to the either relative or absolute quantity
#' @param relative Whether the measure assumes relative or absolute quantity.
#'   Relative quantity is always a number from the \eqn{[0,1]} interval
#' @param strong Whether to use the strong conjunction (`TRUE`) or the weak
#'   conjunction (`FALSE`)
#' @param alg The underlying algebra must be either a string (one from 'lukasiewicz',
#'   'goedel' or 'goguen') or an instance of the S3 class [algebra()].
#' @return A two-argument function, which expects two numeric vectors of equal length
#'   (the vector elements are recycled to ensure equal lengths). The first argument, `x`,
#'   is a vector of membership degrees to be measured, the second argument, `w`, is
#'   the vector of weights.
#'
#'   Let \eqn{U} be the set of input vector indices (1 to `length(x)`). Then the Sugeno integral
#'   computes the truth values according to the following formula:
#'   \eqn{\vee_{z \subseteq U} \wedge_{u \in z} (x[u]) CONJ measure(m_z)},
#'   where
#'   \eqn{m_z = sum(w[z]) / sum(w)} if `relative==TRUE` or \eqn{m_z = sum(w)} if `relative==FALSE`
#'   and where CONJ is a strong conjunction (i.e., `alg$pt`) or a weak conjunction
#'   (i.e., `alg$pi`) according to the `strong` parameter.
#' @author Michal Burda
#' @keywords models robust
#' @seealso [quantifier()], [lingexpr()]
#' @examples
#'   # Dvorak <1> "almost all" quantifier
#'   q <- sugeno(lingexpr(ctx3(), atomic='bi', hedge='ex'))
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   q(x=a, w=1)
#'
#'   # Dvorak <1,1> "almost all" quantifier
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   b <- c(0.2, 1, 0, 0.5, 0.8)
#'   q <- sugeno(lingexpr(ctx3(), atomic='bi', hedge='ex'))
#'   q(x=lukas.residuum(a, b), w=1)
#'
#'   # Murinová <1,1> "almost all" quantifier
#'   a <- c(0.9, 1, 1, 0.2, 1)
#'   b <- c(0.2, 1, 0, 0.5, 0.8)
#'   q <- sugeno(lingexpr(ctx3(), atomic='bi', hedge='ex'))
#'   q(x=lukas.residuum(a, b), w=a)
#' @export
sugeno <- function(measure,
                   relative=TRUE,
                   strong=FALSE,
                   alg=c('lukasiewicz', 'goedel', 'goguen')) {
    .mustBeFunction(measure)
    .mustBeLogicalScalar(relative)
    .mustBeLogicalScalar(strong)
    if (is.character(alg)) {
        alg <- match.arg(alg)
        alg <- algebra(alg)
    }
    .mustBe(is.algebra(alg), "'alg' must be either one of 'goedel', 'goguen', lukasiewicz', or an instance of class 'algebra'")

    relate <- if (relative) function(u) { u / u[length(u)] } else identity
    conj <- if (strong) alg$pt else alg$pi

    function(x, w=1) {
        .mustBeNumericVector(x)
        .mustBeNumericVector(w)
        .mustBe(all(!is.na(w)), "The 'w' argument must not contain NAs")

        l <- max(length(x), length(w))
        x <- rep_len(x, l)
        w <- rep_len(w, l)
        o <- alg$order(x, decreasing=TRUE)
        m <- measure(relate(cumsum(w[o])))
        alg$s(conj(x[o], m))
    }
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Transform data into a `fsets` S3 class using shapes derived from
#' triangles or raised cosines
#'
#' This function creates a set of fuzzy attributes from crisp data. Factors,
#' numeric vectors, matrix or data frame columns are transformed into a set of
#' fuzzy attributes, i.e., columns with membership degrees. Unlike
#' [lcut()], the linguistic approach is not used for transformation,
#' but rather partitioning using regular shapes of the fuzzy sets (such as
#' triangles or raised cosines).
#'
#' The aim of this function is to transform numeric data into a set of fuzzy
#' attributes. The result is in the form of an object of class "fsets", i.e.,
#' a numeric matrix whose columns represent fuzzy sets (fuzzy attributes) with
#' values being the membership degrees.
#'
#' The function behaves differently depending on the type of input `x`.
#'
#' If `x` is a factor or a logical vector (or other non-numeric data) then
#' for each distinct value of an input, a fuzzy set is created, and data would
#' be transformed into crisp membership degrees 0 or 1 only.
#'
#' If `x` is a numeric vector then fuzzy sets are created according to
#' break-points specified in the `breaks` argument with 1st, 2nd and 3rd
#' break-point specifying the first fuzzy set, 2nd, 3rd and 4th break-point
#' specifying the second fuzzy set etc. The shape of the fuzzy set is determined
#' by the `type` argument that may be equal either to a string
#' `'triangle'` or `'raisedcos'` or it could be a function that
#' computes the membership degrees itself (see [triangular()] or
#' [raisedcosine()] functions for details). Additionally, super-sets of
#' these elementary sets may be created by specifying the `merge`
#' argument. Values of this argument specify how many consecutive fuzzy sets
#' should be combined (by using Lukasiewicz's t-conorm) to produce
#' super-sets - see the description of `merge` above.
#'
#' If a matrix (resp. data frame) is provided to this function instead of a
#' single vector, all columns are processed separately as described above and
#' the result is combined with the [cbind.fsets()] function.
#'
#' The function sets up properly the [vars()] and [specs()]
#' properties of the result.
#'
#' @aliases fcut fcut.default fcut.numeric fcut.matrix fcut.data.frame
#' @param x Data to be transformed: a vector, matrix, or data frame.
#' Non-numeric data are allowed.
#' @param breaks This argument determines the break-points of the positions of
#' the fuzzy sets (see also [equidist()]. It should be an ordered vector of
#' numbers such that the \eqn{i}-th index specifies the beginning,
#' \eqn{(i+1)}-th the center, and \eqn{(i+2)}-th the ending of the \eqn{i}-th
#' fuzzy set.
#'
#' I.e. the minimum number of breaks-points is 3; \eqn{n-2} elementary fuzzy
#' sets would be created for \eqn{n} break-points.
#'
#' If considering an i-th fuzzy set (of `type='triangle'`), `x`
#' values lower than \eqn{i}-th break (and greater than \eqn{(i+2)}-th break)
#' would result in zero membership degree, values equal to \eqn{(i+1)}-th break
#' would have membership degree equal 1 and values between them the appropriate
#' membership degree between 0 and 1.
#'
#' The resulting fuzzy sets would be named after the original data by adding
#' dot (".") and a number \eqn{i} of fuzzy set.
#'
#' Unlike [base::cut()], `x` values, that are lower or greater than
#' the given break-points, will have all membership degrees equal to zero.
#'
#' For non-numeric data, this argument is ignored. For `x` being a numeric
#' vector, it must be a vector of numeric values. For `x` being a numeric
#' matrix or data frame, it must be a named list containing a numeric vector
#' for each column - if not, the values are repeated for each column.
#' @param name A name to be added as a suffix to the created fuzzy attribute
#' names. This parameter can be used only if `x` is a vector. If `x`
#' is a matrix or data frame, `name` should be NULL because the fuzzy
#' attribute names are taken from column names of the argument `x`.
#' @param type The type of fuzzy sets to create. Currently, `'triangle'` or
#' `'raisedcos'` may be used. The `type` argument may be also a
#' function with 3 or 4 arguments:
#' * if `type` is a 4-argument function, it is assumed that that it computes
#' membership degrees from values of the first argument while considering
#' the boundaries given by the next 3 arguments;
#' * if `type` is a 3-argument function, it is assumed that it is a factory
#' function similar to [triangular()] or [raisedcosine()], which, from given
#' three boundaries, creates a function that computes membership degrees.
#' @param merge This argument determines whether to derive additional fuzzy
#' sets by merging the elementary fuzzy sets (whose position is determined with
#' the `breaks` argument) into super-sets.  The argument is ignored for
#' non-numeric data in `x`.
#'
#' `merge` may contain any integer number from `1` to
#' `length(breaks) - 2`.  Value `1` means that the elementary fuzzy
#' sets should be present in the output.  Value `2` means that the two
#' consecutive elementary fuzzy sets should be combined by using the Lukasiewic
#' t-conorm, value `3` causes combining three consecutive elementary fuzzy
#' sets etc.
#'
#' The names of the derived (merged) fuzzy sets is derived from the names of
#' the original elementary fuzzy sets by concatenating them with the "|" (pipe)
#' separator.
#' @param parallel Whether the processing should be run in parallel or not.
#' Parallelization is implemented using the [foreach::foreach()] function.
#'  The parallel environment must be set properly in advance, e.g., with
#' the [doMC::registerDoMC()] function.  Currently this argument is
#' applied only if `x` is a matrix or data frame.
#' @param ...  Other parameters to some methods.
#' @return An object of class "fsets" is returned, which is a numeric matrix
#' with columns representing the fuzzy attributes. Each source column of the
#' `x` argument corresponds to multiple columns in the resulting matrix.
#' Columns have names that indicate the name of the source as well as a index
#' \eqn{i} of fuzzy set(s) -- see the description of arguments `breaks`
#' and `merge` above.
#'
#' The resulting object would also have set the [vars()] and
#' [specs()] properties with the former being created from original
#' column names (if `x` is a matrix or data frame) or the `name`
#' argument (if `x` is a numeric vector). The [specs()]
#' incidency matrix would be created to reflect the superset-hood of the merged
#' fuzzy sets.
#' @author Michal Burda
#' @seealso [lcut()], [equidist()], [farules()], [pbld()], [vars()], [specs()],
#'   [cbind.fsets()]
#' @keywords models robust multivariate
#' @examples
#'
#' # fcut on non-numeric data
#' ff <- factor(substring("statistics", 1:10, 1:10), levels = letters)
#' fcut(ff)
#'
#' # transform a single vector into a single fuzzy set
#' x <- runif(10)
#' fcut(x, breaks=c(0, 0.5, 1), name='age')
#'
#' # transform single vector into a partition of the interval 0-1
#' # (the boundary triangles are right-angled)
#' fcut(x, breaks=c(0, 0, 0.5, 1, 1), name='age')
#'
#' # also create supersets
#' fcut(x, breaks=c(0, 0, 0.5, 1, 1), name='age', merge=c(1, 2))
#'
#' # transform all columns of a data frame
#' # with different breakpoints
#' data <- CO2[, c('conc', 'uptake')]
#' fcut(data, breaks=list(conc=c(95, 95, 350, 1000, 1000),
#'                        uptake=c(7, 7, 28.3, 46, 46)))
#'
#' # using a custom 3-argument function (a function factory):
#' f <- function(a, b, c) {
#'   return(function(x) ifelse(a <= x & x <= b, 1, 0))
#' }
#' fcut(x, breaks=c(0, 0.5, 1), name='age', type=f)
#'
#' # using a custom 4-argument function:
#' f <- function(x, a, b, c) {
#'   return(ifelse(a <= x & x <= b, 1, 0))
#' }
#' fcut(x, breaks=c(0, 0.5, 1), name='age', type=f)
#'
#' @export
fcut <- function(x, ...) {
    UseMethod('fcut')
}


#' @rdname fcut
#' @export
fcut.default <- function(x, ...) {
    .stop(paste0("'fcut' not implemented for variable of class '", class(x), "'"))
}


#' @rdname fcut
#' @export
fcut.factor <- function(x,
                        name=deparse(substitute(x)),
                        ...) {
    .mustBeFactor(x)
    .mustNotBeNull(name)
    .mustBeCharacterScalar(name)

    res <- laply(x, function(a) { a == levels(x) }) + 0
    res <- as.matrix(res)
    colnames(res) <- paste(name, levels(x), sep='=')

    theVars <- rep(name, ncol(res))
    theSpecs <- matrix(0, nrow=ncol(res), ncol=ncol(res))
    return(fsets(res,
                 vars=theVars,
                 specs=theSpecs))
}


#' @rdname fcut
#' @export
fcut.logical <- function(x,
                        name=deparse(substitute(x)),
                        ...) {
    .mustBeLogicalVector(x)
    .mustNotBeNull(name)
    .mustBeCharacterScalar(name)

    res <- matrix(c(x+0, (!x)+0), byrow=FALSE, ncol=2)
    colnames(res) <- c(name, paste0('not.', name))

    theVars <- rep(name, ncol(res))
    theSpecs <- matrix(0, nrow=ncol(res), ncol=ncol(res))
    return(fsets(res,
                 vars=theVars,
                 specs=theSpecs))
}


#' @rdname fcut
#' @export
fcut.numeric <- function(x,
                         breaks,
                         name=deparse(substitute(x)),
                         type=c('triangle', 'raisedcos'),
                         merge=1,
                         parallel=FALSE,
                         ...) {
    n <- length(breaks) - 2

    .mustBeNumericVector(x)
    .mustBeNumericVector(breaks)
    .mustBe(length(breaks) >= 3, "'breaks' must be a numeric vector with at least 3 elements")
    .mustNotBeNull(name)
    .mustBeCharacterScalar(name)
    .mustBeNumericVector(merge)
    .mustBe(min(merge) >= 1 && max(merge) <= n, "'merge' must contain integers from 1 to length(breaks)-2")

    func <- NULL
    if (is.function(type)) {
        func <- type
    } else {
        type <- match.arg(type)
        if (type == 'triangle') {
            func <- triangular
        } else {
            func <- raisedcosine
        }
    }
    if (length(names(formals(func))) == 3L) {
        old <- func
        func <- function(x, a, b, c) {
            old(a, b, c)(x)
        }
    }

    # split 'x' according to 'breaks'
    singles <- .rollapply(breaks, 3, function(b) {
        func(x, b[1], b[2], b[3])
    })
    singles <- t(as.matrix(singles))
    colnames(singles) <- paste(name, 1:ncol(singles), sep='=')

    # handle merging
    merge <- as.integer(merge)
    res <- NULL
    if (identical(merge, 1L)) {
        res <- singles
    } else {
        for (w in merge) {
            add <- .rollapply(seq_len(ncol(singles)), w, function(ii) {
                m <- singles[, ii, drop=FALSE]
                l <- split(m, col(m))
                do.call(plukas.tconorm, l)
            })
            add <- t(add)
            colnames(add) <- .rollapply(colnames(singles), w, paste, collapse='|')

            if (is.null(res)) {
                res <- add
            } else {
                res <- cbind(res, add)
            }
        }
    }

    .firstCol <- function(n, i, rowlen) {
        res <- NULL
        k <- n
        while (length(res) < rowlen) {
            if (i <= 1) {
                res <- c(res, rep(0, k))
            } else {
                res <- c(res, rep(1, i), rep(0, k - i))
            }
            k <- k - 1
            i <- i - 1
        }
        return(res)
    }

    .subMat <- function(n, i, rowlen) {
        first <- .firstCol(n, i, rowlen)
        res <- lapply(1:(n-i+1), function(a) {
            r <- c(rep(0, a-1), first)[1:rowlen]
        })
        return(unlist(res))
    }

    .whatToSelect <- function(n, merge) {
        what <- rep(0, n)
        what[merge] <- 1
        counts <- n:1
        res <- sapply(1:n, function(i) {
            rep(what[i], counts[i])
        })
        return(as.logical(unlist(res)))
    }

    # generate vars vector
    vars <- rep(name, ncol(res))

    # generate specs matrix
    rowlen <- (1+n)*n/2
    vec <- unlist(sapply(seq_len(n), function(i) .subMat(n, i, rowlen)))
    specs <- matrix(vec, byrow=FALSE, nrow=rowlen)
    whatToSelect <- .whatToSelect(n, merge)
    specs <- specs[whatToSelect, whatToSelect, drop=FALSE]

    return(fsets(res, vars=vars, specs=specs))
}


#' @rdname fcut
#' @export
fcut.data.frame <- function(x,
                            breaks=NULL,
                            name=NULL,
                            type=c('triangle', 'raisedcos'),
                            merge=1,
                            parallel=FALSE,
                            ...) {

    .mustBeDataFrame(x);
    .mustBe(is.null(name), "If 'x' is a matrix or data frame then 'name' must be NULL")
    .mustBe(!is.null(colnames(x)), "Columns of 'x' must have names")

    if (!is.list(breaks)) {
        breaks <- rep(list(breaks), ncol(x))
        names(breaks) <- colnames(x)
    }
    if (!is.list(merge)) {
        merge <- rep(list(merge), ncol(x))
        names(merge) <- colnames(x)
    }
    if (!is.list(type)) {
        type <- rep(list(type), ncol(x))
        names(type) <- colnames(x)
    }

    loopBody <- function(n) {
        aBreaks <- breaks[[n]]
        aMerge <- merge[[n]]
        aType <- type[[n]]

        res <- fcut(x[[n]],
                    breaks=aBreaks,
                    name=n,
                    type=aType,
                    merge=aMerge,
                    parallel=FALSE,
                    ...)
        return(res)
    }

    n <- NULL
    if (parallel) {
        result <- foreach(n=colnames(x), .combine=cbind.fsets) %dopar% {
            return(loopBody(n))
        }
    } else {
        result <- foreach(n=colnames(x), .combine=cbind.fsets) %do% {
            return(loopBody(n))
        }
    }
    return(result)
}


#' @rdname fcut
#' @export
fcut.matrix <- function(x, ...) {
    result <- fcut(as.data.frame(x), ...)
    return(result)
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.mustBeNumericScalar <- function(x) {
    if (!isTRUE(is.vector(x) && is.numeric(x) && identical(length(x), 1L))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a numeric vector of size 1"),
             call.=FALSE)
    }
 }

.mustBeIntegerScalar <- function(x) {
    if (!isTRUE(is.vector(x) && is.integer(x) && identical(length(x), 1L))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be an integer vector of size 1"),
             call.=FALSE)
    }
 }

.mustBeCharacterScalar <- function(x) {
    if (!isTRUE(is.vector(x) && is.character(x) && identical(length(x), 1L))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a character vector of size 1"),
             call.=FALSE)
    }
}

.mustBeLogicalScalar <- function(x) {
    if (!isTRUE(is.vector(x) && is.logical(x) && identical(length(x), 1L))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a logical vector of size 1"),
             call.=FALSE)
    }
}

.mustBeInInterval <- function(x, a, b) {
    if (!isTRUE(all(x >= a & x <= b))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must contain values within interval ", a, " and ", b),
             call.=FALSE)
    }
}

.mustBeNumericVector <- function(x) {
    if (!isTRUE(is.vector(x) && is.numeric(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a numeric vector"),
             call.=FALSE)
    }
 }

.mustBeCharacterVector <- function(x) {
    if (!isTRUE(is.vector(x) && is.character(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a character vector"),
             call.=FALSE)
    }
 }

.mustBeFactor <- function(x) {
    if (!isTRUE(is.factor(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a factor"),
             call.=FALSE)
    }
 }

.mustBeLogicalVector <- function(x) {
    if (!isTRUE(is.vector(x) && is.logical(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a logical vector"),
             call.=FALSE)
    }
 }

.mustBeList <- function(x) {
    if (!isTRUE(is.list(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a list"),
             call.=FALSE)
    }
 }

.mustBeMatrix <- function(x) {
    if (!isTRUE(is.matrix(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a matrix"),
             call.=FALSE)
    }
 }

.mustBeNumericMatrix <- function(x) {
    if (!isTRUE(is.matrix(x) && is.numeric(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a numeric matrix"),
             call.=FALSE)
    }
 }

.mustBeDataFrame <- function(x) {
    if (!isTRUE(is.data.frame(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a data frame"),
             call.=FALSE)
    }
 }

.mustBeTs <- function(x) {
    if (!isTRUE(stats::is.ts(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a time series object"),
             call.=FALSE)
    }
 }

.mustBeAlgebra <- function(x) {
    if (!isTRUE(is.algebra(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be the algebra object"),
             call.=FALSE)
    }
 }

.mustBeFunction <- function(x) {
    if (!isTRUE(is.function(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be a function"),
             call.=FALSE)
    }
 }

.mustNotBeNull <- function(x) {
    if (!isTRUE(!is.null(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must not be NULL"),
             call.=FALSE)
    }
 }

.mustNotHaveNA <- function(x) {
    if (!isTRUE(!anyNA(x))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must not contain NA values"),
             call.=FALSE)
    }
 }

.mustNotBeZeroLength <- function(x) {
    if (!isTRUE(length(x) > 0)) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must not be of zero length"),
             call.=FALSE)
    }
 }

.stop <- function(msg) {
    stop(paste0(as.list(sys.call(-1))[[1]], ": ", msg), call.=FALSE)
}

.mustBe <- function(cond, msg) {
    if (!isTRUE(as.vector(cond))) {  # as.vector removes all attributes
        stop(paste0(as.list(sys.call(-1))[[1]], ": ", msg), call.=FALSE)
    }
}

.mustBeOneOf <- function(x, vals) {
    if (!isTRUE(all(x %in% vals))) {
        stop(paste0(as.list(sys.call(-1))[[1]],
                    ": '", deparse(substitute(x)), "' must be one of '",
                    paste(vals, collapse="', '"), "'"), call.=FALSE)
    }
}

test_that("mustBeNumericScalar", {
  expect_null(.mustBeNumericScalar(3))
  expect_null(.mustBeNumericScalar(0))
  expect_null(.mustBeNumericScalar(1L))
  expect_null(.mustBeNumericScalar(0.8))
  expect_error(.mustBeNumericScalar('c'))
  expect_error(.mustBeNumericScalar(1:4))
  expect_error(.mustBeNumericScalar(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeNumericScalar(NULL))
  expect_error(.mustBeNumericScalar(TRUE))
})

test_that("mustBeFactor", {
  expect_null(.mustBeFactor(factor(letters)))
  expect_error(.mustBeFactor(3))
  expect_error(.mustBeFactor(0))
  expect_error(.mustBeFactor(0.8))
  expect_error(.mustBeFactor('c'))
  expect_error(.mustBeFactor(1:4))
  expect_error(.mustBeFactor(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeFactor(NULL))
  expect_error(.mustBeFactor(TRUE))
})

test_that("mustBeIntegerScalar", {
  expect_null(.mustBeIntegerScalar(3L))
  expect_null(.mustBeIntegerScalar(0L))
  expect_error(.mustBeIntegerScalar(0.8))
  expect_error(.mustBeIntegerScalar('c'))
  expect_error(.mustBeIntegerScalar(1:4))
  expect_error(.mustBeIntegerScalar(c(1L, 2L)))
  expect_error(.mustBeIntegerScalar(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeIntegerScalar(NULL))
  expect_error(.mustBeIntegerScalar(TRUE))
})

test_that("mustBeCharacterScalar", {
  expect_null(.mustBeCharacterScalar('c'))
  expect_null(.mustBeCharacterScalar('hello'))
  expect_error(.mustBeCharacterScalar(1))
  expect_error(.mustBeCharacterScalar(c('a', 'b')))
  expect_error(.mustBeCharacterScalar(matrix('c', nrow=1, ncol=1)))
  expect_error(.mustBeCharacterScalar(NULL))
  expect_error(.mustBeCharacterScalar(TRUE))
})

test_that("mustBeNumericVector", {
  expect_null(.mustBeNumericVector(3))
  expect_null(.mustBeNumericVector(0))
  expect_null(.mustBeNumericVector(0.8))
  expect_null(.mustBeNumericVector(1:4))
  expect_error(.mustBeNumericVector('c'))
  expect_error(.mustBeNumericVector(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeNumericVector(NULL))
  expect_error(.mustBeNumericVector(TRUE))
  expect_error(.mustBeNumericVector(FALSE))
})

test_that("mustBeLogicalVector", {
  expect_error(.mustBeLogicalVector(3))
  expect_error(.mustBeLogicalVector(0))
  expect_error(.mustBeLogicalVector(0.8))
  expect_error(.mustBeLogicalVector(1:4))
  expect_error(.mustBeLogicalVector('c'))
  expect_error(.mustBeLogicalVector(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeLogicalVector(NULL))
  expect_null(.mustBeLogicalVector(TRUE))
  expect_null(.mustBeLogicalVector(FALSE))
  expect_null(.mustBeLogicalVector(c(T,F,F)))
})

test_that("mustBeCharacterVector", {
  expect_null(.mustBeCharacterVector('c'))
  expect_null(.mustBeCharacterVector('hello'))
  expect_null(.mustBeCharacterVector(c('a', 'b')))
  expect_error(.mustBeCharacterVector(1))
  expect_error(.mustBeCharacterVector(matrix('c', nrow=1, ncol=1)))
  expect_error(.mustBeCharacterVector(NULL))
  expect_error(.mustBeCharacterVector(TRUE))
})

test_that("mustBeList", {
  expect_null(.mustBeList(list()))
  expect_null(.mustBeList(list(a=1, b=2)))
  expect_error(.mustBeList(1))
  expect_error(.mustBeList('a'))
  expect_error(.mustBeList(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeList(NULL))
  expect_error(.mustBeList(TRUE))
  expect_error(.mustBeList(NA))
})

test_that("mustBeMatrix", {
  expect_null(.mustBeMatrix(matrix(1, nrow=1, ncol=1)))
  expect_null(.mustBeMatrix(matrix(1, nrow=3, ncol=5)))
  expect_null(.mustBeMatrix(matrix('a', nrow=1, ncol=1)))
  expect_null(.mustBeMatrix(matrix('a', nrow=3, ncol=5)))
  expect_null(.mustBeMatrix(matrix(TRUE, nrow=1, ncol=1)))

  expect_error(.mustBeMatrix(list(a=3)))
  expect_error(.mustBeMatrix(1:5))
  expect_error(.mustBeMatrix(TRUE))
  expect_error(.mustBeMatrix(NULL))
  expect_error(.mustBeMatrix(letters[1:5]))
})

test_that("mustBeNumericMatrix", {
  expect_null(.mustBeNumericMatrix(matrix(1, nrow=1, ncol=1)))
  expect_null(.mustBeNumericMatrix(matrix(1, nrow=3, ncol=5)))

  expect_error(.mustBeNumericMatrix(matrix('a', nrow=1, ncol=1)))
  expect_error(.mustBeNumericMatrix(matrix('a', nrow=3, ncol=5)))
  expect_error(.mustBeNumericMatrix(matrix(TRUE, nrow=1, ncol=1)))
  expect_error(.mustBeNumericMatrix(list(a=3)))
  expect_error(.mustBeNumericMatrix(1:5))
  expect_error(.mustBeNumericMatrix(TRUE))
  expect_error(.mustBeNumericMatrix(NULL))
  expect_error(.mustBeNumericMatrix(letters[1:5]))
})

test_that("mustBeDataFrame", {
  expect_null(.mustBeDataFrame(data.frame(a=1:5, b=1:5)))

  expect_error(.mustBeDataFrame(matrix(1, nrow=1, ncol=1)))
  expect_error(.mustBeDataFrame(matrix(1, nrow=3, ncol=5)))
  expect_error(.mustBeDataFrame(matrix('a', nrow=1, ncol=1)))
  expect_error(.mustBeDataFrame(matrix('a', nrow=3, ncol=5)))
  expect_error(.mustBeDataFrame(matrix(TRUE, nrow=1, ncol=1)))
  expect_error(.mustBeDataFrame(list(a=3)))
  expect_error(.mustBeDataFrame(1:5))
  expect_error(.mustBeDataFrame(TRUE))
  expect_error(.mustBeDataFrame(NULL))
  expect_error(.mustBeDataFrame(letters[1:5]))
})

test_that("mustBeTs", {
  expect_null(.mustBeTs(stats::ts(1:5)))

  expect_error(.mustBeTs(matrix(1, nrow=3, ncol=5)))
  expect_error(.mustBeTs(matrix('a', nrow=3, ncol=5)))
  expect_error(.mustBeTs(matrix(TRUE, nrow=1, ncol=1)))
  expect_error(.mustBeTs(list(a=3)))
  expect_error(.mustBeTs(1:5))
  expect_error(.mustBeTs(TRUE))
  expect_error(.mustBeTs(NULL))
  expect_error(.mustBeTs(letters[1:5]))
})

test_that("mustBeFunction", {
  expect_null(.mustBeFunction(function(x) {}))
  expect_null(.mustBeFunction(identity))

  expect_error(.mustBeFunction(matrix(1, nrow=3, ncol=5)))
  expect_error(.mustBeFunction(list(a=3)))
  expect_error(.mustBeFunction(1:5))
  expect_error(.mustBeFunction(TRUE))
  expect_error(.mustBeFunction(NULL))
})

test_that("mustNotBeNull", {
  expect_null(.mustNotBeNull(matrix(1, nrow=3, ncol=5)))
  expect_null(.mustNotBeNull(list(a=3)))
  expect_null(.mustNotBeNull(1:5))
  expect_null(.mustNotBeNull(TRUE))

  expect_error(.mustNotBeNull(NULL))
})

test_that("mustNotHaveNA", {
  expect_null(.mustNotHaveNA(matrix(1, nrow=3, ncol=5)))
  expect_null(.mustNotHaveNA(list(a=3)))
  expect_null(.mustNotHaveNA(1:5))
  expect_null(.mustNotHaveNA(TRUE))
  expect_null(.mustNotHaveNA(NULL))

  expect_error(.mustNotHaveNA(NA))
  expect_error(.mustNotHaveNA(c(1:5, NA)))
  expect_error(.mustNotHaveNA(matrix(NA, nrow=3, ncol=5)))
})

test_that("mustNotBeZeroLength", {
  expect_null(.mustNotBeZeroLength(1))
  expect_null(.mustNotBeZeroLength(letters))
  expect_error(.mustNotBeZeroLength(NULL))
})

test_that("stop", {
  expect_error(.stop())
})

test_that("mustBe", {
  expect_null(.mustBe(TRUE, "message"))
  expect_error(.mustBe(FALSE, "message"))
  expect_error(.mustBe(NA, "message"))
  expect_error(.mustBe(c(TRUE, TRUE), "message"))
})

test_that("mustBeOneOf", {
  expect_null(.mustBeOneOf('x', c(letters)))
  expect_error(.mustBeOneOf('X', c(letters)))
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' Reduce the size of rule base
#'
#' From given rule base, select such set of rules that influence mostly the
#' rule base coverage of the input data.
#'
#' From a given rulebase, a rule with greatest coverage is selected. After
#' that, additional rules are selected that increase the rule base coverage the
#' most. Addition stops after the coverage exceeds \eqn{original coverage *
#' ratio}.
#'
#' Note that the size of the resulting rule base is not necessarily minimal
#' because the algorithm does not search all possible combination of rules. It
#' only finds a local minimum of rule base size.
#'
#' @param x Data for the rules to be evaluated on. Could be either a numeric
#' matrix or numeric vector.  If matrix is given then the rules are evaluated
#' on rows. Each value of the vector or column of the matrix represents a
#' predicate - it's numeric value represents the truth values (values in the
#' interval \eqn{[0, 1]}).
#' @param rules Either an object of class "farules" or list of character
#' vectors where each vector is a rule with consequent being the first element
#' of the vector. Elements of the vectors (predicate names) must correspond to
#' the `x`'s names (of columns if `x` is a matrix).
#' @param ratio A percentage of rule base coverage that must be preserved. It
#' must be a value within the \eqn{[0, 1]} interval. Value of 1 means that the
#' rule base coverage of the result must be the same as coverage of input
#' \code{rules}. A sensible value is e.g. 0.9.
#' @param tnorm Which t-norm to use as a conjunction of antecedents. The
#' default is `"goedel"`.
#' @param tconorm Which t-norm to use as a disjunction, i.e., to combine
#' multiple antecedents to get coverage of the rule base. The default is
#' `"goedel"`.
#' @param numThreads How many threads to use for computation. Value higher than
#' 1 causes that the algorithm runs in several parallel threads (using the
#' OpenMP library).
#' @return Function returns an instance of class [farules()] or a
#' list depending on the type of the `rules` argument.
#' @author Michal Burda
#' @seealso [rbcoverage()], [farules()]
#' @references M. Burda, M. Štěpnička, Reduction of Fuzzy Rule Bases Driven by
#' the Coverage of Training Data, in: Proc. 16th World Congress of the
#' International Fuzzy Systems Association and 9th Conference of the European
#' Society for Fuzzy Logic and Technology (IFSA-EUSFLAT 2015), Advances in
#' Intelligent Systems Research, Atlantic Press, Gijon, 2015.
#' @keywords models robust multivariate
#' @export
reduce <- function(x,
                   rules,
                   ratio,
                   tnorm=c("goedel", "goguen", "lukasiewicz"),
                   tconorm=c("goedel", "goguen", "lukasiewicz"),
                   numThreads=1) {
    if (is.vector(x)) {
        x <- matrix(x, nrow=1, dimnames=list(NULL, names(x)))
    }
    .mustBe(is.matrix(x) && is.numeric(x), "'x' must be a numeric vector or matrix")
    .mustBeInInterval(x, 0, 1)

    origRules <- rules
    if (is.farules(origRules)) {
        rules <- origRules$rules
    } else if (is.vector(rules) && is.character(rules)) {
        rules <- list(rules)
    }
    .mustBe(is.list(rules), "'rules' must be a list of rules, a valid farules object, or a character vector")

    .mustBeNumericScalar(numThreads)
    .mustBe(numThreads >= 1, "'numThreads' must be positive integer number")

    tnorm <- match.arg(tnorm)
    if (tnorm == 'goedel') {
        tnorm <- 'minimum'
    } else if (tnorm == 'goguen') {
        tnorm <- 'product'
    }

    tconorm <- match.arg(tconorm)
    if (tconorm == 'goedel') {
        tconorm <- 'maximum'
    } else if (tconorm == 'goguen') {
        tconorm <- 'product'
    }

    lhsSupport <- NA
    if (is.farules(origRules) && ('lhsSupport' %in% colnames(origRules$statistics))) {
        lhsSupport <- origRules$statistics[, 'lhsSupport']
    } else {
        lhsSupport <- rep(1, length(rules))
    }

    lvl <- colnames(x)
    rb <- lapply(rules, function(x) { as.integer(factor(x, levels=lvl)) - 1 })
    config <- list(data=x,
                   rules=rb,
                   lhsSupport=lhsSupport,
                   ratio=ratio,
                   tnorm=tnorm,
                   tconorm=tconorm,
                   numThreads=numThreads)
    result <- .Call("_lfl_reduce", config, PACKAGE="lfl")
    result <- result + 1

    if (is.farules(origRules)) {
        r <- origRules$rules[result]
        s <- origRules$statistics[result, , drop=FALSE]
        return(farules(rules=r, statistics=s))
    } else {
        return(rules[result])
    }
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' A list of the parameters that define the shape of the hedges.
#' @export
defaultHedgeParams <- list('ex'=c(0.77, 0.90, 0.99),
                           'si'=c(0.71, 0.85, 0.962),
                           've'=c(0.66, 0.79, 0.915),
                           'ty'=c(0.88, 0.95, 1),
                           '-' =c(0.45, 0.68, 0.851),
                           'ml'=c(0.43, 0.6, 0.727),
                           'ro'=c(0.4, 0.52, 0.619),
                           'qr'=c(0.3, 0.42, 0.528),
                           'vr'=c(0.1, 0.2, 0.421))



#' Linguistic hedges
#'
#' Returns a function that realizes linguistic hedging - i.e., transformation of linguistic
#' horizon (see [horizon()]) into a linguistic expression.
#'
#' `hedge()` returns a function that realizes the selected linguistic hedge on its parameter:
#' * `ex`: extremely,
#' * `si`: significantly,
#' * `ve`: very,
#' * `ty`: typically,
#' * `-`: empty hedge (no hedging),
#' * `ml`: more or less,
#' * `ro`: roughly,
#' * `qr`: quite roughly,
#' * `vr`: very roughly.
#'
#' This function is quite low-level. Perhaps a more convenient way to create linguistic expressions
#' is to use the [lingexpr()] function.
#'
#' @param type The type of the required linguistic hedge
#' @param hedgeParams Parameters that determine the shape of the hedges
#' @return Returns a function with a single argument, which has to be a numeric
#' vector.
#' @author Michal Burda
#' @seealso [horizon()], [lingexpr()], [fcut()], [lcut()], [ctx()]
#' @keywords models robust
#' @examples
#'     a <- horizon(ctx3(), 'sm')
#'     plot(a)
#'     h <- hedge('ve')
#'     plot(h)
#'     verySmall <- function(x) h(a(x))
#'     plot(verySmall)
#'
#'     # the last plot should be equal to:
#'     plot(lingexpr(ctx3(), atomic='sm', hedge='ve'))
#'
#' @export
hedge <- function(type=c('ex', 'si', 've', 'ty', '-', 'ml', 'ro', 'qr', 'vr'),
                  hedgeParams=defaultHedgeParams) {
    type <- match.arg(type)
    params <- hedgeParams[[type]]
    return(function(x) {
        .Call('_lfl_hedge', as.numeric(as.vector(x)), params, PACKAGE='lfl')
    })
}

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


.elementWisely <- function(f) {
    function(...) {
        elts <- list(...)
        if (length(elts) <= 0L) {
            return(NULL)
        }
        vals <- lapply(elts, as.numeric)
        res <- do.call('mapply', c(list(f), vals))
        mostattributes(res) <- attributes(elts[[1L]])
        return(res)
    }
}


.algebraModification <- function(call, algebra, norm, conorm, resid, neg, invol, ord) {
    resN <- neg(algebra$n)
    resNI <- invol(algebra$ni)
    resT <- norm(algebra$t)
    resPT <- .elementWisely(resT)
    resC <- conorm(algebra$c)
    resR <- resid(algebra$r)
    resI <- norm(algebra$i)
    resPI <- .elementWisely(resI)
    resS <- conorm(algebra$s)
    resB <- function(x, y) { resPI(resR(x, y), resR(y, x)) }
    res <- list(n=resN,
                ni=resNI,
                t=resT,
                pt=resPT,
                c=resC,
                pc=.elementWisely(resC),
                r=resR,
                b=resB,
                i=resI,
                pi=resPI,
                s=resS,
                ps=.elementWisely(resS),
                order=ord,
                algebratype=c(algebra$algebratype, call))
    class(res) <- c('algebra', 'list')
    res
}


.neg0 <- function(f) {
    return(function(x) {
        res <- f(x)
        res[is.na(x)] <- 0
        res
    })
}


.neg1 <- function(f) {
    return(function(x) {
        res <- f(x)
        res[is.na(x)] <- 1
        res
    })
}


.negNA <- function(f) {
    return(function(x) {
        res <- f(x)
        res[is.na(x)] <- NA_real_
        res
    })
}


.normKleene <- function(f) {
    return(function(...) {
        dots <- c(...)
        nonadots <- na.omit(dots)
        res <- f(nonadots)
        if (length(dots) != length(nonadots) && !is.na(res) && res > 0) {
            return(NA_real_)
        }
        res
    })
}


.conormKleene <- function(f) {
    return(function(...) {
        dots <- c(...)
        nonadots <- na.omit(dots)
        res <- f(nonadots)
        if (length(dots) != length(nonadots) && !is.na(res) && res < 1) {
            return(NA_real_)
        }
        res
    })
}


.conormDragon <- function(f) {
    return(function(...) {
        dots <- c(...)
        nonadots <- na.omit(dots)
        res <- f(nonadots)
        if (length(dots) != length(nonadots) && !is.na(res) && res == 0) {
            return(NA_real_)
        }
        res
    })
}


.normUpper <- function(f) {
    return(function(...) {
        dots <- c(...)
        nonadots <- na.omit(dots)
        res <- f(nonadots)
        if (length(dots) != length(nonadots) && !is.na(res) && res == 1) {
            return(NA_real_)
        }
        res
    })
}


.undefinedOrder <- function(x, decreasing=FALSE) {
    stop('Cannot complete the computation, "order" is unimplemented for the algebra.')
}


.dragonflyOrder <- function(x, decreasing=FALSE) {
    large <- sum(!is.na(x) & x > 0)
    zeros <- sum(!is.na(x) & x == 0)
    nas <- sum(is.na(x))
    ordered <- order(x, decreasing=TRUE, na.last=TRUE)
    res <- c(ordered[seq(from=1, length.out=large)],
             ordered[seq(from=large + zeros + 1, length.out=nas)],
             ordered[seq(from=large + 1, length.out=zeros)])
    if (decreasing) {
        return(res)
    } else {
        return(rev(res))
    }
}


.upperBoundOrder <- function(x, decreasing=FALSE) {
    small <- sum(!is.na(x) & x < 1)
    ones <- sum(!is.na(x) & x == 1)
    nas <- sum(is.na(x))
    ordered <- order(x, decreasing=TRUE, na.last=TRUE)
    res <- c(ordered[seq(from=1, length.out=ones)],
             ordered[seq(from=ones + small + 1, length.out=nas)],
             ordered[seq(from=ones + 1, length.out=small)])
    if (decreasing) {
        return(res)
    } else {
        return(rev(res))
    }
}


#' Modify algebra's way of computing with `NA` values.
#'
#' By default, the objects created with the [algebra()] function represent a mathematical
#' algebra capable to work on the \eqn{[0,1]} interval. If `NA` appears as a value instead,
#' it is propagated to the result. That is, any operation with `NA` results in `NA`, by default.
#' This scheme of handling missing values is also known as Bochvar's. To change this default
#' behavior, the following functions may be applied.
#'
#' The `sobocinski()`, `kleene()`, `nelson()`, `dragonfly()`, `lowerEst()` and `lowerBound()`
#'  functions modify the algebra to
#' handle the `NA` in a different way than is the default. Sobocinski's algebra simply ignores `NA` values
#' whereas Kleene's algebra treats `NA` as "unknown value". Dragonfly approach is a combination
#' of Sobocinski's and Bochvar's approach, which preserves the ordering `0 <= NA <= 1`
#' to obtain from compositions (see [compose()])
#' the lower-estimate in the presence of missing values.
#'
#' In detail, the behavior of the algebra modifiers is defined as follows:
#'
#' Sobocinski's negation for `n` being the underlying algebra:
#' \tabular{ll}{
#'   a  \tab n(a)\cr
#'   NA \tab 0
#' }
#'
#' Sobocinski's operation for `op` being one of `t`, `pt`, `c`, `pc`, `i`, `pi`, `s`, `ps`
#' from the underlying algebra:
#' \tabular{lll}{
#'      \tab b        \tab NA\cr
#'   a  \tab op(a, b) \tab a \cr
#'   NA \tab b        \tab NA
#' }
#'
#' Sobocinski's operation for `r` from the underlying algebra:
#' \tabular{lll}{
#'       \tab b       \tab NA  \cr
#'    a  \tab r(a, b) \tab n(a)\cr
#'    NA \tab b       \tab NA
#' }
#'
#' Kleene's negation is identical to `n` from the underlying algebra.
#'
#' Kleene's operation for `op` being one of `t`, `pt`, `i`, `pi` from the underlying algebra:
#' \tabular{llll}{
#'      \tab b        \tab NA \tab 0\cr
#'   a  \tab op(a, b) \tab NA \tab 0\cr
#'   NA \tab NA       \tab NA \tab 0\cr
#'   0  \tab 0        \tab 0  \tab 0
#' }
#'
#' Kleene's operation for `op` being one of `c`, `pc`, `s`, `ps` from the underlying algebra:
#' \tabular{llll}{
#'      \tab b        \tab NA \tab 1\cr
#'   a  \tab op(a, b) \tab NA \tab 1\cr
#'   NA \tab NA       \tab NA \tab 1\cr
#'   1  \tab 1        \tab 1  \tab 1
#' }
#'
#' Kleene's operation for `r` from the underlying algebra:
#' \tabular{llll}{
#'      \tab b        \tab NA \tab 1\cr
#'   a  \tab r(a, b)  \tab NA \tab 1\cr
#'   NA \tab NA       \tab NA \tab 1\cr
#'   0  \tab 1        \tab 1  \tab 1
#' }
#'
#' Dragonfly negation is identical to `n` from the underlying algebra.
#'
#' Dragonfly operation for `op` being one of `t`, `pt`, `i`, `pi` from the underlying algebra:
#' \tabular{lllll}{
#'      \tab b        \tab NA \tab 0 \tab 1\cr
#'   a  \tab op(a, b) \tab NA \tab 0 \tab a\cr
#'   NA \tab NA       \tab NA \tab 0 \tab NA\cr
#'   0  \tab 0        \tab 0  \tab 0 \tab 0\cr
#'   1  \tab b        \tab NA \tab 0 \tab 1
#' }
#'
#' Dragonfly operation for `op` being one of `c`, `pc`, `s`, `ps` from the underlying algebra:
#' \tabular{lllll}{
#'      \tab b        \tab NA \tab 0  \tab 1\cr
#'   a  \tab op(a, b) \tab a  \tab a  \tab 1\cr
#'   NA \tab b        \tab NA \tab NA \tab 1\cr
#'   0  \tab b        \tab NA \tab 0  \tab 1\cr
#'   1  \tab 1        \tab 1  \tab 1  \tab 1
#' }
#'
#' Dragonfly operation for `r` from the underlying algebra:
#' \tabular{lllll}{
#'      \tab b       \tab NA \tab 0    \tab 1\cr
#'   a  \tab r(a, b) \tab NA \tab n(a) \tab 1\cr
#'   NA \tab b       \tab 1  \tab NA   \tab 1\cr
#'   0  \tab 1       \tab 1  \tab 1    \tab 1\cr
#'   1  \tab b       \tab NA \tab 0    \tab 1
#' }
#'
#' LowerEst and LowerBound algebras are equivalent to Dragonfly's algebra in
#' `t`, `pt`, `i`, `pi`, `c`, `pc`, `s`, `ps` operations, but differ in the
#' `r` and `n` operations. `n` is equal to the Sobocinski's negation for both
#' LowerEst and LowerBound algebras.
#'
#' LowerEst's operation for `r` from the underlying algebra:
#'
#' \tabular{lllll}{
#'      \tab b       \tab NA \tab 0    \tab 1\cr
#'   a  \tab r(a, b) \tab NA \tab n(a) \tab 1\cr
#'   NA \tab b       \tab NA \tab 0    \tab 1\cr
#'   0  \tab 1       \tab 1  \tab 1    \tab 1\cr
#'   1  \tab b       \tab NA \tab 0    \tab 1
#' }
#'
#' LowerBound's operation for `r` from the underlying algebra:
#'
#' \tabular{lllll}{
#'      \tab b       \tab NA \tab 0    \tab 1\cr
#'   a  \tab r(a, b) \tab NA \tab n(a) \tab 1\cr
#'   NA \tab 1       \tab 1  \tab 0    \tab 1\cr
#'   0  \tab 1       \tab 1  \tab 1    \tab 1\cr
#'   1  \tab b       \tab NA \tab 0    \tab 1
#' }
#'
#' @param algebra the underlying algebra object to be modified -- see the [algebra()] function
#' @return A list of function of the same structure as is the list returned from the [algebra()] function
#' @author Michal Burda
#' @keywords models robust
#' @examples
#' a <- algebra('lukas')
#' b <- sobocinski(a)
#'
#' a$t(0.3, NA)  # NA
#' b$t(0.3, NA)  # 0.3
#'
#' @export
#' @rdname algebraNA
#' @aliases algebraNA
sobocinski <- function(algebra) {
    .mustBeAlgebra(algebra)

    norm <- function(f) {
        return(function(...) {
            nonadots <- na.omit(c(...))
            if (length(nonadots) <= 0) {
                return(NA_real_)
            }
            f(nonadots)
        })
    }

    resid <- function(f) {
        return(function(x, y) {
            res <- f(x, y)
            xNA <- is.na(x)
            yNA <- is.na(y)
            res[yNA] <- algebra$n(x[yNA])
            res[xNA] <- y[xNA]
            res[xNA & yNA] <- NA_real_
            res
        })
    }

    .algebraModification('sobocinski', algebra, norm, norm, resid, .neg0, .negNA, .undefinedOrder)
}


#' @export
#' @rdname algebraNA
kleene <- function(algebra) {
    .mustBeAlgebra(algebra)

    resid <- function(f) {
        return(function(x, y) {
            res <- f(x, y)
            res[is.na(x)] <- NA_real_
            res[is.na(y)] <- NA_real_
            res[x == 0] <- 1
            res[y == 1] <- 1
            res
        })
    }

    .algebraModification('kleene', algebra, .normKleene, .conormKleene, resid, .negNA, .negNA, .undefinedOrder)
}


#' @export
#' @rdname algebraNA
dragonfly <- function(algebra) {
    .mustBeAlgebra(algebra)

    resid <-function(f) {
        return(function(x, y) {
            res <- f(x, y)
            xNA <- is.na(x)
            yNA <- is.na(y)
            x0 <- !xNA & x == 0
            y0 <- !yNA & y == 0
            res[xNA] <- y[xNA]
            res[yNA] <- NA_real_
            res[xNA & yNA] <- 1
            res[yNA & x0] <- 1
            res[xNA & y0] <- NA_real_
            return(res)
        })
    }

    alg <- .algebraModification('dragonfly', algebra, .normKleene, .conormDragon, resid, .negNA, .negNA, .dragonflyOrder)
    return(alg)
}


#' @export
#' @rdname algebraNA
nelson <- function(algebra) {
    .mustBeAlgebra(algebra)

    resid <- function(f) {
        return(function(x, y) {
            res <- f(x, y)
            xNA <- is.na(x)
            yNA <- is.na(y)
            x0 <- !xNA & x == 0
            y0 <- !yNA & y == 0
            y1 <- !yNA & y == 1
            res[yNA] <- NA_real_
            res[x0 & yNA] <- 1
            res[xNA] <- 1
            res[xNA & !(y0 | y1 | yNA)] <- NA_real_
            res
        })
    }

    .algebraModification('nelson', algebra, .normKleene, .conormKleene, resid, .neg1, .negNA, .undefinedOrder)
}


#' @export
#' @rdname algebraNA
lowerEst <- function(algebra) {
    .mustBeAlgebra(algebra)

    resid <-function(f) {
        return(function(x, y) {
            res <- f(x, y)
            xNA <- is.na(x)
            yNA <- is.na(y)
            x0 <- !xNA & x == 0
            res[yNA] <- NA_real_
            res[yNA & x0] <- 1
            res[xNA] <- y[xNA]
            res
        })
    }

    alg <- .algebraModification('lowerEst', algebra, .normKleene, .conormDragon, resid, .neg0, .negNA, .dragonflyOrder)
    return(alg)
}


#' @export
#' @rdname algebraNA
lowerBound <- function(algebra) {
    .mustBeAlgebra(algebra)

    resid <-function(f) {
        return(function(x, y) {
            res <- f(x, y)
            xNA <- is.na(x)
            yNA <- is.na(y)
            x0 <- !xNA & x == 0
            y0 <- !yNA & y == 0
            res[yNA] <- NA_real_
            res[xNA] <- 1
            res[xNA & y0] <- 0
            res[x0 & yNA] <- 1
            res
        })
    }

    alg <- .algebraModification('lowerBound', algebra, .normKleene, .conormDragon, resid, .neg0, .negNA, .dragonflyOrder)
    return(alg)
}


#' @export
#' @rdname algebraNA
upperBound <- function(algebra) {
    .mustBeAlgebra(algebra)

    resid <-function(f) {
        return(function(x, y) {
            res <- f(x, y)
            xNA <- is.na(x)
            yNA <- is.na(y)
            x1 <- !xNA & x == 1
            y1 <- !yNA & y == 1
            res[xNA] <- y[xNA]
            res[yNA] <- 1
            res[x1 & yNA] <- NA_real_
            res
        })
    }

    alg <- .algebraModification('upperBound', algebra, .normUpper, .conormKleene, resid, .neg0, .negNA, .upperBoundOrder)
    return(alg)
}

set.seed(34523)

test_that('algebraic properties', {
    for (atype in c('goedel', 'goguen', 'lukasiewicz')) {
        a <- algebra(atype)
        alg <- list(bochvar=a,
                    sobocinski=sobocinski(a),
                    kleene=kleene(a),
                    dragonfly=dragonfly(a),
                    nelson=nelson(a),
                    lowerEst=lowerEst(a))

        for (amodif in names(alg)) {
            a <- alg[[amodif]]

            for (x in c(0, 0.5, 1, NA)) {
                # residual negation is derived from residuum
                expect_equal(a$n(x), a$r(x, 0), info=amodif)

                # involutiveness of the involutive negation
                expect_equal(a$ni(a$ni(x)), x, info=amodif)

                for (y in c(0, 0.3, 0.5, 0.8, 1, NA)) {
                    # commutativity
                    expect_equal(a$t(x, y), a$t(y, x), info=amodif)
                    expect_equal(a$pt(x, y), a$pt(y, x), info=amodif)
                    expect_equal(a$c(x, y), a$c(y, x), info=amodif)
                    expect_equal(a$pc(x, y), a$pc(y, x), info=amodif)
                    expect_equal(a$s(x, y), a$s(y, x), info=amodif)
                    expect_equal(a$ps(x, y), a$ps(y, x), info=amodif)
                    expect_equal(a$i(x, y), a$i(y, x), info=amodif)
                    expect_equal(a$pi(x, y), a$pi(y, x), info=amodif)

                    # bi-residuum
                    expect_equal(a$b(x, y), a$i(a$r(x, y), a$r(y, x)), info=amodif)

                    if (!(amodif %in% c('dragonfly', 'lowerEst'))) {
                        # De-Morgan's law
                        expect_equal(a$c(!!!x, !!!y), a$ni(a$t(a$ni(!!!x), a$ni(!!!y))), info=amodif)
                    }
                }
            }
        }
    }
})

#######################################################################
# lfl: Linguistic Fuzzy Logic
# Copyright (C) 2025 Michal Burda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#######################################################################


#' lfl - Linguistic Fuzzy Logic
#'
#' Various algorithms related (not only) to linguistic fuzzy logic: mining for linguistic fuzzy association
#' rules, composition of fuzzy relations, performing perception-based logical deduction (PbLD),
#' and forecasting time-series using fuzzy rule-based ensemble (FRBE).
#'
#' Other methods include fuzzy transform and computation of Sugeno integrals. Also, basic algebraic
#' functions related to fuzzy logic are included, which allow handling missing values using different
#' styles such as Kleene, Bochvar, Sobocinski and others.
#'

#' @keywords internal
"_PACKAGE"

#' @importFrom e1071 skewness
#' @importFrom e1071 kurtosis
#' @importFrom foreach foreach
#' @importFrom foreach %do%
#' @importFrom foreach %dopar%
#' @importFrom forecast auto.arima
#' @importFrom forecast ets
#' @importFrom forecast rwf
#' @importFrom forecast thetaf
#' @importFrom forecast forecast
#' @importFrom plyr colwise
#' @importFrom plyr laply
#' @importFrom Rcpp evalCpp
#' @importFrom stats as.formula
#' @importFrom stats coef
#' @importFrom stats frequency
#' @importFrom stats lm
#' @importFrom stats model.matrix
#' @importFrom stats na.omit
#' @importFrom stats sd
#' @importFrom stats ts
#' @importFrom stats ts.plot
#' @importFrom stats weighted.mean
#' @importFrom tibble as_tibble
#' @importFrom tseries adf.test
#' @importFrom utils str
#' @useDynLib lfl, .registration = TRUE
#'
NULL

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbcoverage.R
\name{rbcoverage}
\alias{rbcoverage}
\title{Compute rule base coverage of data}
\usage{
rbcoverage(
  x,
  rules,
  tnorm = c("goedel", "goguen", "lukasiewicz"),
  onlyAnte = TRUE
)
}
\arguments{
\item{x}{Data for the rules to be evaluated on. Could be either a numeric
matrix or numeric vector.  If matrix is given then the rules are evaluated
on rows. Each value of the vector or column of the matrix represents a
predicate - it's numeric value represents the truth values (values in the
interval \eqn{[0, 1]}).}

\item{rules}{Either an object of class "farules" or list of character
vectors where each vector is a rule with consequent being the first element
of the vector. Elements of the vectors (predicate names) must correspond to
the \code{x}'s names (of columns if \code{x} is a matrix).}

\item{tnorm}{A character string representing a triangular norm to be used
(either \code{"goedel"}, \code{"goguen"}, or \code{"lukasiewicz"}) or an
arbitrary function that takes a vector of truth values and returns a t-norm
computed of them.}

\item{onlyAnte}{TRUE if only antecedent-part of a rule should be evaluated.
Antecedent-part of a rule are all predicates in rule vector starting from
the 2nd position. (First element of a rule is the consequent - see above.)

If FALSE, then the whole rule will be evaluated (antecedent part together
with consequent).}
}
\value{
A numeric value of the rule base coverage of given data.
}
\description{
This function computes rule base coverage, i.e., a an average of maximum
membership degree at which each row of data fires the rules in rule base.
}
\details{
Let \eqn{f_{ij}} be a truth value of \eqn{i}-th rule on \eqn{j}-th row of
data \code{x}.  Then \eqn{m_j = max(f_{.j})} is a maximum truth value that
is reached for the \eqn{j}-th data row with the rule base. Then the rule
base coverage is a mean of that truth values, i.e.  \eqn{rbcoverage =
mean(m_.)}.
}
\examples{

    x <- matrix(1:20 / 20, nrow=2)
    colnames(x) <- letters[1:10]

    rules <- list(c('a', 'c', 'e'),
                  c('b'),
                  c('d', 'a'),
                  c('c', 'a', 'b'))
    rbcoverage(x, rules, "goguen", TRUE)  # returns 1


    rules <- list(c('d', 'a'),
                  c('c', 'a', 'b'))
    rbcoverage(x, rules, "goguen", TRUE)  # returns 0.075)

}
\references{
M. Burda, M. Štěpnička, Reduction of Fuzzy Rule Bases Driven by
the Coverage of Training Data, in: Proc. 16th World Congress of the
International Fuzzy Systems Association and 9th Conference of the European
Society for Fuzzy Logic and Technology (IFSA-EUSFLAT 2015), Advances in
Intelligent Systems Research, Atlantic Press, Gijon, 2015.
}
\seealso{
\code{\link[=fire]{fire()}}, \code{\link[=reduce]{reduce()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quantifier.R
\name{quantifier}
\alias{quantifier}
\title{A quantifier is a function that computes a fuzzy truth value of a claim about
the quantity. This function creates the <1>-type quantifier. (See the examples
below on how to use it as a quantifier of the <1,1> type.)}
\usage{
quantifier(
  quantity = c("all", "almost.all", "most", "many", "few", "several", "some", "at.least"),
  n = NULL,
  alg = c("lukasiewicz", "goedel", "goguen")
)
}
\arguments{
\item{quantity}{the quantity to be evaluated. 'all' computes the degree of
truth to which all elements of the universe have the given property,
'almost.all', #'   'most', and 'many' evaluate whether the property is
present in extremely big, very big, or not small number of elements from
the universe, where these linguistic expressions are internally modeled
using the \code{\link[=lingexpr]{lingexpr()}} function. 'at.least' quantity requires the 'n'
argument to be specified, as it computes the truth value that at least
\eqn{n} elements from the universe have the given property.}

\item{n}{the number of elements in the 'at.least n' quantifier}

\item{alg}{the underlying algebra in which to compute the quantifier.
Note that the algebra must have properly defined the \code{order} function,
as in the case of 'goedel', 'goguen', or 'lukasiewicz' algebra, (see the
\code{\link[=algebra]{algebra()}} function) or as in the \code{\link[=dragonfly]{dragonfly()}} or \code{\link[=lowerEst]{lowerEst()}} algebra.}
}
\value{
A two-argument function, which expects two numeric vectors of equal length
(the vector elements are recycled to ensure equal lengths). The first argument, \code{x},
is a vector of membership degrees to be measured, the second argument, \code{w}, is
the vector of weights to which the element belongs to the universe.

Let \eqn{U} be the set of input vector indices (1 to \code{length(x)}). Then the quantifier
computes the truth values according to the following formula:
\eqn{\vee_{z \subseteq U} \wedge_{u \in z} (x[u]) \wedge measure(m_z)},
where
\eqn{m_z = sum(w)} for \code{"some"} and \verb{"at.least} and \eqn{m_z = sum(w[z]) / sum(w)} otherwise.
See \code{\link[=sugeno]{sugeno()}} for more details on how the quantifier is evaluated.

Setting \code{w} to 1 yields to operation of the <1> quantifier as developed by Dvořák et al.
To compute the <1,1> quantifier as developed by Dvořák et al., e.g. "almost all A are B", \code{w} must
be set again to 1 and \code{x} to the result of the implication \eqn{A \Rightarrow B}.
To compute the <1,1> quantifier as proposed by Murinová et al., e.g. "almost all A are B",
\code{x} must be set to the result of the implication \eqn{A \Rightarrow B} and \code{w} to the membership
degrees of \eqn{A}. See the examples below.
}
\description{
A quantifier is a function that computes a fuzzy truth value of a claim about
the quantity. This function creates the <1>-type quantifier. (See the examples
below on how to use it as a quantifier of the <1,1> type.)
}
\examples{
  # Dvorak <1> "almost all" quantifier
  f <- quantifier('almost.all')
  a <- c(0.9, 1, 1, 0.2, 1)
  f(x=a, w=1)

  # Dvorak <1,1> "almost all" quantifier (w set to 1)
  a <- c(0.9, 1, 1, 0.2, 1)
  b <- c(0.2, 1, 0, 0.5, 0.8)
  f <- quantifier('almost.all')
  f(x=lukas.residuum(a, b), w=1)

  # Murinová <1,1> "almost all" quantifier (note w set to a)
  a <- c(0.9, 1, 1, 0.2, 1)
  b <- c(0.2, 1, 0, 0.5, 0.8)
  f <- quantifier('almost.all')
  f(x=lukas.residuum(a, b), w=a)

  # Murinová <1,1> "some" quantifier
  a <- c(0.9, 1, 1, 0.2, 1)
  b <- c(0.2, 1, 0, 0.5, 0.8)
  f <- quantifier('some')
  f(x=plukas.tnorm(a, b), w=a)
}
\references{
Dvořák, A., Holčapek, M. L-fuzzy quantifiers of type <1> determined by fuzzy measures.
Fuzzy Sets and Systems vol.160, issue 23, 3425-3452, 2009.

Dvořák, A., Holčapek, M. Type <1,1> fuzzy quantifiers determined by fuzzy measures.
IEEE International Conference on Fuzzy Systems (FuzzIEEE), 2010.

Murinová, P., Novák, V. The theory of intermediate quantifiers in fuzzy natural logic
revisited and the model of "Many". Fuzzy Sets and Systems, vol 388, 2020.
}
\seealso{
\code{\link[=sugeno]{sugeno()}}, \code{\link[=lingexpr]{lingexpr()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sugeno.R
\name{sugeno}
\alias{sugeno}
\title{A factory function for creation of sugeno-integrals.}
\usage{
sugeno(
  measure,
  relative = TRUE,
  strong = FALSE,
  alg = c("lukasiewicz", "goedel", "goguen")
)
}
\arguments{
\item{measure}{A non-decreasing function that assigns a truth value from the
\eqn{[0, 1]} interval to the either relative or absolute quantity}

\item{relative}{Whether the measure assumes relative or absolute quantity.
Relative quantity is always a number from the \eqn{[0,1]} interval}

\item{strong}{Whether to use the strong conjunction (\code{TRUE}) or the weak
conjunction (\code{FALSE})}

\item{alg}{The underlying algebra must be either a string (one from 'lukasiewicz',
'goedel' or 'goguen') or an instance of the S3 class \code{\link[=algebra]{algebra()}}.}
}
\value{
A two-argument function, which expects two numeric vectors of equal length
(the vector elements are recycled to ensure equal lengths). The first argument, \code{x},
is a vector of membership degrees to be measured, the second argument, \code{w}, is
the vector of weights.

Let \eqn{U} be the set of input vector indices (1 to \code{length(x)}). Then the Sugeno integral
computes the truth values according to the following formula:
\eqn{\vee_{z \subseteq U} \wedge_{u \in z} (x[u]) CONJ measure(m_z)},
where
\eqn{m_z = sum(w[z]) / sum(w)} if \code{relative==TRUE} or \eqn{m_z = sum(w)} if \code{relative==FALSE}
and where CONJ is a strong conjunction (i.e., \code{alg$pt}) or a weak conjunction
(i.e., \code{alg$pi}) according to the \code{strong} parameter.
}
\description{
A factory function for creation of sugeno-integrals.
}
\examples{
  # Dvorak <1> "almost all" quantifier
  q <- sugeno(lingexpr(ctx3(), atomic='bi', hedge='ex'))
  a <- c(0.9, 1, 1, 0.2, 1)
  q(x=a, w=1)

  # Dvorak <1,1> "almost all" quantifier
  a <- c(0.9, 1, 1, 0.2, 1)
  b <- c(0.2, 1, 0, 0.5, 0.8)
  q <- sugeno(lingexpr(ctx3(), atomic='bi', hedge='ex'))
  q(x=lukas.residuum(a, b), w=1)

  # Murinová <1,1> "almost all" quantifier
  a <- c(0.9, 1, 1, 0.2, 1)
  b <- c(0.2, 1, 0, 0.5, 0.8)
  q <- sugeno(lingexpr(ctx3(), atomic='bi', hedge='ex'))
  q(x=lukas.residuum(a, b), w=a)
}
\seealso{
\code{\link[=quantifier]{quantifier()}}, \code{\link[=lingexpr]{lingexpr()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduce.R
\name{reduce}
\alias{reduce}
\title{Reduce the size of rule base}
\usage{
reduce(
  x,
  rules,
  ratio,
  tnorm = c("goedel", "goguen", "lukasiewicz"),
  tconorm = c("goedel", "goguen", "lukasiewicz"),
  numThreads = 1
)
}
\arguments{
\item{x}{Data for the rules to be evaluated on. Could be either a numeric
matrix or numeric vector.  If matrix is given then the rules are evaluated
on rows. Each value of the vector or column of the matrix represents a
predicate - it's numeric value represents the truth values (values in the
interval \eqn{[0, 1]}).}

\item{rules}{Either an object of class "farules" or list of character
vectors where each vector is a rule with consequent being the first element
of the vector. Elements of the vectors (predicate names) must correspond to
the \code{x}'s names (of columns if \code{x} is a matrix).}

\item{ratio}{A percentage of rule base coverage that must be preserved. It
must be a value within the \eqn{[0, 1]} interval. Value of 1 means that the
rule base coverage of the result must be the same as coverage of input
\code{rules}. A sensible value is e.g. 0.9.}

\item{tnorm}{Which t-norm to use as a conjunction of antecedents. The
default is \code{"goedel"}.}

\item{tconorm}{Which t-norm to use as a disjunction, i.e., to combine
multiple antecedents to get coverage of the rule base. The default is
\code{"goedel"}.}

\item{numThreads}{How many threads to use for computation. Value higher than
1 causes that the algorithm runs in several parallel threads (using the
OpenMP library).}
}
\value{
Function returns an instance of class \code{\link[=farules]{farules()}} or a
list depending on the type of the \code{rules} argument.
}
\description{
From given rule base, select such set of rules that influence mostly the
rule base coverage of the input data.
}
\details{
From a given rulebase, a rule with greatest coverage is selected. After
that, additional rules are selected that increase the rule base coverage the
most. Addition stops after the coverage exceeds \eqn{original coverage *
ratio}.

Note that the size of the resulting rule base is not necessarily minimal
because the algorithm does not search all possible combination of rules. It
only finds a local minimum of rule base size.
}
\references{
M. Burda, M. Štěpnička, Reduction of Fuzzy Rule Bases Driven by
the Coverage of Training Data, in: Proc. 16th World Congress of the
International Fuzzy Systems Association and 9th Conference of the European
Society for Fuzzy Logic and Technology (IFSA-EUSFLAT 2015), Advances in
Intelligent Systems Research, Atlantic Press, Gijon, 2015.
}
\seealso{
\code{\link[=rbcoverage]{rbcoverage()}}, \code{\link[=farules]{farules()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{plot.fsets}
\alias{plot.fsets}
\title{Plot membership degrees stored in the instance of the S3 class
\code{\link[=fsets]{fsets()}} as a line diagram.}
\usage{
\method{plot}{fsets}(x, ...)
}
\arguments{
\item{x}{An instance of class \code{\link[=fsets]{fsets()}}}

\item{...}{Other arguments that are passed to the underlying \code{\link[=ts.plot]{ts.plot()}}
function.}
}
\value{
Result of the \code{\link[=ts.plot]{ts.plot()}} method.
}
\description{
This function plots the membership degrees stored in the instance of the
\code{\link[=fsets]{fsets()}} class. Internally, the membership degrees are
transformed into a time-series object and viewed in a plot using the
\code{\link[=ts.plot]{ts.plot()}} function. This function is useful mainly to see the
shape of fuzzy sets on regularly sampled inputs.
}
\examples{
d <- lcut(0:1000/1000, name='x')
plot(d)

# Additional arguments are passed to the ts.plot method
# Here thick lines represent atomic linguistic expressions,
# i.e. ``small'', ``medium'', and ``big''.
plot(d,
     ylab='membership degree',
     xlab='values',
     gpars=list(lwd=c(rep(1, 3), 5, rep(1, 5), 5, rep(1, 7), 5, rep(1,4))))
}
\seealso{
\code{\link[=fsets]{fsets()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=ts.plot]{ts.plot()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/farules.R
\name{c.farules}
\alias{c.farules}
\title{Take a sequence of instances of S3 class \code{\link[=farules]{farules()}} and combine them into a single
object. An error is thrown if some argument does not inherit from the \code{\link[=farules]{farules()}}
class.}
\usage{
\method{c}{farules}(..., recursive = FALSE)
}
\arguments{
\item{...}{A sequence of objects of class \code{\link[=farules]{farules()}} to be concatenated.}

\item{recursive}{This argument has currently no function and is added here
only for compatibility with generic \code{\link{c}} function.}
}
\value{
An object of class \code{\link[=farules]{farules()}} that is created by merging the
arguments together, i.e.  by concatenating the rules and row-binding the
statistics of given objects.
}
\description{
Take a sequence of instances of S3 class \code{\link[=farules]{farules()}} and combine them into a single
object. An error is thrown if some argument does not inherit from the \code{\link[=farules]{farules()}}
class.
}
\examples{
    ori1 <- farules(rules=list(letters[1:3],
                               letters[2:5]),
                    statistics=matrix(runif(16), nrow=2))
    ori2 <- farules(rules=list(letters[4],
                               letters[3:8]),
                    statistics=matrix(runif(16), nrow=2))
    res <- c(ori1, ori2)
    print(res)
}
\seealso{
\code{\link[=farules]{farules()}}, \code{\link[=searchrules]{searchrules()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frbe.R
\name{frbe}
\alias{frbe}
\title{Fuzzy Rule-Based Ensemble (FRBE) of time-series forecasts}
\usage{
frbe(d, h = 10)
}
\arguments{
\item{d}{A source time-series in the ts time-series format. Note that the
frequency of the time-series must be set properly.}

\item{h}{A forecasting horizon, i.e., the number of values to forecast.}
}
\value{
Result is a list of class \code{frbe} with the following elements:
\itemize{
\item \code{features} - a data frame with computed features of the given time-series;
\item \code{forecasts} - a data frame with forecasts to be ensembled;
\item \code{weights} - weights of the forecasting methods as inferred from the features
and the hard-wired fuzzy rule base;
\item \code{mean} - the resulting ensembled forecast (computed as a weighted sum
of forecasts).
}
}
\description{
This function computes the fuzzy rule-based ensemble of time-series
forecasts. Several forecasting methods are used to predict future values of
a given time-series and a weighted sum is computed from them with weights
being determined from a fuzzy rule base.
}
\details{
This function computes the fuzzy rule-based ensemble of time-series
forecasts. The evaluation comprises the following steps:
\enumerate{
\item Several features are extracted from the given time-series \code{d}:
\itemize{
\item length of the time-series;
\item strength of trend;
\item strength of seasonality;
\item skewness;
\item kurtosis;
\item variation coefficient;
\item stationarity;
\item frequency.
These features are used later to infer weights of the forecasting methods.
}
\item Several forecasting methods are applied on the given time-series \code{d} to
obtain forecasts. Actually, the following methods are used:
\itemize{
\item ARIMA - by calling \code{\link[forecast:auto.arima]{forecast::auto.arima()}};
\item Exponential Smoothing - by calling \code{\link[forecast:ets]{forecast::ets()}};
\item Random Walk with Drift - by calling \code{\link[forecast:rwf]{forecast::rwf()}};
\item Theta - by calling \code{\link[forecast:thetaf]{forecast::thetaf()}}.
}
\item Computed features are input to the fuzzy rule-based inference mechanism
which yields weights of the forecasting methods. The fuzzy rule base is
hardwired in this package and it was obtained by performing data mining with
the use of the \code{\link[=farules]{farules()}} function.
\item A weighted sum of forecasts is computed and returned as a result.
}
}
\examples{
  # prepare data (from the forecast package)
  library(forecast)
  horizon <- 10
  train <- wineind[-1 * (length(wineind)-horizon+1):length(wineind)]
  test <- wineind[(length(wineind)-horizon+1):length(wineind)]

  # perform FRBE
  f <- frbe(ts(train, frequency=frequency(wineind)), h=horizon)

  # evaluate FRBE forecasts
  evalfrbe(f, test)

  # display forecast results
  f$mean

}
\references{
Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
SYST. 2015.
}
\seealso{
\code{\link[=evalfrbe]{evalfrbe()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.ctx.R
\name{print.ctx3}
\alias{print.ctx3}
\alias{print.ctx5}
\alias{print.ctx3bilat}
\alias{print.ctx5bilat}
\title{Print the linguistic context}
\usage{
\method{print}{ctx3}(x, ...)

\method{print}{ctx5}(x, ...)

\method{print}{ctx3bilat}(x, ...)

\method{print}{ctx5bilat}(x, ...)
}
\arguments{
\item{x}{A linguistic context to be printed}

\item{...}{Unused.}
}
\value{
Nothing.
}
\description{
Format an object of the \code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx5]{ctx5()}}, \code{\link[=ctx3bilat]{ctx3bilat()}} and the \code{\link[=ctx5bilat]{ctx5bilat()}}
class into human readable form and print it to the output.
}
\examples{

  context <- ctx3()
  print(context)

}
\seealso{
\code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx5]{ctx5()}}, \code{\link[=ctx3bilat]{ctx3bilat()}}, \code{\link[=ctx5bilat]{ctx5bilat()}}, \code{\link[=minmax]{minmax()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smape.R
\name{smape}
\alias{smape}
\title{Compute Symmetric Mean Absolute Percentage Error (SMAPE)}
\usage{
smape(forecast, validation)
}
\arguments{
\item{forecast}{A numeric vector of forecasted values}

\item{validation}{A numeric vector of actual (real) values}
}
\value{
A Symmetric Mean Absolute Percentage Error (SMAPE)
}
\description{
SMAPE is computed as \eqn{mean(abs(forecast - validation) / ((abs(forecast) + abs(validation)) / 2))}.
}
\seealso{
\code{\link[=rmse]{rmse()}}, \code{\link[=mase]{mase()}}, \code{\link[=frbe]{frbe()}}
}
\author{
Michal Burda
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ft.R
\name{ft}
\alias{ft}
\title{Fuzzy transform}
\usage{
ft(x, xmemb, y, order = 1)
}
\arguments{
\item{x}{the numeric matrix of input values}

\item{xmemb}{the partitioning of input values, i.e., a \code{fsets} object with membership degrees
(see \code{\link[=fcut]{fcut()}})}

\item{y}{the numeric vector of target values}

\item{order}{the order of the fuzzy transform (0, 1, 2, ...)}
}
\value{
the instance of the S3 class \code{ft}
}
\description{
Compute a fuzzy tranform of the given input matrix \code{x}.
}
\examples{

# create the fuzzy transform object
y <- (1:30)^2
x <- as.matrix(data.frame(a = 1:30, b = 30:1))
xmemb <- fcut(x,
              breaks = list(a = equidist(x[, 'a'], 3),
                            b = equidist(x[, 'b'], 3)))
fit <- ft(x, xmemb, y, order = 1)

# obtain function values
x2 <- as.matrix(data.frame(a = 10:20, b = 20:10))
xmemb2 <- fcut(x2,
               breaks = list(a = equidist(x[, 'a'], 3),
                             b = equidist(x[, 'b'], 3)))
y2 <- ftinv(fit, x2, xmemb2)
print(y2)

# compare original values with those obtained by the fuzzy transform
y[10:20] - y2

}
\references{
Perfilieva I. Fuzzy transforms: Theory and applications. FUZZY SET SYST,
volume 157, issue 8, p. 993-1023. 2006.
}
\seealso{
\code{\link[=ftinv]{ftinv()}}, \code{\link[=is.ft]{is.ft()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infer.R
\name{infer}
\alias{infer}
\title{Perform Mamdani or implicative fuzzy inference with given rule-base on given
dataset}
\usage{
infer(
  x,
  rules,
  partition,
  alg = c("goedel", "goguen", "lukasiewicz"),
  type = c("mamdani", "implicative"),
  parallel = FALSE
)
}
\arguments{
\item{x}{Input to the inference. It must be a numeric matrix with columns
representing fuzzy sets (truth values of predicates). Each row represents a
single case of inference. Columns should be named after predicates in rules'
antecedents. The values must be in the interval \eqn{[0, 1]}. \code{x} can also be
an object of class \code{\link[=fsets]{fsets()}} (e.g., created by using the \code{\link[=fcut]{fcut()}} or \code{\link[=lcut]{lcut()}}
functions).}

\item{rules}{A rule-base (a.k.a. linguistic description) either in the form
of the \code{\link[=farules]{farules()}} object or as a list of character vectors where
each element is a fuzzy set name (a predicate) and thus each such vector
forms a rule. The first element of each rule is the consequent predicate,
and the remaining elements are antecedent predicates.}

\item{partition}{A numeric matrix (or a \code{\link[=fsets]{fsets()}} object) with columns that
are consequents in \code{rules}. The column names of \code{partition} must correspond
to consequent predicate names used in \code{rules}. Each row represents a point in the
output domain.}

\item{alg}{The algebra to use for the inference. It can be either a character
string (\code{"goedel"}, \code{"goguen"}, or \code{"lukasiewicz"}) or an instance of the
\code{\link[=algebra]{algebra()}} class.}

\item{type}{The type of inference to use. It can be either \code{"mamdani"} or
\code{"implicative"}.}

\item{parallel}{Whether the processing should be run in parallel or not.
Parallelization is implemented using the \code{\link[foreach:foreach]{foreach::foreach()}}
package. The parallel environment must be set properly in advance, e.g., with
the \code{\link[doMC:registerDoMC]{doMC::registerDoMC()}} function.}
}
\value{
A numeric matrix of inferred membership degrees. The rows of the
result correspond to rows of the \code{x} argument, the columns of the result
correspond to rows of the \code{partition} argument.
}
\description{
Take a set of rules (a rule-base) and perform Mamdani or implicative
inference on each row of a given matrix (or the \code{\link[=fsets]{fsets()}} object) of truth
values.
}
\details{
Perform Mamdani or implicative fuzzy inference with given rule-base \code{rules}
on each row of input \code{x}. Columns of \code{x} are truth values of predicates that
appear in the antecedent part of \code{rules}. The \code{partition} matrix determines
the shape of predicates in consequents: each row represents a point in the
output fuzzy set for which membership degrees are computed.

For both inference types, the predicates in the antecedent part of rules are
combined using a t-norm (\code{alg$pt}) of the given algebra \code{alg} (see also the
\code{fire()} function).

For the Mamdani inference (\code{type = "mamdani"}), the antecedent firing
degrees are combined with the consequent partition using a t-norm (\code{alg$pt})
of the selected algebra \code{alg}, and the resulting fuzzy sets are aggregated
with a t-conorm (\code{alg$ps}).

For the implicative inference (\code{type = "implicative"}), the antecedent
firing degrees are combined with the consequent partition using a residuum
(\code{alg$r}) of the selected algebra \code{alg}, and the resulting fuzzy sets are
aggregated with a t-norm (\code{alg$pi}).
}
\examples{
# Create rules: first element is the consequent, the rest are antecedents
rules <- list(
    c("High",   "temp_high",   "pressure_high"),
    c("Medium", "temp_medium", "pressure_medium"),
    c("Low",    "temp_low")
)

# Create input data (fuzzy set membership degrees)
# (Alternatively, fcut() or lcut() functions can be used)
x <- matrix(
    c(0.9, 0.8, 0.2, 0.3, 0.1,
      0.1, 0.2, 0.4, 0.5, 0.8),
    nrow = 2,
    byrow = TRUE
)
colnames(x) <- c("temp_high", "pressure_high",
                  "temp_medium", "pressure_medium", "temp_low")

# Create output partition
# (Alternatively, fcut() or lcut() functions can be used)
partition <- matrix(
    c(1.0, 0.2, 0.0,
      0.8, 0.5, 0.1,
      0.4, 1.0, 0.4,
      0.1, 0.5, 0.8,
      0.0, 0.1, 1.0),
    nrow = 5,
    byrow = TRUE
)
colnames(partition) <- c("Low", "Medium", "High")

# Mamdani inference with Goedel algebra
infer(x, rules, partition, alg = "goedel", type = "mamdani")

# Implicative inference with Lukasiewicz algebra
infer(x, rules, partition, alg = "lukasiewicz", type = "implicative")

}
\seealso{
\code{\link[=pbld]{pbld()}}, \code{\link[=fire]{fire()}}, \code{\link[=aggregateConsequents]{aggregateConsequents()}}, \code{\link[=defuzz]{defuzz()}}, \code{\link[=algebra]{algebra()}}, \code{\link[=farules]{farules()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/equidist.R
\name{equidist}
\alias{equidist}
\title{Return equidistant breaks}
\usage{
equidist(
  x,
  n,
  left = c("infinity", "same", "none"),
  right = c("infinity", "same", "none")
)
}
\arguments{
\item{x}{A numeric vector of input values}

\item{n}{The number of breaks of \code{x} to find (\code{n} must be at least 2)}

\item{left}{The left border of the returned vector of breaks: \code{"infinity"}, \code{"same"} or \code{"none"}
(see the description below)}

\item{right}{The right border of the returned vector of breaks: \code{"infinity"}, \code{"same"} or \code{"none"}
(see the description below)}
}
\value{
A vector of equidistant breaks, which can be used e.g., in \code{\link[=fcut]{fcut()}}
}
\description{
If both \code{left} and \code{right} equal to \code{"none"}, the function returns a vector of \code{n} values from \code{x}
that divide the range of values in \code{x} into \code{n - 1} equidistant intervals.
}
\details{
If the \code{left} (resp. \code{right}) argument equals to \code{"infinity"}, \code{-Inf} (resp. \code{Inf}) is prepended
(resp. appended) to the result. If it equals to \code{"same"}, the first (resp. last) value is doubled.
Such functionality is beneficial if using the result of this function with e.g., the \code{\link[=fcut]{fcut()}} function:
\code{Inf} values at the beginning (resp. at the end) of the vector of breaks means that the fuzzy set
partition starts with a fuzzy set with kernel going to negative (resp. positive) infinity; the doubled
value at the beginning (resp. end) results in half-cut (trimmed) fuzzy set.
}
\seealso{
\code{\link[=equifreq]{equifreq()}}, \code{\link[=fcut]{fcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/antecedents.R
\name{antecedents}
\alias{antecedents}
\title{Extract antecedent-part (left-hand side) of rules in a list}
\usage{
antecedents(rules)
}
\arguments{
\item{rules}{Either a list of character vectors or an object of class \code{\link[=farules]{farules()}}.}
}
\value{
A list of character vectors.
}
\description{
Given a list of rules or an instance of the S3 \code{\link[=farules]{farules()}} class,
the function returns a list of their antecedents (i.e.
left-hand side of rules).
}
\details{
This function assumes \code{rules} to be a valid \code{\link[=farules]{farules()}} object or
a list of character vectors where
the first element of each vector is a consequent part and the
rest is an antecedent part of rules. Function returns a list of
antecedents.
}
\examples{
    rules <- list(c('a', 'b', 'c'), c('d'), c('a', 'e'))
    antecedents(rules)
}
\seealso{
\code{\link[=consequents]{consequents()}}, \code{\link[=farules]{farules()}}, \code{\link[=searchrules]{searchrules()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rmse.R
\name{rmse}
\alias{rmse}
\title{Compute Root Mean Squared Error (RMSE)}
\usage{
rmse(forecast, validation)
}
\arguments{
\item{forecast}{A numeric vector of forecasted values}

\item{validation}{A numeric vector of actual (real) values}
}
\value{
A Root Mean Squared Error (RMSE)
}
\description{
RMSE is computed as \eqn{sqrt(mean((forecast - validation)^2))}.
}
\seealso{
\code{\link[=smape]{smape()}}, \code{\link[=mase]{mase()}}, \code{\link[=frbe]{frbe()}}
}
\author{
Michal Burda
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mult.R
\name{mult}
\alias{mult}
\title{Callback-based Multiplication of Matrices}
\usage{
mult(x, y, f, ...)
}
\arguments{
\item{x}{A first matrix. The number of columns must match with the number of
rows of the \code{y} matrix.}

\item{y}{A second matrix. The number of rows must match with the number of
columns of the \code{x} matrix.}

\item{f}{A function to be applied to the matrices in order to compute the
multiplication.  It must accept at least two arguments.}

\item{...}{Additional arguments that are passed to the function \code{f}.}
}
\value{
A matrix with \eqn{v} rows and \eqn{w} columns, where \eqn{v} is the
number of rows of \code{x} and \eqn{w} is the number of columns of \code{y}.
}
\description{
Perform a custom multiplication of the matrices \code{x} and \code{y} by
using the callback function \code{f}.
}
\details{
For a matrix \code{x} of size \eqn{(u,v)} and a matrix \code{y} of size
\eqn{(v,w)}, \code{mult} calls the function \code{f} \eqn{uw}-times to
create a resulting matrix of size \eqn{(u,w)}.  Each \eqn{(i,j)}-th element
of the resulting matrix is obtained from a call of the function \code{f}
with \code{x}'s \eqn{i}-th row and \code{y}'s \eqn{j}-th column passed as its arguments.
}
\examples{
    x <- matrix(runif(24, -100, 100), ncol=6)
    y <- matrix(runif(18, -100, 100), nrow=6)

    mult(x, y, function(xx, yy) sum(xx * yy)) # the same as "x \%*\% y"

}
\seealso{
\code{\link[=compose]{compose()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{str.fsets}
\alias{str.fsets}
\title{Compactly display the internal structure of an instance of the \code{\link[=fsets]{fsets()}} class}
\usage{
\method{str}{fsets}(object, ...)
}
\arguments{
\item{object}{An instance of the \code{\link[=fsets]{fsets()}} class}

\item{...}{Unused.}
}
\value{
Nothing
}
\description{
This function is a method for the generic function \code{\link[=str]{str()}} for objects of the
\code{\link[=fsets]{fsets()}} class. It is a compact way to display the internal structure of an
instance of the \code{\link[=fsets]{fsets()}} class.
}
\seealso{
\code{\link[=fsets]{fsets()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algebra.R
\name{print.algebra}
\alias{print.algebra}
\title{Print an instance of the \code{\link[=algebra]{algebra()}} S3 class in a human readable form.}
\usage{
\method{print}{algebra}(x, ...)
}
\arguments{
\item{x}{An instance of the \code{\link[=algebra]{algebra()}} S3 class}

\item{...}{Unused.}
}
\value{
None.
}
\description{
Print an instance of the \code{\link[=algebra]{algebra()}} S3 class in a human readable form.
}
\seealso{
\code{\link[=algebra]{algebra()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicesDeprecated.R
\name{slices}
\alias{slices}
\title{Return vector of values from given interval}
\usage{
slices(from, to, n)
}
\arguments{
\item{from}{The lower bound of the interval.}

\item{to}{The upper bound of the interval.}

\item{n}{The length of the vector to be produced.}
}
\value{
A vector of numbers in the given interval and size.
}
\description{
Returns an ordered vector of values from given interval, of given size,
generated by equal steps.
}
\details{
Returns a vector of values from \code{from} to \code{to} (inclusive), with
equal difference between two consecutive values, with total length \code{n}.
Function is useful e.g., together with the \code{\link{pbld}} or
\code{\link{defuzz}} functions (for the \code{values} argument; see also
\code{\link{lcut}} or \code{\link{fcut}}) or \code{\link{defuzz}}).
}
\examples{
\dontrun{
    slices(1, 5, 10) # 1, 1.5, 2, 2.5, 3, 3.5 4, 4.5, 5
}
# is the same as
seq(1, 5, length.out=10)


}
\seealso{
\code{\link{pbld}}, \code{\link{defuzz}}, \code{\link{fcut}},
\code{\link{lcut}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.specific.R
\name{is.specific}
\alias{is.specific}
\title{Determine whether the first set \code{x} of predicates is more specific (or equal)
than \code{y} with respect to \code{vars} and \code{specs}.}
\usage{
is.specific(x, y, fsets, vars = NULL, specs = NULL)
}
\arguments{
\item{x}{The first character vector of predicates.}

\item{y}{The second character vector of predicates.}

\item{fsets}{A valid instance of the \code{\link[=fsets]{fsets()}} class such that all values in \code{x} and \code{y}
can be found in \code{colnames(fsets)}}

\item{vars}{Deprecated parameter must be \code{NULL}.}

\item{specs}{Deprecated parameter must be \code{NULL}.}
}
\value{
TRUE or FALSE (see description).
}
\description{
The function takes two character vectors of predicates and determines whether
\code{x} is more specific (or equal w.r.t. the specificity) than \code{y}. The
specificity relation is fully determined with the values of the \code{\link[=vars]{vars()}} vector
and the \code{\link[=specs]{specs()}} incidence matrix that is encapsulated in the given \code{fsets} object.
}
\details{
Let \eqn{x_i} and \eqn{y_j} represent some predicates of vectors \code{x} and \code{y},
respectively. Function assumes that each vector \code{x} and \code{y} does not
contain two or more predicates with the same value of \code{\link[=vars]{vars()}}.

This function returns TRUE iff all of the following conditions hold:
\itemize{
\item for any \eqn{y_j} there exists \eqn{x_i} such that \eqn{vars[y_j] = vars[x_i]};
\item for any \eqn{x_i} there either does not exist \eqn{y_j} such that
\eqn{vars[x_i] = vars[y_j]}, or \eqn{x_i = y_j}, or \eqn{specs[x_i, y_j] = 1}.
}
}
\examples{
    # prepare fsets object
    v <- c(rep('a', 3), rep('b', 3), rep('c', 3), rep('d', 3))
    s <- matrix(c(0,1,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,

                  0,0,0, 0,1,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,

                  0,0,0, 0,0,0, 0,1,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,

                  0,0,0, 0,0,0, 0,0,0, 0,1,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0,
                  0,0,0, 0,0,0, 0,0,0, 0,0,0),
                byrow=TRUE,
                ncol=12)
    m <- matrix(0, nrow=1, ncol=12)
    colnames(m) <- paste(rep(c('VeSm', 'Sm', 'Bi'), times=4),
                         rep(c('a', 'b', 'c', 'd'), each=3),
                         sep='.')
    f <- fsets(m, v, s)


    # returns TRUE
    is.specific(c('VeSm.a', 'Bi.c'),
                c('VeSm.a', 'Bi.c'),
                f)

    # returns TRUE (x and y swapped return FALSE)
    is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                c('Sm.a', 'Bi.c', 'Sm.d'),
                f)

    # returns TRUE (x and y swapped return FALSE)
    is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                c('VeSm.a', 'Bi.c'),
                f)

    # returns TRUE (x and y swapped return FALSE)
    is.specific(c('VeSm.a', 'Bi.c', 'Sm.d'),
                character(),
                f)

    # returns FALSE
    is.specific(c('Sm.a'), c('Bi.c'), f)

    # returns FALSE
    is.specific(c('VeSm.a', 'Sm.c'),
                c('Sm.a', 'Bi.c'),
                f)
}
\seealso{
\code{\link[=perceive]{perceive()}}, \code{\link[=pbld]{pbld()}}, \code{\link[=fsets]{fsets()}}, \code{\link[=vars]{vars()}}, \code{\link[=specs]{specs()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exprs.R
\name{triangular}
\alias{triangular}
\alias{raisedcosine}
\title{Factories for functions that convert numeric data into membership degrees of fuzzy sets}
\usage{
triangular(lo, center, hi)

raisedcosine(lo, center, hi)
}
\arguments{
\item{lo}{A lower bound (can be -Inf).}

\item{center}{A peak value.}

\item{hi}{An upper bound (can be Inf).}
}
\value{
A function with single argument \code{x} that should be a numeric vector to be converted.
}
\description{
These functions create functions with a single argument \code{x} that compute membership degrees of \code{x} to a fuzzy set
of either triangular or raised-cosine shape that is defined by \code{lo}, \code{center}, and \code{hi}.
}
\details{
The arguments must satisfy \verb{lo <= center <= hi}. Functions compute membership degrees of triangular or
raised-cosine fuzzy sets. \code{x} values equal to \verb{center obtain membership degree equal to 1, }x\verb{values lower than}lo\verb{or greater than}hi\verb{obtain membership degree equal to 0. A transition of the triangular (resp. raised cosine) shape (with peak at}center\verb{is computed for}x\verb{values between}lo\code{and}hi`.

If \code{lo == -Inf} then any value that is lower or equal to center gets membership degree 1.  Similarly, if \code{hi == Inf}
then any value that is greater or equal to center gets membership degree 1. \code{NA} and \code{NaN} values remain unchanged.

\code{triangular()} produces fuzzy sets of a triangular shape (with peak at \code{center}), \code{raisedcosine()} produces
fuzzy sets defined as a raised cosine hill.
}
\examples{

tr <- triangular(1, 2, 3)
tr(1:30 / 3)

rc <- raisedcosine(1, 2, 3)
rc(1:30 / 3)

plot(triangular(-1, 0, 1), from=-2, to=3)
plot(triangular(-1, 0, 2), from=-2, to=3)
plot(triangular(-Inf, 0, 1), from=-2, to=3)
plot(triangular(-1, 0, Inf), from=-2, to=3)

plot(raisedcosine(-1, 0, 1), from=-2, to=3)
plot(raisedcosine(-1, 0, 2), from=-2, to=3)
plot(raisedcosine(-Inf, 0, 1), from=-2, to=3)
plot(raisedcosine(-1, 0, Inf), from=-2, to=3)

}
\seealso{
\code{\link[=fcut]{fcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defuzz.R
\name{defuzz}
\alias{defuzz}
\title{Convert fuzzy set into a crisp numeric value}
\usage{
defuzz(
  degrees,
  values,
  type = c("mom", "fom", "lom", "dee", "cog", "expun", "expw1", "expw2")
)
}
\arguments{
\item{degrees}{A fuzzy set in the form of a numeric vector of membership
degrees of values provided as the \code{values} argument.}

\item{values}{A universe for the fuzzy set.}

\item{type}{Type of the requested defuzzification method. The possibilities are:
\itemize{
\item \code{'mom'}: Mean of Maxima - maximum membership degrees are
found and a mean of values that correspond to that degrees is returned;
\item \code{'fom'}: First of Maxima - first value with maximum membership
degree is returned;
\item \code{'lom'}: Last of Maxima - last value with maximum membership degree
is returned;
\item \code{'dee'}: Defuzzification of Evaluative Expressions - method used
by the \code{\link[=pbld]{pbld()}} inference mechanism that combines the former three
approaches according to the shape of the \code{degrees} vector:
If \code{degrees} is non-increasing then \code{'lom'} type is used,
if it is non-decreasing then \code{'fom'} is applied, else \code{'mom'} is selected;
\item \code{'cog'}: Center of Gravity - the result is a mean of \code{values} weighted by \code{degrees};
\item \code{'exp1'}: Experimental 1.
}}
}
\value{
A defuzzified value.
}
\description{
Take a discretized fuzzy set (i.e., a vector of membership degrees and a vector
of numeric values that correspond to that degrees) and perform a selected
type of defuzzification, i.e., conversion of the fuzzy set into a single
crisp value.
}
\details{
Function converts input fuzzy set into a crisp value. The definition of
input fuzzy set is provided by the arguments \code{degrees} and
\code{values}. These arguments should be numeric vectors of the same length,
the former containing membership degrees in the interval \eqn{[0, 1]} and
the latter containing the corresponding crisp values: i.e., \code{values[i]}
has a membership degree \code{degrees[i]}.
}
\examples{

# returns mean of maxima, i.e., mean of 6, 7, 8
defuzz(c(0, 0, 0, 0.1, 0.3, 0.9, 0.9, 0.9, 0.2, 0),
       1:10,
       type='mom')

}
\seealso{
\code{\link[=fire]{fire()}}, \code{\link[=aggregateConsequents]{aggregateConsequents()}}, \code{\link[=perceive]{perceive()}}, \code{\link[=pbld]{pbld()}}, \code{\link[=infer]{infer()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/farules.R
\name{farules}
\alias{farules}
\title{Create an instance of S3 class \code{farules} which represents a set of fuzzy
association rules and their statistical characteristics.}
\usage{
farules(rules, statistics)
}
\arguments{
\item{rules}{A list of character vectors, where each vector represents a
rule and each value of the vector represents a predicate. The first value of
the vector is assumed to be a rule's consequent, the rest is a rule's antecedent.}

\item{statistics}{A numeric matrix of various statistical characteristics of
the rules. Each column of that matrix corresponds to some statistic (such as
support, confidence, etc.) and each row corresponds to a rule in the list of
\code{rules}.}
}
\value{
Returns an object of class \code{farules}.
}
\description{
This function is a constructor that returns an instance of the \code{farules} S3 class.
To search for fuzzy association rules, refer to the \code{\link[=searchrules]{searchrules()}} function.
}
\seealso{
\code{\link[=searchrules]{searchrules()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mase.R
\name{mase}
\alias{mase}
\title{Compute Mean Absolute Scaled Error (MASE)}
\usage{
mase(forecast, validation)
}
\arguments{
\item{forecast}{A numeric vector of forecasted values}

\item{validation}{A numeric vector of actual (real) values}
}
\value{
A Mean Absolute Scaled Error (MASE)
}
\description{
MASE is computed as \eqn{sum(abs(validation - forecast)) / sum(abs(validation[-1] - validation[-n])) / (n/(n-1))}.
}
\seealso{
\code{\link[=rmse]{rmse()}}, \code{\link[=smape]{smape()}}, \code{\link[=frbe]{frbe()}}
}
\author{
Michal Burda
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/searchrules.R
\name{searchrules}
\alias{searchrules}
\title{Searching for fuzzy association rules}
\usage{
searchrules(
  d,
  lhs = 2:ncol(d),
  rhs = 1,
  tnorm = c("goedel", "goguen", "lukasiewicz"),
  n = 100,
  best = c("confidence"),
  minSupport = 0.02,
  minConfidence = 0.75,
  maxConfidence = 1,
  maxLength = 4,
  numThreads = 1,
  trie = (maxConfidence < 1)
)
}
\arguments{
\item{d}{An object of class \code{\link[=fsets]{fsets()}} - it is basically a matrix
where columns represent the fuzzy sets and values are the membership
degrees. For creation of such object, use \code{\link[=fcut]{fcut()}} or \code{\link[=lcut]{lcut()}} function.}

\item{lhs}{Indices of fuzzy attributes that may appear on the left-hand-side
(LHS) of association rules, i.e., in the antecedent.}

\item{rhs}{Indices of fuzzy attributes that may appear on the
right-hand-side (RHS) of association rules, i.e., in the consequent.}

\item{tnorm}{A t-norm to be used for computation of conjunction of fuzzy
attributes. (Only the starting letters of "lukasiewicz",
"goedel" and "goguen" are allowed).}

\item{n}{The non-negative number of rules to be found. If zero, the function
returns all rules satisfying the given conditions. If positive, only the
\code{n} best rules are returned. The criterion of what is "best" is
specified with the \code{best} argument.}

\item{best}{Specifies the measure according to which the rules are ordered
from best to worst. This argument is used mainly in combination with the
\code{n} argument. Currently, only a single value ("confidence") can be used.}

\item{minSupport}{The minimum support degree of a rule. Rules with support
below that number are filtered out. It must be a numeric value from interval
\eqn{[0, 1]}. See below for details on how the support degree is computed.}

\item{minConfidence}{The minimum confidence degree of a rule. Rules with
confidence below that number are filtered out.  It must be a numeric value
from interval \eqn{[0, 1]}. See below for details on how the confidence degree is
computed.}

\item{maxConfidence}{Maximum confidence threshold. After finding a rule that
has confidence degree above the \code{maxConfidence} threshold, no other
rule is produced based on adding some additional attribute to its antecedent
part. I.e., if "Sm.age & Me.age => Sm.height" has confidence above the
\code{maxConfidence} threshold, no other rule containing "Sm.age & Me.age"
will be produced regardless of its interest measures.

If you want to disable this feature, set \code{maxConfidence} to 1.}

\item{maxLength}{Maximum allowed length of the rule, i.e., maximum
number of predicates that are allowed on the left-hand + right-hand side of the rule. If
negative, the maximum length of rules is unlimited.}

\item{numThreads}{Number of threads used to perform the algorithm in
parallel. If greater than 1, the OpenMP library (not to be confused with
Open MPI) is used for parallelization. Please note that there are known
problems of using OpenMP together with other means of parallelization that
may be used within R. Therefore, if you plan to use the \code{searchrules}
function with some external parallelization mechanisms such as the
\code{doMC} library, make sure that \code{numThreads} equals 1. This
feature is available only on systems that have the OpenMP library installed.}

\item{trie}{Whether or not to use the internal mechanism of Tries. If FALSE,
then a rule may appear in the output that is a descendant of a rule that
has confidence above the \code{maxConfidence} threshold.

Tries consume a large amount of memory, so if you encounter problems with
insufficient memory, set this argument to FALSE. On the other hand, the size
of the result (if \code{n} is set to 0) can be very high if trie is set to
FALSE.}
}
\value{
A list of the following elements: \code{rules} and \code{statistics}.

\code{rules} is a list of mined fuzzy association rules. Each element of
that list is a character vector with consequent attribute being on the first
position.

\code{statistics} is a data frame of statistical characteristics about mined
rules. Each row corresponds to a rule in the \code{rules} list. Let us
consider a rule "a & b => c", let \eqn{\otimes} be a t-norm specified with
the \code{tnorm} parameter and \eqn{i} goes over all rows of a data table
\code{d}. Then columns of the \code{statistics} data frame are as follows:
\itemize{
\item support: a rule's support degree: \eqn{1/nrow(d) * \sum_{\forall i} a(i) \otimes b(i) \otimes c(i)}
\item lhsSupport: a support of rule's antecedent (LHS): \eqn{1/nrow(d) * \sum_{\forall i} a(i) \otimes b(i)}
\item rhsSupport: a support of rule's consequent (RHS): \eqn{1/nrow(d) * \sum_{\forall i} c(i)}
\item confidence: a rule's confidence degree: \eqn{support / lhsSupport}
}
}
\description{
This function searches the given \code{\link[=fsets]{fsets()}} object \code{d} for all
fuzzy association rules that satisfy defined constraints. It returns a list
of fuzzy association rules together with some statistics characterizing them
(such as support, confidence etc.).
}
\details{
The function searches data frame \code{d} for fuzzy association rules that
satisfy conditions specified by the parameters.
}
\examples{

  d <- lcut(CO2)
  searchrules(d, lhs=1:ncol(d), rhs=1:ncol(d))

}
\seealso{
\code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=farules]{farules()}}, \code{\link[=fsets]{fsets()}}, \code{\link[=pbld]{pbld()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{fsets}
\alias{fsets}
\alias{vars}
\alias{vars<-}
\alias{specs}
\alias{specs<-}
\title{S3 class representing a set of fuzzy sets on the fixed universe}
\usage{
fsets(
  x,
  vars = rep(deparse(substitute(x)), ncol(x)),
  specs = matrix(0, nrow = ncol(x), ncol = ncol(x))
)

vars(f)

vars(f) <- value

specs(f)

specs(f) <- value
}
\arguments{
\item{x}{A matrix of membership degrees. Columns of the matrix represent
fuzzy sets, colnames are names of the fuzzy sets (and must not be NULL). Rows
of the matrix represent elements of the universe.}

\item{vars}{A character vector that must correspond to the
columns of \code{x}. It is a vector of names of original variables that the
fuzzy sets were created from. In other words, the \code{vars} vector should
contain the same value for each \code{x}'s column that corresponds to the same
variable. Names of the \code{vars} vector are ignored.
For instance, an \code{\link[=fcut]{fcut()}} function can transform a single numeric
vector into several different fuzzy sets. To indicate that all of them in
fact describe the same original variable, the same name is stored on
appropriate positions of the \code{vars} vector.}

\item{specs}{A square numeric matrix containing values from \verb{\{0, 1\}}.
It is a specificity matrix, for which both rows and columns correspond to
\code{x}'s columns and where \code{specs[i][j] == 1} if and only if \code{i}-th fuzzy
set (i.e. \code{x[, i]}) is more specific (is a subset or equal to) than \code{j}-th fuzzy
set (i.e. \code{x[, j]}).}

\item{f}{An instance of S3 class \code{fsets}.}

\item{value}{Attribute values to be set to the object.}
}
\value{
\code{\link[=fsets]{fsets()}} returns an object of S3 class \code{fsets}. \code{\link[=vars]{vars()}} returns
a vector of original variable names of the \code{fsets} object. \code{specs}
returns the specificity matrix.
}
\description{
The aim of the \code{fsets} S3 class is to store several fuzzy sets in the
form of numeric matrix where columns represent fuzzy sets, rows are
elements from the universe, and therefore a value of \code{i}-th row and \code{j}-th
column is a membership degree of \code{i}-th element of the universe to \code{j}-th fuzzy
set. The \code{fsets} object also stores the information about
the origin of the fuzzy sets as well as a relation of specificity among
them.
}
\details{
The \code{fsets()} function is a constructor of an object of type \code{fsets}.
Each object stores two attributes: \code{vars} and \code{specs}. The functions \code{\link[=vars]{vars()}}
and \code{\link[=specs]{specs()}}). can be used to access these attributes.

It is assumed that the fuzzy sets
are derived from some raw variables, e.g., numeric vectors or factors. \code{vars}
attribute is a character vector of names of raw variables with size equal
to the number of fuzzy sets in \code{fsets} object. It is then assumed that
two fuzzy sets with the same name in \code{\link[=vars]{vars()}} attribute are derived from
the same variable.

\code{specs} attribute gives a square numeric matrix of size equal to the number
of fuzzy sets in \code{fsets}. \code{specs[i][j] == 1} if and only if the \code{i}-th fuzzy
set is more specific than \code{j}-th fuzzy set. Specificity of fuzzy sets means
the nestedness of fuzzy set: for instance, \verb{very small} is more specific than
\code{small}; similarly, \verb{extremely big} is more specific than \verb{very big}; on the
other hand, \verb{very big} and \verb{extremely small} are incomparable. A necessary
condition for specificity is subsethood.
}
\examples{
    # create a matrix of random membership degrees
    m <- matrix(runif(30), ncol=5)
    colnames(m) <- c('a1', 'a2', 'a12', 'b1', 'b2')

    # create vars - first three (a1, a2, a3) and next two (b1, b2)
    # fuzzy sets originate from the same variable
    v <- c('a', 'a', 'a', 'b', 'b')
    names(v) <- colnames(m)

    # create specificity matrix - a1 and a2 are more specific than a12,
    # the rest is incomparable
    s <- matrix(c(0, 0, 1, 0, 0,
                  0, 0, 1, 0, 0,
                  0, 0, 0, 0, 0,
                  0, 0, 0, 0, 0,
                  0, 0, 0, 0, 0), byrow=TRUE, ncol=5)
    colnames(s) <- colnames(m)
    rownames(s) <- colnames(m)

    # create a valid instance of the fsets class
    o <- fsets(m, v, s)

}
\seealso{
\code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=is.specific]{is.specific()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exprsDeprecated.R
\name{triangle}
\alias{triangle}
\alias{raisedcos}
\title{Deprecated functions to compute membership degrees of numeric fuzzy sets}
\usage{
triangle(x, lo, center, hi)

raisedcos(x, lo, center, hi)
}
\arguments{
\item{x}{A numeric vector to be transformed.}

\item{lo}{A lower bound (can be -Inf).}

\item{center}{A peak value.}

\item{hi}{An upper bound (can be Inf).}
}
\value{
A numeric vector of membership degrees of \code{x} to a fuzzy set with the shape
determined with \code{lo}, \code{center}, \code{hi}.
}
\description{
These functions compute membership degrees of numeric fuzzy sets with
triangular or raised-cosine shape. These functions are \emph{deprecated}.
Please use \code{\link[=triangular]{triangular()}} or \code{\link[=raisedcosine]{raisedcosine()}} functions instead.
}
\seealso{
\code{\link[=triangular]{triangular()}}, \code{\link[=raisedcosine]{raisedcosine()}}
}
\author{
Michal Burda
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/farules.R
\name{as.data.frame.farules}
\alias{as.data.frame.farules}
\title{Convert the instance of the \code{\link[=farules]{farules()}} S3 class into a data frame.
Empty \code{\link[=farules]{farules()}} object is converted into an empty \code{\link[=data.frame]{data.frame()}}.}
\usage{
\method{as.data.frame}{farules}(x, ...)
}
\arguments{
\item{x}{An instance of class \code{\link[=farules]{farules()}} to be transformed.}

\item{...}{Unused.}
}
\value{
A data frame of statistics of the rules that are stored in the given
\code{\link[=farules]{farules()}} object. Row names of the resulting data frame are in
the form: \verb{A1 & A2 & ... & An => C}, where \code{Ai} are antecedent
predicates and \code{C} is a consequent. An empty \code{\link[=farules]{farules()}} object
is converted into an empty \code{\link[=data.frame]{data.frame()}} object.
}
\description{
Convert the instance of the \code{\link[=farules]{farules()}} S3 class into a data frame.
Empty \code{\link[=farules]{farules()}} object is converted into an empty \code{\link[=data.frame]{data.frame()}}.
}
\seealso{
\code{\link[=farules]{farules()}}, \code{\link[=searchrules]{searchrules()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pbld.R
\name{pbld}
\alias{pbld}
\title{Perform a Perception-based Logical Deduction (PbLD) with given rule-base on
given dataset}
\usage{
pbld(
  x,
  rules,
  partition,
  values,
  type = c("global", "local"),
  parallel = FALSE
)
}
\arguments{
\item{x}{Input to the inference. It should be an object of class
\code{\link[=fsets]{fsets()}} (e.g., created by using the \code{\link[=fcut]{fcut()}} or \code{\link[=lcut]{lcut()}} functions).
It is basically a matrix with columns representing fuzzy sets.

Each row represents a single case of inference. Columns should be named
after predicates in rules' antecedents.}

\item{rules}{A rule-base (a.k.a. linguistic description) either in the form
of the \code{\link[=farules]{farules()}} object or as a list of character vectors where
each element is a fuzzy set name (a predicate) and thus each such vector
forms a rule.}

\item{partition}{A \code{\link[=fsets]{fsets()}} object with columns that are
consequents in \code{rules}. These membership degrees must correspond to
\code{values}.}

\item{values}{Crisp values that correspond to rows of membership degrees in
the \code{partition} matrix.  Function assumes that the values are sorted in
the ascending order.}

\item{type}{The type of inference to use. It can be either \code{"local"} or
\code{"global"} (default).}

\item{parallel}{Whether the processing should be run in parallel or not.
Parallelization is implemented using the \code{\link[foreach:foreach]{foreach::foreach()}}
package. The parallel environment must be set properly in advance, e.g., with
the \code{\link[doMC:registerDoMC]{doMC::registerDoMC()}} function.}
}
\value{
A vector of inferred defuzzified values. The number of resulting
values corresponds to the number of rows of the \code{x} argument.
}
\description{
Take a set of rules (a rule-base) and perform a Perception-based Logical
Deduction (PbLD) on each row of a given \code{\link[=fsets]{fsets()}} object.
}
\details{
Perform a Perception-based Logical Deduction (PbLD) with given rule-base
\code{rules} on each row of input \code{x}. Columns of \code{x} are truth
values of predicates that appear in the antecedent part of \code{rules},
\code{partition} together with \code{values} determine the shape of
predicates in consequents: each element in \code{values} corresponds to a
row of membership degrees in \code{partition}.
}
\examples{

# --- TRAINING PART ---
# custom context of the RHS variable
uptakeContext <- ctx3(7, 28.3, 46)

# convert data into fuzzy sets
d <- lcut(CO2, context=list(uptake=uptakeContext))

# split data into the training and testing set
testingIndices <- 1:5
trainingIndices <- setdiff(seq_len(nrow(CO2)), testingIndices)
training <- d[trainingIndices, ]
testing <- d[testingIndices, ]

# search for rules
r <- searchrules(training, lhs=1:38, rhs=39:58, minConfidence=0.5)

# --- TESTING PART ---
# prepare values and partition
v <- seq(uptakeContext[1], uptakeContext[3], length.out=1000)
p <- lcut(v, name='uptake', context=uptakeContext)

# do the inference
pbld(testing, r, p, v)

}
\references{
A. Dvořák, M. Štěpnička, On perception-based logical deduction
and its variants, in: Proc. 16th World Congress of the International Fuzzy
Systems Association and 9th Conference of the European Society for Fuzzy
Logic and Technology (IFSA-EUSFLAT 2015), Advances in Intelligent Systems
Research, Atlantic Press, Gijon, 2015.
}
\seealso{
\code{\link[=infer]{infer()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=searchrules]{searchrules()}}, \code{\link[=fire]{fire()}}, \code{\link[=aggregateConsequents]{aggregateConsequents()}}, \code{\link[=defuzz]{defuzz()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compose.R
\name{compose}
\alias{compose}
\title{Composition of Fuzzy Relations}
\usage{
compose(
  x,
  y,
  e = NULL,
  alg = c("goedel", "goguen", "lukasiewicz"),
  type = c("basic", "sub", "super", "square"),
  quantifier = NULL,
  sorting = sort
)
}
\arguments{
\item{x}{A first fuzzy relation to be composed. It must be a numeric matrix
with values within the \eqn{[0,1]} interval. The number of columns must
match with the number of rows of the \code{y} matrix.}

\item{y}{A second fuzzy relation to be composed. It must be a numeric matrix
with values within the \eqn{[0,1]} interval. The number of columns must
match with the number of rows of the \code{x} matrix.}

\item{e}{Deprecated. An excluding fuzzy relation. If not NULL,
it must be a numeric matrix with dimensions equal to the \code{y} matrix.}

\item{alg}{An algebra to be used for composition. It must be one of
\code{'goedel'} (default), \code{'goguen'}, or \code{'lukasiewicz'}, or an instance of class \code{algebra}
(see \code{\link[=algebra]{algebra()}}).}

\item{type}{A type of a composition to be performed. It must be one of
\code{'basic'} (default), \code{'sub'}, \code{'super'}, or \code{'square'}.}

\item{quantifier}{Deprecated. If not NULL, it must be a function taking a single
argument, a vector of relative cardinalities, that would be translated into
membership degrees. A result of the \code{\link[=lingexpr]{lingexpr()}} function is a
good candidate for that. Note that the vector of relative cardinalities contains also
two attributes, \code{x} and \code{y}, which carry the original \code{R}'s data row (in \code{x}) and \code{S}'s
feature column (in \code{y}). These attributes are accessible using the standard \code{\link[base:attr]{base::attr()}}
function. Find examples below that define some quantifiers.}

\item{sorting}{Deprecated. Sorting function used within quantifier application. The given function
must sort the membership degrees and allow the \code{decreasing} argument as in \code{\link[base:sort]{base::sort()}}.
This function have to be explicitly specified typically if performing compositions that
handle \code{NA} values.}
}
\value{
A matrix with \eqn{v} rows and \eqn{w} columns, where \eqn{v} is the
number of rows of \code{x} and \eqn{w} is the number of columns of \code{y}.
}
\description{
Composition of Fuzzy Relations
}
\details{
Function composes a fuzzy relation \code{x} (i.e., a numeric matrix of size
\eqn{(u,v)}) with a fuzzy relation \code{y} (i.e., a numeric matrix of size
\eqn{(v,w)}) and possibly with the deprecated use of an exclusion fuzzy relation
\code{e} (i.e., a numeric matrix of size \eqn{(v,w)}).

The style of composition is determined by the algebra \code{alg}, the
composition type \code{type}, and possibly also by a deprecated \code{quantifier}.

This function performs four main composition types, the basic composition (
also known as direct product), the Bandler-Kohout subproduct (also subdirect
product), the Bandler-Kohout superproduct (also supdirect product), and finally,
the Bandler-Kohout square product. More complicated composition operations
may be performed by using the \code{\link[=mult]{mult()}} function and/or by combining multiple
composition results with the \code{\link[=algebra]{algebra()}} operations.
}
\examples{
    R <- matrix(c(0.1, 0.6, 1, 0, 0, 0,
                  0, 0.3, 0.7, 0.9, 1, 1,
                  0, 0, 0.6, 0.8, 1, 0,
                  0, 1, 0.5, 0, 0, 0,
                  0, 0, 1, 1, 0, 0), byrow=TRUE, nrow=5)

    S <- matrix(c(0.9, 1, 0.9, 1,
                  1, 1, 1, 1,
                  0.1, 0.2, 0, 0.2,
                  0, 0, 0, 0,
                  0.7, 0.6, 0.5, 0.4,
                  1, 0.9, 0.7, 0.6), byrow=TRUE, nrow=6)

    RS <- matrix(c(0.6, 0.6, 0.6, 0.6,
                   1, 0.9, 0.7, 0.6,
                   0.7, 0.6, 0.5, 0.4,
                   1, 1, 1, 1,
                   0.1, 0.2, 0, 0.2), byrow=TRUE, nrow=5)

    compose(R, S, alg='goedel', type='basic') # should be equal to RS
}
\seealso{
[algebra(), \code{\link[=mult]{mult()}}, \code{\link[=lingexpr]{lingexpr()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{cbind.fsets}
\alias{cbind.fsets}
\title{Combine several 'fsets' objects into a single one}
\usage{
\method{cbind}{fsets}(..., deparse.level = 1, warn = TRUE)
}
\arguments{
\item{...}{A sequence of objects of class 'fsets' to be merged by columns.}

\item{deparse.level}{This argument has currently no function and is added
here only for compatibility with generic \code{\link[=cbind]{cbind()}} function.}

\item{warn}{Whether to issue warning when combining two fsets having the same vars
about the fact that specs may not be accurate}
}
\value{
An object of class 'fsets' that is created by merging the arguments
by columns.  Also the arguments' attributes \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}} are merged together.
}
\description{
Take a sequence of objects of class 'fsets' and combine them by columns.
This version of cbind takes care of the \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}}
attributes of the arguments and merges them to the result. If some argument
does not inherit from the 'fsets' class, an error is thrown.
}
\details{
The \code{\link[=vars]{vars()}} attribute is merged by concatenating the \code{\link[=vars]{vars()}} attributes
of each argument. Also the \code{\link[=specs]{specs()}} attributes of the arguments are merged together.
}
\examples{
    d1 <- fcut(CO2[, 1:2])
    d2 <- fcut(CO2[, 3:4], breaks=list(conc=1:4*1000/4))
    r <- cbind(d1, d2)

    print(colnames(d1))
    print(colnames(d2))
    print(colnames(r))

    print(vars(d1))
    print(vars(d2))
    print(vars(r))

    print(specs(d1))
    print(specs(d2))
    print(specs(r))

}
\seealso{
\code{\link[=vars]{vars()}}, \code{\link[=specs]{specs()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hedge.R
\name{defaultHedgeParams}
\alias{defaultHedgeParams}
\title{A list of the parameters that define the shape of the hedges.}
\usage{
defaultHedgeParams
}
\description{
A list of the parameters that define the shape of the hedges.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.frbe.R
\name{print.frbe}
\alias{print.frbe}
\title{Print an instance of the \code{\link[=frbe]{frbe()}} class}
\usage{
\method{print}{frbe}(x, ...)
}
\arguments{
\item{x}{An instance of \code{\link[=frbe]{frbe()}} class}

\item{...}{Unused.}
}
\value{
None.
}
\description{
Format an object of the \code{\link[=frbe]{frbe()}} class into human readable form
and print it to the output.
}
\details{
Format an object of the \code{\link[=frbe]{frbe()}} class into human readable form
and print it to the output.
}
\examples{

  # prepare data (from the forecast package)
  library(forecast)
  horizon <- 10
  train <- wineind[-1 * (length(wineind)-horizon+1):length(wineind)]
  test <- wineind[(length(wineind)-horizon+1):length(wineind)]
  f <- frbe(ts(train, frequency=frequency(wineind)), h=horizon)
  print(f)
  print(test)

}
\references{
Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
SYST. 2015.
}
\seealso{
\code{\link[=frbe]{frbe()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hedge.R
\name{hedge}
\alias{hedge}
\title{Linguistic hedges}
\usage{
hedge(
  type = c("ex", "si", "ve", "ty", "-", "ml", "ro", "qr", "vr"),
  hedgeParams = defaultHedgeParams
)
}
\arguments{
\item{type}{The type of the required linguistic hedge}

\item{hedgeParams}{Parameters that determine the shape of the hedges}
}
\value{
Returns a function with a single argument, which has to be a numeric
vector.
}
\description{
Returns a function that realizes linguistic hedging - i.e., transformation of linguistic
horizon (see \code{\link[=horizon]{horizon()}}) into a linguistic expression.
}
\details{
\code{hedge()} returns a function that realizes the selected linguistic hedge on its parameter:
\itemize{
\item \code{ex}: extremely,
\item \code{si}: significantly,
\item \code{ve}: very,
\item \code{ty}: typically,
\item \code{-}: empty hedge (no hedging),
\item \code{ml}: more or less,
\item \code{ro}: roughly,
\item \code{qr}: quite roughly,
\item \code{vr}: very roughly.
}

This function is quite low-level. Perhaps a more convenient way to create linguistic expressions
is to use the \code{\link[=lingexpr]{lingexpr()}} function.
}
\examples{
    a <- horizon(ctx3(), 'sm')
    plot(a)
    h <- hedge('ve')
    plot(h)
    verySmall <- function(x) h(a(x))
    plot(verySmall)

    # the last plot should be equal to:
    plot(lingexpr(ctx3(), atomic='sm', hedge='ve'))

}
\seealso{
\code{\link[=horizon]{horizon()}}, \code{\link[=lingexpr]{lingexpr()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=ctx]{ctx()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lcut.R
\name{lcut}
\alias{lcut}
\alias{lcut.default}
\alias{lcut.factor}
\alias{lcut.logical}
\alias{lcut.numeric}
\alias{lcut.data.frame}
\alias{lcut.matrix}
\title{Transform data into a \code{fsets} S3 class of linguistic fuzzy attributes}
\usage{
lcut(x, ...)

\method{lcut}{default}(x, ...)

\method{lcut}{factor}(x, name = deparse(substitute(x)), ...)

\method{lcut}{logical}(x, name = deparse(substitute(x)), ...)

\method{lcut}{numeric}(
  x,
  context = minmax,
  atomic = c("sm", "me", "bi", "lm", "um", "ze", "neg.sm", "neg.me", "neg.bi", "neg.lm",
    "neg.um"),
  hedges = c("ex", "si", "ve", "ty", "-", "ml", "ro", "qr", "vr"),
  name = deparse(substitute(x)),
  hedgeParams = defaultHedgeParams,
  ...
)

\method{lcut}{data.frame}(
  x,
  context = minmax,
  atomic = c("sm", "me", "bi", "lm", "um", "ze", "neg.sm", "neg.me", "neg.bi", "neg.lm",
    "neg.um"),
  hedges = c("ex", "si", "ve", "ty", "-", "ml", "ro", "qr", "vr"),
  ...
)

\method{lcut}{matrix}(x, ...)
}
\arguments{
\item{x}{Data to be transformed: if it is a numeric vector, matrix, or data
frame, then the creation of linguistic fuzzy attributes takes place. For
other data types the \code{\link[=fcut]{fcut()}} function is called implicitly.}

\item{...}{Other parameters to some methods.}

\item{name}{A name to be added as a suffix to the created fuzzy attribute
names. This parameter can be used only if \code{x} is a numeric or logical
vector or a factor. If \code{x} is a matrix or data frame, \code{name} should be NULL
because the fuzzy attribute names are taken from column names of parameter
\code{x}. The \code{name} is also used as a value for the \code{vars} attribute of the
resulting \code{\link[=fsets]{fsets()}} instance.}

\item{context}{A definition of context of a numeric attribute. It must be an
instance of an S3 class \code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx5]{ctx5()}}, \code{\link[=ctx3bilat]{ctx3bilat()}} or \code{\link[=ctx5bilat]{ctx5bilat()}}.

If \code{x} is a matrix or data frame then \code{context} should be a named list of
contexts for each \code{x}'s column.}

\item{atomic}{A vector of atomic linguistic expressions to be used for
creation of fuzzy attributes.}

\item{hedges}{A vector of linguistic hedges to be used for creation of fuzzy
attributes.}

\item{hedgeParams}{Parameters that determine the shape of the hedges}
}
\value{
An object of S3 class \code{fsets} is returned, which is a numeric matrix
with columns representing the fuzzy attributes. Each source column of the
\code{x} argument corresponds to multiple columns in the resulting matrix.
Columns will have names derived from used \eqn{hedges}, atomic expression,
and \eqn{name} specified as the optional parameter.

The resulting object would also have set the \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}}
properties with the former being created from original column names (if \code{x}
is a matrix or data frame) or the \code{name} argument (if \code{x} is a numeric
vector). The \code{\link[=specs]{specs()}} incidency matrix would be created to reflect the
following order of the hedges: \eqn{"ex" < "si" < "ve" < "-" < "ml" < "ro"
  < "qr" < "vr"} and \eqn{"ty" < "" < "ml" < "ro" < "qr" < "vr"}.  Fuzzy
attributes created from the same source numeric vector (or column) would be
ordered that way, with other fuzzy attributes (from the other source) being
incomparable.
}
\description{
This function creates a set of linguistic fuzzy attributes from crisp data.
Numeric vectors, matrix or data frame columns are transformed into a set of
fuzzy attributes, i.e., columns with membership degrees. Factors and other
data types are transformed to fuzzy attributes by calling the \code{\link[=fcut]{fcut()}}
function.
}
\details{
The aim of this function is to transform numeric data into a set of fuzzy
attributes.  The resulting fuzzy attributes have direct linguistic
interpretation. This is a unique variant of fuzzification that is suitable
for the inference mechanism based on Perception-based Linguistic Description
(PbLD) -- see \code{\link[=pbld]{pbld()}}.

A numeric vector is transformed into a set of fuzzy attributes according to
the following scheme:

\eqn{<hedge> <atomic expression>}

where \eqn{<atomic expression>} is an atomic linguistic expression, a value
from the following possibilities (note that the allowance of atomic
expressions is influenced with \code{context} being used - see \link{ctx} for details):
\itemize{
\item \code{neg.bi}: big negative (far from zero)
\item \code{neg.um}: upper medium negative (between medium negative and big negative)
\item \code{neg.me}: medium negative
\item \code{neg.lm}: lower medium negative (between medium negative and small
negative)
\item \code{neg.sm}: small negative (close to zero)
\item \code{ze}: zero
\item \code{sm}: small
\item \code{lm}: lower medium
\item \code{me}: medium
\item \code{um}: upper medium
\item \code{bi}: big
}

A \eqn{<hedge>} is a modifier that further concretizes the atomic expression
(note that not each combination of hedge and atomic expression is allowed -
see \link{allowed.lingexpr} for more details):
\itemize{
\item \code{ex}: extremely,
\item \code{si}: significantly,
\item \code{ve}: very,
\item \code{ty}: typically,
\item \code{-}: empty hedge (no hedging),
\item \code{ml}: more or less,
\item \code{ro}: roughly,
\item \code{qr}: quite roughly,
\item \code{vr}: very roughly.
}

Accordingly to the theory developed by Novak (2008), not every hedge is
suitable with each atomic expression (see the description of the \code{hedges}
argument).  The hedges to be used can be selected with the \code{hedges} argument.
Function takes care of not to use hedge together with an unapplicable atomic
expression by itself.

Obviously, distinct data have different meaning of what is "small", "medium",
or "big" etc.  Therefore, a \code{context} has to be set that specifies sensible
values for these linguistic expressions.

If a matrix (resp. data frame) is provided to this function instead of a
single vector, all columns are processed the same way.

The function also sets up properly the \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}} properties of
the result.
}
\examples{

# transform a single vector
x <- runif(10)
lcut(x, name='age')

# transform single vector with a custom context
lcut(x, context=ctx5(0, 0.2, 0.5, 0.7, 1), name='age')

# transform all columns of a data frame
# and do not use any hedges
data <- CO2[, c('conc', 'uptake')]
lcut(data)


# definition of custom contexts for different columns
# of a data frame while selecting only "ve" and "ro" hedges.
lcut(data,
     context=list(conc=minmax,
                  uptake=ctx3(0, 25, 50)),
     hedges=c('ve', 'ro'))


# lcut on non-numeric data is the same as fcut()
ff <- factor(substring("statistics", 1:10, 1:10), levels = letters)
lcut(ff)

}
\references{
V. Novak, A comprehensive theory of trichotomous evaluative
linguistic expressions, Fuzzy Sets and Systems 159 (22) (2008) 2939--2969.
}
\seealso{
\code{\link[=fcut]{fcut()}}, \code{\link[=fsets]{fsets()}}, \code{\link[=vars]{vars()}}, \code{\link[=specs]{specs()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lfl-package.R
\docType{package}
\name{lfl-package}
\alias{lfl}
\alias{lfl-package}
\title{lfl - Linguistic Fuzzy Logic}
\description{
Various algorithms related (not only) to linguistic fuzzy logic: mining for linguistic fuzzy association
rules, composition of fuzzy relations, performing perception-based logical deduction (PbLD),
and forecasting time-series using fuzzy rule-based ensemble (FRBE).
}
\details{
Other methods include fuzzy transform and computation of Sugeno integrals. Also, basic algebraic
functions related to fuzzy logic are included, which allow handling missing values using different
styles such as Kleene, Bochvar, Sobocinski and others.
}
\author{
\strong{Maintainer}: Michal Burda \email{michal.burda@osu.cz} (\href{https://orcid.org/0000-0002-4182-4407}{ORCID})

Authors:
\itemize{
  \item Michal Burda \email{michal.burda@osu.cz} (\href{https://orcid.org/0000-0002-4182-4407}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{is.fsets}
\alias{is.fsets}
\title{Test whether \code{x} is a valid object of the S3 \code{fsets} class}
\usage{
is.fsets(x)
}
\arguments{
\item{x}{An object being tested.}
}
\value{
TRUE if \code{x} is a valid \code{fsets} object and FALSE otherwise.
}
\description{
This function tests whether \code{x} inherits from S3 \code{fsets} class.
}
\seealso{
\code{\link[=fsets]{fsets()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/horizon.R
\name{horizon}
\alias{horizon}
\title{Create a function that computes linguistic horizons}
\usage{
horizon(
  context,
  atomic = c("sm", "me", "bi", "lm", "um", "ze", "neg.sm", "neg.me", "neg.bi", "neg.lm",
    "neg.um")
)
}
\arguments{
\item{context}{A context of linguistic expressions (see \code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx5]{ctx5()}}, \code{\link[=ctx3bilat]{ctx3bilat()}} or \code{\link[=ctx5bilat]{ctx5bilat()}})}

\item{atomic}{An atomic expression whose horizon we would like to obtain}
}
\value{
A function of single argument that must be a numeric vector
}
\description{
Based on given \code{context} and \code{atomic} expression, this function returns a function that computes a linguistic
horizon, i.e., a triangular function representing basic limits of what humans treat as "small", "medium", "big" etc.
within given \code{context}. Linguistic horizon stands as a base for creation of linguistic expressions. A linguistic
expression is created by applying a \code{\link[=hedge]{hedge()}} on horizon. (Atomic linguistic expression is created from horizon by
applying an empty (\code{-}) hedge).
}
\details{
The values of the \code{atomic} parameter have the following meaning (in ascending order):
\itemize{
\item \code{neg.bi}: big negative (far from zero)
\item \code{neg.um}: upper medium negative (between medium negative and big negative)
\item \code{neg.me}: medium negative
\item \code{neg.lm}: lower medium negative (between medium negative and small negative)
\item \code{neg.sm}: small negative (close to zero)
\item \code{ze}: zero
\item \code{sm}: small
\item \code{lm}: lower medium
\item \code{me}: medium
\item \code{um}: upper medium
\item \code{bi}: big
}

Based on the context type, the following atomic expressions are allowed:
\itemize{
\item \code{\link[=ctx3]{ctx3()}} (trichotomy): small, medium, big;
\item \code{\link[=ctx5]{ctx5()}} (pentachotomy): small, lower medium, medium, upper medium, big;
\item \code{\link[=ctx3bilat]{ctx3bilat()}} (bilateral trichotomy): negative big, negative medium, negative small,
zero, small, medium, big;
\item \code{\link[=ctx5bilat]{ctx5bilat()}} (bilateral pentachotomy): negative big, negative medium, negative
small, zero, small, medium, big.
}

This function is quite low-level. Perhaps a more convenient way to create linguistic expressions
is to use the \code{\link[=lingexpr]{lingexpr()}} function.
}
\examples{
    plot(horizon(ctx3(), 'sm'), from=-1, to=2)
    plot(horizon(ctx3(), 'me'), from=-1, to=2)
    plot(horizon(ctx3(), 'bi'), from=-1, to=2)

    a <- horizon(ctx3(), 'sm')
    plot(a)
    h <- hedge('ve')
    plot(h)
    verySmall <- function(x) h(a(x))
    plot(verySmall)
}
\seealso{
\code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx5]{ctx5()}}, \code{\link[=ctx3bilat]{ctx3bilat()}}, \code{\link[=ctx5bilat]{ctx5bilat()}}, \code{\link[=hedge]{hedge()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algebraNA.R
\name{sobocinski}
\alias{sobocinski}
\alias{algebraNA}
\alias{kleene}
\alias{dragonfly}
\alias{nelson}
\alias{lowerEst}
\alias{lowerBound}
\alias{upperBound}
\title{Modify algebra's way of computing with \code{NA} values.}
\usage{
sobocinski(algebra)

kleene(algebra)

dragonfly(algebra)

nelson(algebra)

lowerEst(algebra)

lowerBound(algebra)

upperBound(algebra)
}
\arguments{
\item{algebra}{the underlying algebra object to be modified -- see the \code{\link[=algebra]{algebra()}} function}
}
\value{
A list of function of the same structure as is the list returned from the \code{\link[=algebra]{algebra()}} function
}
\description{
By default, the objects created with the \code{\link[=algebra]{algebra()}} function represent a mathematical
algebra capable to work on the \eqn{[0,1]} interval. If \code{NA} appears as a value instead,
it is propagated to the result. That is, any operation with \code{NA} results in \code{NA}, by default.
This scheme of handling missing values is also known as Bochvar's. To change this default
behavior, the following functions may be applied.
}
\details{
The \code{sobocinski()}, \code{kleene()}, \code{nelson()}, \code{dragonfly()}, \code{lowerEst()} and \code{lowerBound()}
functions modify the algebra to
handle the \code{NA} in a different way than is the default. Sobocinski's algebra simply ignores \code{NA} values
whereas Kleene's algebra treats \code{NA} as "unknown value". Dragonfly approach is a combination
of Sobocinski's and Bochvar's approach, which preserves the ordering \verb{0 <= NA <= 1}
to obtain from compositions (see \code{\link[=compose]{compose()}})
the lower-estimate in the presence of missing values.

In detail, the behavior of the algebra modifiers is defined as follows:

Sobocinski's negation for \code{n} being the underlying algebra:
\tabular{ll}{
a  \tab n(a)\cr
NA \tab 0
}

Sobocinski's operation for \code{op} being one of \code{t}, \code{pt}, \code{c}, \code{pc}, \code{i}, \code{pi}, \code{s}, \code{ps}
from the underlying algebra:
\tabular{lll}{
\tab b        \tab NA\cr
a  \tab op(a, b) \tab a \cr
NA \tab b        \tab NA
}

Sobocinski's operation for \code{r} from the underlying algebra:
\tabular{lll}{
\tab b       \tab NA  \cr
a  \tab r(a, b) \tab n(a)\cr
NA \tab b       \tab NA
}

Kleene's negation is identical to \code{n} from the underlying algebra.

Kleene's operation for \code{op} being one of \code{t}, \code{pt}, \code{i}, \code{pi} from the underlying algebra:
\tabular{llll}{
\tab b        \tab NA \tab 0\cr
a  \tab op(a, b) \tab NA \tab 0\cr
NA \tab NA       \tab NA \tab 0\cr
0  \tab 0        \tab 0  \tab 0
}

Kleene's operation for \code{op} being one of \code{c}, \code{pc}, \code{s}, \code{ps} from the underlying algebra:
\tabular{llll}{
\tab b        \tab NA \tab 1\cr
a  \tab op(a, b) \tab NA \tab 1\cr
NA \tab NA       \tab NA \tab 1\cr
1  \tab 1        \tab 1  \tab 1
}

Kleene's operation for \code{r} from the underlying algebra:
\tabular{llll}{
\tab b        \tab NA \tab 1\cr
a  \tab r(a, b)  \tab NA \tab 1\cr
NA \tab NA       \tab NA \tab 1\cr
0  \tab 1        \tab 1  \tab 1
}

Dragonfly negation is identical to \code{n} from the underlying algebra.

Dragonfly operation for \code{op} being one of \code{t}, \code{pt}, \code{i}, \code{pi} from the underlying algebra:
\tabular{lllll}{
\tab b        \tab NA \tab 0 \tab 1\cr
a  \tab op(a, b) \tab NA \tab 0 \tab a\cr
NA \tab NA       \tab NA \tab 0 \tab NA\cr
0  \tab 0        \tab 0  \tab 0 \tab 0\cr
1  \tab b        \tab NA \tab 0 \tab 1
}

Dragonfly operation for \code{op} being one of \code{c}, \code{pc}, \code{s}, \code{ps} from the underlying algebra:
\tabular{lllll}{
\tab b        \tab NA \tab 0  \tab 1\cr
a  \tab op(a, b) \tab a  \tab a  \tab 1\cr
NA \tab b        \tab NA \tab NA \tab 1\cr
0  \tab b        \tab NA \tab 0  \tab 1\cr
1  \tab 1        \tab 1  \tab 1  \tab 1
}

Dragonfly operation for \code{r} from the underlying algebra:
\tabular{lllll}{
\tab b       \tab NA \tab 0    \tab 1\cr
a  \tab r(a, b) \tab NA \tab n(a) \tab 1\cr
NA \tab b       \tab 1  \tab NA   \tab 1\cr
0  \tab 1       \tab 1  \tab 1    \tab 1\cr
1  \tab b       \tab NA \tab 0    \tab 1
}

LowerEst and LowerBound algebras are equivalent to Dragonfly's algebra in
\code{t}, \code{pt}, \code{i}, \code{pi}, \code{c}, \code{pc}, \code{s}, \code{ps} operations, but differ in the
\code{r} and \code{n} operations. \code{n} is equal to the Sobocinski's negation for both
LowerEst and LowerBound algebras.

LowerEst's operation for \code{r} from the underlying algebra:

\tabular{lllll}{
\tab b       \tab NA \tab 0    \tab 1\cr
a  \tab r(a, b) \tab NA \tab n(a) \tab 1\cr
NA \tab b       \tab NA \tab 0    \tab 1\cr
0  \tab 1       \tab 1  \tab 1    \tab 1\cr
1  \tab b       \tab NA \tab 0    \tab 1
}

LowerBound's operation for \code{r} from the underlying algebra:

\tabular{lllll}{
\tab b       \tab NA \tab 0    \tab 1\cr
a  \tab r(a, b) \tab NA \tab n(a) \tab 1\cr
NA \tab 1       \tab 1  \tab 0    \tab 1\cr
0  \tab 1       \tab 1  \tab 1    \tab 1\cr
1  \tab b       \tab NA \tab 0    \tab 1
}
}
\examples{
a <- algebra('lukas')
b <- sobocinski(a)

a$t(0.3, NA)  # NA
b$t(0.3, NA)  # 0.3

}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evalfrbe.R
\name{evalfrbe}
\alias{evalfrbe}
\title{Evaluate the performance of the FRBE forecast}
\usage{
evalfrbe(fit, real, error = c("smape", "mase", "rmse"))
}
\arguments{
\item{fit}{A FRBE model of class \code{frbe} as returned by the \code{\link[=frbe]{frbe()}} function.}

\item{real}{A numeric vector of real (known) values. The vector must
correspond to the values being forecasted, i.e., the length must be the same
as the horizon forecasted by \code{\link[=frbe]{frbe()}}.}

\item{error}{Error measure to be computed. It can be either Symmetric Mean
Absolute Percentage Error (\code{smape}), Mean Absolute Scaled Error (\code{mase}), or
Root Mean Squared Error (\code{rmse}). See also \code{\link[=smape]{smape()}}, \code{\link[=mase]{mase()}}, and \code{\link[=rmse]{rmse()}}
for more details.}
}
\value{
Function returns a data.frame with single row and columns
corresponding to the error of the individual forecasting methods that the
FRBE is computed from. Additionally to this, a column "avg" is added with
error of simple average of the individual forecasting methods and a column
"frbe" with error of the FRBE forecasts.
}
\description{
Take a FRBE forecast and compare it with real values using arbitrary error
function.
}
\details{
Take a FRBE forecast and compare it with real values by evaluating a given
error measure.  FRBE forecast should be made for a horizon of the same value
as length of the vector of real values.
}
\examples{

  # prepare data (from the forecast package)
  library(forecast)
  horizon <- 10
  train <- wineind[-1 * (length(wineind)-horizon+1):length(wineind)]
  test <- wineind[(length(wineind)-horizon+1):length(wineind)]
  f <- frbe(ts(train, frequency=frequency(wineind)), h=horizon)
  evalfrbe(f, test)

}
\references{
Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
SYST. 2015.
}
\seealso{
\code{\link[=frbe]{frbe()}}, \code{\link[=smape]{smape()}}, \code{\link[=mase]{mase()}}, \code{\link[=rmse]{rmse()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcut.R
\name{fcut}
\alias{fcut}
\alias{fcut.default}
\alias{fcut.numeric}
\alias{fcut.matrix}
\alias{fcut.data.frame}
\alias{fcut.factor}
\alias{fcut.logical}
\title{Transform data into a \code{fsets} S3 class using shapes derived from
triangles or raised cosines}
\usage{
fcut(x, ...)

\method{fcut}{default}(x, ...)

\method{fcut}{factor}(x, name = deparse(substitute(x)), ...)

\method{fcut}{logical}(x, name = deparse(substitute(x)), ...)

\method{fcut}{numeric}(
  x,
  breaks,
  name = deparse(substitute(x)),
  type = c("triangle", "raisedcos"),
  merge = 1,
  parallel = FALSE,
  ...
)

\method{fcut}{data.frame}(
  x,
  breaks = NULL,
  name = NULL,
  type = c("triangle", "raisedcos"),
  merge = 1,
  parallel = FALSE,
  ...
)

\method{fcut}{matrix}(x, ...)
}
\arguments{
\item{x}{Data to be transformed: a vector, matrix, or data frame.
Non-numeric data are allowed.}

\item{...}{Other parameters to some methods.}

\item{name}{A name to be added as a suffix to the created fuzzy attribute
names. This parameter can be used only if \code{x} is a vector. If \code{x}
is a matrix or data frame, \code{name} should be NULL because the fuzzy
attribute names are taken from column names of the argument \code{x}.}

\item{breaks}{This argument determines the break-points of the positions of
the fuzzy sets (see also \code{\link[=equidist]{equidist()}}. It should be an ordered vector of
numbers such that the \eqn{i}-th index specifies the beginning,
\eqn{(i+1)}-th the center, and \eqn{(i+2)}-th the ending of the \eqn{i}-th
fuzzy set.

I.e. the minimum number of breaks-points is 3; \eqn{n-2} elementary fuzzy
sets would be created for \eqn{n} break-points.

If considering an i-th fuzzy set (of \code{type='triangle'}), \code{x}
values lower than \eqn{i}-th break (and greater than \eqn{(i+2)}-th break)
would result in zero membership degree, values equal to \eqn{(i+1)}-th break
would have membership degree equal 1 and values between them the appropriate
membership degree between 0 and 1.

The resulting fuzzy sets would be named after the original data by adding
dot (".") and a number \eqn{i} of fuzzy set.

Unlike \code{\link[base:cut]{base::cut()}}, \code{x} values, that are lower or greater than
the given break-points, will have all membership degrees equal to zero.

For non-numeric data, this argument is ignored. For \code{x} being a numeric
vector, it must be a vector of numeric values. For \code{x} being a numeric
matrix or data frame, it must be a named list containing a numeric vector
for each column - if not, the values are repeated for each column.}

\item{type}{The type of fuzzy sets to create. Currently, \code{'triangle'} or
\code{'raisedcos'} may be used. The \code{type} argument may be also a
function with 3 or 4 arguments:
\itemize{
\item if \code{type} is a 4-argument function, it is assumed that that it computes
membership degrees from values of the first argument while considering
the boundaries given by the next 3 arguments;
\item if \code{type} is a 3-argument function, it is assumed that it is a factory
function similar to \code{\link[=triangular]{triangular()}} or \code{\link[=raisedcosine]{raisedcosine()}}, which, from given
three boundaries, creates a function that computes membership degrees.
}}

\item{merge}{This argument determines whether to derive additional fuzzy
sets by merging the elementary fuzzy sets (whose position is determined with
the \code{breaks} argument) into super-sets.  The argument is ignored for
non-numeric data in \code{x}.

\code{merge} may contain any integer number from \code{1} to
\code{length(breaks) - 2}.  Value \code{1} means that the elementary fuzzy
sets should be present in the output.  Value \code{2} means that the two
consecutive elementary fuzzy sets should be combined by using the Lukasiewic
t-conorm, value \code{3} causes combining three consecutive elementary fuzzy
sets etc.

The names of the derived (merged) fuzzy sets is derived from the names of
the original elementary fuzzy sets by concatenating them with the "|" (pipe)
separator.}

\item{parallel}{Whether the processing should be run in parallel or not.
Parallelization is implemented using the \code{\link[foreach:foreach]{foreach::foreach()}} function.
The parallel environment must be set properly in advance, e.g., with
the \code{\link[doMC:registerDoMC]{doMC::registerDoMC()}} function.  Currently this argument is
applied only if \code{x} is a matrix or data frame.}
}
\value{
An object of class "fsets" is returned, which is a numeric matrix
with columns representing the fuzzy attributes. Each source column of the
\code{x} argument corresponds to multiple columns in the resulting matrix.
Columns have names that indicate the name of the source as well as a index
\eqn{i} of fuzzy set(s) -- see the description of arguments \code{breaks}
and \code{merge} above.

The resulting object would also have set the \code{\link[=vars]{vars()}} and
\code{\link[=specs]{specs()}} properties with the former being created from original
column names (if \code{x} is a matrix or data frame) or the \code{name}
argument (if \code{x} is a numeric vector). The \code{\link[=specs]{specs()}}
incidency matrix would be created to reflect the superset-hood of the merged
fuzzy sets.
}
\description{
This function creates a set of fuzzy attributes from crisp data. Factors,
numeric vectors, matrix or data frame columns are transformed into a set of
fuzzy attributes, i.e., columns with membership degrees. Unlike
\code{\link[=lcut]{lcut()}}, the linguistic approach is not used for transformation,
but rather partitioning using regular shapes of the fuzzy sets (such as
triangles or raised cosines).
}
\details{
The aim of this function is to transform numeric data into a set of fuzzy
attributes. The result is in the form of an object of class "fsets", i.e.,
a numeric matrix whose columns represent fuzzy sets (fuzzy attributes) with
values being the membership degrees.

The function behaves differently depending on the type of input \code{x}.

If \code{x} is a factor or a logical vector (or other non-numeric data) then
for each distinct value of an input, a fuzzy set is created, and data would
be transformed into crisp membership degrees 0 or 1 only.

If \code{x} is a numeric vector then fuzzy sets are created according to
break-points specified in the \code{breaks} argument with 1st, 2nd and 3rd
break-point specifying the first fuzzy set, 2nd, 3rd and 4th break-point
specifying the second fuzzy set etc. The shape of the fuzzy set is determined
by the \code{type} argument that may be equal either to a string
\code{'triangle'} or \code{'raisedcos'} or it could be a function that
computes the membership degrees itself (see \code{\link[=triangular]{triangular()}} or
\code{\link[=raisedcosine]{raisedcosine()}} functions for details). Additionally, super-sets of
these elementary sets may be created by specifying the \code{merge}
argument. Values of this argument specify how many consecutive fuzzy sets
should be combined (by using Lukasiewicz's t-conorm) to produce
super-sets - see the description of \code{merge} above.

If a matrix (resp. data frame) is provided to this function instead of a
single vector, all columns are processed separately as described above and
the result is combined with the \code{\link[=cbind.fsets]{cbind.fsets()}} function.

The function sets up properly the \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}}
properties of the result.
}
\examples{

# fcut on non-numeric data
ff <- factor(substring("statistics", 1:10, 1:10), levels = letters)
fcut(ff)

# transform a single vector into a single fuzzy set
x <- runif(10)
fcut(x, breaks=c(0, 0.5, 1), name='age')

# transform single vector into a partition of the interval 0-1
# (the boundary triangles are right-angled)
fcut(x, breaks=c(0, 0, 0.5, 1, 1), name='age')

# also create supersets
fcut(x, breaks=c(0, 0, 0.5, 1, 1), name='age', merge=c(1, 2))

# transform all columns of a data frame
# with different breakpoints
data <- CO2[, c('conc', 'uptake')]
fcut(data, breaks=list(conc=c(95, 95, 350, 1000, 1000),
                       uptake=c(7, 7, 28.3, 46, 46)))

# using a custom 3-argument function (a function factory):
f <- function(a, b, c) {
  return(function(x) ifelse(a <= x & x <= b, 1, 0))
}
fcut(x, breaks=c(0, 0.5, 1), name='age', type=f)

# using a custom 4-argument function:
f <- function(x, a, b, c) {
  return(ifelse(a <= x & x <= b, 1, 0))
}
fcut(x, breaks=c(0, 0.5, 1), name='age', type=f)

}
\seealso{
\code{\link[=lcut]{lcut()}}, \code{\link[=equidist]{equidist()}}, \code{\link[=farules]{farules()}}, \code{\link[=pbld]{pbld()}}, \code{\link[=vars]{vars()}}, \code{\link[=specs]{specs()}},
\code{\link[=cbind.fsets]{cbind.fsets()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minmax.R
\name{minmax}
\alias{minmax}
\title{Creating linguistic context directly from values}
\usage{
minmax(x, type = c("ctx3", "ctx5", "ctx3bilat", "ctx5bilat"), ...)
}
\arguments{
\item{x}{A numeric vector to compute the context from}

\item{type}{A type of the context to be returned. Must be one of:
\code{ctx3}, \code{ctx5}, \code{ctx3bilat} or \code{ctx5bilat}}

\item{...}{other parameters to be passed to the appropriate constructor
(\code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx3bilat]{ctx3bilat()}}, \code{\link[=ctx5]{ctx5()}}, and \code{\link[=ctx5bilat]{ctx5bilat()}}) that is called internally.
These values overwrite the defaults computed by \code{minmax} -- see the examples.}
}
\description{
This function creates a context (i.e., an instance of S3 class
\code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx3bilat]{ctx3bilat()}}, \code{\link[=ctx5]{ctx5()}}, or \code{\link[=ctx5bilat]{ctx5bilat()}}) based on values
of the numeric vector \code{x}. In default, the context is based on minimum
and maximum value of \code{x} in the following way:
\itemize{
\item \code{ctx3}, \code{ctx5}: low = minimum, high = maximum value of \code{x};
\item \code{ctx3bilat}, \code{ctx5bilat}: negMax = minimum, max = maximum value of \code{x},
origin = mean of minimum and maximum.
}
}
\details{
Other values are computed according to the defaults as defined in the constructors
\code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx3bilat]{ctx3bilat()}}, \code{\link[=ctx5]{ctx5()}}, and \code{\link[=ctx5bilat]{ctx5bilat()}}).
}
\examples{
  minmax(0:100)                # returns ctx3: 0, 50, 100
  minmax(0:100, high=80)       # returns ctx3: 0, 40, 80
  minmax(0:100, relCenter=0.4) # returns ctx3: 0, 40, 100
  minmax(0:100, type='ctx5')   # returns ctx5: 0, 25, 50, 75, 100
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.frbe.R
\name{is.frbe}
\alias{is.frbe}
\title{Test whether \code{x} is a valid object of the S3 \code{frbe} class}
\usage{
is.frbe(x)
}
\arguments{
\item{x}{An object being tested.}
}
\value{
\code{TRUE} if \code{x} is a valid \code{frbe} object and \code{FALSE} otherwise.
}
\description{
Test whether \code{x} has a valid format for the objects of the S3 \code{frbe} class.
}
\details{
This function tests whether \code{x} inherits from \code{frbe} i.e., whether
it is a list with the following elements: \code{forecasts} data frame,
\code{features} data frame, \code{weights} vector, and \code{mean} vector.
Instances of the S3 class \code{frbe} are usually created by the \code{\link[=frbe]{frbe()}} function.
}
\references{
Štěpnička, M., Burda, M., Štěpničková, L. Fuzzy Rule Base
Ensemble Generated from Data by Linguistic Associations Mining. FUZZY SET
SYST. 2015.
}
\seealso{
\code{\link[=frbe]{frbe()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/perceive.R
\name{perceive}
\alias{perceive}
\title{From a set of rules, remove each rule for which another rule exists that is
more specific.}
\usage{
perceive(
  rules,
  fsets,
  type = c("global", "local"),
  fired = NULL,
  vars = NULL,
  specs = NULL
)
}
\arguments{
\item{rules}{A list of character vectors where each element is a fuzzy set
name (a predicate) and thus each such vector forms a rule.}

\item{fsets}{A valid instance of the \code{\link[=fsets]{fsets()}} class such that all predicates
in \code{rules} (i.e., all values of all character vectors in \code{rules$rules})
can be found in \code{colnames(fsets)}}

\item{type}{The type of perception to use. It can be either \code{"local"}
or \code{"global"} (default).}

\item{fired}{If \code{type=="global"} then this argument can be NULL. If
\code{type} is \code{"local"} then \code{fired} must be a numeric vector of
values in the interval \eqn{[0,1]} indicating the truth values of all rules,
i.e., the length of the vector must be equal to the number of rules in the
\code{rules} argument.}

\item{vars}{A deprecated parameter that must be \code{NULL}. Formerly, it was
a named (typically character) vector that determined which
predicates originate from the same variable, i.e., which of them semantically
deal with the same property.  For that purpose, each value from any vector
stored in the \code{rules} list must be present in \code{names(vars)}.  See
also \code{\link[=vars]{vars()}} function of the \code{\link[=fsets]{fsets()}} class.}

\item{specs}{A deprecated parameter that must be \code{NULL}. Formerly, it was
a square numeric matrix containing values from \eqn{\{0, 1\}}.
It is a specificity matrix for which each row and column corresponds to an
\code{rules}'es predicate \code{specs[i][j] = 1} if and only if the
\eqn{i}-th predicate is more specific (i.e., the corresponding fuzzy set is a
subset of) than the \eqn{j}-th predicate (i.e. \code{x[, j]}).  See also
\code{\link[=specs]{specs()}} function of the \code{\link[=fsets]{fsets()}} class.}
}
\value{
A modified list of rules for which no other more specific rule
exists. (Each rule is a vector.)
}
\description{
Examine rules in a list and remove all of them for which other more specific
rules are present in the list. The specificity is determined by calling the
\code{\link[=is.specific]{is.specific()}} function.  This operation is a part of the
\code{\link[=pbld]{pbld()}} inference mechanism.
}
\details{
In other words, for each rule \code{x} in the \code{rules} list, it searches for another
rule \code{y} such that \code{is.specific(y, x)} returns TRUE. If yes then
\code{x} is removed from the list.
}
\examples{
# prepare fsets
f <- lcut(data.frame(a=0:1, b=0:1, c=0:1, d=0:1))

# run perceive function: (sm.a, bi.c) has
# more specific rule (ve.sm.a, bi.c)
perceive(list(c('sm.a', 'bi.c'),
              c('ve.sm.a', 'bi.c'),
              c('sm.b', 'sm.d')),
         f)

}
\seealso{
\code{\link[=is.specific]{is.specific()}}, \code{\link[=fsets]{fsets()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/farules.R
\name{is.farules}
\alias{is.farules}
\title{Test whether \code{x} inherits from the S3 \code{farules} class.}
\usage{
is.farules(x)
}
\arguments{
\item{x}{An object being tested.}
}
\value{
\code{TRUE} if \code{x} is a valid \code{\link[=farules]{farules()}} object and \code{FALSE} otherwise.
}
\description{
Test whether \code{x} inherits from the S3 \code{farules} class.
}
\seealso{
\code{\link{farules}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ftinv.R
\name{ftinv}
\alias{ftinv}
\title{Inverse of the fuzzy transform}
\usage{
ftinv(fit, x, xmemb)
}
\arguments{
\item{fit}{The fuzzy transform object as the instance of the \code{ft} S3 class}

\item{x}{The numeric matrix of input values, for which the inverse fuzzy transform
has to be computed}

\item{xmemb}{the partitioning of input values, i.e., a \code{fsets} object with membership degrees
(see \code{\link[=fcut]{fcut()}}). Such partitioning must correspond to the \code{xmemb} partitioning used to
create \code{fit} using the \code{\link[=ft]{ft()}} function.}
}
\value{
The inverse of the fuzzy transform \code{fit}, i.e., the approximated values
of the original function that was the subject of the fuzzy transform
}
\description{
Compute an inverse of fuzzy transform \code{fit} for values \code{x} with corresponding
membership degrees \code{xmemb}.
}
\examples{

# create the fuzzy transform object
y <- (1:30)^2
x <- as.matrix(data.frame(a = 1:30, b = 30:1))
xmemb <- fcut(x,
              breaks = list(a = equidist(x[, 'a'], 3),
                            b = equidist(x[, 'b'], 3)))
fit <- ft(x, xmemb, y, order = 1)

# obtain function values
x2 <- as.matrix(data.frame(a = 10:20, b = 20:10))
xmemb2 <- fcut(x2,
               breaks = list(a = equidist(x[, 'a'], 3),
                             b = equidist(x[, 'b'], 3)))
y2 <- ftinv(fit, x2, xmemb2)
print(y2)

# compare original values with those obtained by the fuzzy transform
y - y2

}
\references{
Perfilieva I. Fuzzy transforms: Theory and applications. FUZZY SET SYST,
volume 157, issue 8, p. 993-1023. 2006.
}
\seealso{
\code{\link[=ft]{ft()}}, \code{\link[=is.ft]{is.ft()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fire.R
\name{fire}
\alias{fire}
\title{Evaluate rules and obtain truth-degrees}
\usage{
fire(
  x,
  rules,
  tnorm = c("goedel", "goguen", "lukasiewicz"),
  onlyAnte = TRUE,
  parallel = FALSE
)
}
\arguments{
\item{x}{Truth degrees of predicates. \code{x} could be either a numeric
matrix or a numeric vector. If vector is given then each named element represents
a truth value of a predicate. If matrix is given then each row of the matrix
is evaluated sequentially as a vector. The values must be in the interval
\eqn{[0, 1]}.}

\item{rules}{Either an object of S3 class \code{\link[=farules]{farules()}} or a list of character
vectors where each vector is a rule in a conjunctive form. Elements of these
character vectors (i.e., predicate names) must correspond to
the \code{x}'s names (of elements resp. columns if \code{x} is a vector resp. matrix).}

\item{tnorm}{A character string representing a triangular norm to be used
(either \code{"goedel"}, \code{"goguen"}, or \code{"lukasiewicz"}) or an arbitrary function
that performs element-wise computation on arbitrary number of vector parameters
similarly as e.g. \code{\link[=pgoedel.tnorm]{pgoedel.tnorm()}}, \code{\link[=pgoguen.tnorm]{pgoguen.tnorm()}} or \code{\link[=plukas.tnorm]{plukas.tnorm()}}.}

\item{onlyAnte}{\code{TRUE} is useful if rules store both the antecedent and consequent
and if only the antecedent-part of a rule should be included into the evaluated
conjunction. Antecedent-part of a rule are all predicates in the vector starting from
the 2nd position. \code{TRUE} value in this parameter causes the first element of each
rule to be ignored.

If \code{FALSE}, all predicates in a rule will be included in the conjunction.}

\item{parallel}{Deprecated parameter. Computation is done sequentially.}
}
\value{
If \code{x} is a matrix then the result of this function is a list
of numeric vectors with truth values of each rule, i.e., each element of the
resulting list corresponds to a rule and each value of the vector in the resulting
list corresponds to a row of the original data matrix \code{x}.

\code{x} as a vector is treated as a single-row matrix.
}
\description{
Given truth degrees of predicates, compute the truth value of given list of rules.
}
\details{
The aim of this function is to compute the truth value of each rule in a
\code{rules} list by assigning truth values to rule's predicates given by data \code{x}.

\code{x} is a numeric vector or numeric matrix of truth values of predicates. If
\code{x} is vector then \code{names(x)} must correspond to the predicate names
in \code{rules}. If \code{x} is a matrix then each column should represent a predicate
and thus \code{colnames(x)} must correspond to predicate names in \code{rules}.
Values of \code{x} are interpreted as truth values, i.e., they must be from the
interval \eqn{[0, 1]}. If matrix is given, the resulting truth values are
computed row-wisely.

\code{rules} may be a list of character vectors or an instance of the S3 class
\code{\link[=farules]{farules()}}. The character vectors in the \code{rules} list represent formulae
in conjunctive form. If \code{onlyAnte=FALSE}, \code{fire()} treats the rule as
a conjunction of all predicates, i.e., a conjunction of all predicates is
computed. If \code{onlyAnte=TRUE}, the first element of each rule is removed
prior evaluation, i.e., a conjunction of all predicates except the first
are computed: this is useful if \code{rules} is a \code{\link[=farules]{farules()}} object, since
\code{\link[=farules]{farules()}} objects save a rule's consequent as the first element (see also
\code{\link[=antecedents]{antecedents()}} and \code{\link[=consequents]{consequents()}} functions).

The type of  conjunction to be computed can be specified with the \code{tnorm} parameter.
}
\examples{

# fire whole rules on a vector
x <- 1:10 / 10
names(x) <- letters[1:10]
rules <- list(c('a', 'c', 'e'),
              c('b'),
              c('d', 'a'),
              c('c', 'a', 'b'))
fire(x, rules, tnorm='goguen', onlyAnte=FALSE)

# fire antecedents of the rules on a matrix
x <- matrix(1:20 / 20, nrow=2)
colnames(x) <- letters[1:10]
rules <- list(c('a', 'c', 'e'),
              c('b'),
              c('d', 'a'),
              c('c', 'a', 'b'))
fire(x, rules, tnorm='goedel', onlyAnte=TRUE)

# the former command should be equal to
fire(x, antecedents(rules), tnorm='goedel', onlyAnte=FALSE)

}
\seealso{
\code{\link[=aggregateConsequents]{aggregateConsequents()}}, \code{\link[=defuzz]{defuzz()}}, \code{\link[=perceive]{perceive()}}, \code{\link[=pbld]{pbld()}}, \code{\link[=infer]{infer()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=farules]{farules()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{multivariate}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateConsequents.R
\name{aggregateConsequents}
\alias{aggregateConsequents}
\title{Aggregation of fired consequents into a resulting fuzzy set}
\usage{
aggregateConsequents(
  conseq,
  degrees,
  partition,
  firing = lukas.residuum,
  aggreg = pgoedel.tnorm
)
}
\arguments{
\item{conseq}{A character vector of consequents. Each value in the vector
must correspond to a name of some column of the \code{partition} matrix.
The length of this vector must be the same as of the \code{degrees}
argument.}

\item{degrees}{A numeric vector of membership degrees at which the
corresponding consequents (see the \code{conseq} argument) are fired.}

\item{partition}{A matrix of membership degrees that describes the meaning of
the consequents in vector \code{conseq}: each column of the matrix
corresponds to a fuzzy set that models a single consequent (of a name given
by column names of the matrix), each row corresponds to a single crisp
value (which is not important for this function), hence each cell
corresponds to a membership degree in which the crisp value is a member of
a fuzzy set modeling the consequent.  Each consequent in \code{conseq}
must correspond to some column of this matrix. Such matrix may be created
e.g., by using \code{\link[=fcut]{fcut()}} or \code{\link[=lcut]{lcut()}} functions.}

\item{firing}{A two-argument function used to compute the resulting truth value of the consequent.
Function is evaluated for each consequent in \code{conseq}, with corresponding \code{degrees} value
as the first argument and corresponding truth-value of the consequent (from \code{partition})
as the second argument. In default, the Lukasiewicz residuum (\code{\link[=lukas.residuum]{lukas.residuum()}}) is
evaluated that way.}

\item{aggreg}{An aggregation function to be used to combine fuzzy sets resulting from firing
the consequents with the \code{firing} function. The function should accept multiple
numeric vectors of membership degrees as its arguments.
In default, the Goedel t-norm (\code{\link[=pgoedel.tnorm]{pgoedel.tnorm()}}) is evaluated.}
}
\value{
A vector of membership degrees of fuzzy set elements that correspond
to rows in the \code{partition} matrix. If empty vector of consequents is
provided, vector of 1's is returned. The length of the resulting vector
equals to the number of rows of the \code{partition} matrix.
}
\description{
Take a character vector of consequent names, a numeric vector representing
the degree of consequents' firing and a matrix that models fuzzy sets
corresponding to the consequent names, and perform an aggregation of the
consequents into a resulting fuzzy set.
}
\details{
This function is typically used within an inference mechanism, after a set of
firing rules is determined and membership degrees of their antecedents are
computed, to combine the consequents of the firing rules into a resulting
fuzzy set. The result of this function is then typically defuzzified
(see \code{\link[=defuzz]{defuzz()}}) to  obtain a crisp result of the inference.

Function assumes a set of rules with antecedents firing at degrees given in
\code{degrees} and with consequents in \code{conseq}. The meaning of the consequents is
modeled with fuzzy sets whose membership degree values are captured in the
\code{partition} matrix.

With default values of \code{firing} and \code{aggreg} parameters, the function
computes a fuzzy set that results from a conjunction (Goedel minimum t-norm)
of all provided implicative (Lukasiewicz residuum) rules.

In detail, the function first computes the fuzzy set of each fired consequent
by calling \verb{part\[i\] <- firing(degrees\[i\], partition\[, conseq\[i\]\])} for each
\code{i}-th consequent and the results are aggregated using the \code{aggreg}
parameter: \verb{aggreg(part\[1\], part\[2\], ...)}. In order to aggregate consequents
in a Mamdani-Assilian's fashion, set \code{firing} to \code{\link[=pgoedel.tnorm]{pgoedel.tnorm()}} and \code{aggreg}
to \code{\link[=pgoedel.tconorm]{pgoedel.tconorm()}}.
}
\examples{
    # create a partition matrix
    partition <- matrix(c(0:10/10, 10:0/10, rep(0, 5),
                          rep(0, 5), 0:10/10, 10:0/10,
                          0:12/12, 1, 12:0/12),
                        byrow=FALSE,
                        ncol=3)
    colnames(partition) <- c('a', 'b', 'c')

    # the result of aggregation is equal to:
    # pmin(1, partition[, 1] + (1 - 0.5), partition[, 2] + (1 - 0.8))
    aggregateConsequents(c('a', 'b'), c(0.5, 0.8), partition)

}
\seealso{
\code{\link[=fire]{fire()}}, \code{\link[=perceive]{perceive()}}, \code{\link[=pbld]{pbld()}}, \code{\link[=infer]{infer()}}, \code{\link[=defuzz]{defuzz()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algebra.R
\name{algebra}
\alias{algebra}
\alias{is.algebra}
\alias{goedel.tnorm}
\alias{lukas.tnorm}
\alias{goguen.tnorm}
\alias{pgoedel.tnorm}
\alias{plukas.tnorm}
\alias{pgoguen.tnorm}
\alias{goedel.tconorm}
\alias{lukas.tconorm}
\alias{goguen.tconorm}
\alias{pgoedel.tconorm}
\alias{plukas.tconorm}
\alias{pgoguen.tconorm}
\alias{goedel.residuum}
\alias{lukas.residuum}
\alias{goguen.residuum}
\alias{goedel.biresiduum}
\alias{lukas.biresiduum}
\alias{goguen.biresiduum}
\alias{invol.neg}
\alias{strict.neg}
\title{Algebra for Fuzzy Sets}
\usage{
algebra(name, stdneg = FALSE, ...)

is.algebra(a)

goedel.tnorm(...)

lukas.tnorm(...)

goguen.tnorm(...)

pgoedel.tnorm(...)

plukas.tnorm(...)

pgoguen.tnorm(...)

goedel.tconorm(...)

lukas.tconorm(...)

goguen.tconorm(...)

pgoedel.tconorm(...)

plukas.tconorm(...)

pgoguen.tconorm(...)

goedel.residuum(x, y)

lukas.residuum(x, y)

goguen.residuum(x, y)

goedel.biresiduum(x, y)

lukas.biresiduum(x, y)

goguen.biresiduum(x, y)

invol.neg(x)

strict.neg(x)
}
\arguments{
\item{name}{The name of the algebra to be created. Must be one of: "goedel",
"lukasiewicz", "goguen" (or an unambiguous abbreviation).}

\item{stdneg}{(Deprecated.) \code{TRUE} to force the use of a "standard" negation (i.e.,
involutive negation). Otherwise, the appropriate negation is used in the
algebra (e.g., strict negation in Goedel and Goguen algebras and involutive
negation in Lukasiewicz algebra).}

\item{...}{For t-norms and t-conorms, these arguments are numeric vectors
of values to compute t-norms or t-conorms from.  Values outside the
\eqn{[0,1]} interval cause an error. NA values are also permitted.

For the \code{algebra()} function, these arguments are passed to the factory
functions that create the algebra. (Currently unused.)}

\item{a}{An object to be checked if it is a valid algebra (i.e., a list
returned by the \code{algebra} function).}

\item{x}{Numeric vector of values to compute a residuum or bi-residuum from.
Values outside the \eqn{[0,1]} interval cause an error. NA values are also
permitted.}

\item{y}{Numeric vector of values to compute a residuum or bi-residuum from.
Values outside the \eqn{[0,1]} interval cause an error. NA values are also
permitted.}
}
\value{
Functions for t-norms and t-conorms (such as \code{goedel.tnorm})
return a numeric vector of size 1 that is the result of the appropriate
t-norm or t-conorm applied on all values of all arguments.

Element-wise versions of t-norms and t-conorms (such as \code{pgoedel.tnorm})
return a vector of results after applying the appropriate t-norm or t-conorm
on argument in an element-wise (i.e., by indices) way. The
resulting vector is of length of the longest argument (shorter arguments are
recycled).

Residua and bi-residua functions return a numeric vector of length of the
longest argument (shorter argument is recycled).

\code{strict.neg} and \code{invol.neg} compute negations and return a
numeric vector of the same size as the argument \code{x}.

\code{algebra} returns a list of functions of the requested algebra:
\code{"n"} (residual negation), \code{"ni"} (involutive negation), \code{"t"} (t-norm),
\code{"pt"} (element-wise t-norm),
\code{"c"} (t-conorm), \code{"pc"} (element-wise t-conorm), \code{"r"} (residuum),
\code{"b"} (bi-residuum), \code{"s"} (supremum),
\code{"ps"} (element-wise supremum), \code{"i"} (infimum), and
\code{"pi"} (element-wise infimum).

For Lukasiewicz algebra, the elements \code{"n"} and \code{"ni"} are the same, i.e.
the \code{invol.neg} function. For Goedel and Goguen algebra, \code{"n"} (the residual
negation) equals \code{strict.neg} and \code{"ni"} (the involutive negation) equals
\code{invol.neg}.

\code{"s"}, \code{"ps"}, \code{"i"}, \code{"pi"} are the same for each type of algebra:
\code{goedel.conorm}, \code{pgoedel.conorm}, \code{goedel.tnorm}, and \code{pgoedel.tnorm}.
}
\description{
Compute triangular norms (t-norms), triangular conorms (t-conorms), residua,
bi-residua, and negations.
}
\details{
\code{goedel.tnorm}, \code{lukas.tnorm}, and \code{goguen.tnorm} compute the
Goedel, Lukasiewicz, and Goguen triangular norm (t-norm) from all values in
the arguments. If the arguments are vectors they are combined together
firstly so that a numeric vector of length 1 is returned.

\code{pgoedel.tnorm}, \code{plukas.tnorm}, and \code{pgoguen.tnorm} compute
the same t-norms, but in an element-wise manner. I.e. the values
with indices 1 of all arguments are used to compute the t-norm, then the
second values (while recycling the vectors if they do not have the same
size) so that the result is a vector of values.

\code{goedel.tconorm}, \code{lukas.tconorm}, \code{goguen.tconorm}, are
similar to the previously mentioned functions, except that they compute
triangular conorms (t-conorms).  \code{pgoedel.tconorm},
\code{plukas.tconorm}, and \code{pgoguen.tconorm} are their element-wise alternatives.

\code{goedel.residuum}, \code{lukas.residuum}, and \code{goguen.residuum}
compute residua (i.e., implications) and \code{goedel.biresiduum},
\code{lukas.biresiduum}, and \code{goguen.biresiduum} compute bi-residua. Residua and
bi-residua are computed in an element-wise manner, for each corresponding
pair of values in \code{x} and \code{y} arguments.

\code{invol.neg} and \code{strict.neg} compute the involutive and strict
negation, respectively.

Let \eqn{a}, \eqn{b} be values from the interval \eqn{[0, 1]}. The realized functions
can be defined as follows:
\itemize{
\item Goedel t-norm: \eqn{min{a, b}};
\item Goguen t-norm: \eqn{ab};
\item Lukasiewicz t-norm: \eqn{max{0, a+b-1}};
\item Goedel t-conorm: \eqn{max{a, b}};
\item Goguen t-conorm: \eqn{a+b-ab};
\item Lukasiewicz t-conorm: \eqn{min{1, a+b}};
\item Goedel residuum (standard Goedel implication): \eqn{1} if \eqn{a \le b} and
\eqn{b} otherwise;
\item Goguen residuum (implication): \eqn{1} if \eqn{a \le b} and \eqn{b/a}
otherwise;
\item Lukasiewicz residuum (standard Lukasiewicz implication): \eqn{1} if
\eqn{a \le b} and \eqn{1-a+b} otherwise;
\item Involutive negation: \eqn{1-x};
\item Strict negation: \eqn{1} if \eqn{x=0} and \eqn{0} otherwise.
}

Bi-residuum \eqn{B} is derived from residuum \eqn{R}
as follows: \deqn{B(a, b) = inf(R(a, b), R(b, a)),}
where \eqn{inf} is the operation of infimum, which for all three algebras
corresponds to the \eqn{min} operation.

The arguments have to be numbers from the interval \eqn{[0, 1]}. Values
outside that range cause an error. NaN values are treated as NAs.

If some argument is NA or NaN, the result is NA. For other handling of missing values,
see \link{algebraNA}.

Selection of a t-norm may serve as a basis for definition of other operations.
From the t-norm, the operation of a residual implication may be defined, which
in turn allows the definition of a residual negation. If the residual negation
is not involutive, the involutive negation is often added as a new operation
and together with the t-norm can be used to define the t-conorm. Therefore,
the \code{algebra} function returns a named list of operations derived from the selected
Goedel, Goguen, or Lukasiewicz t-norm. Concretely:
\itemize{
\item \code{algebra("goedel")}: returns the strict negation as the residual negation,
the involutive negation, and also the Goedel t-norm, t-conorm, residuum, and bi-residuum;
\item \code{algebra("goguen")}: returns the strict negation as the residual negation,
the involutive negation, and also the Goguen t-norm, t-conorm, residuum, and bi-residuum;
\item \code{algebra("lukasiewicz")}: returns involutive negation as both residual and involutive
negation, and also the Lukasiewicz t-norm, t-conorm, residuum, and bi-residuum.
}

Moreover, \code{algebra} returns the supremum and infimum functions computed as maximum and minimum,
respectively.

\code{is.algebra} tests whether the given \code{a} argument is a valid
algebra, i.e., a list returned by the \code{algebra} function.
}
\examples{
    # direct and element-wise version of functions
    goedel.tnorm(c(0.3, 0.2, 0.5), c(0.8, 0.1, 0.5))  # 0.1
    pgoedel.tnorm(c(0.3, 0.2, 0.5), c(0.8, 0.1, 0.5)) # c(0.3, 0.1, 0.5)

    # algebras
    x <- runif(10)
    y <- runif(10)
    a <- algebra('goedel')
    a$n(x)     # residual negation
    a$ni(x)    # involutive negation
    a$t(x, y)  # t-norm
    a$pt(x, y) # element-wise t-norm
    a$c(x, y)  # t-conorm
    a$pc(x, y) # element-wise t-conorm
    a$r(x, y)  # residuum
    a$b(x, y)  # bi-residuum
    a$s(x, y)  # supremum
    a$ps(x, y) # element-wise supremum
    a$i(x, y)  # infimum
    a$pi(x, y) # element-wise infimum

    is.algebra(a) # TRUE
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ctx.R
\name{ctx}
\alias{ctx}
\alias{ctx3}
\alias{ctx3bilat}
\alias{ctx5}
\alias{ctx5bilat}
\alias{as.ctx3}
\alias{as.ctx3.ctx3}
\alias{as.ctx3.ctx3bilat}
\alias{as.ctx3.ctx5}
\alias{as.ctx3.ctx5bilat}
\alias{as.ctx3.default}
\alias{as.ctx3bilat}
\alias{as.ctx3bilat.ctx3bilat}
\alias{as.ctx3bilat.ctx3}
\alias{as.ctx3bilat.ctx5}
\alias{as.ctx3bilat.ctx5bilat}
\alias{as.ctx3bilat.default}
\alias{as.ctx5}
\alias{as.ctx5.ctx5}
\alias{as.ctx5.ctx3}
\alias{as.ctx5.ctx3bilat}
\alias{as.ctx5.ctx5bilat}
\alias{as.ctx5.default}
\alias{as.ctx5bilat}
\alias{as.ctx5bilat.ctx5bilat}
\alias{as.ctx5bilat.ctx3}
\alias{as.ctx5bilat.ctx3bilat}
\alias{as.ctx5bilat.ctx5}
\alias{as.ctx5bilat.default}
\alias{is.ctx3}
\alias{is.ctx3bilat}
\alias{is.ctx5}
\alias{is.ctx5bilat}
\title{Context for linguistic expressions}
\usage{
ctx3(
  low = 0,
  center = low + (high - low) * relCenter,
  high = 1,
  relCenter = 0.5
)

ctx3bilat(
  negMax = -1,
  negCenter = origin + (negMax - origin) * relCenter,
  origin = 0,
  center = origin + (max - origin) * relCenter,
  max = 1,
  relCenter = 0.5
)

ctx5(
  low = 0,
  lowerCenter = mean(c(low, center)),
  center = low + (high - low) * relCenter,
  upperCenter = mean(c(center, high)),
  high = 1,
  relCenter = 0.5
)

ctx5bilat(
  negMax = -1,
  negUpperCenter = mean(c(negCenter, negMax)),
  negCenter = origin + (negMax - origin) * relCenter,
  negLowerCenter = mean(c(origin, negCenter)),
  origin = 0,
  lowerCenter = mean(c(origin, center)),
  center = origin + (max - origin) * relCenter,
  upperCenter = mean(c(center, max)),
  max = 1,
  relCenter = 0.5
)

as.ctx3(x)

\method{as.ctx3}{ctx3}(x)

\method{as.ctx3}{ctx3bilat}(x)

\method{as.ctx3}{ctx5}(x)

\method{as.ctx3}{ctx5bilat}(x)

\method{as.ctx3}{default}(x)

as.ctx3bilat(x)

\method{as.ctx3bilat}{ctx3bilat}(x)

\method{as.ctx3bilat}{ctx3}(x)

\method{as.ctx3bilat}{ctx5}(x)

\method{as.ctx3bilat}{ctx5bilat}(x)

\method{as.ctx3bilat}{default}(x)

as.ctx5(x)

\method{as.ctx5}{ctx5}(x)

\method{as.ctx5}{ctx3}(x)

\method{as.ctx5}{ctx3bilat}(x)

\method{as.ctx5}{ctx5bilat}(x)

\method{as.ctx5}{default}(x)

as.ctx5bilat(x)

\method{as.ctx5bilat}{ctx5bilat}(x)

\method{as.ctx5bilat}{ctx3}(x)

\method{as.ctx5bilat}{ctx3bilat}(x)

\method{as.ctx5bilat}{ctx5}(x)

\method{as.ctx5bilat}{default}(x)

is.ctx3(x)

is.ctx3bilat(x)

is.ctx5(x)

is.ctx5bilat(x)
}
\arguments{
\item{low}{Lowest value of an unilateral context.}

\item{center}{A positive middle value of a bilateral context, or simply a
middle value of an unilateral context.}

\item{high}{Highest value of an unilateral context.}

\item{relCenter}{A relative quantity used to compute the \code{negCenter}
and/or \code{center}, if they are not specified explicitly. The sensible
value is 0.5 for context symmetric around center, or 0.42 as proposed by
Novak.}

\item{negMax}{Lowest negative value of a bilateral context.}

\item{negCenter}{A negative middle value.}

\item{origin}{Origin, i.e., the initial point of the bilateral context. It is
typically a value of zero.}

\item{max}{Highest value of a bilateral context.}

\item{lowerCenter}{A typical positive value between origin and center.}

\item{upperCenter}{A typical positive value between center and maximum.}

\item{negUpperCenter}{A typical negative value between \code{negMax} and
\code{negCenter}.}

\item{negLowerCenter}{A typical negative value between \code{negCenter} and
\code{negOrigin}.}

\item{x}{A value to be examined or converted. For \verb{as.ctx*}, it can be
an instance of any \verb{ctx*} class or a numeric vector of size equal to
the number of points required for the given context type.}
}
\value{
\verb{ctx*} and \verb{as.ctx*} return an instance of the appropriate
class. \verb{is.ctx*} returns \code{TRUE} or \code{FALSE}.
}
\description{
A context describes a range of allowed values for a data column.
}
\details{
A context describes a range of allowed values for a data column. For that,
only the borders of the interval, i.e., minimum and maximum, are usually
needed, but we use contexts to hold more additional information that is
crucial for the construction of linguistic expressions.

Currently, four different contexts are supported that determine the types of
possible linguistic expressions, as constructed with \code{\link[=lingexpr]{lingexpr()}}.
Unilateral or bilateral context is allowed in the variants of trichotomy or
pentachotomy. Trichotomy distinguishes three points in the interval: the
lowest value, highest value, and center. Pentachotomy adds lower center and
upper center to them. As opposite to unilateral, the bilateral context
handles explicitly the negative values. That is, bilateral context expects
some middle point, the origin (usually 0), around which the positive and
negative values are placed.

Concretely, the type of the context determines the allowed atomic
expressions as follows:
\itemize{
\item \code{ctx3}: trichotomy (low, center, high) enables atomic expressions:
small, medium, big;
\item \code{ctx5}: pentachotomy (low, lowerCenter, center, upperCenter, high) enables
atomic expressions: small, lower medium, medium, upper medium, big;
\item \code{ctx3bilat}: bilateral trichotomy (negMax, negCenter, origin, center, max)
enables atomic expressions: negative big, negative medium, negative small,
zero, small, medium, big;
\item \code{ctx5bilat}: bilateral pentachotomy (negMax, negCenter, origin, center,
max) enables atomic expressions: negative big, negative medium, negative
small, zero, small, medium, big.
}

The \verb{as.ctx*} functions return instance of the appropriate class. The
functions perform the conversion so that missing points of the new context
are computed from the old context that is being transformed. In the
subsequent table, rows represent compatible values of different context
types:

\tabular{llll}{
ctx3   \tab ctx5        \tab ctx3bilat \tab ctx5bilat      \cr
\tab             \tab negMax    \tab negMax         \cr
\tab             \tab           \tab negUpperCenter \cr
\tab             \tab negCenter \tab negCenter      \cr
\tab             \tab           \tab negLowerCenter \cr
low    \tab low         \tab origin    \tab origin         \cr
\tab lowerCenter \tab           \tab lowerCenter    \cr
center \tab center      \tab center    \tab center \cr
\tab upperCenter \tab           \tab upperCenter \cr
high   \tab high        \tab max       \tab max
}

The \verb{as.ctx*} conversion is performed by replacing values by rows, as
indicated in the table above.  When converting from a context with less
points to a context with more points (e.g., from unilateral to bilateral, or
from trichotomy to pentachotomy), missing points are computed as follows:
\itemize{
\item \code{center} is computed as a mean of \code{origin} (or \code{low}) and \code{max} (or \code{high}).
\item \code{lowerCenter} is computed as a mean of \code{origin} (or \code{low}) and \code{center}.
\item \code{upperCenter} is computed as a mean of \code{mas} (or \code{high}) and \code{center}.
\item negative points (such as \code{negMax}, \code{negCenter} etc.) are computed
symmetrically around \code{origin} to the corresponding positive points.
}

The code \verb{as.ctx*} functions allow the parameter to be also a numeric
vector of size equal to the number of points required for the given context
type, i.e. 3 (\code{ctx3}), 5 (\code{ctx3bilat}, \code{ctx5}), or 9 (\code{ctx5bilat}).
}
\examples{
    ctx3(low=0, high=10)
    as.ctx3bilat(ctx3(low=0, high=10))
}
\seealso{
\code{\link[=minmax]{minmax()}}, \code{\link[=lingexpr]{lingexpr()}}, \code{\link[=horizon]{horizon()}}, \code{\link[=hedge]{hedge()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{as.data.frame.fsets}
\alias{as.data.frame.fsets}
\alias{as.matrix.fsets}
\title{Convert an object of \code{fsets} class into a matrix or data frame
This function converts an instance of S3 class \link{fsets} into a
matrix or a data frame. The \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}} attributes
of the original object are deleted.}
\usage{
\method{as.data.frame}{fsets}(x, ...)

\method{as.matrix}{fsets}(x, ...)
}
\arguments{
\item{x}{An instance of class \link{fsets} to be converted}

\item{...}{arguments further passed to \code{as.data.frame} after converting
to matrix in \code{as.data.frame.fsets}. Unused in \code{as.matrix.fsets}.}
}
\value{
A numeric matrix or data frame of membership degrees.
}
\description{
Convert an object of \code{fsets} class into a matrix or data frame
This function converts an instance of S3 class \link{fsets} into a
matrix or a data frame. The \code{\link[=vars]{vars()}} and \code{\link[=specs]{specs()}} attributes
of the original object are deleted.
}
\seealso{
\code{\link[=fsets]{fsets()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fsets.R
\name{print.fsets}
\alias{print.fsets}
\title{Print an instance of the \code{\link[=fsets]{fsets()}} class}
\usage{
\method{print}{fsets}(x, ...)
}
\arguments{
\item{x}{An instance of the \code{\link[=fsets]{fsets()}} class}

\item{...}{Unused.}
}
\value{
Nothing
}
\description{
Format an object of the \code{\link[=fsets]{fsets()}} class into human readable form
and print it to the output.
}
\examples{
    d <- fcut(CO2[, 1:2])
    print(d)

}
\seealso{
\code{\link[=fsets]{fsets()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lingexpr.R
\name{lingexpr}
\alias{lingexpr}
\alias{allowed.lingexpr}
\title{Creator of functions representing linguistic expressions}
\usage{
lingexpr(
  context,
  atomic = c("sm", "me", "bi", "lm", "um", "ze", "neg.sm", "neg.me", "neg.bi", "neg.lm",
    "neg.um"),
  hedge = c("ex", "si", "ve", "ty", "-", "ml", "ro", "qr", "vr"),
  negated = FALSE,
  hedgeParams = defaultHedgeParams
)

allowed.lingexpr
}
\arguments{
\item{context}{A context of linguistic expressions (see \code{\link[=ctx3]{ctx3()}}, \code{\link[=ctx5]{ctx5()}}, \code{\link[=ctx3bilat]{ctx3bilat()}} or \code{\link[=ctx5bilat]{ctx5bilat()}})}

\item{atomic}{An atomic expression whose horizon we would like to obtain}

\item{hedge}{The type of the required linguistic hedge ('-' for no hedging)}

\item{negated}{Negate the expression? (For instance, "not very small".) Negation
is done using the \code{\link[=invol.neg]{invol.neg()}} function.}

\item{hedgeParams}{Parameters that determine the shape of the hedges}
}
\value{
Returns a function with a single argument, which has to be a numeric vector.
}
\description{
A linguistic expression represents vague human terms such as "very small", "extremely big" etc. Such notions are
always reasoned within a given context. \code{lingexpr} returns a function that models a selected linguistic expression.
Accordingly to the given \code{context}, \code{atomic} expression (such as "small", "big") and a linguistic \code{hedge} (such as
\code{very}, \code{extremely}), the returned function transforms numeric values into degrees (from \verb{[0, 1]} interval),
at which the values correspond to the expression.
}
\details{
Based on the context type, the following atomic expressions are allowed:
\itemize{
\item \code{\link[=ctx3]{ctx3()}} (trichotomy): small, medium, big;
\item \code{\link[=ctx5]{ctx5()}} (pentachotomy): small, lower medium, medium, upper medium, big;
\item \code{\link[=ctx3bilat]{ctx3bilat()}} (bilateral trichotomy): negative big, negative medium, negative small,
zero, small, medium, big;
\item \code{\link[=ctx5bilat]{ctx5bilat()}} (bilateral pentachotomy): negative big, negative medium, negative
small, zero, small, medium, big.
}

The values of the \code{atomic} parameter have the following meaning (in ascending order):
\itemize{
\item \code{neg.bi}: big negative (far from zero)
\item \code{neg.um}: upper medium negative (between medium negative and big negative)
\item \code{neg.me}: medium negative
\item \code{neg.lm}: lower medium negative (between medium negative and small negative)
\item \code{neg.sm}: small negative (close to zero)
\item \code{ze}: zero
\item \code{sm}: small
\item \code{lm}: lower medium
\item \code{me}: medium
\item \code{um}: upper medium
\item \code{bi}: big
}

\code{hedge} parameter has the following meaning:
\itemize{
\item \code{ex}: extremely,
\item \code{si}: significantly,
\item \code{ve}: very,
\item \code{ty}: typically,
\item \code{-}: empty hedge,
\item \code{ml}: more or less,
\item \code{ro}: roughly,
\item \code{qr}: quite roughly,
\item \code{vr}: very roughly.
}

Accordingly to the theory of linguistic expressions by Novak, not every hedge is applicable to each atomic
expression. The combinations of allowed pairs can be found in \link{allowed.lingexpr}. Trying to create forbidden
combination results in error.
}
\examples{
    small <- lingexpr(ctx3(0, 0.5, 1), atomic='sm', hedge='-')
    small(0)   # 1
    small(0.8) # 0
    plot(small)

    verySmall <- lingexpr(ctx3(0, 0.5, 1), atomic='sm', hedge='ve')
    plot(verySmall)
}
\seealso{
\code{\link[=horizon]{horizon()}}, \code{\link[=hedge]{hedge()}}, \code{\link[=fcut]{fcut()}}, \code{\link[=lcut]{lcut()}}, \code{\link[=ctx]{ctx()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ft.R
\name{is.ft}
\alias{is.ft}
\title{Test whether \code{x} is a valid object of the S3 \code{ft} class}
\usage{
is.ft(x)
}
\arguments{
\item{x}{An object to be tested}
}
\value{
\code{TRUE} if \code{x} is a valid \code{ft} object and \code{FALSE} otherwise.
}
\description{
Test whether \code{x} has a valid format for objects of the S3 \code{ft} class that represents
the Fuzzy Transform.
}
\details{
This function tests whether \code{x} is an instance of the \code{ft} class and whether it is a list
with the following elements: \code{inputs} character vector, \code{partitions} list, \code{order} number,
\code{antecedents} matrix and \code{consequents} matrix.
}
\seealso{
\code{\link[=ft]{ft()}}, \code{\link[=ftinv]{ftinv()}}
}
\author{
Michal Burda
}
\keyword{models}
\keyword{robust}
